Package: tinyplot
Type: Package
Title: Lightweight Extension of the Base R Graphics System
Version: 0.6.1
Date: 2026-03-26
Authors@R: 
  c(
    person(
      given = "Grant",
      family = "McDermott",
      role = c("aut", "cre"),
      email = "contact@grantmcdermott.com",
      comment = c(ORCID = "0000-0001-7883-8573")
    ),
    person(
      given = "Vincent",
      family = "Arel-Bundock",
      role = "aut",
      email = "vincent.arel-bundock@umontreal.ca",
      comment = c(ORCID = "0000-0003-1995-6531")
    ),
    person(
      given = "Achim",
      family = "Zeileis",
      role = "aut",
      email = "Achim.Zeileis@R-project.org",
      comment = c(ORCID = "0000-0003-0918-3766")
     ),
    person(
      given = "Etienne",
      family = "Bacher",
      role = "ctb",
      email = "etienne.bacher@protonmail.com"
    )
  )
Description: Lightweight extension of the base R graphics system, with support
      for automatic legends, facets, themes, and various other enhancements.
License: Apache License (>= 2)
Depends: R (>= 4.0.0)
Imports: graphics, grDevices, stats, tools, utils
Suggests: fontquiver, png, rsvg, svglite (>= 2.2.0), tinytest,
        tinysnapshot (>= 0.0.3), knitr
Config/Needs/website: altdoc (>= 0.7.2), future.apply
Encoding: UTF-8
RoxygenNote: 7.3.3
URL: https://grantmcdermott.com/tinyplot/
BugReports: https://github.com/grantmcdermott/tinyplot/issues
NeedsCompilation: no
Packaged: 2026-03-27 04:50:26 UTC; gmcd
Author: Grant McDermott [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7883-8573>),
  Vincent Arel-Bundock [aut] (ORCID:
    <https://orcid.org/0000-0003-1995-6531>),
  Achim Zeileis [aut] (ORCID: <https://orcid.org/0000-0003-0918-3766>),
  Etienne Bacher [ctb]
Maintainer: Grant McDermott <contact@grantmcdermott.com>
Repository: CRAN
Date/Publication: 2026-03-28 06:10:16 UTC
Built: R 4.6.0; ; 2026-04-26 15:00:45 UTC; unix

Package: tinyplot
Type: Package
Title: Lightweight Extension of the Base R Graphics System
Version: 0.7.0
Date: 2026-07-02
Authors@R: 
  c(
    person(
      given = "Grant",
      family = "McDermott",
      role = c("aut", "cre"),
      email = "contact@grantmcdermott.com",
      comment = c(ORCID = "0000-0001-7883-8573")
    ),
    person(
      given = "Vincent",
      family = "Arel-Bundock",
      role = "aut",
      email = "vincent.arel-bundock@umontreal.ca",
      comment = c(ORCID = "0000-0003-1995-6531")
    ),
    person(
      given = "Achim",
      family = "Zeileis",
      role = "aut",
      email = "Achim.Zeileis@R-project.org",
      comment = c(ORCID = "0000-0003-0918-3766")
     ),
    person(
      given = "Etienne",
      family = "Bacher",
      role = "ctb",
      email = "etienne.bacher@protonmail.com"
    )
  )
Description: Lightweight extension of the base R graphics system, with support
      for automatic legends, facets, themes, and various other enhancements.
License: Apache License (>= 2)
Depends: R (>= 4.0.0)
Imports: graphics, grDevices, stats, tools, utils
Suggests: fontquiver, png, rsvg, svglite (>= 2.2.0), tinytest,
        tinysnapshot (>= 0.0.3), knitr
Config/Needs/website: altdoc (>= 0.7.2), future.apply
Encoding: UTF-8
URL: https://grantmcdermott.com/tinyplot/
BugReports: https://github.com/grantmcdermott/tinyplot/issues
Config/roxygen2/version: 8.0.0
NeedsCompilation: no
Packaged: 2026-07-03 02:03:38 UTC; gmcd
Author: Grant McDermott [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-7883-8573>),
  Vincent Arel-Bundock [aut] (ORCID:
    <https://orcid.org/0000-0003-1995-6531>),
  Achim Zeileis [aut] (ORCID: <https://orcid.org/0000-0003-0918-3766>),
  Etienne Bacher [ctb]
Maintainer: Grant McDermott <contact@grantmcdermott.com>
Repository: CRAN
Date/Publication: 2026-07-03 06:40:02 UTC

# Generated by roxygen2: do not edit by hand

S3method(tinyplot,default)
S3method(tinyplot,density)
S3method(tinyplot,formula)
export(draw_legend)
export(get_saved_par)
export(plt)
export(plt_add)
export(tinylabel)
export(tinyplot)
export(tinyplot_add)
export(tinytheme)
export(tpar)
export(type_abline)
export(type_area)
export(type_barplot)
export(type_boxplot)
export(type_density)
export(type_errorbar)
export(type_function)
export(type_glm)
export(type_hist)
export(type_histogram)
export(type_hline)
export(type_jitter)
export(type_lines)
export(type_lm)
export(type_loess)
export(type_pointrange)
export(type_points)
export(type_polygon)
export(type_polypath)
export(type_qq)
export(type_rect)
export(type_ribbon)
export(type_ridge)
export(type_rug)
export(type_segments)
export(type_spineplot)
export(type_spline)
export(type_summary)
export(type_text)
export(type_violin)
export(type_vline)
importFrom(grDevices,adjustcolor)
importFrom(grDevices,as.raster)
importFrom(grDevices,axisTicks)
importFrom(grDevices,cairo_pdf)
importFrom(grDevices,col2rgb)
importFrom(grDevices,colorRampPalette)
importFrom(grDevices,convertColor)
importFrom(grDevices,dev.cur)
importFrom(grDevices,dev.list)
importFrom(grDevices,dev.new)
importFrom(grDevices,dev.off)
importFrom(grDevices,extendrange)
importFrom(grDevices,gray.colors)
importFrom(grDevices,hcl)
importFrom(grDevices,hcl.colors)
importFrom(grDevices,hcl.pals)
importFrom(grDevices,jpeg)
importFrom(grDevices,nclass.Sturges)
importFrom(grDevices,palette)
importFrom(grDevices,palette.colors)
importFrom(grDevices,palette.pals)
importFrom(grDevices,pdf)
importFrom(grDevices,png)
importFrom(grDevices,recordGraphics)
importFrom(grDevices,svg)
importFrom(grDevices,xy.coords)
importFrom(graphics,Axis)
importFrom(graphics,abline)
importFrom(graphics,arrows)
importFrom(graphics,axTicks)
importFrom(graphics,axis)
importFrom(graphics,box)
importFrom(graphics,boxplot)
importFrom(graphics,grconvertX)
importFrom(graphics,grconvertY)
importFrom(graphics,hist)
importFrom(graphics,legend)
importFrom(graphics,lines)
importFrom(graphics,mtext)
importFrom(graphics,par)
importFrom(graphics,plot.default)
importFrom(graphics,plot.new)
importFrom(graphics,plot.window)
importFrom(graphics,points)
importFrom(graphics,polygon)
importFrom(graphics,polypath)
importFrom(graphics,rasterImage)
importFrom(graphics,rect)
importFrom(graphics,rug)
importFrom(graphics,segments)
importFrom(graphics,strwidth)
importFrom(graphics,text)
importFrom(graphics,title)
importFrom(stats,aggregate)
importFrom(stats,approx)
importFrom(stats,as.formula)
importFrom(stats,ave)
importFrom(stats,bw.SJ)
importFrom(stats,bw.bcv)
importFrom(stats,bw.nrd)
importFrom(stats,bw.nrd0)
importFrom(stats,bw.ucv)
importFrom(stats,density)
importFrom(stats,dnorm)
importFrom(stats,glm)
importFrom(stats,lm)
importFrom(stats,loess)
importFrom(stats,loess.control)
importFrom(stats,median)
importFrom(stats,model.frame)
importFrom(stats,na.omit)
importFrom(stats,ppoints)
importFrom(stats,predict)
importFrom(stats,qnorm)
importFrom(stats,qt)
importFrom(stats,quantile)
importFrom(stats,setNames)
importFrom(stats,spline)
importFrom(stats,terms)
importFrom(stats,weighted.mean)
importFrom(tools,file_ext)
importFrom(utils,globalVariables)
importFrom(utils,head)
importFrom(utils,modifyList)
importFrom(utils,tail)

<img src="https://grantmcdermott.com/tinyplot/logo.svg" align="right" width="120" alt="tinyplot logo" />

# tinyplot

<!-- badges: start -->

<a href="https://CRAN.R-project.org/package=tinyplot"><img src="https://www.r-pkg.org/badges/version/tinyplot" class="img-fluid" alt="CRAN version"></a>
<a href="https://grantmcdermott.r-universe.dev"><img src="https://grantmcdermott.r-universe.dev/badges/tinyplot" class="img-fluid" alt="R-universe version"></a>
<a href="https://CRAN.R-project.org/package=tinyplot"><img src="https://tinyverse.netlify.app/badge/tinyplot" class="img-fluid" alt="Dependencies"></a>
<a href="https://github.com/grantmcdermott/tinyplot/actions/workflows/R-CMD-check.yaml"><img src="https://github.com/grantmcdermott/tinyplot/actions/workflows/R-CMD-check.yaml/badge.svg" class="img-fluid" alt="GitHub Actions"></a>
<a href="https://grantmcdermott.com/tinyplot/LICENSE.html"><img src="https://img.shields.io/badge/license-Apache2.0-blue" class="img-fluid" alt="License"></a>
<a href="https://grantmcdermott.com/tinyplot/index.html"><img src="https://img.shields.io/badge/docs-homepage-blue.svg" class="img-fluid" alt="Docs"></a>
<!-- badges: end -->

## What

A lightweight extension of the base R graphics system, with support for
automatic grouping, legends, facets, themes, and various other
enhancements.

The stable version of **tinyplot** is available on
[CRAN](https://CRAN.R-project.org/package=tinyplot).

``` r
install.packages("tinyplot")
```

Or, you can grab the latest development version from
[R-universe](https://grantmcdermott.R-universe.dev/).

``` r
install.packages("tinyplot", repos = "https://grantmcdermott.r-universe.dev")
```

## Why

R users are spoiled for choice when it comes to visualization
frameworks. The options, of course, include **ggplot2** and **lattice**;
say nothing of the bewildering array of extensions built around, on top
of, and in between these amazing packages.[^1]

Given the wealth of options, it is perhaps understandable that even avid
R users can overlook the base R **graphics** system. This is
unfortunate, because base R offers very powerful and flexible plotting
facilities.[^2] The downside of this power and flexibility is that base
R plotting can require a lot of manual tinkering. A case in point is
plotting grouped data with an appropriate legend. Doing so with the
generic `plot()` function can require several function calls or a loop,
fiddling with your plot regions, and then generating the legend
manually.

The **tinyplot** package aims to remove this overhead. It provides a
lightweight extension of the base R graphics system that preserves the
same core logic and syntax, but with numerous convenience features to
make base R plotting a more user-friendly experience. For example, the
core `tinyplot()` function—or its shorthand alias `plt()`—makes it easy
to plot grouped datasets and generate automatic legends in a single
function call. Or, you can display groups in separate facets without
having to worry about manually setting plot regions. While **tinyplot**
offers these and various other enhancements, it tries as far as possible
to be a drop-in replacement for the equivalent base plotting function.
Users should generally be able to swap out a valid `plot()` call for
`tinyplot()`/`plt()` without any changes to the expected output.

It is worth emphasizing that **tinyplot** requires *only* base R. It has
zero recursive dependencies and we have been careful to keep its
installation size down to a minimum. **tinyplot** should thus provide an
attractive and lightweight option for package developers (or regular R
users!) who would like to produce convenient and sophisticated plots,
combined with a minimal footprint.

## Quickstart

The **tinyplot** website includes a detailed [introductory
tutorial](https://grantmcdermott.com/tinyplot/vignettes/introduction.html),
with numerous examples. But here are some quickstart examples of the
package in action.

``` r
library(tinyplot)
```

Grouped scatterplot with automatic legend:

``` r
# with(iris, tinyplot(x = Petal.Length, y = Sepal.Length, by = Species)) # atomic
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris)             # formula
```

<img src="man/figures/README-quickstart2-1.png" style="width:70.0%" />

If you would prefer to save on a few keystrokes, you can use the
shorthand `plt()` alias instead of typing out `tinyplot()` in full.
Here’s the same plot with this shorthand `plt()` alias, as well as an
added `"lm"` layer and a few aesthetic tweaks:

``` r
plt(
  Sepal.Length ~ Petal.Length | Species, 
  data = iris,
  palette = "dark", pch = 16,
  grid = TRUE, frame = FALSE
)
plt_add(type = "lm")
```

<img src="man/figures/README-quickstart3-1.png" style="width:70.0%" />

Speaking of aesthetic tweaks, **tinyplot** also provides a set of
built-in themes for convenient plot customization:

``` r
tinytheme("clean2")

plt(Sepal.Length ~ Petal.Length | Species, data = iris)
plt_add(type = "lm")
```

<img src="man/figures/README-quickstart_theme-1.png"
style="width:70.0%" />

Themes are persistent and will be applied to subsequent plots. For
example, here is a grouped density plot:

``` r
plt(
  ~ Petal.Length | Species,
  data = iris,
  type = "density",
  fill = "by",
  main = "Distribution of petal lengths",
  sub = "Grouped by species"
)
```

<img src="man/figures/README-quickstart4-1.png" style="width:70.0%" />

Grouped scatterplot with (continuous) gradient legend, combined with
facet layout:

``` r
plt(
  Sepal.Length ~ Petal.Length | Sepal.Length, data = iris,
  facet = ~Species, pch = 19,
  main = "Faceted flowers", sub = "Brought to you by tinyplot"
)
```

<img src="man/figures/README-quickstart5-1.png" style="width:70.0%" />

``` r
# reset the theme 
tinytheme()
```

Hopefully, these have been enough to pique your interest. Head over to
the [intro
tutorial](https://grantmcdermott.com/tinyplot/vignettes/introduction.html)
for many more examples, including model-based and custom plot types.

[^1]: You could make a strong case that **ggplot2**, in particular, is
    the most influential graphics library of the last two decades,
    across any programming language. Both **ggplot2** and **lattice**
    are built on top of the lower-level **grid** package, which was
    incorporated into the base R distribution way back in R 2.0.0.

[^2]: Just type `demo(graphics)` or `demo(persp)` into your R console to
    get an idea. Or, take a look at
    [these](https://github.com/karoliskoncevicius/r_notes/blob/main/baseplotting.md)
    [two](https://poisonalien.github.io/basegraphics/baseplots.html)
    excellent tutorials.

sanitize_type = function(settings) {
  env2env(settings, environment(), c("type", "dots", "x", "y"))

  if (inherits(type, "tinyplot_type")) {
    settings$type = type$name
    settings$type_draw = type$draw
    settings$type_data = type$data
    return(invisible(NULL))
  }

  known_types = c(
    "p", "l", "o", "b", "c", "h", "j", "s", "S", "n",
    "abline",
    "area",
    "bar", "barplot",
    "box", "boxplot",
    "chull",
    "density",
    "ellipse",
    "errorbar",
    "function",
    "glm",
    "hist", "histogram",
    "hline",
    "j", "jitter",
    "lines",
    "lm",
    "loess",
    "pointrange",
    "points",
    "polygon", "polypath",
    "qq",
    "rect",
    "ribbon",
    "ridge",
    "rug",
    "segments",
    "spine", "spineplot",
    "spline",
    "summary",
    "text",
    "violin",
    "vline"
  )
  assert_choice(type, known_types, null.ok = TRUE)

  if (is.null(type)) {
    if (is.null(x) && !(is.factor(y) || is.character(y))) {
      # enforce histogram type for y ~ 1
      settings$x = y
      settings$y = NULL
      type = type_hist
    } else if (is.null(x) && (is.factor(y) || is.character(y))) {
      # enforce barplot type for factor(y) ~ 1
      settings$x = y
      settings$y = NULL
      type = type_barplot
    } else if ((is.factor(x) || is.character(x)) && is.null(y)) {
      # enforce barplot type for ~ factor(y)
      type = type_barplot
    } else if (!is.null(x) && (is.factor(x) || is.character(x)) && !(is.factor(y) || is.character(y))) {
      # enforce boxplot type for y ~ factor(x)
      type = type_boxplot
    } else if (!is.null(x) && (is.factor(y) || is.character(y))) {
      # enforce spineplot type for factor(y) ~ x
      type = type_spineplot
    } else {
      type = "p"
    }
  }

  if (is.character(type)) {
    type = switch(type,
      "abline"     = type_abline,
      "area"       = type_area,
      "bar"        = type_barplot,
      "barplot"    = type_barplot,
      "box"        = type_boxplot,
      "boxplot"    = type_boxplot,
      "chull"      = type_chull,
      "density"    = type_density,
      "ellipse"    = type_ellipse,
      "errorbar"   = type_errorbar,
      "function"   = type_function,
      "glm"        = type_glm,
      "hist"       = type_histogram,
      "histogram"  = type_histogram,
      "hline"      = type_hline,
      "j"          = type_jitter,
      "jitter"     = type_jitter,
      "lm"         = type_lm,
      "loess"      = type_loess,
      "p"          = type_points,
      "pointrange" = type_pointrange,
      "points"     = type_points,
      "polygon"    = type_polygon,
      "polypath"   = type_polypath,
      "qq"         = type_qq,
      "rect"       = type_rect,
      "ribbon"     = type_ribbon,
      "ridge"      = type_ridge,
      "rug"        = type_rug,
      "segments"   = type_segments,
      "spine"      = type_spineplot,
      "spineplot"  = type_spineplot,
      "spline"     = type_spline,
      "summary"    = type_summary,
      "text"       = type_text,
      "violin"     = type_violin,
      "vline"      = type_vline,
      type # default case (incl. line-family chars, handled below)
    )
  }

  if (is.function(type)) {
    args = intersect(names(formals(type)), names(dots))
    args = if (length(args) >= 1L) dots[args] else list()
    type = do.call(type, args)
    type$dots = dots[setdiff(names(dots), names(args))]
  } else if (is.character(type) && type %in% c("l", "lines", "o", "b", "c", "h", "s", "S")) {
    # Line-family convenience characters route through type_lines(), preserving
    # the requested plot character (so e.g. "b" stays points + lines, rather than
    # defaulting to "l").
    args = dots[intersect(names(formals(type_lines)), names(dots))]
    args[["type"]] = if (type == "lines") "l" else type
    type = do.call(type_lines, args)
  }

  if (inherits(type, "tinyplot_type")) {
    settings$type = type$name
    settings$type_draw = type$draw
    settings$type_data = type$data
  }
}

## auxiliary functions for formula/facet parsing

tinyformula = function(formula, facet = NULL) {
  ## input
  ## - formula:       y ~ x or y ~ x | z or ~ x or ~ x | z
  ## - facet:         ~ a or ~ a + b or b ~ a
  ##
  ## output:
  ## - x:      ~ x
  ## - y:      NULL or ~ y
  ## - by:     NULL or ~ z or ~ z1 + z2 + ... (use interaction of all)
  ## - xfacet: NULL or ~ a or ~ a + b etc.
  ## - yfacet: NULL or ~ b
  ## - full:   e.g. ~ x + y + z + a + b

  ## preliminaries
  if (!inherits(formula, "formula")) formula = as.formula(formula)
  nf = length(formula)

  ## basic formula types
  x     = ~ x
  y     = if (nf == 2L) NULL else ~ y
  by    = if (!inherits(formula[[nf]], "call") || formula[[nf]][[1L]] != as.name("|")) NULL else ~ z
  if (is.null(facet) || !inherits(facet, "formula")) {
    xfacet = NULL
    yfacet = NULL
  } else {
    xfacet = ~ a
    yfacet = if (length(facet) == 2L) NULL else ~ b
  }

  ## fill with actual terms
  environment(x) = environment(formula)
  if (!is.null(y)) {
    environment(y) = environment(formula)
    y[[2L]] = formula[[2L]]
  }
  if (is.null(by)) {
    x[[2L]] = formula[[nf]]
  } else {
    environment(by) = environment(formula)
    by[[2L]] = formula[[nf]][[3L]]
    x[[2L]] = formula[[nf]][[2L]]
  }
  if (!is.null(xfacet)) {
    environment(xfacet) = environment(formula)
    xfacet[[2L]] = facet[[length(facet)]]
  }
  if (!is.null(yfacet)) {
    environment(yfacet) = environment(formula)
    yfacet[[2L]] = facet[[2L]]
  }

  ## combine everything
  full = x
  if (!is.null(y))      full[[2L]] = call("+", full[[2L]], y[[2L]])
  if (!is.null(by))     full[[2L]] = call("+", full[[2L]], by[[2L]])
  if (!is.null(xfacet)) full[[2L]] = call("+", full[[2L]], xfacet[[2L]])
  if (!is.null(yfacet)) full[[2L]] = call("+", full[[2L]], yfacet[[2L]])

  ## return list of all formulas
  return(list(
    x = x,
    y = y,
    by = by,
    xfacet = xfacet,
    yfacet = yfacet,
    full = full
  ))
}

tinyframe = function(formula, data, drop = FALSE) {
  ## input
  ## - formula: (sub-)formula
  ## - data: model.frame from full formula
  if (is.null(formula)) return(NULL)
  vars = attr(terms(formula), "variables")[-1L]
  if (is.null(vars)) return(data[, 0L, drop = drop])
  names = sapply(vars, deparse, width.cutoff = 500L)
  data[, names, drop = drop]
}

sanitize_facet = function(settings) {
  env2env(
    settings,
    environment(),
    c("facet", "by", "null_facet", "facet_attr", "facet_by")
  )

  # flag if facet=="by" (i.e., facet matches the grouping variable)
  facet_by = FALSE
  if (!is.null(facet) && length(facet) == 1 && facet == "by") {
    by = as.factor(by) ## if by==facet, then both need to be factors
    facet = by
    facet_by = TRUE
  } else if (!is.null(facet) && inherits(facet, "formula")) {
    facet = get_facet_fml(facet, data = data)
    if (isTRUE(attr(facet, "facet_grid"))) {
      facet.args[["nrow"]] = attr(facet, "facet_nrow")
    }
  }
  facet_attr = attributes(facet) # TODO: better way to restore facet attributes?
  null_facet = is.null(facet)
  
  # update settings
  env2env(
    environment(),
    settings,
    c("facet", "null_facet", "facet_attr", "facet_by", "by")
  )
}

#' Set or Reset Plot Themes for `tinyplot`
#'
#' @md
#' @description
#' The `tinytheme` function sets or resets the theme for plots created with
#' `tinyplot`. Themes control the appearance of plots, such as text alignment,
#' font styles, axis labels, and even dynamic margin adjustment to reduce
#' whitespace.
#'
#' @param theme A character string specifying the name of the theme to apply.
#'   Themes are arranged in an approximate hierarchy, adding or subtracting
#'   elements in the order presented below. Note that several themes are
#'   _dynamic_, in the sense that they attempt to reduce whitespace in a way
#'   that is responsive to the length of axes labels, tick marks, etc. These
#'   dynamic plots are marked with an asterisk (*) below.
#'   
#'   - `"default"`: inherits the user's default base graphics settings.
#'   - `"basic"`: light modification of `"default"`, only adding filled points, a panel background grid, and light gray background to facet titles.
#'   - `"dynamic"` (*): builds on `"basic"` by enabling dynamic margin adjustment with tighter default margins, horizontal axis labels, and the subtitle moved above the plotting area. Turns off the panel grid. Provides the foundation for all other dynamic themes and is a good starting point for users who wish to build custom dynamic themes.
#'     - `"clean"` (*): builds on `"dynamic"` by re-enabling the panel background grid and setting different default palettes ("Tableau 10" for discrete colors and "agSunset" for gradient colors).
#'       - `"clean2"` (*): removes the plot frame (box) from `"clean"`.
#'     - `"classic"` (*): builds on `"dynamic"` with L-shaped axes (removing the top and right-hand edges of the plot frame), smaller axis text, tighter axis spacing, and the "Okabe-Ito" palette as a default for discrete colors. Inspired by the **ggplot2** theme of the same name.
#'     - `"bw"` (*): similar to `"clean"`, except uses thinner lines for the plot frame (box), solid fine grid lines, smaller axis text, tighter axis spacing, and sets the "Okabe-Ito" palette as a default for discrete colors. Inspired by the **ggplot2** theme of the same name.
#'       - `"linedraw"` (*): builds on `"bw"` with black facet title strips and white strip text. Inspired by the **ggplot2** theme of the same name.
#'       - `"minimal"` (*): removes the plot frame (box) from `"bw"`, as well as the background for facet titles. Inspired by the **ggplot2** theme of the same name.
#'         - `"ipsum"` (*): builds on `"minimal"` with bold titles, no ticks, fine grid, edge-aligned axis titles, and a custom muted palette. Inspired by the **hrbrthemes** theme of the same name for **ggplot2**.
#'         - `"ipsum2"` (*): a lighter variant of `"ipsum"` that retains the original italic subtitle and edge-aligned axis titles, but without the additional spacing and palette changes.
#'         - `"dark"` (*): similar to `"minimal"`, but set against a dark background with foreground and a palette colours lightened for appropriate contrast.
#'         - `"socviz"` (*): builds on `"minimal"` with L-shaped axes, very light grid lines, and larger axis text. Inspired by Kieran Healy's **socviz** package theme for **ggplot2**.
#'     - `"broadsheet"` (*): a publication/newspaper style with only horizontal grid lines, no frame, short x-axis ticks, and muted secondary text (subtitle, caption). Compact axis spacing.
#'       - `"nber"` (*): builds on `"broadsheet"` for an NBER working paper style with a light blue-grey background, grey text, italic axis titles and captions, and a blue-grey discrete palette.
#'     - `"web"` (*): a FiveThirtyEight-inspired style with a light grey device background, no frame or axis lines, and bold grid lines. Suited to web/online publication.
#'     - `"tufte"` (*): floating axes and minimalist plot artifacts in the style of Edward Tufte.
#'       - `"float"` (*): builds on `"tufte"` with outward ticks, fewer tick marks, and a "dark" qualitative palette.
#'     - `"void"` (*): switches off all axes, titles, legends, etc.
#'     - `"ridge"` (*): a specialized theme for ridge plots (see [`type_ridge()`]). Builds off of `"clean"`, but adds ridge-specific tweaks (e.g. default "Zissou 1" palette for discrete colors, solid horizontal grid lines, and minor adjustments to y-axis labels). Not recommended for non-ridge plots.
#'       - `"ridge2"` (*): removes the plot frame (box) from `"ridge"`, but retains the x-axis line. Again, not recommended for non-ridge plots.
#' @param ... Named arguments to override specific theme settings. These
#'   arguments are passed to `tpar()` and take precedence over the predefined
#'   settings in the selected theme.
#' @param register Optional character string. If provided, the theme (with any
#'   `...` overrides) is registered under this name via [`tinytheme_register()`]
#'   and simultaneously activated. This is a shortcut for calling
#'   [`tinytheme_register()`] and `tinytheme()` separately.
#'
#' @details
#' Sets a list of graphical parameters using `tpar()`
#'
#' **Persistent vs. ephemeral themes.** Calling `tinytheme("<theme>")` triggers
#' a persistent theme, which will be applied to all subsequent graphics (incl.
#' base `plot`). To reset the theme to your default settings, call `tinytheme()`
#' without arguments. Alternatively, invoke the `tinyplot(..., theme = <theme>)`
#' argument for an ephemeral theme that is automatically reset at the end of the
#' plot call.
#' 
#' ```
#' # Set a persistent theme
#' tinytheme("clean")
#' tinyplot(1:10)  # uses "clean"
#' tinyplot(10:1)  # still uses "clean"
#' tinytheme()     # reset to default
#' 
#' # Use an ephemeral theme for a single plot (incl. added components)
#' tinyplot(1:10, theme = "dark")  # uses "dark" ephemerally
#' tinyplot_add(type = "S")        # same plot, so retains theme
#' tinyplot(10:1)                  # new plot, so back to default
#' ```
#' 
#' **Custom overrides.** To customize a theme's parameters (e.g., `mar`, `las`,
#' etc.), either pass them directly as additional args to
#' `tinytheme(<theme>, ...)` or as a list arguments to
#' `tinyplot(..., theme = list(<theme>, ...))`.
#' Please note that passing overrides through `tpar()` or `par()` _won't_ work,
#' because themes work by installing a persistent hook, which means that an
#' active theme will override any subsequent calls for parameters that the theme
#' controls.
#' 
#' ```
#' # Do this
#' tinytheme("clean", mar = c(5, 5, 2, 2))
#' <some plot>
#' 
#' # Or this (ephemeral version)
#' tinyplot(..., theme = list("clean", mar = c(5, 5, 2, 2)))
#' 
#' # Don't do this (the theme hook will overwrite your changes)
#' tinytheme("clean")
#' tpar(mar = c(5, 5, 2, 2))
#' <some plot>
#' ```
#' 
#' **Spacing primitives.** Dynamic themes compute `mgp` (margin line positions)
#' automatically from two spacing primitives, rather than requiring users to
#' reason about how `mar`, `mgp`, and `tcl` combine:
#'
#' - `gap.axis`: the gap in margin lines between the tick tip and the near edge
#'   of the tick label. Default `0.2`.
#' - `gap.lab`: the gap in margin lines between the far edge of the tick label
#'   and the near edge of the axis title. Default `1.0`.
#'
#' These primitives scale automatically with `cex.axis` and `cex.lab`, so the
#' visible spacing between elements remains constant regardless of text size.
#' To adjust spacing, pass them as overrides:
#'
#' ```
#' # Tighter spacing between tick labels and axis titles
#' tinytheme("clean", gap.lab = 0.5)
#'
#' # More room between ticks and tick labels
#' tinytheme("clean", gap.axis = 0.5)
#' ```
#'
#' If you supply an explicit `mgp` value, it is used as-is and the primitives
#' are ignored.
#'
#' @return The function returns nothing. It is called for its side effects.
#' 
#' @seealso [`tpar`] which does the heavy lifting under the hood;
#'   [tinytheme_register()] for registering custom named themes.
#'
#' @examples
#' # Reusable plot function
#' p = function() tinyplot(
#'   lat ~ long | depth, data = quakes,
#'   main = "Earthquakes off Fiji",
#'   sub = "Data courtesy of the Harvard PRIM-H project"
#' )
#' p()
#' 
#' # Set a theme
#' tinytheme("dark")
#' p()
#' 
#' #  A set theme is persistent and will apply to subsequent plots
#' tinyplot(0:10)
#'
#' # Try a different theme
#' tinytheme("clean")
#' p()
#'          
#' # Customize the theme by overriding default settings
#' tinytheme("clean",
#'           adj.xlab = 1, adj.ylab = 1,
#'           cex.lab = 0.75, cex.axis = 0.9,
#'           font.sub = 3,
#'           gap.axis = 0, gap.lab = 0.5,
#'           tcl = -0.1)
#' p()
#' 
#' # Another custom theme example, including a different font
#' tinytheme("bw", font.main = 2, col.axis = "darkcyan", family = "HersheyScript")
#' p()
#' 
#' # Aside: One or two specialized themes are only meant for certain plot types
#' tinytheme("ridge2")
#' tinyplot(I(cut(lat, 10)) ~ depth, data = quakes, type = "ridge")
#'
#' # Reset the theme
#' tinytheme()
#' p()
#' 
#' # For an ephemeral theme, use `tinyplot(..., theme = <theme>)` directly
#' tinyplot(0:10, theme = "clean", main = "This theme is ephemeral")
#' tinyplot(10:0, main = "See, no more theme")
#' 
#' # Customize an ephemeral theme by passing arguments as a list
#' tinyplot(0:10, main = "Custom", theme = list("clean", grid.col = "hotpink"))
#' 
#' # Themes showcase
#' ## We'll use a slightly more intricate plot (long y-axis labs and facets)
#' ## to demonstrate dynamic margin adjustment etc.
#' 
#' thms = eval(formals(tinytheme)$theme)
#' 
#' for (thm in thms) {
#'   tinytheme(thm)
#'   tinyplot(
#'     I(Sepal.Length*1e4) ~ Petal.Length | Species, facet = "by", data = iris,
#'     yaxl = ",", 
#'     main = paste0('tinytheme("', thm, '")'),
#'     sub = "A subtitle",
#'     cap = "A caption"
#'   )
#'   box("outer", lty = 2)
#' }
#' 
#' # Reset
#' tinytheme()
#'
#' @export
tinytheme = function(
    theme = c(
      "default", "basic", "dynamic",
      "clean", "clean2", "bw", "linedraw", "classic",
      "minimal", "ipsum", "ipsum2", "dark",
      "socviz", "broadsheet", "nber", "web",
      "ridge", "ridge2",
      "tufte", "float", "void"
    ),
    ...,
    register = NULL
    ) {
  
  if (length(theme) > 1) theme = theme[1]

  registered = names(get_environment_variable(".registered_themes"))
  assert_choice(theme, c(builtin_themes, registered))

  # in notebooks, we don't want to close the device because no image.
  # init_tpar() tries to be smart, but may fail.
  init_tpar(rm_hook = TRUE)

  settings = switch(theme,
    "default" = theme_default,
    "basic" = theme_basic,
    "broadsheet" = theme_broadsheet,
    "bw" = theme_bw,
    "linedraw" = theme_linedraw,
    "classic" = theme_classic,
    "clean" = theme_clean,
    "clean2" = theme_clean2,
    "dark" = theme_dark,
    "dynamic" = theme_dynamic,
    "ipsum" = theme_ipsum,
    "ipsum2" = theme_ipsum2,
    "minimal" = theme_minimal,
    "nber" = theme_nber,
    "ridge" = theme_ridge,
    "ridge2" = theme_ridge2,
    "socviz" = theme_socviz,
    "tufte" = theme_tufte,
    "float" = theme_float,
    "void" = theme_void,
    "web" = theme_web,
    get_environment_variable(".registered_themes")[[theme]]
  )

  dots = list(...)
  for (n in names(dots)) {
    settings[[n]] = dots[[n]]
  }

  # `col.default` < 0 gotcha for user-supplied palettes.(#627)
  if ("palette.qualitative" %in% names(dots) &&
      !("col.default" %in% names(dots)) &&
      is.numeric(settings[["col.default"]]) &&
      settings[["col.default"]] < 0L) {
    settings[["col.default"]] = abs(settings[["col.default"]])
  }

  # Compute mgp from spacing primitives when dynmar is active and the user
  # didn't provide an explicit mgp override. The near edge of margin text
  # (facing the plot region) aligns with the half-cell boundary (0.5*cex from
  # center). Using 0.5*cex in mgp keeps the visible gap between adjacent text
  # elements constant regardless of cex scaling.
  if (isTRUE(settings[["dynmar"]]) && !("mgp" %in% names(dots))) {
    .ga = settings[["gap.axis"]] %||% 0.2
    .gl = settings[["gap.lab"]] %||% 1.0
    .ca = settings[["cex.axis"]] %||% 1
    # FIXME: mgp is shared across sides, so we use the larger label cex to
    # avoid clipping on either axis. Ideally we'd set side-specific mgp when
    # cex.xlab and cex.ylab differ.
    .cl = max(
      settings[["cex.lab"]] %||% 1,
      settings[["cex.xlab"]] %||% 0,
      settings[["cex.ylab"]] %||% 0
    )
    .mgp2 = .ga + 0.5 * .ca
    .mgp1 = .mgp2 + .gl + 0.5 * .cl
    settings[["mgp"]] = c(.mgp1, .mgp2, 0)
  }

  if (!is.null(register)) {
    tinytheme_register(register, theme = theme, ...)
    settings[["tinytheme"]] = register
  }

  if (length(settings) > 0) {
    if (theme == "default") {
      # for default theme, we want to revert the original pars and turn off the
      # before.new.plot hook (otherwise manual par(x = y) changes won't work)
      tpar(settings, hook = FALSE)
      old_hooks = get_environment_variable(".tpar_hooks")
      remove_hooks(old_hooks)
    } else {
      tpar(settings, hook = TRUE)
    }
  }

  return(invisible(NULL))
}



#
## Themes (these are read and set at initial load time)

builtin_themes = c(
  "default", "basic", "dynamic",
  "clean", "clean2", "bw", "linedraw", "classic",
  "minimal", "ipsum", "ipsum2", "dark",
  "socviz", "broadsheet", "nber", "web",
  "ridge", "ridge2",
  "tufte", "float", "void"
)

# theme_default = list()

theme_default = list(
  tinytheme = "default",
  adj = par("adj"), # 0.5,
  adj.main = par("adj"), # 0.5,
  adj.cap = par("adj"), # 0.5,
  adj.sub = par("adj"), # 0.5,
  bg = "white", # par("bg") # "white"
  bty = par("bty"), #"o",
  cex = par("cex"), #1,
  cex.axis = par("cex.axis"), #1,
  cex.lab = par("cex.lab"), #1,
  cex.main = par("cex.main"), #1.2,
  cex.cap = 1,
  cex.sub = par("cex.sub"), #1,
  cex.xlab = NULL, # defer to par("cex.lab") unless set explicitly
  cex.ylab = NULL, # defer to par("cex.lab") unless set explicitly
  col = par("col"), #"black",
  col.axis = par("col.axis"), #1,
  col.xaxs = par("col.axis"), #1,
  col.yaxs = par("col.axis"), #1,
  col.lab = par("col.lab"), #"black",
  col.main = par("col.main"), #"black",
  col.cap = "black",
  col.sub = par("col.sub"), #"black",
  dynmar = FALSE,
  facet.bg = NULL,
  facet.border = NA,
  family = par("family"), # ""
  fg = par("fg"),
  font = par("font"), # 1,
  font.axis = par("font.axis"), # 1,
  font.lab = par("font.lab"), # 1,
  font.main = par("font.main"), # 2,
  font.cap = 1,
  font.sub = par("font.sub"), # 2,
  grid = FALSE,
  grid.col = "lightgray",
  grid.lty = "dotted",
  grid.lwd = 1,
  ljust = "left",
  lab = par("lab"), # c(5, 5, 7),
  las = par("las"), # 0,
  lwd = par("lwd"), # 1,
  lwd.axis = par("lwd"), # 1,
  mar = c(5.1, 4.1, 4.1, 2.1), ## test
  mgp = par("mgp"),
  # palette.qualitative = "R4",
  # palette.sequential = "Viridis",
  pch = par("pch"), # 1,
  ribbon.alpha = 0.2,
  side.sub = 1,
  tck = NA,
  tcl = par("tcl"), # -0.5
  xaxt = "standard",
  yaxt = "standard"
)

# derivatives of "default" 
# - basic
# - tufte
# - void

theme_basic = modifyList(theme_default, list(
  tinytheme = "basic",
  adj.cap = 1,
  facet.bg = "gray90",
  facet.border = "black",
  grid = TRUE,
  pch = 16
))

# derivatives of "basic"
# - dynamic

theme_dynamic = modifyList(theme_basic, list(
  ## Notes:
  ##  - Dynamic themes start from an (almost) zero `mar` baseline. The
  ##    `draw_facet_window()` helper builds each side's margin up from this
  ##    pad, adding only what the plot actually needs (tick row, axis label,
  ##    main, sub). See `dynmar_side()` in utils.R.
  ##  - `mgp` is computed in `tinytheme()` from the spacing primitives
  ##    (gap.axis, gap.lab) and the active cex.axis/cex.lab values. If the
  ##    user provides an explicit `mgp`, it is used as-is and the primitives
  ##    are ignored.
  ##  - `side.sub = 3` moves the sub-caption above the plot (below main).
  ##  - `tcl = -0.3` tightens axis tick marks relative to the base default.
  ##
  tinytheme = "dynamic",
  adj.main = 0,
  adj.sub = 0,
  dynmar = TRUE,
  gap.axis = 0.2,  # gap (lines) between tick tip and tick label cell edge
  gap.lab = 1.0,   # gap (lines) from tick label cell edge to title cell edge
  gap.main = 0.7,  # gap (lines) from main baseline to element below (plot or sub top)
  gap.sub = 0.7,   # gap (lines) from sub baseline to plot box
  grid = FALSE,
  las = 1,
  mar = c(0.1, 0.1, 0.6, 0.6),
  mgp = NULL,      # computed from gap.axis/gap.lab in tinytheme()
  side.sub = 3,
  tcl = -0.3
))

# derivatives of "dynamic"
# - clean
# - classic

theme_clean = modifyList(theme_dynamic, list(
  tinytheme = "clean",
  cex.cap = 0.8,
  grid = TRUE,
  palette.qualitative = "Tableau 10",
  palette.sequential = "ag_Sunset"
))

theme_classic = modifyList(theme_dynamic, list(
  tinytheme = "classic",
  bty = "l",
  cex.axis = 0.8,
  cex.cap = 0.8,
  col.default = -1L,  # black single-group; drop it from the grouped palette
  facet.bg = NULL,
  font.main = 1,
  gap.axis = 0.1,
  gap.lab = 0.4,
  palette.qualitative = "ggplot2"
))

# derivatives of "clean"
# - clean2
# - bw

theme_clean2 = modifyList(theme_clean, list(
  tinytheme = "clean2",
  facet.border = "gray90",
  xaxt = "labels",
  yaxt = "labels"
))

theme_bw = modifyList(theme_clean, list(
  tinytheme = "bw",
  cex.axis = 0.8,
  col.default = -1L,  # black single-group; drop it from the grouped palette
  font.main = 1,
  gap.axis = 0.1,
  gap.lab = 0.4,
  grid = "xy",
  grid.lty = 1,
  grid.lwd = 0.5,
  lwd = 0.5,
  lwd.axis = 0.5,
  palette.qualitative = "ggplot2"
))

theme_linedraw = modifyList(theme_bw, list(
  tinytheme = "linedraw",
  facet.bg = "black",
  facet.col = "white",
  grid.col = "black",
  grid.lwd = 0.1
))

# derivatives of "bw"
# - minimal
# - ipsum
# - dark
 
theme_minimal = modifyList(theme_bw, list(
  tinytheme = "minimal",
  bty = "n",
  facet.bg = NULL,
  facet.border = NULL,
  xaxt = "labels",
  yaxt = "labels"
))

theme_ipsum = modifyList(theme_minimal, list(
  tinytheme = "ipsum",
  adj.xlab = 1,
  adj.ylab = 1,
  cex.lab = 0.8,
  col.default = "black",  # bespoke palette doesn't lead with black; pin it
  font.main = 2,
  font.sub = 1,
  font.cap = 3,
  gap.axis = 0,
  gap.lab = 0.7,
  grid = "xy",
  palette.qualitative = c("#D18975", "#8FD175", "#3F2D54", "#75B8D1",
                          "#2D543D", "#C9D175", "#D1AB75", "#D175B8", "#758BD1"),
  tcl = 0
))

theme_ipsum2 = modifyList(theme_minimal, list(
  tinytheme = "ipsum2",
  bty = "n",
  font.sub = 3,
  adj.ylab = 1,
  adj.xlab = 1,
  col.default = -1L,  # black single-group; drop it from the grouped palette
  palette.qualitative = "Okabe-Ito"  # keep Okabe-Ito, not the ggplot2 palette
))

theme_dark = modifyList(theme_minimal, list(
  tinytheme = "dark",
  bg = "#1A1A1A",
  col.default = NULL,  # no fixed default: single-group uses palette[1] (Set 2)
  fg = "#BBBBBB",
  # col = "white",
  col.xaxs = "#BBBBBB",
  col.yaxs = "#BBBBBB",
  col.lab = "#BBBBBB",
  col.main = "#BBBBBB",
  col.cap = "#BBBBBB",
  col.sub = "#BBBBBB",
  col.axis = "#BBBBBB",
  # facet.bg = "gray20",
  grid.col = "#6D6D6D",
  palette.qualitative = "Set 2",
  palette.sequential = "Sunset",
  ribbon.alpha = 0.5
))

# derivatives of clean/clean2

theme_ridge = modifyList(theme_clean, list(
  tinytheme = "ridge",
  col.default = "black",  # keep black ridgelines; Zissou is for gradient fills
  palette.qualitative = "Zissou 1",
  grid = FALSE
))
theme_ridge2 = modifyList(theme_clean2, list(
  tinytheme = "ridge2",
  col.default = "black",  # keep black ridgelines; Zissou is for gradient fills
  palette.qualitative = "Zissou 1",
  grid = FALSE
))

# derivatives of "minimal"
# - socviz

theme_socviz = modifyList(theme_minimal, list(
  tinytheme = "socviz",
  bty = "l",
  cex.axis = 1.1,
  cex.cap = 0.75,
  cex.lab = 1,
  cex.main = 1.4,
  cex.sub = 1.05,
  col.default = "black",  # bespoke palette doesn't lead with black; pin it
  col.xaxs = "gray10",
  col.yaxs = "gray10",
  facet.bg = NULL,
  facet.col = "grey10",
  font.main = 2,
  grid.col = "gray90",
  grid.lty = 1,
  grid.lwd = 0.3,
  lwd = 1,
  lwd.axis = 1,
  palette.qualitative = c("#E69F00", "#56B4E9", "#009E73", "#D55E00",
                          "#CC79A7", "#0072B2", "#F0E442", "#000000"),
  tcl = -0.25,
  xaxt = "standard",
  yaxt = "standard"
))

# derivatives of "dynamic"
# - broadsheet
# - web

theme_broadsheet = modifyList(theme_dynamic, list(
  tinytheme = "broadsheet",
  bty = "n",
  cex.cap = 0.8,
  col.cap = "gray40",
  col.default = -1L,  # black single-group; drop it from the grouped palette
  col.sub = "gray40",
  font.main = 2,
  gap.axis = 0.1,
  gap.lab = 0.5,
  grid = "Y",
  grid.col = "gray85",
  grid.lty = 1,
  grid.lwd = 0.5,
  palette.qualitative = "Okabe-Ito",
  tcl = -0.2,
  yaxt = "labels"
))

# derivatives of "broadsheet"
# - nber

theme_nber = modifyList(theme_broadsheet, list(
  tinytheme = "nber",
  bg = "#F2F7FB",
  col.default = NULL,  # override broadsheet's black: nber palette leads with blue
  cex.cap = 1,
  cex.main = 1.4,
  cex.sub = 1,
  col.axis = "#4C4D4F",
  col.lab = "#6D6E72",
  col.main = "#4C4D4F",
  col.sub = "#6D6E72",
  col.cap = "#6D6E72",
  col.xaxs = "#6D6E72",
  facet.bg = "white",
  facet.col = "#6D6E72",
  font.cap = 3,
  font.lab = 3,
  font.sub = 3,
  grid.col = "#BCBFC3",
  palette.qualitative = c(
    "#0063A7", "#6D6E70", "#941A22", 
    "#EDAF48", "#2E8B57", "#7B5EA7")
))

theme_web = modifyList(theme_dynamic, list(
  tinytheme = "web",
  bg = "#F0F0F0",
  bty = "n",
  cex.cap = 0.8,
  col.cap = "gray40",
  col.sub = "gray40",
  font.main = 2,
  grid = TRUE,
  grid.col = "#D2D2D2",
  grid.lty = 1,
  grid.lwd = 0.5,
  palette.qualitative = c(
    "#008FD5", "#FF2700", "#77AB43", "#5D2E8C",
    "#FF8C00", "#00B2A9", "#A67C52", "#C71585"),
  tcl = 0,
  xaxt = "labels",
  yaxt = "labels"
))

# standalone dynamic derivatives
# - tufte
# - void

theme_tufte = modifyList(theme_dynamic, list(
  tinytheme = "tufte",
  bty = "n",
  facet.bg = NULL,
  facet.border = NA,
  font.main = 1,
  gap.axis = -0.2,
  gap.lab = 0.5,
  grid = FALSE,
  lab = c(10, 10, 7),
  tcl = 0.2
))

theme_float = modifyList(theme_tufte, list(
  tinytheme = "float",
  col.default = "black",
  facet.bg = "gray95",
  gap.axis = 0,
  gap.lab = 0.7,
  lab = c(5, 5, 7),
  palette.qualitative = "dark",
  tcl = -0.2
))

theme_void = modifyList(theme_dynamic, list(
  tinytheme = "void",
  col.default = "black",
  facet.bg = NULL,
  facet.border = NA,
  font.main = 1,
  grid = FALSE,
  palette.qualitative = "Tableau 10",
  palette.sequential = "ag_Sunset",
  xaxt = "none",
  yaxt = "none"
))


#
## Theme registry helpers
#

# Internal: unified theme lookup (registered first, then built-in)
get_theme_def = function(name) {
  if (is.null(name) || name == "default") return(theme_default)
  registry = get_environment_variable(".registered_themes")
  if (!is.null(registry[[name]])) return(registry[[name]])
  obj_name = paste0("theme_", name)
  if (exists(obj_name, envir = asNamespace("tinyplot"), inherits = FALSE)) {
    return(get(obj_name, envir = asNamespace("tinyplot")))
  }
  NULL
}


#' Register, List, and Unregister Custom Themes
#'
#' @md
#' @description
#' `tinytheme_register()` registers a custom theme so it can be used by name
#' with `tinytheme()` or `tinyplot(..., theme = )`. Custom themes inherit from
#' a base theme and apply user-specified overrides. Registered themes are
#' session-scoped: they persist across plots but not across R sessions. To make
#' a custom theme permanently available, register it in your `.Rprofile`.
#'
#' `tinytheme_list()` returns the names of all available themes (built-in and
#' registered).
#'
#' `tinytheme_unregister()` removes a previously registered theme from the
#' registry. Does not reset an active theme.
#'
#' @param name Character string. The name for your custom theme. Cannot clash
#'   with or overwrite a built-in theme name (`"default"`, `"clean"`, etc.)
#' @param theme Character string or list. The base theme to inherit from. If a
#'   string, it must reference a built-in or previously-registered theme. If a
#'   list, it is used directly as the base definition. Default is `"default"`.
#' @param ... Named arguments to override specific theme settings. These are
#'   the same parameters accepted by `tpar()`.
#'
#' @return `tinytheme_register()` returns the theme definition list (invisibly).
#'   `tinytheme_list()` returns a named list with character vectors `builtin`
#'   and `registered`. `tinytheme_unregister()` returns `NULL` (invisibly).
#'
#' @seealso [tinytheme()]
#'
#' @examples
#' # Register a custom theme based on "float" but with a grid
#' tinytheme_register("float2", theme = "float", grid = TRUE)
#'
#' # Use it
#' tinyplot(1:5, theme = "float2")
#'
#' # List all themes
#' tinytheme_list()
#'
#' # Unregister
#' tinytheme_unregister("float2")
#'
#' @export
tinytheme_register = function(name, theme = "default", ...) {
  if (!is.character(name) || length(name) != 1 || nchar(name) == 0) {
    stop("`name` must be a single non-empty character string.", call. = FALSE)
  }
  builtins = builtin_themes
  if (name %in% builtins) {
    stop(
      sprintf("'%s' is a built-in theme and cannot be overridden.", name),
      call. = FALSE
    )
  }

  if (is.character(theme) && length(theme) == 1) {
    base_theme = get_theme_def(theme)
    if (is.null(base_theme)) {
      stop(sprintf("Base theme '%s' not found.", theme), call. = FALSE)
    }
  } else if (is.list(theme)) {
    base_theme = theme
  } else {
    stop("`theme` must be a theme name (string) or a list.", call. = FALSE)
  }

  dots = list(...)
  new_theme = if (length(dots) > 0) modifyList(base_theme, dots) else base_theme
  new_theme[["tinytheme"]] = name

  registry = get_environment_variable(".registered_themes") %||% list()
  registry[[name]] = new_theme
  set_environment_variable(.registered_themes = registry)
  invisible(new_theme)
}


#' @rdname tinytheme_register
#' @export
tinytheme_list = function() {
  builtins = builtin_themes
  registered = names(get_environment_variable(".registered_themes"))
  list(builtin = builtins, registered = registered)
}


#' @rdname tinytheme_register
#' @export
tinytheme_unregister = function(name) {
  registry = get_environment_variable(".registered_themes")
  if (!name %in% names(registry)) {
    warning(
      sprintf("Theme '%s' is not registered. Nothing to remove.", name),
      call. = FALSE
    )
    return(invisible(NULL))
  }
  registry[[name]] = NULL
  set_environment_variable(.registered_themes = registry)
  invisible(NULL)
}

draw_title = function(main, sub, cap, xlab, ylab, legend, legend_args, opar,
                      xlab_line_offset = 0,
                      ylab_line_offset = 0) {
  # main title
  # Note that we include a special catch for the main title if legend is
  # "top!" (and main is specified in the first place).
  legend_eval = tryCatch(eval(legend), error = function(e) NULL)
  # Extra bit of footwork if user passed legend = legend(...) instead of
  # legend = list(...), since the call environment is tricky
  if (is.null(legend_eval)) {
    legend_eval = tryCatch(paste0(legend)[[2]], error = function(e) NULL)
  }

  adj_title = !is.null(legend) && ((is.character(legend) && legend == "top!") || (!is.null(legend_args[["x"]]) && legend_args[["x"]] == "top!") || (is.list(legend_eval) && legend_eval[[1]] == "top!"))

  # For the "top!" legend case, bump main title up to make space for the
  # legend beneath it: Take the normal main title line gap (i.e., 1.7 lines)
  # and add the difference between original top margin and new one (i.e.,
  # which should equal the height of the new legend). Note that we also
  # include a 0.1 epsilon bump, which we're using to reset the tinyplot
  # window in case of recursive "top!" calls. (See draw_legend code.)

  if (isTRUE(adj_title) && isTRUE(get_tpar("dynmar", FALSE))) {
    gap_main = get_tpar("gap.main", 0.7)
    top_soma = get_environment_variable(".top_legend_soma") %||% 0
    line_main = top_soma + gap_main - 0.1
  } else if (isTRUE(adj_title)) {
    line_main = par("mar")[3] - opar[["mar"]][3] + 1.7 + 0.1
  } else if (isTRUE(get_tpar("dynmar", FALSE))) {
    gap_main = get_tpar("gap.main", 0.7)
    line_main = get_tpar("mgp")[3] + gap_main - 0.1
  } else {
    line_main = NULL
  }

  # When sub sits on top (side.sub == 3), push main up by the sub block
  # height so main is above sub rather than overlapping it.
  sub_lines = text_line_count(sub)
  if (sub_lines > 0L && isTRUE(get_tpar("side.sub", 1) == 3)) {
    if (is.null(line_main)) {
      gap_main = get_tpar("gap.main", 0.7)
      line_main = get_tpar("mgp")[3] + gap_main - 0.1
    }
    cex_sub = get_tpar("cex.sub", 1.2)
    gap_sub = get_tpar("gap.sub", 0.7)
    line_main = line_main + gap_sub + (sub_lines - 1 + 0.6) * cex_sub
  }

  if (!is.null(sub)) {
    if (isTRUE(get_tpar("side.sub", 1) == 3)) {
      gap_sub = get_tpar("gap.sub", 0.7)
      line_sub = get_tpar("line.sub", gap_sub - 0.1)
      if (isTRUE(adj_title) && isTRUE(get_tpar("dynmar", FALSE))) {
        top_soma = get_environment_variable(".top_legend_soma") %||% 0
        line_sub = top_soma + gap_sub - 0.1
      }
    } else {
      line_sub = get_tpar("line.sub", 4)
    }
    args = list(
      text = sub,
      line = line_sub,
      cex = get_tpar("cex.sub", 1.2),
      col = get_tpar("col.sub", "black"),
      adj = get_tpar(c("adj.sub", "adj")),
      font = get_tpar("font.sub", 1),
      side = get_tpar("side.sub", 1),
      las = 1
    )
    args = Filter(function(x) !is.null(x), args)
    do.call(mtext, args, quote = TRUE)
  }

  if (!is.null(main)) {
    # title() stacks multi-line main *above* `line_main` (line N at
    # `line_main`, extra lines extend upward). The reserved top margin
    # already accounts for (N-1)*cex_main extra lines, so no line-shift
    # adjustment is needed here.
    # For "top!" + dynmar, line_main exceeds par("mar")[3] so we need xpd=NA
    # to allow drawing outside the clipping region.
    .oxpd = par("xpd")
    if (!is.null(line_main) && line_main > par("mar")[3] - 1) par(xpd = NA)
    args = list(
      main = main,
      line = line_main,
      cex.main = get_tpar("cex.main", 1.4),
      col.main = get_tpar("col.main", "black"),
      font.main = get_tpar("font.main", 2),
      adj = get_tpar(c("adj.main", "adj"), 3))
    args = Filter(function(x) !is.null(x), args)
    do.call(title, args, quote = TRUE)
    par(xpd = .oxpd)
  }


  cap_in_legend = !is.null(legend_args[["x"]]) && grepl("bottom!$", legend_args[["x"]])
  if (!is.null(cap) && !cap_in_legend) {
    cex_cap = get_tpar("cex.cap", 1)
    line_cap = get_tpar("line.cap", NULL)
    if (is.null(line_cap)) {
      line_cap = par("mar")[1] - 1
    }
    args = list(
      text = cap,
      line = line_cap,
      cex = cex_cap,
      col = get_tpar("col.cap", "black"),
      adj = get_tpar(c("adj.cap", "adj")),
      font = get_tpar("font.cap", 1),
      side = 1,
      las = 1
    )
    args = Filter(function(x) !is.null(x), args)
    do.call(mtext, args, quote = TRUE)
  }

  # Axis titles. For multi-line labels, base R places line 1 at
  # `line = mgp[1] - (N-1)*cex`, which pushes line 1 up into the tick-label
  # zone. Shift `line` down so line 1 aligns with where a single-line xlab
  # would be (and the extra lines extend below).
  # Also push down by xlab_line_offset (= .whtsbp[1]) when dynmar has
  # reserved extra space for rotated (las=2/3) x-tick labels.
  args = list(xlab = xlab)
  xlab_lines = text_line_count(xlab)
  cex_xlab = get_tpar(c("cex.xlab", "cex.lab"), 1)
  # Compute expression height excess (same logic as dynmar_side)
  xlab_expr_excess = 0
  if (is.language(xlab)) {
    expr_lines = strheight(xlab, units = "inches", cex = cex_xlab) / par("csi")
    text_lines = strheight("X", units = "inches", cex = cex_xlab) / par("csi")
    xlab_expr_excess = max(0, expr_lines - text_lines)
  }
  if (xlab_lines > 1L || xlab_line_offset != 0 || xlab_expr_excess > 0) {
    args[["line"]] = get_tpar("mgp")[1] +
                     (xlab_lines - 1) * cex_xlab +
                     xlab_line_offset +
                     xlab_expr_excess
  }
  args[["adj"]] = get_tpar(c("adj.xlab", "adj"))
  args[["cex.lab"]] = cex_xlab
  do.call(title, args, quote = TRUE)

  # ylab: base R already places multi-line text correctly (outermost line at
  # mgp[1], subsequent lines closer to the plot), so no line shift needed.
  args = list(ylab = ylab)
  cex_ylab = get_tpar(c("cex.ylab", "cex.lab"), 1)
  if (ylab_line_offset != 0) {
    args[["line"]] = get_tpar("mgp")[1] + ylab_line_offset
  }
  args[["adj"]] = get_tpar(c("adj.ylab", "adj"))
  args[["cex.lab"]] = cex_ylab
  do.call(title, args, quote = TRUE)
}

#' Ridge plot type
#'
#' @description Type function for producing ridge plots (also known as joy plots),
#' which display density distributions for multiple groups with vertical offsets.
#' This function uses `tinyplot` scaffolding, which enables added functionality
#' such as grouping and faceting.
#'
#' The line color is controlled by the `col` argument in the `tinyplot()` call.
#' The fill color is controlled by the `bg` argument in the `tinyplot()` call.
#'
#' @param scale Numeric. Controls the scaling factor of each plot.
#' Values greater than 1 means that plots overlap.
#' @param joint.max character indicating how to scale the maximum of the densities:
#'   The default `"all"` indicates that all densities are scaled jointly relative to
#'   the same maximum so that the areas of all densities are comparable.
#'   Alternatively, `"facet"` indicates that the maximum is computed within
#'   each facet so that the areas of the densities are comparable within each
#'   facet but not necessarily across facets. Finally, `"by"` indicates that
#'   each row (in each facet) is scaled separately, so that the areas of the
#'   densities for `by` groups in the same row are comparable but not necessarily
#'   across rows.
#' @param breaks Numeric. If a color gradient is used for shading, the
#' breaks between the colors can be modified. The default is to use
#' equidistant breaks spanning the range of the `x` variable.
#' @param probs Numeric. Instead of specifying the same `breaks` on the
#' x-axis for all groups, it is possible to specify group-specific quantiles
#' at the specified `probs`. The quantiles are computed based on the density
#' (rather than the raw original variable). Only one of `breaks` or
#' `probs` must be specified.
#' @param ylevels a character or numeric vector specifying in which order
#' the levels of the y-variable should be plotted.
#' @inheritParams stats::density
#' @param bw the smoothing \code{\link[stats:bw.nrd]{bandwidth}} to be used,
#'   see \code{\link[stats]{density}} for details and options.
#' @param kernel a character string giving the smoothing kernel to be used. This
#'   must partially match one of `"gaussian"`, `"rectangular"`, `"triangular"`,
#'   `"epanechnikov"`, `"biweight"`, `"cosine"` or `"optcosine"`, with default
#'   `"gaussian"`, and may be abbreviated to a unique prefix (single letter).
#'
#'   `"cosine"` is smoother than `"optcosine"`, which is the usual 'cosine'
#'   kernel in the literature and almost MSE-efficient. However, `"cosine"` is
#'   the version used by S.
#' @param joint.bw character string indicating whether (and how) the smoothing
#'   bandwidth should be computed from the joint data distribution. The default
#'   of `"mean"` will compute the joint bandwidth as the mean of the individual
#'   subgroup bandwidths (weighted by their number of observations). Choosing
#'   `"full"` will result in a joint bandwidth computed from the full
#'   distribution (merging all subgroups). For `"none"` the individual bandwidth
#'   will be computed independently for each subgroup. Also accepts a logical
#'   argument, where `TRUE` maps to `"mean"` and `FALSE` maps to `"none"`. See
#'   \code{\link{type_density}} for some discussion of practical considerations.
#' @param gradient Logical or character. Should a gradient fill be used to
#'   shade the area under the density? If a character specification is used,
#'   then it can either be of length 1 and specify the palette to be used with
#'   `gradient = TRUE` corresponding to `gradient = "viridis"`. If a character
#'   vector of length greater than 1 is used, then it should specify the
#'   colors in the palette, e.g., `gradient = hcl.colors(512)`.
#' @param raster Logical. Should the `gradient` fill be drawn using
#'   \code{\link[graphics]{rasterImage}}? Defaults to `FALSE`, in which case the
#'   `gradient` fill will instead be drawn using
#'   \code{\link[graphics]{polygon}}. See the `Technical note on gradient fills`
#'   section below.
#' @param col Character string denoting the outline (border) color for all
#' of the ridge densities. Note that a singular value is expected; if multiple
#' colors are provided then only the first will be used. This argument is mostly
#' useful for the aesthetic effect of drawing a common outline color in
#' combination with gradient fills. See Examples. If left unset here, the
#' outline color can also be supplied via the top-level `tinyplot(..., col =)`
#' call (the `type_ridge(col =)` argument takes precedence if both are given).
#' When neither is supplied, the outline defaults to the active theme's
#' `col.default` (see [tpar]), falling back to the first color of the theme's
#' qualitative palette, or black if no theme is set.
#' @param alpha Numeric in the range `[0,1]` for adjusting the alpha
#' transparency of the density fills. In most cases, will default to a value of
#' 1, i.e. fully opaque. But for some `by` grouped plots (excepting the special
#' cases where `by==y` or `by==x`), will default to 0.6.
#'
#' @section Technical note on gradient fills:
#'
#' `tinyplot` uses two basic approaches for drawing gradient fills in ridge line
#' plots, e.g., if `type_ridge(gradient = TRUE)`.
#'
#' The first (and default) polygon-based approach involves dividing up the main
#' density region into many smaller polygons along the x-axis. Each of these
#' smaller polygons inherits a different color "segment" from the underlying
#' palette swatch, which in turn creates the effect of a continuous gradient
#' when they are all plotted together. Internally, this polygon-based approach
#' is vectorized (i.e., all of the sub-polygons are plotted simultaneously). It
#' is thus efficient from a plotting perspective and generally also performs
#' well from an aesthetic perspective. However, it can occasionally produce
#' undesirable plotting artifacts on some graphics devices---e.g., thin but
#' visible vertical lines---if alpha transparency is being used at the same
#' time.
#'
#' For this reason, we also offer an alternative raster-based approach for
#' gradient fills that users can invoke via
#' `type_ridge(gradient = TRUE, raster = TRUE)`. The essential idea is that we
#' coerce the density polygon into a raster representation (using
#' \code{\link[graphics]{rasterImage}}) and achieve the gradient effect via
#' color interpolation. The trade-off this time is potential smoothness
#' artifacts around the top of the ridge densities at high resolutions, since we
#' have converted a vector object into a raster object.
#'
#' Again, we expect that the choice between these two approaches will only
#' matter for ridge plots that combine gradient fills with alpha transparency
#' (and on certain graphics devices). We recommend that users experiment to
#' determine which approach is optimal for their device.
#'
#' @examples
#' aq = transform(
#'   airquality,
#'   Month = factor(month.abb[Month], levels = month.abb[5:9]),
#'   Month2 = factor(month.name[Month], levels = month.name[5:9]),
#'   Late = ifelse(Day > 15, "Late", "Early")
#'   )
#'
#' # default ridge plot (using the "ridge" convenience string)
#' tinyplot(Month ~ Temp, data = aq, type = "ridge")
#'
#' # for ridge plots, we recommend pairing with the dedicated theme(s), which
#' # facilitate nicer y-axis labels, grid lines, etc.
#'
#' tinytheme("ridge")
#' tinyplot(Month ~ Temp, data = aq, type = "ridge")
#'
#' tinytheme("ridge2") # removes the plot frame (but keeps x-axis line)
#' tinyplot(Month ~ Temp, data = aq, type = "ridge")
#'
#' # the "ridge(2)" themes are especially helpful for long y labels, due to
#' # dynamic plot adjustment
#' tinyplot(Month2 ~ Temp, data = aq, type = "ridge")
#'
#' # pass customization arguments through type_ridge()... for example, use
#' # the scale argument to change/avoid overlap of densities (more on scaling
#' # further below)
#'
#' tinyplot(Month ~ Temp, data = aq, type = type_ridge(scale = 1))
#'
#' ## by grouping is also supported. two special cases of interest:
#'
#' # 1) by == y (color by y groups)
#' tinyplot(Month ~ Temp | Month, data = aq, type = "ridge")
#'
#' # 2) by == x (gradient coloring along x)
#' tinyplot(Month ~ Temp | Temp, data = aq, type = "ridge")
#'
#' # aside: pass explicit `type_ridge(col = <col>)` arg to set a different
#' # border color
#' tinyplot(Month ~ Temp | Temp, data = aq, type = type_ridge(col = "white"))
#'
#' # gradient coloring along the x-axis can also be invoked manually without
#' # a legend (the next two tinyplot calls are equivalent)
#'
#' # tinyplot(Month ~ Temp, data = aq, type = type_ridge(gradient = "agsunset"))
#' tinyplot(Month ~ Temp, data = aq, type = type_ridge(gradient = TRUE))
#'
#' # aside: when combining gradient fill with alpha transparency, it may be
#' # better to use the raster-based approach (test on your graphics device)
#'
#' tinyplot(Month ~ Temp, data = aq,
#'   type = type_ridge(gradient = TRUE, alpha = 0.5),
#'   main = "polygon fill (default)")
#' tinyplot(Month ~ Temp, data = aq,
#'   type = type_ridge(gradient = TRUE, alpha = 0.5, raster = TRUE),
#'   main = "raster fill")
#'
#' # highlighting only the center 50% of the density (i.e., 25%-75% quantiles)
#' tinyplot(Month ~ Temp, data = aq, type = type_ridge(
#'   gradient = hcl.colors(3, "Dark Mint")[c(2, 1, 2)],
#'   probs = c(0.25, 0.75), col = "white"))
#'
#' # highlighting the probability distribution by color gradient
#' # (darkest point = median)
#' tinyplot(Month ~ Temp, data = aq, type = type_ridge(
#'   gradient = hcl.colors(250, "Dark Mint")[c(250:1, 1:250)],
#'   probs = 0:500/500))
#'
#' # faceting also works, although we recommend switching (back) to the "ridge"
#' # theme for faceted ridge plots
#'
#' tinytheme("ridge")
#' tinyplot(Month ~ Ozone, facet = ~ Late, data = aq,
#'   type = type_ridge(gradient = TRUE))
#'
#' ## use the joint.max argument to vary the maximum density used for
#' ## determining relative scaling...
#'
#' # jointly across all densities (default) vs. per facet
#' tinyplot(Month ~ Temp, facet = ~ Late, data = aq,
#'   type = type_ridge(scale = 1))
#' tinyplot(Month ~ Temp, facet = ~ Late, data = aq,
#'   type = type_ridge(scale = 1, joint.max = "facet"))
#'
#' # jointly across all densities (default) vs. per by row
#' tinyplot(Month ~ Temp | Late, data = aq,
#'   type = type_ridge(scale = 1))
#' tinyplot(Month ~ Temp | Late, data = aq,
#'   type = type_ridge(scale = 1, joint.max = "by"))
#'
#' # restore the default theme
#' tinytheme()
#'
#' @export
type_ridge = function(
    scale = 1.5,
    joint.max = c("all", "facet", "by"),
    breaks = NULL,
    probs = NULL,
    ylevels = NULL,
    bw = "nrd0",
    joint.bw =  c("mean", "full", "none"),
    adjust = 1,
    kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"),
    n = 512,
    # more args from density here?
    gradient = FALSE,
    raster = FALSE,
    col = NULL,
    alpha = NULL
    ) {

  kernel = match.arg(kernel, c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"))
  if (is.logical(joint.bw)) {
    joint.bw = ifelse(joint.bw, "mean", "none")
  }
  joint.bw = match.arg(joint.bw, c("mean", "full", "none"))

  out = list(
    draw = draw_ridge(),
    data = data_ridge(bw = bw, adjust = adjust, kernel = kernel, n = n,
                      joint.bw = joint.bw,
                      scale = scale,
                      joint.max = joint.max,
                      gradient = gradient,
                      breaks = breaks,
                      probs = probs,
                      ylevels = ylevels,
                      raster = raster,
                      col = col,
                      alpha = alpha
                      ),
    name = "ridge"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#
## Underlying data_ridge function
data_ridge = function(bw = "nrd0", adjust = 1, kernel = "gaussian", n = 512,
                      joint.bw = "mean",
                      scale = 1.5,
                      joint.max = "all",
                      gradient = FALSE,
                      breaks = NULL,
                      probs = NULL,
                      ylevels = NULL,
                      raster = FALSE,
                      col = NULL,
                      alpha = NULL
                      ) {
  fun = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "yaxt", "xaxt", "null_by"))

    # `col` may arrive either via the top-level `tinyplot(..., col =)` call
    # (stored in settings) or via the `type_ridge(col =)` constructor arg. The
    # constructor arg takes precedence; otherwise fall back to the settings
    # value so that a user-supplied `col` is respected. (#598)
    if (is.null(col)) col = settings[["col"]]

    #  catch for special cases
    anyby = !null_by
    x_by = anyby && identical(datapoints$x, datapoints$by)
    y_by = anyby && identical(datapoints$y, datapoints$by)
    if (x_by) {
      gradient = TRUE
      datapoints$by = ""
    } else if (y_by) {
      datapoints$by = ""
    } else if (anyby && is.null(alpha)) {
      alpha = 0.6
    }
    # flag for (non-gradient) interior fill adjustment
    fill_by = anyby || y_by
    if (isTRUE(x_by)) fill_by = FALSE
    # if (isTRUE(anyby) && is.null(alpha)) alpha = 0.6

    ## reorder levels of y-variable if requested
    if (!is.null(ylevels)) {
      if (!is.factor(datapoints$y)) datapoints$y = factor(datapoints$y)
      datapoints$y = factor(datapoints$y, levels = if(is.numeric(ylevels)) levels(datapoints$y)[ylevels] else ylevels)
      if (y_by) datapoints$by = datapoints$y
    }

    ##
    datapoints = split(datapoints, list(datapoints$y, datapoints$by, datapoints$facet))

    if (joint.bw == "none" || is.numeric(bw)) {
        dens_bw = bw
    } else {
        if (joint.bw == "mean") {
            # Use weighted mean of subgroup bandwidths
            bws = sapply(datapoints, function(dat) bw_fun(kernel = bw, dat$x))
            ws = sapply(datapoints, nrow)
            dens_bw = weighted.mean(bws, ws)
        } else if (joint.bw == "full") {
            dens_bw = bw_fun(kernel = bw, unlist(sapply(datapoints, `[[`, "x")))
        }
    }

    datapoints = lapply(datapoints, function(dat) {
      dens = density(dat$x, bw = dens_bw, kernel = kernel, n = n)
      out = data.frame(
        by = dat$by[1], # already split
        facet = dat$facet[1], # already split
        x = dens$x,
        y = dat$y[1],
        ymin = 0L,
        ymax = dens$y
      )
      return(out)
    })
    datapoints = do.call(rbind, datapoints)

    if (is.character(joint.max)) {
      joint.max = match.arg(joint.max, c("all", "facet", "by"))
      joint.max = switch(joint.max,
        "all" = rep.int(1, nrow(datapoints)),
        "facet" = datapoints$facet,
        "by" = interaction(datapoints$facet, datapoints$y)
      )
      joint.max = ave(datapoints$ymax, joint.max, FUN = max)
    }
    datapoints$ymax = datapoints$ymax / joint.max * scale

    datapoints = split(datapoints, datapoints$facet)
    offset_z = function(k) {
      ksplit = split(k, k$y)
      for (idx in seq_along(ksplit)) {
        ksplit[[idx]]$ymax = ksplit[[idx]]$ymax + idx - 1
        ksplit[[idx]]$ymin = ksplit[[idx]]$ymin + idx - 1
      }
      k = do.call(rbind, ksplit)
      return(k)
    }
    datapoints = do.call(rbind, lapply(datapoints, offset_z))

    # Store y-level offsets for added layers (e.g., jitter)
    ylevs = unique(datapoints$y)
    group_offsets = setNames(seq_along(ylevs) - 1L, ylevs)
    offsets_axis = "y"

    if (y_by) {
      datapoints$y = factor(datapoints$y)
      datapoints$by = factor(datapoints$y, levels = rev(levels(datapoints$y)))
    } else if (x_by) {
      datapoints$by = datapoints$x
    }

    # Manual breaks flag. Only used if gradient is on
    manbreaks = !is.null(breaks) || !is.null(probs)

    ## use color gradient?
    xlim = range(datapoints$x, na.rm = TRUE)
    if (!is.null(probs)) {
      if (!is.null(breaks)) {
        warning("only one of 'breaks' and 'quantile' must be specified")
        probs = NULL
      } else {
        if (probs[1L] > 0) probs = c(0, probs)
        if (probs[length(probs)] < 1) probs = c(probs, 1)
      }
    }
    if (!isFALSE(gradient)) {
      dotspal = list(...)[["palette"]]
      palette = if (!is.null(dotspal)) dotspal else gradient
      gradient = TRUE
      if (isTRUE(palette)) {
        palette = if (!is.null(.tpar[["palette.sequential"]])) .tpar[["palette.sequential"]] else "viridis"
      }

      if (length(palette) > 1L || !is.character(palette)) {
        ## color vector already given
        if (is.null(breaks) && is.null(probs)) {
          breaks = seq(from = xlim[1L], to = xlim[2L], length.out = length(palette) + 1L)
        } else {
          npal = pmax(length(breaks), length(probs)) - 1L
          if (length(palette) != npal) {
            warning("length of 'palette' does not match 'breaks'/'probs'")
            palette = rep_len(palette, npal)
          }
          if (isTRUE(raster)) raster = npal > 20L
        }
      } else {
        ## only palette name given
        npal = if (is.null(breaks) && is.null(probs)) 512L else pmax(length(breaks), length(probs)) - 1L
        palette = hcl.colors(npal, palette = palette)
        if (is.null(breaks) && is.null(probs)) breaks = seq(from = xlim[1L], to = xlim[2L], length.out = npal + 1L)
        if (isTRUE(raster)) raster = npal > 20L
      }
    } else {
      palette = NULL
      if (!is.null(breaks) || !is.null(probs)) gradient = TRUE
    }
    if (!is.null(breaks)) {
      breaks[1L] = pmin(breaks[1L], xlim[1L])
      breaks[length(breaks)] = pmax(breaks[length(breaks)], xlim[2L])
    }

    # Single-group (or x_by) ridges: default the outline colour consistently
    # with the other plot types. An explicit `col.default` wins; otherwise fall
    # back to the first colour of the active qualitative palette (e.g. blue under
    # "clean"), or base palette()[1] (black) when no theme palette is set. (#598)
    if (is.null(col) && (!anyby || x_by)) {
      col = get_tpar("col.default", default = NULL)
      if (is.null(col)) {
        pal_q = .tpar[["palette.qualitative"]]
        col = if (!is.null(pal_q)) {
          resolve_palette_spec(
            pal_q, ngrps = 1L, gradient = FALSE, ordered = FALSE,
            alpha = 1, adjustcolor = adjustcolor
          )
        } else {
          palette()[1L]
        }
      }
    }

    # For ridge themes without groups, a numeric bg (e.g. bg = 0.2) should
    # produce transparent gray, not a transparent palette colour. (#547)
    ridge_theme = identical(.tpar[["tinytheme"]], "ridge") || identical(.tpar[["tinytheme"]], "ridge2")
    if (ridge_theme && !anyby) {
      ubg = settings[["bg"]]
      if (!is.null(ubg) && length(ubg) == 1 && is.numeric(ubg) && ubg >= 0 && ubg <= 1) {
        settings[["bg"]] = adjustcolor("gray", alpha.f = ubg)
      }
    }

    # Save original yaxt for type_info before overwriting
    yaxt_orig = yaxt
    yaxt = "n"
    ylim = c(min(datapoints$ymin), max(datapoints$ymax))
    type_info = list(
      gradient = gradient,
      palette = palette,
      breaks = breaks,
      probs = probs,
      manbreaks = manbreaks,
      yaxt = yaxt_orig,
      raster = raster,
      ridge_theme = ridge_theme,
      x_by = x_by,
      y_by = y_by,
      fill_by = fill_by,
      col = col,
      alpha = alpha
    )

    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25

    env2env(environment(), settings, c(
      "datapoints",
      "yaxt",
      "ylim",
      "type_info",
      "group_offsets",
      "offsets_axis"
    ))
  }
  return(fun)
}


#
## Underlying draw_ridge function
draw_ridge = function() {
  fun = function(ix, iy, iz, ibg, icol, iymin, iymax, type_info, ...) {
    ridge_theme = isTRUE(type_info[["ridge_theme"]])
    d = data.frame(x = ix, y = iy, ymin = iymin, ymax = iymax)
    dsplit = split(d, d$y)
    if (!is.null(type_info[["col"]])) icol = type_info[["col"]]
    if (is.null(ibg)) {
      pal_q = .tpar[["palette.qualitative"]]
      # For non-ridge themes with a palette, derive fill from the first palette
      # colour. The theme palette may be a name (e.g. "Tableau 10") or a vector
      # of colours (e.g. the ipsum/socviz themes), so resolve it to an actual
      # colour rather than assuming a name. (#547, #598)
      default_bg = if (!ridge_theme && !is.null(pal_q)) {
        pal_q1 = resolve_palette_spec(
          pal_q, ngrps = 1L, gradient = FALSE, ordered = FALSE,
          alpha = 1, adjustcolor = adjustcolor
        )
        seq_palette(pal_q1, n = 2)[2]
      } else {
        "gray"
      }
      ibg = if (isTRUE(type_info[["fill_by"]])) seq_palette(icol, n = 2)[2] else default_bg
    }
    if (!is.null(type_info[["alpha"]]) && is.null(type_info[["palette"]])) {
      ibg = adjustcolor(ibg, alpha.f = type_info[["alpha"]])
    }
    lab = if (is.factor(d$y)) levels(d$y) else unique(d$y)
    if (isTRUE(type_info[["y_by"]])) {
      # avoid duplicating the y-axis labs for the special y==by case
      # val = match(lab, levels(d$y)) - 1
      val = match(d$y[1], levels(d$y))
      lab = lab[val]
      val = val - 1
    } else {
      val = cumsum(rep(1, length(lab))) - 1
    }
    if (ridge_theme) abline(h = val, col = .tpar[["grid.col"]])
    draw_segments = if (type_info[["raster"]]) segmented_raster else segmented_polygon
    for (i in rev(seq_along(dsplit))) {
      if (type_info[["gradient"]]) {
        with(
          dsplit[[i]],
          draw_segments(
            x, ymax, ymin = ymin[1L],
            breaks = type_info[["breaks"]],
            probs = type_info[["probs"]],
            manbreaks = type_info[["manbreaks"]],
            col = if (is.null(type_info[["palette"]])) ibg else type_info[["palette"]],
            # border = if (is.null(type_info[["palette"]])) icol else "transparent",
            alpha = type_info[["alpha"]]
          )
        )
      }
      with(dsplit[[i]], polygon(x, ymax, col = if (type_info[["gradient"]]) "transparent" else ibg, border = NA))
      with(dsplit[[i]], lines(x, ymax, col = icol))
    }
    # tinyAxis(x = d$y, side = 2, at = val, labels = lab, type = type_info[["yaxt"]], padj = padj)
    if (ridge_theme) {
      tinyAxis(x = d$y, side = 2, at = val, labels = lab, type = type_info[["yaxt"]],
               padj = 0,
               mgp = c(3, 1, 0) - c(0.5, 0.5 + 0.3, 0),
               tcl = 0)
      if (identical(.tpar[["tinytheme"]], "ridge2")) axis(1, labels = FALSE)
    } else {
      tinyAxis(x = d$y, side = 2, at = val, labels = lab, type = type_info[["yaxt"]])
    }
  }
  return(fun)
}


#
## Auxiliary functions

## auxiliary function for drawing shaded segmented polygon
segmented_polygon = function(x, y, ymin = 0, breaks = range(x), probs = NULL, manbreaks = FALSE, col = "lightgray", border = "transparent", alpha = NULL) {

  if (!is.null(probs)) {
    ## map quantiles to breaks
    if (!(missing(breaks) || is.null(breaks))) stop("only one of 'breaks' and 'probs' must be specified")
    breaks = quantile.density(list(x = x, y = y - ymin), probs = probs)
  }

  ## sanity check
  if (breaks[1L] > x[1L] || breaks[length(breaks)] < x[length(x)]) stop("'breaks' do no span range of 'x'")

  # ## recycle color (if necessary) rather use colorRampPalette below
  # col = rep_len(col, length(breaks) - 1L)

  # Create individual polygons
  if (isFALSE(manbreaks)) {
    # Special case for length(breaks)==length(x). We can take a fully vectorised
    # shortcut
    xx = c(rbind(x[-length(x)], x[-1], x[-1], x[-length(x)], NA))
    yy = c(rbind(y[-length(y)], y[-1], ymin, ymin, NA))
  } else {
    # For other cases, we'll do a bit more work to make sure that the polygons
    # overlap
    bvals = do.call(c, sapply(seq_along(breaks[-1]), function(b) tail(x[x<breaks[b]], 1)))
    bidx = match(bvals, x)
    bidx = c(1, bidx, length(x))
    xx = do.call(c, sapply(
      seq_along(bidx[-1]),
      function(b) {
        xx = x[bidx[b]:bidx[b+1]]
        c(xx, rev(xx), NA)
      }
    ))
    yy = do.call(c, sapply(
      seq_along(bidx[-1]),
      function(b) {
        yy = y[bidx[b]:bidx[b+1]]
        c(yy, rep(ymin, length(yy)), NA)
      }
    ))
    # Catch for missing lower breaks if individual density doesn't span the full
    # range of x (and discrete color fills are supplied)
    # Aside: We only need to fix the lower missing breaks, since missing top
    # breaks will be ignored in the final polygon call anyway
    lwrmissings = breaks < x[bidx][1]
    if (any(lwrmissings[-1])) {
      lwrmissings = breaks[lwrmissings]
      xxtra = c(rbind(lwrmissings[-length(lwrmissings)], lwrmissings[-1], lwrmissings[-1], lwrmissings[-length(lwrmissings)], NA))
      yytra = c(rbind(rep(ymin, length(lwrmissings)-1), ymin, ymin, ymin, NA))
      xx = c(xxtra, xx)
      yy = c(yytra, yy)
    }
  }

  # Drop trailing NAs
  xx = xx[1:(length(xx)-1)]
  yy = yy[1:(length(yy)-1)]

  if (!is.null(alpha)) col = adjustcolor(col, alpha.f = alpha)

  # Color ramp for cases where the breaks don't match
  if (isFALSE(length(col)==1)) {
    col = rev(col) ## uncomment to make extreme cols dark
    if (isTRUE(manbreaks)) {
      if (!is.null(breaks) && length(breaks)-1 != length(col)) {
        xrange = range(xx, na.rm = TRUE)
        idx = which(breaks >= xrange[1] & breaks < xrange[2])
        idx = c(idx, length(idx)+1)
        col = col[idx]
        col = colorRampPalette(col, alpha = TRUE)(length(x)) # support alpha?
      }
    } else if (isFALSE(manbreaks) || length(col) > length(x) || length(x) %% length(col) != 0) {
      xrange = range(xx, na.rm = TRUE)
      idx = which(breaks >= xrange[1] & breaks < xrange[2])
      idx = c(idx, length(idx)+1)
      col = col[idx]
      col = colorRampPalette(col, alpha = TRUE)(length(x)) # support alpha?
    }
  }
  border = if (is.null(alpha)) col else adjustcolor(col = col, alpha.f = alpha/2)

  ## draw all polygons
  polygon(xx, yy, col = col, border = border, lwd = 0.5)
}

#' @importFrom graphics rasterImage
#' @importFrom grDevices as.raster
segmented_raster = function(x, y, ymin = 0, breaks = range(x), probs = NULL, manbreaks = FALSE, col = "lightgray", border = "transparent", alpha = NULL) {
  ## set up raster matrix on x-grid and 500 y-pixels
  n = length(x) - 1L
  m = 500L ## FIXME: hard-coded?
  r = matrix(1:n, ncol = n, nrow = m, byrow = TRUE)

  ## map quantiles to breaks
  if (!is.null(probs)) {
    if (!(missing(breaks) || is.null(breaks))) stop("only one of 'breaks' and 'probs' must be specified")
    breaks = quantile.density(list(x = x, y = y - ymin), probs = probs)
  }

  if (!is.null(alpha)) col = adjustcolor(col, alpha.f = alpha)
  col = rev(col) ## uncomment to make extreme cols dark
  ## map colors to intervals and fill colors by column
  col = col[cut(x, breaks = breaks, include.lowest = TRUE)]
  r[] = col[r]

  ## clip raster pixels above density line
  ymax = max(y)
  ix = cbind(as.vector(row(r)), as.vector(col(r)))
  ix = ix[seq(from = ymax, to = ymin, length.out = m)[row(r)] > y[col(r)], , drop = FALSE]
  r[ix] = NA

  ## plot density and add raster gradient
  rasterImage(as.raster(r), min(x), ymin, max(x), ymax, interpolate = length(breaks) >= 20L) ## FIXME: improve quality for "few" breaks?
}

## auxiliary function for determining quantiles based on density function

#' @importFrom stats median approx
quantile.density = function(x, probs = seq(0, 1, 0.25), ...) {
  ## sanity check for probabilities
  if (any(probs < 0 | probs > 1)) stop("'probs' outside [0,1]")

  ## probability density function, extrapolated to zero, use midpoints
  n = length(x$x)
  pdf = x$y
  pdf = c(0, pdf, 0)

  ## x variable, also extrapolated, use midpoints
  x = x$x
  delta = median(diff(x))
  x = c(x[1L] - delta, x, x[n] + delta)

  ## numerical integration of density
  cdf = c(0, cumsum(diff(x) * (pdf[-1L] + pdf[-(n + 2L)])/2))
  cdf = cdf/cdf[n + 2L]

  ## approximate quantiles
  approx(cdf, x, xout = probs, rule = 2)$y
}

#' Operations on package load
#' @importFrom utils globalVariables
#' @importFrom graphics legend
#' @param libname library name
#' @param pkgname package name name
#' @keywords internal
#' @noRd
.onLoad = function(libname, pkgname) {
  # https://stackoverflow.com/questions/12598242/global-variables-in-packages-in-r
  # https://stackoverflow.com/questions/49056642/r-how-to-make-variable-available-to-namespace-at-loading-time?noredirect=1&lq=1

  globalVariables(c(
    "add",
    "alpha",
    "asp",
    "axes",
    "bg",
    "bubble",
    "bubble_alpha",
    "bubble_bg_alpha",
    "bubble_cex",
    "by_continuous",
    "by_dep",
    "by_ordered",
    "cap",
    "cex",
    "cex_dep",
    "cex_fct_adj",
    "data",
    "datapoints",
    "dots",
    "draw",
    "facet",
    "facet.args",
    "facet_attr",
    "facet_bg",
    "facet_border",
    "facet_by",
    "facet_col",
    "facet_font",
    "facet_newlines",
    "facet_rect",
    "facet_text",
    "facets",
    "fill",
    "flip",
    "frame.plot",
    "has_legend",
    "height",
    "iby",
    "ifacet",
    "labels",
    "lighten",
    "lty",
    "lwd",
    "nfacet_cols",
    "nfacet_rows",
    "nfacets",
    "ngrps",
    "null_by",
    "null_facet",
    "null_palette",
    "null_xlim",
    "multi_legend",
    "legend",
    "lgnd_labs",
    "lgnd_cex",
    "has_cap",
    "has_sub",
    "legend_draw_flag",
    "multi_legend",
    "legend_args",
    "null_ylim",
    "lgby",
    "lgbub",
    "oxaxis",
    "oyaxis",
    "pch",
    "rev_x",
    "rev_y",
    "ribbon.alpha",
    "split_data",
    "tpars",
    "type",
    "type_info",
    "was_area_type",
    "weights",
    "width",
    "x",
    "x_by",
    "x_dep",
    "xaxb",
    "xaxl",
    "xaxs",
    "xaxt",
    "xlab",
    "xlabs",
    "xlim",
    "xlvls",
    "xmax",
    "xmax_dep",
    "xmin",
    "xmin_dep",
    "y",
    "y_dep",
    "yaxb",
    "yaxl",
    "yaxs",
    "yaxt",
    "ygroup",
    "ylab",
    "ylabs",
    "ylim",
    "ymax",
    "ymax_dep",
    "ymin",
    "ymin_dep"
  ))
}

#' Histogram plot type
#'
#' @md
#' @description Type function for histogram plots. `type_hist` is an alias for
#'   `type_histogram`.
#' @param breaks Passed to \code{\link[graphics]{hist}}. One of:
#' - a vector giving the breakpoints between histogram cells,
#' - a function to compute the vector of breakpoints,
#' - a single number giving the number of cells for the histogram,
#' - a character string naming an algorithm to compute the number of cells (see ‘Details’ of \code{\link[graphics]{hist}}),
#' - a function to compute the number of cells.
#' In the last three cases the number is a suggestion only; as the breakpoints
#' will be set to pretty values, the number is limited to 1e6 (with a warning if
#' it was larger). If breaks is a function, the x vector is supplied to it as
#' the only argument (and the number of breaks is only limited by the amount of
#' available memory).
#' @param free.breaks Logical indicating whether the breakpoints should be
#'   computed separately for each group or facet? Default is `FALSE`, meaning
#'   that the breakpoints are computed from the full dataset; thus ensuring
#'   common bin widths across each group/facet. Can also use `free` as an
#'   acceptable argument alias. Ignored if there are no groups and/or facets.
#' @param drop.zeros Logical indicating whether bins with zero counts should be
#'   dropped before plotting. Default is `TRUE`. Note that switching to `FALSE`
#'   may interfere with faceted plot behaviour if `facet.args = list(free)`,
#'   since the `x` variable is effectively recorded over the full range of the
#'   x-axis (even if it does not extend over this range for every group).
#' @inheritParams graphics::hist
#' @examples
#' # "histogram"/"hist" type convenience string(s)
#' tinyplot(Nile, type = "histogram")
#'
#' # Use `type_histogram()` to pass extra arguments for customization
#' tinyplot(Nile, type = type_histogram(breaks = 30))
#' tinyplot(Nile, type = type_histogram(breaks = 30, freq = FALSE))
#' # etc.
#'
#' # Grouped histogram example
#' tinyplot(
#'     ~ Petal.Width | Species,
#'     type = "histogram",
#'     data = iris
#' )
#'
#' # Faceted version
#' tinyplot(
#'     ~Petal.Width,
#'     facet = ~Species,
#'     type = "histogram",
#'     data = iris
#' )
#'
#' # For visualizing faceted histograms across varying scales, you may also wish
#' # to impose free histogram breaks too (i.e., calculate breaks separately for
#' # each group). Compare:
#'
#' # free facet scales + shared histogram breaks, versus...
#' tinyplot(
#'     ~Petal.Width,
#'     facet = ~Species,
#'     facet.args = list(free = TRUE),
#'     type = type_histogram(),
#'     data = iris
#' )
#' # ... free facet scales + free histogram breaks
#' tinyplot(
#'     ~Petal.Width,
#'     facet = ~Species,
#'     facet.args = list(free = TRUE),
#'     type = type_histogram(free = TRUE),
#'     data = iris
#' )
#'
#' @export
type_histogram = function(breaks = "Sturges",
                          freq = NULL, right = TRUE,
                          free.breaks = FALSE, drop.zeros = TRUE) {
    out = list(
        data = data_histogram(
            breaks = breaks,
            free.breaks = free.breaks, drop.zeros = drop.zeros,
            freq = freq, right = right),
        draw = draw_rect(),
        name = "histogram"
    )
    class(out) = "tinyplot_type"
    return(out)
}
#' @export
#' @name type_hist
#' @rdname type_histogram
type_hist = type_histogram


data_histogram = function(breaks = "Sturges",
                          free.breaks = FALSE, drop.zeros = TRUE,
                          freq = NULL, right = TRUE) {
    hbreaks = breaks
    hfree.breaks = free.breaks
    hdrop.zeros = drop.zeros
    hfreq = freq
    hright = right

    fun = function(settings, .breaks = hbreaks, .freebreaks = hfree.breaks, .freq = hfreq, .right = hright, .drop.zeros = hdrop.zeros, ...) {
        env2env(settings, environment(), c("palette", "bg", "col", "plot", "datapoints", "ymin", "ymax", "xmin", "xmax", "freq", "ylab", "xlab", "facet", "ribbon.alpha", "by", "null_by", "null_palette"))

        has_weights = !is.null(datapoints[["weights"]])
        if (has_weights) settings$weights_used = TRUE

        hbreaks = ifelse(!sapply(.breaks, is.null), .breaks, "Sturges")

        # Multi-group displays fill from the palette at `ribbon.alpha`
        # transparency via the `ribbon.alpha` keyword. For single-group displays
        # with a theme palette active we leave `bg = NULL` so the fill tracks the
        # resolved border colour (see by_bg), which honours `col.default`. With
        # no theme palette, single-group uses the neutral "lightgray" shared by
        # all single-group area fills (matches base R hist()).
        if (is.null(bg) && !null_by) {
          bg = ribbon.alpha
        } else if (is.null(bg) && null_by && null_palette && is.null(get_tpar("palette.qualitative", default = NULL))) {
          bg = "lightgray"
        }

        if (!.freebreaks) xbreaks = hist(datapoints$x, breaks = hbreaks, right = .right, plot = FALSE)$breaks
        datapoints = split(datapoints, list(datapoints$by, datapoints$facet))
        datapoints = Filter(function(k) nrow(k) > 0, datapoints)

        datapoints = lapply(datapoints, function(k) {
            if (.freebreaks) xbreaks = breaks
            h = hist(k$x, breaks = xbreaks, right = .right, plot = FALSE)
            # weighted histogram: hist() has no weights arg, so recompute the bin
            # counts as weighted sums and rescale densities to integrate to 1.
            if (has_weights) {
                bin = findInterval(k$x, h$breaks, rightmost.closed = TRUE, left.open = .right)
                bin[bin < 1L] = 1L
                wsum = vapply(seq_along(h$mids), function(b) sum(k$weights[bin == b]), numeric(1))
                bw = diff(h$breaks)
                h$counts = wsum
                h$density = wsum / sum(k$weights) / bw
            }
            # zero count cases
            if (.drop.zeros) {
                nzidx = which(h$counts > 0)
                h$density = h$density[nzidx]
                h$counts = h$counts[nzidx]
                h$breaks = h$breaks[c(1, nzidx + 1)]
                h$mids = h$mids[nzidx]
            }
            freq = if (!is.null(.freq)) .freq else is.null(.freq) && h$equidist
            out = data.frame(
                by = k$by[1], # already split
                facet = k$facet[1], # already split
                ymin = 0,
                ymax = if (freq) h$counts else h$density,
                xmin = h$breaks[-1],
                xmax = h$mids + (h$mids - h$breaks[-1]),
                freq = freq
            )
            return(out)
        })
        datapoints = do.call(rbind, datapoints)

        if (is.null(ylab)) {
            ylab = ifelse(datapoints$freq[1], "Frequency", "Density")
        }

        x = c(datapoints$xmin, datapoints$xmax)
        y = c(datapoints$ymin, datapoints$ymax)
        ymin = datapoints$ymin
        ymax = datapoints$ymax
        xmin = datapoints$xmin
        xmax = datapoints$xmax
        by = if (length(unique(datapoints$by)) == 1) by else datapoints$by
        facet = if (length(unique(datapoints$facet)) == 1) facet else datapoints$facet
        
        # legend customizations
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% par("lwd")
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, c(
            "x",
            "y",
            "ymin",
            "ymax",
            "xmin",
            "xmax",
            "ylab",
            "col",
            "bg",
            "datapoints",
            "by",
            "facet"
        ))
    }
    return(fun)
}

#' Density plot type
#' 
#' @md
#' @description Type function for density plots.
#' @inheritParams stats::density
#' @param bw the smoothing \code{\link[stats:bw.nrd]{bandwidth}} to be used,
#'   see \code{\link[stats]{density}} for details and options.
#' @param kernel a character string giving the smoothing kernel to be used. This
#'   must partially match one of `"gaussian"`, `"rectangular"`, `"triangular"`,
#'   `"epanechnikov"`, `"biweight"`, `"cosine"` or `"optcosine"`, with default
#'   `"gaussian"`, and may be abbreviated to a unique prefix (single letter).
#'
#'   `"cosine"` is smoother than `"optcosine"`, which is the usual 'cosine'
#'   kernel in the literature and almost MSE-efficient. However, `"cosine"` is
#'   the version used by S.
#' @inheritParams type_ribbon
#' @param joint.bw character string indicating whether (and how) the smoothing
#'   bandwidth should be computed from the joint data distribution when there
#'   are multiple subgroups. The options are `"mean"` (the default), `"full"`,
#'   and `"none"`. Also accepts a logical argument, where `TRUE` maps to
#'   `"mean"` and `FALSE` maps to `"none"`. See the "Bandwidth selection"
#'   section below for a discussion of practical considerations.
#' @inherit stats::density details
#' @section Bandwidth selection: While the choice of smoothing bandwidth will
#'   always stand to affect a density visualization, it gains an added
#'   importance when multiple densities are drawn simultaneously (e.g., for
#'   subgroups with respect to `by` or `facet`). Allowing each subgroup to
#'   compute its own separate bandwidth independently offers greater flexibility
#'   in capturing the unique characteristics of each subgroup, particularly when
#'   distributions differ substantially in location and/or scale. However, this
#'   approach may overemphasize small random variations and make it harder to
#'   visually compare densities across subgroups. Hence, it is often useful to
#'   employ the same ("joint") bandwidth across all subgroups. The following
#'   strategies are available via the `joint.bw` argument:
#' 
#'   - The default `joint.bw = "mean"` first computes the individual bandwidths
#'   for each group but then computes their mean, weighted by the number of
#'   observations in each group. This will work well when all groups have
#'   similar amounts of scatter (similar variances), even when they have
#'   potentially rather different locations. The weighted averaging stabilizes
#'   potential fluctuations in the individual bandwidths, especially when some
#'   subgroups are rather small.
#' 
#'   - Alternatively, `joint.bw = "full"` can be used to compute the joint
#'   bandwidth from the full joint distribution (merging all groups). This will
#'   yield an even more robust bandwidth, especially when the groups overlap
#'   substantially (i.e., have similar locations and scales). However, it may
#'   lead to too large bandwidths and thus too much smoothing, especially when
#'   the locations of the groups differ substantially.
#' 
#'   - Finally, `joint.bw = "none"` disables the joint bandwidth so that each
#'   group just employs its individual bandwidth. This is often the best choice
#'   if the amounts of scatter differ substantially between the groups, thus
#'   necessitating different amounts of smoothing.
#' @section Titles: This tinyplot method for density plots differs from the base
#'   \code{\link[stats]{plot.density}} function in its treatment of titles. The
#'   x-axis title displays only the variable name, omitting details about the
#'   number of observations and smoothing bandwidth. Additionally, the main
#'   title is left blank by default for a cleaner appearance.
#' @examples
#' # "density" type convenience string
#' tinyplot(~Sepal.Length, data = iris, type = "density")
#' 
#' # grouped density example
#' tinyplot(~Sepal.Length | Species, data = iris, type = "density")
#' 
#' # use `bg = "by"` (or, equivalent `fill = "by"`) to get filled densities
#' tinyplot(~Sepal.Length | Species, data = iris, type = "density", fill = "by")
#' 
#' # use `type_density()` to pass extra arguments for customization
#' tinyplot(
#'   ~Sepal.Length | Species, data = iris,
#'   type = type_density(bw = "SJ"),
#'   main = "Bandwidth computed using Sheather & Jones (1991)"
#' )
#' 
#' # The default for grouped density plots is to use the mean of the
#' # individual subgroup bandwidths (weighted by group size) as the
#' # joint bandwidth. Alternatively, the bandwidth from the "full"
#' # data or separate individual bandwidths ("none") can be used.
#' tinyplot(~Sepal.Length | Species, data = iris,
#'     ylim = c(0, 1.25), type = "density")        # mean (default)
#' tinyplot_add(joint.bw = "full", lty = 2)        # full data
#' tinyplot_add(joint.bw = "none", lty = 3)        # none (individual)
#' legend("topright", c("Mean", "Full", "None"), lty = 1:3, bty = "n", title = "Joint BW")
#' 
#' @importFrom stats density weighted.mean
#' @importFrom stats bw.SJ bw.bcv bw.nrd bw.nrd0 bw.ucv 
#' @export
type_density = function(
        bw = "nrd0",
        joint.bw =  c("mean", "full", "none"),
        adjust = 1,
        kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"),
        n = 512,
        # more args from density here?
        alpha = NULL
    ) {
    kernel = match.arg(kernel, c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"))
    if (is.logical(joint.bw)) {
        joint.bw = ifelse(joint.bw, "mean", "none")
    }
    joint.bw = match.arg(joint.bw, c("mean", "full", "none"))
    out = list(
        data = data_density(bw = bw, adjust = adjust, kernel = kernel, n = n,
                            joint.bw = joint.bw, alpha = alpha),
        draw = NULL,
        name = "density"
    )
    class(out) = "tinyplot_type"
    return(out)
}

data_density = function(bw = "nrd0", adjust = 1, kernel = "gaussian", n = 512,
                        joint.bw = "none", alpha = NULL) {
    fun = function(settings, ...) {
        env2env(settings, environment(), c("by", "bg", "facet", "ylab", "col", "ribbon.alpha", "datapoints"))
        ribbon.alpha = if (is.null(alpha)) .tpar[["ribbon.alpha"]] else (alpha)
        has_weights = !is.null(datapoints[["weights"]])
        if (has_weights) settings$weights_used = TRUE
        
        if (is.null(ylab)) ylab = "Density"
        
        datapoints = split(datapoints, list(datapoints$by, datapoints$facet))
        datapoints = Filter(function(k) nrow(k) > 0, datapoints)
        
        if (joint.bw == "none" || is.numeric(bw)) {
            dens_bw = bw
        } else {
            if (joint.bw == "mean") {
                # Use weighted mean of subgroup bandwidths
                bws = sapply(datapoints, function(dat) bw_fun(kernel = bw, dat$x))
                ws = sapply(datapoints, nrow)
                dens_bw = weighted.mean(bws, ws)
            } else if (joint.bw == "full") {
                dens_bw = bw_fun(kernel = bw, unlist(sapply(datapoints, `[[`, "x")))
            }
        }
        
        datapoints = lapply(datapoints, function(dat) {
            wts = if (has_weights) dat[["weights"]] / sum(dat[["weights"]]) else NULL
            d = density(dat$x, bw = dens_bw, kernel = kernel, n = n, weights = wts)
            out = data.frame(
                by = dat$by[1], # already split
                facet = dat$facet[1], # already split
                y = d$y,
                x = d$x
            )
            return(out)
        })
        datapoints = do.call(rbind, datapoints)
        datapoints$ymax = datapoints$y
        datapoints$ymin = rep.int(0, nrow(datapoints))
        
        # flags for legend and fill
        dtype = if (!is.null(bg)) "ribbon" else "l"

        type = dtype
        by = if (length(unique(datapoints$by)) == 1) by else datapoints$by
        facet = if (length(unique(datapoints$facet)) == 1) facet else datapoints$facet
        
        # legend customizations (only for filled density plots)
        if (!is.null(bg)) {
            settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
            settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
            settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% par("lwd")
            settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
            settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
            settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        }
        
        env2env(environment(), settings, c(
            "ylab",
            "type",
            "ribbon.alpha",
            "datapoints",
            "by",
            "facet"
        ))
    }
    return(fun)
}

#' Convex hull plot type
#'
#' @description Type function for drawing convex hulls around grouped points.
#'   Uses \code{\link[grDevices]{chull}} to compute the convex hull of each
#'   group and draws the result as a polygon.
#'
#' @inheritParams type_polygon
#'
#' @examples
#' # "chull" type convenience character string
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, type = "chull")
#'
#' # layer filled convex hull(s) on top of points
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic")
#' tinyplot_add(type = "chull", fill = 0.2)
#'
#' @export
type_chull = function(density = NULL, angle = 45) {
  out = list(
    draw = draw_polygon(density = density, angle = angle),
    data = data_chull(),
    name = "chull"
  )
  class(out) = "tinyplot_type"
  return(out)
}


data_chull = function() {
  fun = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "by", "facet"))

    datapoints = split(datapoints, list(datapoints$by, datapoints$facet))
    datapoints = Filter(function(k) nrow(k) > 0, datapoints)

    datapoints = lapply(datapoints, function(dat) {
      idx = grDevices::chull(dat$x, dat$y)
      idx = c(idx, idx[1L])
      data.frame(
        x = dat$x[idx],
        y = dat$y[idx],
        by = dat$by[1L],
        facet = dat$facet[1L]
      )
    })
    datapoints = do.call(rbind, datapoints)

    by = if (length(unique(datapoints$by)) == 1) by else datapoints$by
    facet = if (length(unique(datapoints$facet)) == 1) facet else datapoints$facet

    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25

    env2env(environment(), settings, c("datapoints", "by", "facet"))
  }
  return(fun)
}

#' Plot summary values of `y` at unique values of `x`
#'
#' @md
#' @description
#' Applies a summary function to `y` along unique values of `x`. For example,
#' plot the mean `y` value for each `x` value. Internally,
#' `type_summary()` applies a thin wrapper around \code{\link[stats]{ave}} and
#' then passes the result to [`type_lines`] for drawing.
#'
#' @param fun summarizing function. Should be compatible with
#'   \code{\link[stats]{ave}}. Defaults to \code{\link[base]{mean}}.
#' @param ... Additional arguments are passed to the `lines()` function,
#' ex: `type="p"`, `col="pink"`.
#' @seealso [`ave`] which performs the summarizing (averaging) behind the
#'   scenes.
#' @examples
#' # Plot the mean chick weight over time
#' tinyplot(weight ~ Time, data = ChickWeight, type = "summary")
#'
#' # Note: "mean" is the default function, so these are also equivalent:
#' # tinyplot(weight ~ Time, data = ChickWeight, type = type_summary())
#' # tinyplot(weight ~ Time, data = ChickWeight, type = type_summary(mean))
#'
#' # Plot the median instead
#' tinyplot(weight ~ Time, data = ChickWeight, type = type_summary(median))
#'
#' # Works with groups and/or facets too
#' tinyplot(weight ~ Time | Diet, facet = "by", data = ChickWeight, type = "summary")
#'
#' # Custom/complex function example
#' tinyplot(
#'   weight ~ Time | Diet,
#'   facet = "by", data = ChickWeight,
#'   type = type_summary(function(y) quantile(y, probs = 0.9) / max(y))
#' )
#'
#' @importFrom stats ave
#' @export
type_summary = function(fun = mean, ...) {
  assert_function(fun)
  lines_args = list(...)
  data_summary = function(fun) {
    funky = function(settings, ...) {
      env2env(settings, environment(), c("datapoints", "by", "facet"))

      datapoints = split(datapoints, list(datapoints$facet, datapoints$by), drop = TRUE)
      datapoints = lapply(datapoints, function(dat) {
        newy = ave(dat$y, dat$x, FUN = fun)
        dat$y = newy
        dat = dat[order(dat$x), ]
        return(dat)
      })
      datapoints = do.call(rbind, datapoints)
      env2env(environment(), settings, "datapoints")
    }
    return(funky)
  }
  out = list(
    draw = draw_lines(...),
    data = data_summary(fun = fun),
    name = "l"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' Barplot type
#'
#' @description Type function for producing barplots. For formulas of type
#'   `~ x` (without left-hand side) the barplot visualizes the counts (absolute
#'   frequencies) of the levels of `x`. For formulas of type `y ~ x` the value
#'   of `y` within each level of `x` is visualized, if necessary aggregated
#'   using some function (default: mean).
#'
#' @param width numeric, optional vector of bar widths. (The distance between
#'   the midpoints of the bars is always 1.)
#' @param beside logical. In case of a `by` grouping variable, should bars be
#'   juxtaposed? Default is to use stacked bars instead.
#' @param center logical or numeric. In case of stacked barplots (`beside = FALSE`)
#'   should the bars be centered (or all start at zero, default)? If set to
#'   `TRUE` the center is at the mid-point of the middle category (in case of
#'   uneven number of categories) or between the two middle categories (in case
#'   of an even number). Additionally it is possible to set `center = 2` or
#'   `center = 2.5` to indicate that centering should be after the second category
#'   or the mid-way in the third category, respectively.
#' @param FUN a function to compute the summary statistic for `y` within each
#'   group of `x` in case of using a two-sided formula `y ~ x` (default: mean).
#' @param xlevels a character or numeric vector specifying the ordering of the
#'   levels of the `x` variable (if character) or the corresponding indexes
#'   (if numeric) for the plot.
#' @param xaxlabels a character vector with the axis labels for the `x` variable,
#'   defaulting to the levels of `x`.
#' @param offset optional specification for shifting bar baselines, accepting
#'   one of two distinct forms. See the Examples for illustrations of both.
#' 
#'   - *Positions* via an unnamed numeric scalar or vector. Bars start at the
#'   offset value(s) rather than zero, matched per x-level after any `xlevels`
#'   reordering (a scalar is applied to all bars). Useful for waterfall charts.
#'   The positional form cannot be combined with `center`.
#'   - *Category* via a character vector such as `offset = "Unsure"`, or a
#'   named numeric vector such as `offset = c(Unsure = 1.1)`. The named
#'   level(s) of the `by` grouping are "set aside", i.e. pulled out of the
#'   (optionally centered) stack and drawn as standalone bars. This is useful
#'   for diverging/Likert plots where a neutral category (e.g. "Unsure") is
#'   shown apart from the diverging stack. The category form requires a `by`
#'   grouping and `beside = FALSE`, but can be combined with `center`.
#' @param drop.zeros logical. Should bars with zero height be dropped? If set
#'   to `FALSE` (default) a zero height bar is still drawn for which the border
#'   lines will still be visible.
#' @param lighten logical. Should the fills use a lighter, opaque tint of the
#'   series colour(s)? Default is `TRUE`, which keeps single- and multi-group
#'   displays consistent and lets the fill read cleanly over grid lines. Set to
#'   `FALSE` to use the fully-saturated palette colour(s) instead.
#'
#' @examples
#' # Basic examples of frequency tables (without y variable)
#' tinyplot(~ cyl, data = mtcars, type = "barplot")
#' tinyplot(~ cyl | vs, data = mtcars, type = "barplot")
#' tinyplot(~ cyl | vs, data = mtcars, type = "barplot", beside = TRUE)
#' 
#' # Reorder x variable categories either by their character levels or numeric indexes
#' tinyplot(~ cyl, data = mtcars, type = "barplot", xlevels = c("8", "6", "4"))
#' tinyplot(~ cyl, data = mtcars, type = "barplot", xlevels = 3:1)
#' 
#' # Note: Above we used automatic argument passing for `beside`. But this
#' # wouldn't work for `width`, since it would conflict with the top-level
#' # `tinyplot(..., width = <width>)` argument. It's safer to pass these args
#' # through the `type_barplot()` functional equivalent.
#' tinyplot(
#'   ~ cyl | vs, data = mtcars,
#'   type = type_barplot(beside = TRUE, drop.zeros = TRUE, width = 0.65)
#' )
#' 
#' # Example for numeric y aggregated by x (default: FUN = mean) + facets
#' tinyplot(
#'   extra ~ ID | group, facet = "by", data = sleep,
#'   type = "barplot",
#'   theme = "clean2"
#' )
#' 
#' # Fancy frequency table:
#' tinyplot(
#'   Freq ~ Sex | Survived, data = as.data.frame(Titanic),
#'   facet = ~ Class, facet.args = list(nrow = 1),
#'   type = "barplot", flip = TRUE,
#'   theme = "clean2"
#' )
#'
#' # Centered barplot for conditional proportions of dark (black/brown) vs.
#' # light (red/blond) hair color, conditional on eye color and sex.
#' # Aside: use `lighten = FALSE` to avoid lightening the bar fill colors.
#' hec = as.data.frame(proportions(HairEyeColor, 2:3))
#' hcols = c("black", "sienna", "indianred", "goldenrod")
#' tinyplot(
#'   Freq ~ Eye | Hair, data = hec,
#'   facet = ~ Sex, facet.args = list(ncol = 1),
#'   type = type_barplot(center = TRUE, lighten = FALSE),
#'   flip = TRUE, yaxl = "percent",
#'   theme = list("clean2", palette.qualitative = hcols)
#' )
#'
#' # Use cases for the `offset` argument
#'
#' # 1. Waterfall plot
#' d = data.frame(item = c("Sales", "Services", "Costs", "Returns", "TOTAL"),
#'                value = c(100, 40, -80, -10, 50))
#' d$item = factor(d$item, levels = d$item)
#' d$offset = c(0, cumsum(d$value[1:3]), 0)
#' tinyplot(
#'   value ~ item | I(value < 0), data = d,
#'   type = type_barplot(offset = d$offset, lighten = FALSE),
#'   col = NA, # (optional: turn off border)
#'   legend = FALSE
#' )
#' tinyplot_add(type = type_vline(4.5), lty = 2, col = "grey50")
#'
#' # 2. Diverging/Likert layout: a character (or named numeric) offset "sets
#' # aside" the named category, pulling it out of the centered stack and drawing
#' # it as a standalone bar. Here a neutral "Unsure" response is shown apart from
#' # the diverging agree/disagree scale.
#' lik = expand.grid(
#'   question = c("Pay", "Workload", "Manager", "Culture"),
#'   response = c("Strong disagree", "Disagree", "Agree", "Strong agree", "Unsure")
#' )
#' lik$response = factor(lik$response, levels = unique(lik$response))
#' lik$share = c( # proportions summing to 1 within each question
#'   .10, .25, .05, .15,
#'   .20, .30, .15, .20,
#'   .35, .20, .40, .30,
#'   .25, .15, .35, .20,
#'   .10, .10, .05, .15
#' )
#' # diverging palette: reds (disagree) -> blues (agree), grey for "Unsure"
#' pal = c("#b2182b", "#ef8a62", "#67a9cf", "#2166ac", "grey")
#' tinyplot(
#'   share ~ question | response, data = lik,
#'   type = type_barplot(center = TRUE, offset = "Unsure", lighten = FALSE),
#'   flip = TRUE, xlab = NA, ylab = NA, yaxl = "percent",
#'   legend = list("top!", title = FALSE),
#'   theme = list("clean2", palette.qualitative = pal),
#'   main = "Hypothetical Likert example with category offset"
#' )
#' tinyplot_add(type = "vline")
#'
#' @export
type_barplot = function(width = 5/6, beside = FALSE, center = FALSE, offset = NULL, FUN = NULL, xlevels = NULL, xaxlabels = NULL, drop.zeros = FALSE, lighten = TRUE) {
  out = list(
    data = data_barplot(width = width, beside = beside, center = center, offset = offset, FUN = FUN, xlevels = xlevels, xaxlabels = xaxlabels, drop.zeros = drop.zeros, lighten = lighten),
    draw = draw_rect(),
    name = "barplot"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' @importFrom stats aggregate
data_barplot = function(width = 5/6, beside = FALSE, center = FALSE, offset = NULL, FUN = NULL, xlevels = NULL, xaxlabels = NULL, drop.zeros = FALSE, lighten = TRUE) {
    fun = function(settings, ...) {
        env2env(
          settings,
          environment(),
          c(
            "datapoints", "null_by", "facet_by",
            "xlab", "ylab", "xlim", "ylim", "yaxl", "xaxt",
            "null_palette", "col", "bg"
          )
        )

        ## tabulate/aggregate datapoints
        if (is.null(datapoints$y)) {
          if (is.null(xlab) || identical(xlab, "Index")) xlab = ylab
          if (is.null(settings$y_dep) && is.null(ylab)) ylab = "Count"
          datapoints$y = numeric(nrow(datapoints))          
          if (!is.null(FUN)) warning("without 'y' variable 'FUN' specification is ignored")
          FUN = length
        } else {
          if (is.null(FUN)) FUN = function(x, ...) mean(x, ..., na.rm = TRUE)
        }
        if (!is.factor(datapoints$x)) datapoints$x = factor(datapoints$x)
        if (!is.null(xlevels)) {
          xlevels = if(is.numeric(xlevels)) levels(datapoints$x)[xlevels] else xlevels
          if (anyNA(xlevels) || !all(xlevels %in% levels(datapoints$x))) warning("not all 'xlevels' correspond to levels of 'x'")
          datapoints$x = factor(datapoints$x, levels = xlevels)
        }
        if (!is.null(xaxlabels)) levels(datapoints$x) = xaxlabels
        datapoints = aggregate(datapoints[, "y", drop = FALSE], datapoints[, c("x", "by", "facet")], FUN = FUN, drop = FALSE)
        datapoints$y[is.na(datapoints$y)] = 0 #FIXME: always?#
        if (!is.factor(datapoints$by)) datapoints$by = factor(datapoints$by)
        if (!is.factor(datapoints$facet)) datapoints$facet = factor(datapoints$facet)
        
        ## `offset` accepts two distinct forms:
        ##  - unnamed numeric -> positional, keyed by x-level (waterfall)
        ##  - character or *named* numeric -> keyed by `by`-level: those groups
        ##    are "set aside", i.e. pulled out of the (optionally centered) stack
        ##    and drawn as standalone bars (diverging/Likert layout).
        aside = NULL # named numeric of set-aside `by`-levels -> baseline value
        if (!is.null(offset)) {
          offset_bylevel = is.character(offset) ||
            (is.numeric(offset) && !is.null(names(offset)) && any(nzchar(names(offset))))
          if (offset_bylevel) {
            if (isTRUE(null_by)) {
              stop("a character or named 'offset' requires a 'by' grouping variable")
            }
            if (isTRUE(facet_by)) {
              stop("a character or named 'offset' is not supported when 'facet' is the 'by' grouping")
            }
            if (beside) {
              stop("a character or named 'offset' requires stacked bars; set 'beside = FALSE'")
            }
            nm = if (is.character(offset)) offset else names(offset)
            bad = setdiff(nm, levels(datapoints$by))
            if (length(bad)) {
              stop(sprintf(
                "'offset' must name levels of the 'by' grouping; unknown: %s",
                paste(bad, collapse = ", ")
              ))
            }
            if (is.character(offset)) {
              ## auto-placement: baseline = max full-column total (incl. set-aside)
              col_tot = tapply(
                datapoints$y, interaction(datapoints$x, datapoints$facet),
                sum, na.rm = TRUE
              )
              base_auto = max(col_tot, na.rm = TRUE)
              aside = stats::setNames(rep(base_auto, length(offset)), offset)
            } else {
              aside = offset
            }
            offset = NULL # disable the positional post-hoc shift below
          } else {
            if (!is.numeric(offset)) stop("'offset' must be numeric")
            if (!isFALSE(center)) {
              warning("'offset' cannot be combined with 'center'; ignoring 'center'")
              center = FALSE
            }
            nx_levels = nlevels(datapoints$x)
            if (length(offset) == 1L) {
              offset = rep(offset, nx_levels)
            } else if (length(offset) != nx_levels) {
              stop(sprintf(
                "'offset' must be length 1 or %d (number of x levels), got %d",
                nx_levels, length(offset)
              ))
            }
          }
        }
        if (is.null(offset) && is.null(aside) && isFALSE(null_by) && isFALSE(facet_by) && !beside && any(datapoints$y < 0)) {
          warning("'beside' must be TRUE if there are negative 'y' values")
          beside = TRUE
        }
        if (beside & !isFALSE(center)) {
          warning("'center' is currently only supported for 'beside = FALSE'")
        }
        null_ylim = is.null(ylim)
        offset_sum = function(z, center = TRUE, na.rm = TRUE) {
          n = length(z)
          if (isFALSE(center) || n < 1L) return(0)
          mid = if (isTRUE(center)) n/2 else center
          z[floor(mid) + 1L] = (mid - floor(mid)) * z[floor(mid) + 1L]
          sum(z[0L:floor(mid) + 1L], na.rm = TRUE)
        }
        if (is.null(xlim)) xlim = c(1, nlevels(datapoints$x)) + c(-0.5, 0.5) * width
        if (is.null(ylim)) ylim = if (beside || length(unique(datapoints$by)) == 1L) {
          c(pmin(0, min(datapoints$y, na.rm = TRUE) * 1.02), pmax(0, max(datapoints$y, na.rm = TRUE) * 1.02))
        } else {
          is_off = if (is.null(aside)) rep(FALSE, nrow(datapoints)) else datapoints$by %in% names(aside)
          ## range of the centered/stacked retained (non-set-aside) categories
          stack_range = range(unlist(tapply(
            seq_len(nrow(datapoints)),
            interaction(datapoints$x, datapoints$facet),
            function(idx) {
              z = datapoints$y[idx]
              keep = !is_off[idx]
              zc = z; zc[!keep] = 0
              c(0, sum(zc, na.rm = TRUE)) - offset_sum(z[keep], center = center)
            }
          )))
          ## range of the standalone set-aside bars (stacked from their baselines)
          off_range = NULL
          if (any(is_off)) {
            base = aside[as.character(datapoints$by[is_off])]
            ytop = unlist(tapply(
              seq_len(sum(is_off)),
              interaction(datapoints$x[is_off], datapoints$facet[is_off]),
              function(j) base[j] + cumsum(datapoints$y[is_off][j])
            ))
            off_range = range(c(base, ytop), na.rm = TRUE)
          }
          range(c(stack_range, off_range), na.rm = TRUE) * 1.02
        }

        ## fill lightening (see by_bg)
        settings[["lighten"]] = lighten

        ## default color palette
        ngrps = length(unique(datapoints$by))
        if (ngrps == 1L && null_palette) {
          # With a theme palette active, leave bg = NULL so the fill tracks
          # the resolved border colour (see by_bg). Otherwise use the neutral
          # "lightgray" shared by all single-group area fills (matches base R
          # hist()/boxplot()).
          if (is.null(bg) && is.null(get_tpar("palette.qualitative", default = NULL))) bg = "lightgray"
        } else {
          if (is.null(bg)) bg = "by"
        }

        ## calculate bar rectangles per facet 
        sdat = split(datapoints, datapoints$facet)
        datapoints = lapply(sdat, function(df)  {
          
          df = df[order(df$x), , drop = FALSE]
          nx = nlevels(df$x)
          nb = nlevels(df$by)
          
          if (beside) {
            xl = as.numeric(df$x) - width/2 + (as.numeric(df$by) - 1) * width/nb * as.numeric(!facet_by)
            xr = if (facet_by) xl + width else xl + width/nb
            yb = 0
            yt = df$y
          } else {
            is_off = if (is.null(aside)) rep(FALSE, nrow(df)) else df$by %in% names(aside)
            xl = as.numeric(df$x) - width/2
            xr = xl + width
            ## stack/center only the retained categories: set-aside rows are
            ## zeroed so they don't grow or shift the centered stack, but the
            ## centering midpoint is computed from the retained subset alone.
            cs = tapply(seq_len(nrow(df)), df$x, function(idx) {
              z = df$y[idx]
              keep = !is_off[idx]
              zc = z; zc[!keep] = 0
              cumsum(c(0, zc)) - offset_sum(z[keep], center = center)
            })
            yb = if (facet_by) 0 else unlist(lapply(cs, `[`, -(nb + 1L)))
            yt = if (facet_by) df$y else unlist(lapply(cs, `[`, -1L))
            ## set-aside rows: standalone bars stacked from their baseline value
            if (any(is_off)) {
              base = aside[as.character(df$by[is_off])]
              ot = unlist(tapply(
                seq_len(sum(is_off)),
                df$x[is_off],
                function(j) base[j] + cumsum(df$y[is_off][j])
              ))
              ob = ot - df$y[is_off]
              yb[is_off] = ob
              yt[is_off] = ot
            }
          }
          
          df$xmin = xl
          df$xmax = xr
          df$ymin = yb
          df$ymax = yt
          df$nx = nx
          
          if (drop.zeros) {
            yb = rep_len(yb, length(yt))
            yok = abs(yt - yb) > 0
            df = df[yok,  , drop = FALSE]
          }
          
          return(df)
        })
        datapoints = do.call("rbind", datapoints)
        nx = datapoints$nx[1]
        datapoints$nx = NULL
        xlabs = 1L:nx
        names(xlabs) = levels(datapoints$x)

        # Apply offset: shift bar baselines after rectangle computation
        if (!is.null(offset)) {
          off = offset[as.numeric(datapoints$x)]
          datapoints$ymin = datapoints$ymin + off
          datapoints$ymax = datapoints$ymax + off
          if (null_ylim) {
            ylim = range(c(0, datapoints$ymin, datapoints$ymax), na.rm = TRUE) * 1.02
          }
        }

        if (!isFALSE(center)) {
          if (is.null(yaxl)) {
            yaxl = abs
          } else if (is.character(yaxl)) {
            yaxl = paste0("abs_", yaxl)
          }
        }

        axes = TRUE
        frame.plot = FALSE
        xaxs = "r"
        xaxt = if (xaxt == "s") "l" else xaxt
        yaxs = "i"
        
        # legend customizations
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, c(
          "datapoints",
          "xlab",
          "ylab",
          "xlim",
          "ylim",
          "axes",
          "xlabs",
          "frame.plot",
          "xaxs",
          "xaxt",
          "yaxl",
          "yaxs",
          "col",
          "bg"
        ))
    }
    return(fun)
}

swap_elements = function(env, a, b) {
  # Swap two elements in an environment
  val_a = if (exists(a, envir = env, inherits = FALSE)) env[[a]] else NULL
  val_b = if (exists(b, envir = env, inherits = FALSE)) env[[b]] else NULL

  assign(a, val_b, envir = env)
  assign(b, val_a, envir = env)
}


flip_datapoints = function(settings) {
  env2env(settings, environment(), c("flip", "type", "datapoints", "log"))

  assert_flag(flip)
  if (isTRUE(flip)) {
    if (type == "boxplot") {
      # boxplot: let horizontal=TRUE do most work; only swap labels
      swap_elements(settings, "xlab", "ylab")
    } else {
      datapoints = swap_columns(datapoints, "xmin", "ymin")
      datapoints = swap_columns(datapoints, "xmax", "ymax")
      datapoints = swap_columns(datapoints, "x", "y")

      # Swap all the x/y settings in the environment
      swap_elements(settings, "x", "y")
      swap_elements(settings, "xaxb", "yaxb")
      swap_elements(settings, "xaxl", "yaxl")
      swap_elements(settings, "xaxs", "yaxs")
      swap_elements(settings, "xaxt", "yaxt")
      swap_elements(settings, "xlab", "ylab")
      swap_elements(settings, "xlabs", "ylabs")
      swap_elements(settings, "xlim", "ylim")
      swap_elements(settings, "rev_x", "rev_y")
      swap_elements(settings, "xmax", "ymax")
      swap_elements(settings, "xmin", "ymin")

      if (!is.null(log)) {
        log = chartr("xy", "yx", log)
        assign("log", log, envir = settings)
      }

      # Copy modified datapoints back
      assign("datapoints", datapoints, envir = settings)
    }
  }
}

check_dependency = function(library_name) {
  flag = requireNamespace(library_name, quietly = TRUE)
  if (isFALSE(flag)) {
    msg = sprintf("Please install the `%s` package.", library_name)
    return(msg)
  } else {
    return(TRUE)
  }
}

assert_dependency = function(library_name) {
  flag = check_dependency(library_name)
  if (!isTRUE(flag)) stop(flag, call. = FALSE)
  return(invisible())
}

assert_choice = function(x, choice, null.ok = FALSE, name = as.character(substitute(x))) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (is.character(x) && length(x) == 1 && x %in% choice) {
    return(TRUE)
  }
  msg = sprintf(
    "`%s` must be one of: %s",
    name,
    paste(choice, collapse = ", ")
  )
  stop(msg, call. = FALSE)
}

check_true = function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  if (isTRUE(x)) {
    return(invisible(TRUE))
  }
  return(FALSE)
}

assert_true = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  msg = sprintf("`%s` must be true.", name)
  if (!isTRUE(check_true(x, null.ok = null.ok))) {
    stop(msg, call. = FALSE)
  }
}

check_string = function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  if (is.character(x) && length(x) == 1) {
    return(invisible(TRUE))
  }
  return(FALSE)
}

assert_string = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  msg = sprintf("`%s` must be a string.", name)
  if (!isTRUE(check_string(x, null.ok = null.ok))) {
    stop(msg, call. = FALSE)
  }
}

check_flag = function(x, null.ok = FALSE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (is.logical(x) && length(x) == 1) {
    return(TRUE)
  }
  return(FALSE)
}

assert_flag = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  msg = sprintf("`%s` must be a logical flag.", name)
  if (!isTRUE(check_flag(x, null.ok = null.ok))) {
    stop(msg, call. = FALSE)
  }
}

assert_length = function(x, len = 1, null.ok = FALSE, name = as.character(substitute(x))) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  msg = sprintf("`%s` must be one of these lengths: %s", name, paste(len, collapse = ", "))
  if (!length(x) %in% len) {
    stop(msg, call. = FALSE)
  }
}

assert_grid = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  if (is.null(x) && isTRUE(null.ok)) return(invisible(TRUE))
  if (is.logical(x) && length(x) == 1) return(invisible(TRUE))
  valid_strings = c("x", "y", "X", "Y", "xy", "xY", "Xy", "XY")
  if (is.character(x) && length(x) == 1 && x %in% valid_strings) return(invisible(TRUE))
  msg = sprintf("`%s` must be a logical flag or a character string combining x/X and/or y/Y (e.g., \"xy\", \"XY\", \"xY\")", name)
  stop(msg, call. = FALSE)
}

assert_logical = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(invisible(TRUE))
  }
  msg = sprintf("`%s` must be a logical vector", name)
  if (!is.logical(x)) stop(msg, call. = FALSE)
}


check_integerish = function(x, len = NULL, lower = NULL, upper = NULL, null.ok = TRUE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (!is.numeric(x)) {
    return(FALSE)
  }
  x = stats::na.omit(x)
  if (!is.null(len) && length(x) != len) {
    return(FALSE)
  }
  if (!is.null(lower) && any(x < lower)) {
    return(FALSE)
  }
  if (!is.null(upper) && any(x > upper)) {
    return(FALSE)
  }
  if (isTRUE(any(abs(x - round(x)) > (.Machine$double.eps)^0.5))) {
    return(FALSE)
  }
  return(TRUE)
}

assert_integerish = function(x, len = NULL, lower = NULL, upper = NULL, null.ok = FALSE, name = as.character(substitute(x))) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible())
  }
  msg = sprintf("`%s` must be integer-ish", name)
  if (is.null(x) && !isTRUE(null.ok)) stop(sprintf("%s should not be NULL.", name), call. = FALSE)
  if (!isTRUE(check_integerish(x, len = len, lower = lower, upper = upper, null.ok = null.ok))) {
    if (!is.numeric(x)) msg = paste0(msg, "; it is not numeric")
    if (!is.null(len) && length(x) != len) msg = paste0(msg, sprintf("; its length must be %s", len))
    if (!is.null(lower) && any(x < lower)) msg = paste0(msg, sprintf("; all values must be greater than or equal to %s", lower))
    if (!is.null(upper) && any(x > upper)) msg = paste0(msg, sprintf("; all values must be less than or equal to %s", upper))
    if (isTRUE(any(abs(x - round(x)) > (.Machine$double.eps)^0.5))) msg = paste0(msg, "; all values must be close to integers")
    stop(msg, call. = FALSE)
  }
}

check_numeric = function(x, len = NULL, lower = NULL, upper = NULL, null.ok = TRUE) {
  if (is.null(x) && isTRUE(null.ok)) {
    return(TRUE)
  }
  if (!is.numeric(x)) {
    return(FALSE)
  }
  if (!is.null(len) && length(x) != len) {
    return(FALSE)
  }
  if (!is.null(lower) && any(x < lower)) {
    return(FALSE)
  }
  if (!is.null(upper) && any(x > upper)) {
    return(FALSE)
  }
  return(TRUE)
}

assert_numeric = function(x, len = NULL, lower = NULL, upper = NULL, null.ok = FALSE, name = as.character(substitute(x))) {
  msg = sprintf("`%s` must be numeric", name)
  if (!isTRUE(check_numeric(x, len = len, lower = lower, upper = upper, null.ok = null.ok))) {
    if (!is.null(len) && length(x) != len) msg = paste0(msg, sprintf("; its length must be %s", len))
    if (!is.null(lower) && any(x < lower)) msg = paste0(msg, sprintf("; all values must be greater than or equal to %s", lower))
    if (!is.null(upper) && any(x > upper)) msg = paste0(msg, sprintf("; all values must be less than or equal to %s", upper))
    stop(msg, call. = FALSE)
  }
}

assert_data_frame = function(x, min_rows = 0, min_cols = 0, name = as.character(substitute(x))) {
  msg = sprintf("`%s` must be a data.frame.", name)
  if (!is.data.frame(x)) stop(msg, call. = FALSE)
  msg = sprintf("Number of rows in `%s` must be at least `%s`", name, min_rows)
  if (nrow(x) < min_rows) stop(msg, call. = FALSE)
  msg = sprintf("Number of columns in `%s` must be at least `%s`", name, min_cols)
  if (ncol(x) < min_cols) stop(msg, call. = FALSE)
}


check_character = function(x, len = NULL, null.ok = FALSE, name = as.character(substitute(x))) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(TRUE)
  } else if (!is.character(x)) {
    msg = sprintf("`%s` must be character.", name)
    return(msg)
  } else if (!is.null(len) && length(x) != len) {
    msg = sprintf("`%s` must have length %s.", name, len)
    return(msg)
  }
  return(TRUE)
}

assert_character = function(x, len = NULL, null.ok = FALSE, name = as.character(substitute(x))) {
  flag = check_character(x, len = len, null.ok = null.ok, name = name)
  if (!isTRUE(flag)) {
    stop(flag, call. = FALSE)
  } else {
    return(invisible(TRUE))
  }
}

assert_list = function(x, named = FALSE, len = NULL, null.ok = FALSE, name = as.character(substitute(x))) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  if (!is.list(x)) stop("Input is not a list.", call. = FALSE)
  if (isTRUE(named)) {
    if (is.null(names(x))) {
      stop(sprintf("`%s` should be named list.", name), call. = FALSE)
    }
  }
  if (!is.null(len)) {
    if (length(x) != len) {
      stop(sprintf("`%s` must be of length %s.", name, len), call. = FALSE)
    }
  }
}

assert_function = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  if (!is.function(x)) {
    msg = sprintf("`%s` must be a function.", name)
    stop(msg, call. = FALSE)
  }
}

check_atomic_vector = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  if (isTRUE(null.ok) && is.null(x)) {
    return(invisible(TRUE))
  }
  # doesn't work on glue::glue() output
  # flag = is.atomic(x) && is.vector(x) && !is.list(x)
  flag = is.atomic(x) && is.null(dim(x)) && length(x) > 0 && !is.list(x)
  if (flag) {
    out = TRUE
  } else if (is.factor(x) && is.null(dim(x))) {
    out = TRUE
  } else {
    out = sprintf("`%s` must be an atomic vector.", name)
  }
  return(out)
}

assert_atomic_vector = function(x, null.ok = FALSE, name = as.character(substitute(x))) {
  flag = check_atomic_vector(x, null.ok = null.ok, name = name)
  if (!isTRUE(flag)) {
    stop(flag, call. = FALSE)
  } else {
    return(invisible(TRUE))
  }
}

assert_class = function(x, classname) {
  if (!inherits(x, classname)) {
    msg = sprintf("`x` must be of class `%s`.", classname)
    stop(msg, call. = FALSE)
  }
}

#' @title Retrieve the saved graphical parameters
#'   
#' @description Convenience function for retrieving the graphical parameters
#'   (i.e., the full list of `tag = value` pairs held in
#'   \code{\link[graphics]{par}}) from either immediately before or
#'   immediately after the most recent [tinyplot] call.
#'
#' @param when character. From when should the saved parameters be retrieved?
#'   Either "before" (the default) or "after" the preceding `tinyplot` call.
#'   
#' @details A potential side-effect of [tinyplot] is that it can change a user's
#'   \code{\link[graphics]{par}} settings. For example, it may adjust the inner
#'   and outer plot margins to make space for an automatic legend; see
#'   [draw_legend]. While it is possible to immediately restore the original
#'   \code{\link[graphics]{par}} settings upon exit via the
#'   `tinyplot(..., restore.par = TRUE)` argument, this is not the default
#'   behaviour. The reason being that we need to preserve the adjusted parameter
#'   settings in case users want to add further graphical annotations to their
#'   plot (e.g., \code{\link[graphics]{abline}}, \code{\link[graphics]{text}},
#'   etc.) Nevertheless, it may still prove desirable to recall and reset these
#'   original graphical parameters after the fact (e.g., once all these extra
#'   annotations have been added). That is the purpose of this [get_saved_par]
#'   function.
#'   
#'   Of course, users may prefer to manually capture and reset graphical
#'   parameters, as per the standard method described in the
#'   \code{\link[graphics]{par}} documentation. For example:
#'   
#'   ```
#'   op = par(no.readonly = TRUE)  # save current par settings 
#'   # <do lots of (tiny)plotting>
#'   par(op)                       # reset original pars
#'   ```
#'   
#'   This standard manual approach may be safer than [get_saved_par] because it
#'   offers more precise control. Specifically, the value of [get_saved_par] 
#'   itself will be reset after ever new [tinyplot] call; i.e. it may inherit an
#'   already-changed set of parameters. Users should bear these trade-offs in
#'   mind when deciding which approach to use. As a general rule,
#'   [get_saved_par] offers the convenience of resetting the original
#'   \code{\link[graphics]{par}} settings even if a user forgot to save them
#'   beforehand. But one should avoid invoking it after a series of consecutive
#'   [tinyplot] calls.
#'   
#'   Finally, note that users can always call \code{\link[grDevices]{dev.off}}
#'   to reset all \code{\link[graphics]{par}} settings to their defaults.
#' 
#' @returns A list of \code{\link[graphics]{par}} settings.
#' 
#' @examples
#' #
#' # Contrived example where we draw a grouped scatterplot with a legend and
#' # manually add corresponding best fit lines for each group...
#' #
#' 
#' # First draw the grouped scatterplot
#' tinyplot(Sepal.Length ~ Petal.Length | Species, iris)
#' 
#' # Preserving adjusted par settings is good for adding elements to our plot
#' for (s in levels(iris$Species)) {
#'   abline(
#'     lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
#'     col = which(levels(iris$Species)==s)
#'   )
#' }
#' 
#' # Get saved par from before the preceding tinyplot call (but don't use yet)
#' sp = get_saved_par("before")
#' 
#' # Note the changed margins will affect regular plots too, which is probably
#' # not desirable
#' plot(1:10)
#' 
#' # Reset the original parameters (could use `par(sp)` here)
#' tpar(sp)
#' # Redraw our simple plot with our corrected right margin
#' plot(1:10)
#' 
#' #
#' # Quick example going the other way, "correcting" for par.restore = TRUE...
#' #
#' 
#' tinyplot(Sepal.Length ~ Petal.Length | Species, iris, restore.par = TRUE)
#' # Our added best lines will be wrong b/c of misaligned par
#' for (s in levels(iris$Species)) {
#'   abline(
#'     lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
#'     col = which(levels(iris$Species)==s), lty = 2
#'   )
#' }
#' # grab the par settings from the _end_ of the preceding tinyplot call to fix
#' tpar(get_saved_par("after"))
#' # now the best lines are correct
#' for (s in levels(iris$Species)) {
#'   abline(
#'     lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
#'     col = which(levels(iris$Species)==s)
#'   )
#' }
#' 
#' # reset again to original saved par settings before exit
#' tpar(sp)
#'
#' @export
get_saved_par = function(when = c("before", "after", "first")) {
  when = match.arg(when)
  par_env_name = paste0(".saved_par_", when)
  return(get(par_env_name, envir = get(".tinyplot_env", envir = parent.env(environment()))))
}

# (non-exported) companion function(s) for setting the original pars
set_saved_par = function(when = c("before", "after", "first"), value) {
  when = match.arg(when)
  par_env_name = paste0(".saved_par_", when)
  assign(par_env_name, value, envir = get(".tinyplot_env", envir = parent.env(environment())))
}

sanitize_xylab = function(settings) {
  env2env(
    settings,
    environment(),
    c(
      "type",
      "x", "xlab", "x_dep", "xmin_dep", "xmax_dep",
      "y", "ylab", "y_dep", "ymin_dep", "ymax_dep"
    )
  )

  out_xlab = NULL
  out_ylab = NULL

  is_boxplot = type %in% c("boxplot")
  is_density = type %in% c("density")
  is_frequency = type %in% c("histogram", "barplot", "function")
  is_function = type %in% c("function")
  is_range = type %in% c("rect", "segments", "pointrange")
  is_ribbon = type %in% c("ribbon")
  is_index = !is_frequency && !is_ribbon && !is_density

  ##### xlab
  if (!is.null(xlab)) {
    out_xlab = xlab
  } else if (!is.null(xmin_dep) && !is.null(xmax_dep)) {
    out_xlab = sprintf("[%s, %s]", xmin_dep, xmax_dep)
  } else if (is_boxplot && is.null(y)) {
    out_xlab = ""
  } else if (is_index && is.null(y) && !is.null(x)) {
    out_xlab = "Index"
  } else {
    out_xlab = x_dep
  }

  ##### ylab
  if (!is.null(ylab)) {
    out_ylab = ylab
  } else if (is_frequency && is.null(y) && !is.null(x)) {
    out_ylab = if (type == "barplot") "Count" else "Frequency"
  } else if (is_density && is.null(y) && !is.null(x)) {
    out_ylab = "Density"
  } else if (is_ribbon) {
    if (!is.null(y_dep)) {
      out_ylab = y_dep
    } else if (!is.null(ymin_dep) && !is.null(ymax_dep)) {
      out_ylab = sprintf("[%s, %s]", ymin_dep, ymax_dep)
    }
  } else if ((is_range || is_ribbon) && !is.null(ymin_dep) && !is.null(ymax_dep)) {
    out_ylab = sprintf("[%s, %s]", ymin_dep, ymax_dep)
  } else if (!is.null(y_dep)) {
    out_ylab = y_dep
  } else if (is.null(y) && !is.null(x_dep)) {
    out_ylab = x_dep
  } else {
    out_ylab = NULL
  }

  settings$xlab = out_xlab
  settings$ylab = out_ylab
}

#' Plot a function
#'
#' @details
#' When using `type_function()` in a `tinyplot()` call, the `x` value indicates
#' the range of values to plot on the x-axis.
#'
#' @param fun Function of `x` to plot. Defaults to \code{\link[stats]{dnorm}}.
#' @param args List of additional arguments to be passed to `fun`.
#' @param n Number of points to interpolate on the x axis.
#' @param ... Additional arguments are passed to the `lines()` function,
#' ex: `type="p"`, `col="pink"`.
#' @importFrom stats dnorm
#'
#' @examples
#' # Plot the normal density (default function)
#' tinyplot(x = -4:4, type = "function")
#' # tinyplot(x = -4:4, type = type_function()) # same
#'
#' # Customize by passing explicit arguments to your function
#' tinyplot(x = -1:10, type = type_function(
#'   fun = dnorm, args = list(mean = 3)
#' ))
#'
#' # Additional arguments are passed to the `lines()` function.
#' tinyplot(x = -4:4, type = type_function(
#'   fun = dnorm,
#'   col = "pink", type = "p", pch = 3
#' ))
#'
#' # Custom function example
#' ## (Here using `function(x)`, but you could also use the shorter `\(x)`
#' ## anonymous function syntax introduced in R 4.1.0)
#' tinyplot(x = -4:4, type = type_function(fun = function(x) 0.5 * exp(-abs(x))))
#'
#' @export
type_function = function(fun = dnorm, args = list(), n = 101, ...) {
  assert_function(fun)
  lines_args = list(...)
  data_function = function(args, fun) {
    funky = function(settings, ...) {
      env2env(settings, environment(), c("xlim", "ylim", "datapoints"))

      if (nrow(datapoints) == 0 || !"x" %in% names(datapoints)) {
        stop("Need to provide x values to plot the function.", call. = FALSE)
      }
      if (is.null(xlim)) {
        xlim = range(datapoints[["x"]])
      }
      if (is.null(ylim)) {
        tmp = c(list(datapoints[["x"]]), args)
        tmp = range(tmp)
        tmp = seq(tmp[1], tmp[2], length.out = n)
        tmp = c(list(tmp), args)
        tmp = do.call(fun, tmp)
        ylim = c(min(tmp), max(tmp))
      }
      env2env(environment(), settings, c("xlim", "ylim"))
    }
  }
  draw_function = function() {
    funky = function(ifacet, data_facet, ...) {
      xrange = range(data_facet[[ifacet]][["x"]])
      x = seq(xrange[1], xrange[2], length.out = n)
      y = do.call(fun, c(list(x), args))
      tmp = c(list(x, y), lines_args)
      do.call(lines, tmp)
    }
    return(funky)
  }
  out = list(
    draw = draw_function(),
    data = data_function(args = args, fun = fun),
    name = "function"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#
## main orchestration function -----
#

by_aesthetics = function(settings) {
  env2env(
    settings,
    environment(),
    c(
      "datapoints", "by", "type", "null_by", "pch", "bg", "lty", "lwd",
      "bubble", "cex", "alpha", "col", "fill", "ribbon.alpha", "lighten"
    )
  )

  # Detect grouping characteristics
  by_ordered = FALSE
  by_continuous = !null_by && inherits(datapoints$by, c("numeric", "integer"))
  if (isTRUE(by_continuous) && type %in% c("l", "b", "o", "ribbon", "polygon", "polypath", "boxplot", "chull")) {
    # Only warn if a legend would actually be drawn: the reversion to a discrete
    # legend still needs to happen for correct grouping, but it's not worth
    # flagging when the user has suppressed the legend anyway (#656).
    if (!isFALSE(settings$legend)) {
      warning("\nContinuous legends not supported for this plot type. Reverting to discrete legend.")
    }
    by_continuous = FALSE
  } else if (!null_by) {
    by_ordered = is.ordered(by)
  }

  ngrps = if (null_by) {
    1L
  } else if (is.factor(by)) {
    nlevels(by)
  } else if (by_continuous) {
    100L
  } else {
    length(unique(by))
  }

  pch = by_pch(ngrps = ngrps, type = type, pch = pch)
  lty = by_lty(ngrps = ngrps, type = type, lty = lty)
  lwd = by_lwd(ngrps = ngrps, type = type, lwd = lwd)
  cex = by_cex(ngrps = ngrps, type = type, bubble = bubble, cex = cex)

  col = by_col(
    col = col,
    palette = settings$palette, # must use unevaluated palette
    alpha = alpha,
    by_ordered = by_ordered,
    by_continuous = by_continuous,
    ngrps = ngrps,
    adjustcolor = adjustcolor
  )

  bg = by_bg(
    bg = bg,
    fill = fill,
    col = col,
    palette = settings$palette, # must use unevaluated palette
    alpha = alpha,
    by_ordered = by_ordered,
    by_continuous = by_continuous,
    ngrps = ngrps,
    type = type,
    by = by,
    ribbon.alpha = ribbon.alpha,
    lighten = lighten,
    adjustcolor = adjustcolor
  )

  # update settings
  env2env(
    environment(),
    settings,
    c("by_continuous", "by_ordered", "ngrps", "pch", "lty", "lwd", "cex", "col", "bg")
  )
}


#
## subsidiary functions -----
#

by_col = function(col, palette, alpha, by_ordered, by_continuous, ngrps, adjustcolor) {
  ordered = if (is.null(by_ordered)) FALSE else by_ordered
  gradient = if (is.null(by_continuous)) FALSE else by_continuous
  assert_logical(ordered)
  assert_logical(gradient)

  if (is.null(alpha)) alpha = 1
  if (gradient) ngrps = 100L

  if (is_by_keyword(col)) col = NULL

  pal_theme = if (ordered || gradient) {
    get_tpar("palette.sequential", default = NULL)
  } else {
    get_tpar("palette.qualitative", default = NULL)
  }

  # Resolve a (possibly relative) col.default against the qualitative palette.
  # A relative (negative) index also drops the chosen colour from the grouped
  # palette, so e.g. `col.default = -1` uses the palette's leading colour for
  # single-group displays and the palette-minus-that-colour for grouped ones
  # (see resolve_col_default). col.default only applies to qualitative displays.
  if (!ordered && !gradient) {
    cd = resolve_col_default(get_tpar("col.default", default = NULL), pal_theme)
    pal_theme = cd[["palette"]]
    # Single-group default colour (theme-settable via tpar). When NULL, defer to
    # the usual palette resolution below (first qualitative colour, or palette()[1]).
    if (is.null(col) && ngrps == 1L) {
      col = cd[["col_default"]]
    }
  }

  cols = resolve_manual_colors(col, ngrps, gradient, ordered, alpha, adjustcolor)
  if (!is.null(cols)) {
    return(cols)
  }

  cols = resolve_palette_colors(
    palette = palette,
    theme_palette = pal_theme,
    ngrps = ngrps,
    ordered = ordered,
    gradient = gradient,
    alpha = alpha,
    adjustcolor = adjustcolor
  )

  cols
}


by_bg = function(bg, fill, col, palette, alpha, by_ordered, by_continuous, ngrps, type, by, ribbon.alpha, lighten, adjustcolor) {
  lighten = if (is.null(lighten)) TRUE else isTRUE(lighten)
  if (is.null(bg) && !is.null(fill)) bg = fill

  # A numeric bg/fill in [0,1] is an alpha *request*, not a colour: remember it
  # as the transparency to layer on top of the resolved (and possibly lightened)
  # base fill, and defer to the "by" palette for the base colour itself. Applying
  # alpha last -- after lightening -- means a fill request blends a light tint
  # toward the background rather than darkening a saturated base (issue #646).
  if (!is.null(bg) && length(bg) == 1 && is.numeric(bg) && bg >= 0 && bg <= 1) {
    alpha = bg
    bg = "by"
  }

  ordered = if (is.null(by_ordered)) FALSE else by_ordered
  gradient = if (is.null(by_continuous)) FALSE else by_continuous

  # Whether the base fill comes from palette resolution (the "by" keyword, incl.
  # a numeric fill request) or the single-group default, vs. an explicit colour
  # such as `fill = "bisque"`. Only the former are eligible for lightening; an
  # explicit colour is always honoured verbatim.
  bg_from_palette = !is.null(bg) && length(bg) == 1 && is_by_keyword(bg)
  palette_fill = bg_from_palette

  if (bg_from_palette) {
    pal_theme = if (ordered || gradient) {
      get_tpar("palette.sequential", default = NULL)
    } else {
      get_tpar("palette.qualitative", default = NULL)
    }
    # Mirror by_col(): a relative col.default trims the grouped palette, so
    # grouped fills stay in step with grouped line colours.
    if (!ordered && !gradient) {
      pal_theme = resolve_col_default(get_tpar("col.default", default = NULL), pal_theme)[["palette"]]
    }
    # Resolve opaque (alpha = 1); any user alpha is layered on below.
    bg = resolve_palette_colors(
      palette = palette,
      theme_palette = pal_theme,
      ngrps = ngrps,
      ordered = ordered,
      gradient = gradient,
      alpha = 1,
      adjustcolor = adjustcolor
    )
  } else if (ngrps == 1L && is.null(bg) && type %in% c("boxplot", "violin", "barplot", "histogram")) {
    # Single-group fill tracks the theme's *default* colour (col.default ->
    # palette[1] -> black) so that themed box/violin/bar/histogram plots match
    # their own multi-group counterparts. We resolve this default independently
    # of `col` so that a user-supplied outline colour (e.g. `col = "white"`)
    # doesn't bleed into the fill.
    bg = by_col(
      col = NULL, palette = palette, alpha = 1, by_ordered = by_ordered,
      by_continuous = by_continuous, ngrps = 1L, adjustcolor = adjustcolor
    )
    palette_fill = TRUE
  } else if (length(bg) != ngrps) {
    bg = rep(bg, ngrps)
  }

  # The solid categorical area types take the lighter-but-opaque tint, but only
  # for qualitative groupings: lightening a sequential/continuous ramp would
  # crush the contrast between adjacent levels.
  area_light_types = c("barplot", "boxplot", "violin")
  light_ok = lighten && type %in% area_light_types && !ordered && !gradient

  if (palette_fill) {
    # For a *chromatic* base the fill is a lighter-but-opaque tint of that colour
    # (via seq_palette, the same HCL ramp used for ridge fills and legend
    # swatches), so it reads cleanly over grid lines unlike alpha blending. For
    # an *achromatic* base (typically black, e.g. the "bw"/"classic"/"ipsum"
    # themes and the plain default) seq_palette's light endpoint can't reach the
    # neutral "lightgray" used by the no-theme path and base R's hist()/boxplot()
    # -- so we fall back to that literal. Histograms only lighten in the
    # single-group case; grouped histograms keep their semi-transparent fill so
    # that overlapping distributions remain legible (issue #646).
    if (light_ok) {
      bg = lighten_fill(bg)
    } else if (type == "histogram" && ngrps == 1L) {
      bg = lighten_fill(bg)
    }
    # Layer any user-supplied transparency on top of the (lightened) base fill.
    if (!is.null(alpha)) bg = apply_alpha(bg, alpha, adjustcolor)
  }

  # Semi-transparent fill for ribbon-type plots (the default look for ribbons).
  # Other area types only become transparent if the user explicitly asks (e.g.
  # `fill = 0.3`), which is handled via the `alpha` layering above.
  if (type == "ribbon") {
    if (!is.null(bg)) {
      bg = adjustcolor(bg, ribbon.alpha)
    } else if (!is.null(col)) {
      bg = adjustcolor(col, ribbon.alpha)
    }
  }

  bg
}


by_pch = function(ngrps, type, pch = NULL) {
  no_pch = FALSE
  if (identical(type, "text")) {
    pch = rep(15, ngrps)
  } else if (!type %in% c("p", "b", "o", "pointrange", "errorbar", "boxplot", "qq")) {
    no_pch = TRUE
    pch = NULL

    # special "by" convenience keyword
  } else if (is_by_keyword(pch)) {
    no_pch = TRUE # skip checks below
    pch = 1:ngrps + par("pch") - 1
    # correctly recycle if over max pch type
    if (max(pch) > 25L) {
      pch_below = pch[pch <= 25L]
      pch_above = pch[pch > 25L]
      pch_above = rep_len(0:25, length(pch_above))
      pch = c(pch_below, pch_above)
    }

    # return NULL if not a valid point type
  } else if (is.null(pch)) {
    pch = par("pch")
  }

  if (!no_pch) {
    assert_len_1_or_ngrps(pch, ngrps, "pch", allow_character = TRUE)
    if (length(pch) == 1) pch = rep(pch, ngrps)
  }

  return(pch)
}


by_lty = function(ngrps, type, lty = NULL) {
  # We only care about line types, otherwise return NULL
  if (!type %in% c("l", "b", "o", "c", "h", "s", "S", "ribbon", "barplot", "boxplot", "rect", "spineplot", "segments", "qq", "abline", "hline", "vline")) {
    lty = NULL

    # special "by" convenience keyword
  } else if (is_by_keyword(lty)) {
    lty_dict = c("solid", "dashed", "dotted", "dotdash", "longdash", "twodash")
    par_lty = par("lty")

    if (!par_lty %in% lty_dict) {
      warning(
        "\nBespoke lty specifications (i.e., using string combinations) are not ",
        "currently supported alongside the lty='by' keyword argument. ",
        "Defaulting to 1 and looping from there.\n"
      )
      par_lty = 1
    } else {
      par_lty = which(par_lty == lty_dict)
    }
    lty = 1:ngrps + par_lty - 1
    # correctly recycle if over max lty type
    if (max(lty) > 6L) {
      lty_below = lty[lty <= 6L]
      lty_above = lty[lty > 6L]
      lty_above = rep_len(1:6, length(lty_above))
      lty = c(lty_below, lty_above)
    }

    # NULL -> solid (or default) line
  } else if (is.null(lty)) {
    if (!identical(type, "boxplot")) {
      lty = rep(par("lty"), ngrps)
    }

    # atomic vector: sanity check length
  } else if (is.atomic(lty) && is.vector(lty)) {
    assert_len_1_or_ngrps(lty, ngrps, "lty")
    if (length(lty) == 1) lty = rep(lty, ngrps)
  }

  lty
}


by_lwd = function(ngrps, type, lwd = NULL) {
  lwd_base = par("lwd")
  lwd_floor = lwd_base / min(5, max((ngrps - 1), 1))
  lwd_ceiling = lwd_base * min(5, ngrps)

  no_lwd = FALSE
  # special "by" convenience keyword
  if (is_by_keyword(lwd)) {
    no_lwd = TRUE # skip checks below
    lwd = seq(lwd_floor, lwd_ceiling, length.out = ngrps)
  } else if (is.null(lwd)) {
    no_lwd = TRUE
    lwd = NULL
  }

  if (!no_lwd) {
    assert_len_1_or_ngrps(lwd, ngrps, "lwd")
    if (length(lwd) == 1) lwd = rep(lwd, ngrps)
  }

  return(lwd)
}


by_cex = function(ngrps, type, bubble = FALSE, cex = NULL) {
  no_cex = FALSE
  # special "by" convenience keyword
  if (is_by_keyword(cex)) {
    no_cex = TRUE # skip checks below
    cex = rescale_num(c(1:ngrps), to = c(1, 2.5))
  } else if (is.null(cex)) {
    no_cex = TRUE
    # cex = NULL
    # can't leave cex as NULL otherwise JIT cex_fct_adj adjustment in
    # draw_legend() won't work later
    cex = 1
    cex = rep(cex, ngrps)
  }

  # placehodler
  if (bubble) no_cex = TRUE

  if (!no_cex) {
    assert_len_1_or_ngrps(cex, ngrps, "cex")
    if (length(cex) == 1) cex = rep(cex, ngrps)
  }

  return(cex)
}


#
## helper functions -----
#

apply_alpha = function(cols, alpha, adjustcolor) {
  if (is.null(cols) || is.null(alpha) || identical(alpha, 0)) {
    return(cols)
  }
  adjustcolor(cols, alpha.f = alpha)
}

# Apply the lighter-but-opaque tint (PR #614) to each colour in a vector.
lighten_fill = function(cols) {
  if (is.null(cols)) return(cols)
  vapply(
    cols,
    function(cc) if (is_achromatic(cc)) "lightgray" else seq_palette(cc, n = 3)[3],
    character(1),
    USE.NAMES = FALSE
  )
}

is_by_keyword = function(x) {
  is.character(x) && length(x) == 1 && !is.na(x) && identical(x, "by")
}

warn_recycle_colors = function(ncols, ngrps) {
  warning(
    "\nFewer colours (", ncols, ") provided than there are groups (",
    ngrps, "). Recycling to make up the shortfall."
  )
}

expand_colors_to_ngrps = function(values, ngrps, gradient) {
  if (length(values) == 1) {
    return(rep(values, ngrps))
  }
  if (length(values) >= ngrps) {
    return(values[seq_len(ngrps)])
  }
  if (gradient) {
    return(colorRampPalette(colors = values, alpha = TRUE)(ngrps))
  }
  warn_recycle_colors(length(values), ngrps)
  rep_len(values, ngrps)
}

assert_len_1_or_ngrps = function(x, ngrps, name, allow_character = FALSE) {
  types = if (allow_character) "numeric or character" else "numeric"
  valid_type = is.numeric(x) || (allow_character && is.character(x))
  valid = is.atomic(x) && is.vector(x) && valid_type && (length(x) == 1 || length(x) == ngrps)
  if (!valid) {
    stop(sprintf("`%s` must be `NULL`, or a %s vector of length 1 or %s.", name, types, ngrps), call. = FALSE)
  }
}

match_palette_name = function(name, candidates) {
  normalize = function(x) tolower(gsub("[-, _, \\,, (, ), \\ , \\.]", "", x))
  charmatch(normalize(name), normalize(candidates))
}

## Resolve a palette spec to its full concrete colour vector (or NULL if it
## can't be resolved to a discrete set, e.g. a continuous hcl palette). Used to
## support relative `col.default` indices.
resolve_palette_to_colors = function(palette) {
  if (is.null(palette)) return(NULL)
  if (is.character(palette) && length(palette) > 1) return(unname(palette))
  if (is.character(palette) && length(palette) == 1) {
    discrete_pals = palette.pals()
    idx = match_palette_name(palette, discrete_pals)
    if (!is.na(idx) && idx >= 1L) {
      return(unname(palette.colors(palette = discrete_pals[idx])))
    }
  }
  NULL
}

## Resolve a (possibly relative) `col.default` against the qualitative palette.
## `col_default` may be NULL, a character colour, or a numeric index into the
## palette. A negative index additionally *drops* that colour from the palette
## returned for grouped displays. So `col.default = -1` with an "Okabe-Ito"
## palette uses black (the leading colour) for single-group plots and an
## Okabe-Ito-minus-black palette for grouped plots, avoiding the need to keep a
## separate "no-black" palette copy around (#598, #614). Returns a list with the
## resolved single-group `col_default` and the (possibly trimmed) `palette`.
resolve_col_default = function(col_default, palette_spec) {
  if (!is.numeric(col_default)) {
    return(list(col_default = col_default, palette = palette_spec))
  }
  full = resolve_palette_to_colors(palette_spec)
  if (is.null(full)) {
    return(list(col_default = NULL, palette = palette_spec))
  }
  i = as.integer(col_default)
  out_col = full[abs(i)]
  if (i < 0L) palette_spec = full[-abs(i)]
  list(col_default = out_col, palette = palette_spec)
}

## Handle direct color input via `col` arg. Returns colors or NULL if not applicable.
resolve_manual_colors = function(col, ngrps, gradient, ordered, alpha, adjustcolor) {
  if (is.null(col) || !is.atomic(col) || !is.vector(col)) {
    return(NULL)
  }

  cols = col
  if (length(cols) == 1) {
    cols = rep(cols, ngrps)
  } else if (length(cols) < ngrps) {
    cols = expand_colors_to_ngrps(cols, ngrps, gradient)
  }

  # Map numeric indices to palette colors (unless ordered)
  if (!ordered && is.numeric(cols)) {
    base_pal = palette()
    cols = if (ngrps <= length(base_pal)) {
      base_pal[cols]
    } else {
      hcl.colors(max(cols))[cols]
    }
  }

  if (gradient) cols = rev(cols)
  apply_alpha(cols, alpha, adjustcolor)
}

## High-level palette resolution: theme fallback, defaults, then delegate to resolve_palette_spec.
resolve_palette_colors = function(palette, theme_palette, ngrps, ordered, gradient, alpha, adjustcolor) {
  palette_choice = palette

  # Pick theme palette if no explicit palette provided
  if (is.null(palette_choice) && !is.null(theme_palette)) {
    palette_choice = theme_palette
    if (is.character(theme_palette) && length(theme_palette) == 1) {
      # Check if theme palette needs to switch to sequential
      use_sequential = FALSE
      idx = match_palette_name(theme_palette, palette.pals())
      if (!is.na(idx) && idx >= 1L) {
        max_colors = length(palette.colors(palette = palette.pals()[idx]))
        use_sequential = ngrps >= max_colors || ordered
      } else {
        idx = match_palette_name(theme_palette, hcl.pals())
        use_sequential = !is.na(idx) && idx >= 1L && gradient
      }
      if (use_sequential) {
        palette_choice = get_tpar("palette.sequential", default = NULL)
      }
    }
  }

  if (is.null(palette_choice)) {
    # Default palette selection (alpha applied at end)
    base_pal = palette()
    if (ngrps <= length(base_pal) && !ordered && !gradient) {
      cols = base_pal[seq_len(ngrps)]
    } else if (ngrps <= 8 && !ordered) {
      cols = palette.colors(n = ngrps, palette = "R4")
    } else if (!gradient && !ordered) {
      cols = hcl.colors(n = ngrps, palette = "Viridis")
    } else {
      # Restricted viridis for gradient/ordered (excludes extreme ends)
      cols = colorRampPalette(
        hcl.colors(n = 100, palette = "Viridis")[11:90],
        alpha = TRUE
      )(ngrps)
    }
    cols = apply_alpha(cols, alpha, adjustcolor)
  } else {
    cols = resolve_palette_spec(
      palette = palette_choice,
      ngrps = ngrps,
      gradient = gradient,
      ordered = ordered,
      alpha = alpha,
      adjustcolor = adjustcolor
    )
  }

  if (gradient || ordered) cols = rev(cols)
  cols
}

## Parse palette arg (vector, string, call, or function) into colors.
resolve_palette_spec = function(palette, ngrps, gradient, ordered, alpha, adjustcolor) {
  cols = NULL
  if (is.character(palette) && length(palette) > 1) {
    # Direct color vector
    cols = palette
  } else if (is.character(palette)) {
    # Named palette string - try palette.pals() then hcl.pals()
    discrete_pals = palette.pals()
    idx = match_palette_name(palette, discrete_pals)

    if (!is.na(idx)) {
      if (idx < 1L) stop("'palette' is ambiguous")
      matched_name = discrete_pals[idx]
      max_colors = length(palette.colors(palette = matched_name))

      if (gradient) {
        cols = colorRampPalette(palette.colors(palette = matched_name))(ngrps)
      } else if (ngrps >= max_colors) {
        warn_recycle_colors(max_colors, ngrps)
        cols = palette.colors(n = ngrps, palette = matched_name, recycle = TRUE)
      } else {
        cols = palette.colors(n = ngrps, palette = matched_name)
      }
    } else {
      hcl_pals = hcl.pals()
      idx = match_palette_name(palette, hcl_pals)
      if (!is.na(idx)) {
        if (idx < 1L) stop("'palette' is ambiguous")
        cols = hcl.colors(n = ngrps, palette = palette)
      } else {
        stop(
          "\nPalette string not recognized. Must be a value produced by either ",
          "`palette.pals()` or `hcl.pals()`.\n",
          call. = FALSE
        )
      }
    }
  } else if (inherits(palette, c("call", "name"))) {
    # Expression or symbol
    if (inherits(palette, "name")) {
      eval_palette = tryCatch(eval(palette), error = function(e) NULL)
      if (is.character(eval_palette)) {
        cols = eval_palette
      }
    }
    if (is.null(cols)) {
      args = as.list(palette)
      fun_name = paste(args[[1]])
      args[[1]] = NULL
      if (fun_name %in% c("c", "list")) {
        cols = unlist(args, recursive = TRUE, use.names = FALSE)
      } else {
        args[["n"]] = ngrps
        if (any(names(args) == "")) args[which(names(args) == "")] = NULL
        cols = tryCatch(
          do.call(fun_name, args),
          error = function(e) do.call(eval(palette), args)
        )
      }
    }
  } else if (inherits(palette, "function")) {
    cols = palette(ngrps)
  } else {
    stop(
      "\nInvalid palette argument. Must be a recognized keyword, or a ",
      "palette-generating function with named arguments.\n"
    )
  }

  # Uniform post-processing
  cols = expand_colors_to_ngrps(cols, ngrps, gradient)
  apply_alpha(cols, alpha, adjustcolor)
}

#' Lines plot type
#'
#' @description Type function for plotting lines.
#' 
#' @inheritParams graphics::plot.default
#' @inheritParams dodge_positions
#' 
#' @examples
#' # "l" type convenience character string
#' tinyplot(circumference ~ age | Tree, data = Orange, type = "l")
#' 
#' # Use `type_lines()` to pass extra arguments for customization
#' tinyplot(circumference ~ age | Tree, data = Orange, type = type_lines(type = "s"))
#' 
#' # Direct legend labels are a good option for grouped lined plots (assuming
#' # there aren't too many groups and the data are sorted along the x-axis)
#' tinyplot(
#'   circumference ~ age | Tree, data = Orange, type = "l",
#'   legend = "direct"
#' )
#' 
#' # Fancier version(s) that use a theme and repel overlapping labels
#' Orange2 = transform(Orange, Tree = paste("Tree", Tree))
#' tinyplot(
#'   circumference ~ age | Tree, data = Orange2, type = "l",
#'   legend = list("direct", repel = TRUE), # auto repel
#'   theme = "socviz"
#' )
#' tinyplot(
#'   circumference ~ age | Tree, data = Orange2, type = "l",
#'   legend = list("direct", nudge_y = c("Tree 1" = 3, "Tree 3" = -5)), # manual
#'   theme = "socviz"
#' )
#' 
#' @export
type_lines = function(type = "l", dodge = 0, fixed.dodge = FALSE) {
  out = list(
    draw = draw_lines(type = type),
    data = data_lines(dodge = dodge, fixed.dodge = fixed.dodge),
    name = type
  )
  class(out) = "tinyplot_type"
  return(out)
}


data_lines = function(dodge = 0, fixed.dodge = FALSE) {
  fun = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "xlabs"))

    if (is.character(datapoints$x)) {
      datapoints$x = as.factor(datapoints$x)
    }
    if (is.factor(datapoints$x)) {
      # honour pre-ordered factors; otherwise fall back to first-appearance order
      xlvls = if (is.ordered(datapoints$x)) levels(datapoints$x) else unique(datapoints$x)
      datapoints$x = factor(datapoints$x, levels = xlvls)
      xlabs = seq_along(xlvls)
      names(xlabs) = xlvls
      datapoints$x = as.integer(datapoints$x)
    }

    # dodge
    if (dodge != 0) {
      datapoints = dodge_positions(datapoints, dodge, fixed.dodge)
    }

    x = datapoints$x
    env2env(environment(), settings, c(
      "x",
      "xlabs",
      "datapoints"
    ))
  }
  fun
}


draw_lines = function(type = "l") {
    fun = function(ix, iy, icol, ipch, ibg, ilty, ilwd, icex = 1, ...) {
        lines(
            x = ix,
            y = iy,
            col = icol,
            type = type,
            pch = ipch,
            bg = ibg,
            lty = ilty,
            lwd = ilwd,
            cex = icex
        )
    }
    return(fun)
}

#' Prepare multi-legend specifications
#'
#' @description Sets up multiple legend specifications for multi-legends
#'   (e.g., color grouping + bubble size). Creates `lgby` and `lgbub` objects
#'   that will be passed to draw_multi_legend().
#'
#' @param settings Settings environment from tinyplot
#'
#' @returns NULL (modifies settings environment in-place)
#'
#' @keywords internal
prepare_legend_multi = function(settings) {
  env2env(
    settings,
    environment(),
    c(
      "legend",
      "legend_args",
      "by_dep",
      "lgnd_labs",
      "type",
      "pch",
      "lty",
      "lwd",
      "col",
      "bg",
      "by_continuous",
      "lgnd_cex",
      "cex_dep",
      "bubble_cex",
      "cex_fct_adj",
      "bubble_alpha",
      "bubble_bg_alpha",
      "has_sub"
    )
  )

  # Legend for grouping variable (by)
  lgby = list(
    legend_args = modifyList(
      legend_args,
      list(x.intersp = 1, y.intersp = 1),
      keep.null = TRUE
    ),
    by_dep = by_dep,
    lgnd_labs = lgnd_labs,
    type = type,
    pch = pch,
    lty = lty,
    lwd = lwd,
    col = col,
    bg = bg,
    gradient = by_continuous,
    cex = lgnd_cex,
    has_sub = has_sub
  )

  # Legend for bubble sizes
  lgbub = list(
    legend_args = modifyList(
      legend_args,
      list(title = cex_dep, ncol = 1),
      keep.null = TRUE
    ),
    lgnd_labs = names(bubble_cex),
    type = type,
    pch = pch,
    lty = lty,
    lwd = lwd,
    col = adjustcolor(par("col"), alpha.f = bubble_alpha),
    bg = adjustcolor(par("col"), alpha.f = bubble_bg_alpha),
    cex = bubble_cex * cex_fct_adj,
    has_sub = has_sub,
    draw = FALSE
  )

  env2env(environment(), settings, c("legend_args", "lgby", "lgbub"))
}


#
## Multi-Legend Rendering -----
#

#' Draw multiple legends with automatic positioning
#'
#' @description Handles multiple legends (e.g., color grouping + bubble size) by:
#'   1. Extracting dimensions from fake legends
#'   2. Calculating sub-positioning based on dimensions
#'   3. Drawing legends in ascending order of width (widest last)
#'
#' @md
#' @param legend_list A list of legend arguments, where each element is itself a
#'   list of arguments that can be passed on to [draw_legend]. Legends will be
#'   drawn vertically (top to bottom) in the order that they are provided. Note
#'   that we currently only support 2 legends, i.e. the top-level list has a
#'   maximum length of 2.
#' @param position String indicating the base keyword position for the
#'   multi-legend. Currently only `"right!"` and `"left!"` are supported.
#'
#' @returns No return value, called for side effect of drawing multiple legends.
#'
#' @seealso [draw_legend]
#'
#' @keywords internal
#'
#' @examples
#' \dontrun{
#' oldmar = par("mar")
#'
#' # Multi-legend example (color + bubble)
#'
#' l1 = list(
#'   lgnd_labs = c("Red", "Blue", "Green"),
#'   legend_args = list(title = "Colors"),
#'   pch = 16,
#'   col = c("red", "blue", "green"),
#'   type = "p"
#' )
#'
#' l2 = list(
#'   lgnd_labs = c("Tiny", "Small", "Medium", "Large", "Huge"),
#'   legend_args = list(title = "Size"),
#'   pch = 16,
#'   col = "black",
#'   cex = seq(0.5, 2.5, length.out = 5),
#'   type = "p"
#' )
#'
#' # Draw together
#' draw_multi_legend(list(l1, l2), position = "right!")
#'
#' par(mar = oldmar)
#' }
#'
#' @keywords internal
draw_multi_legend = function(
    legend_list,
    position = "right!"
) {

  # Validate inputs
  if (!is.list(legend_list) || length(legend_list) != 2) {
    stop("Currently only 2 legends are supported in multi-legend mode")
  }

  # Currently only support right!/left! positioning
  if (!grepl("right!$|left!$", position)) {
    warning(
      '\nMulti-legends currently only work with "right!" or "left!" keyword positioning.\n',
      'Reverting to "right!" default\n'
    )
    position = "right!"
  }

  # Determine sub-positions based on main position
  if (grepl("right!$", position)) {
    sub_positions = c("bottomright!", "topright!")
  } else if (grepl("left!$", position)) {
    sub_positions = c("bottomleft!", "topleft!")
  }

  # Assign positions of individual legends. Re-sanitize so the new per-legend
  # position string ("bottomright!" etc.) populates legend_args[["x"]], which
  # downstream code (e.g. build_legend_args) reads without further normalization.
  for (ll in seq_along(legend_list)) {
    legend_list[[ll]][["legend"]] = sub_positions[ll]
    legend_list[[ll]][["legend_args"]][["x"]] = NULL
    legend_list[[ll]][["legend_args"]] = sanitize_legend(
      sub_positions[ll], legend_list[[ll]][["legend_args"]]
    )
  }

  #
  ## Step 1: Extract legend dimensions (by drawing fake legends)
  #

  legend_dims = vector("list", length(legend_list))
  for (ll in seq_along(legend_list)) {
    legend_ll = legend_list[[ll]]
    legend_ll$new_plot = ll == 1  # Only draw new plot for first legend
    legend_ll$draw = FALSE
    legend_dims[[ll]] = do.call(draw_legend, legend_ll)
  }

  #
  ## Step 2: Calculate sub-positioning based on dimensions
  #

  # Extract dimensions
  lwidths = sapply(legend_dims, function(x) x$rect$w)
  lheights = sapply(legend_dims, function(x) x$rect$h)
  # For inset adjustment, default to 0.5 unless one or more of the two legends
  # is bigger than half the plot height.
  linset = if (any(lheights > 0.5)) lheights[2] / sum(lheights) else 0.5

  # Shared soma from widest legend so all legends use the same outer margin
  max_w = max(lwidths)
  lmar_vals = legend_list[[1]]$lmar %||% tpar("lmar")
  soma_target = (grconvertX(max_w, to = "lines") - grconvertX(0, to = "lines")) +
    sum(lmar_vals)

  # Determine ljust mode (shared across dual legends)
  ljust_mode = legend_list[[1]]$legend_args[["ljust"]] %||% tpar("ljust") %||% "left"
  ljust_mode = match.arg(ljust_mode, c("left", "center", "l", "c"))
  if (ljust_mode == "l") ljust_mode = "left"
  if (ljust_mode == "c") ljust_mode = "center"

  #
  ## Step 3: Reposition (via adjusted inset arg) and draw legends
  #

  # Note: we draw the legends in ascending order of width (i.e., widest legend
  #   last) in order to correctly set the overall plot dimensions.
  ## FIXME: current logic only works for "right!"/"left!" legend
  width_order = order(lwidths)

  # Quick idx for original order (needed for vertical legend placement)
  for (i in seq_along(legend_list)) legend_list[[i]]$idx = i

  for (o in seq_along(width_order)) {
    io = width_order[o]
    legend_o = legend_list[[io]]
    legend_o$new_plot = FALSE
    legend_o$draw = TRUE
    legend_o$soma_target = if (ljust_mode == "center") NULL else soma_target
    legend_o$legend_args$inset = c(0, 0)
    legend_o$legend_args$inset[1] = if (ljust_mode == "center") {
      if (o == 1) -abs(diff(lwidths)) / 2 else 0
    } else {
      0
    }
    legend_o$legend_args$inset[2] = if (legend_o$idx == 1) linset + 0.01 else 1 - linset + 0.01
    legend_o$idx = NULL
    do.call(draw_legend, legend_o)
  }

  invisible(NULL)
}

#' Polypath polygon type
#' 
#' @description Type function for plotting polygons.
#' Arguments are passed to \code{\link[graphics]{polypath}}.
#' 
#' @inheritParams graphics::polypath
#' 
#' @examples
#' # "polypath" type convenience character string
#' tinyplot(
#'     c(.1, .1, .6, .6, NA, .4, .4, .9, .9),
#'     c(.1, .6, .6, .1, NA, .4, .9, .9, .4),
#'     type = "polypath", fill = "grey"
#' )
#' 
#' # Use `type_polypath()` to pass extra arguments for customization
#' tinyplot(
#'     c(.1, .1, .6, .6, NA, .4, .4, .9, .9),
#'     c(.1, .6, .6, .1, NA, .4, .9, .9, .4),
#'     type = type_polypath(rule = "evenodd"), fill = "grey"
#' )
#' @export
type_polypath = function(rule = "winding") {
    out = list(
        draw = draw_polypath(rule = rule),
        data = data_polypath(),
        name = "polypath"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_polypath = function() {
  fun = function(settings, ...) {
    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
  }
  return(fun)
}


draw_polypath = function(rule = "winding") {
    fun =  function(ix, iy, icol, ibg, ilty, ilwd, dots, ...) {
        polypath(
            x = ix,
            y = iy,
            border = icol,
            col = ibg,
            lty = ilty,
            lwd = ilwd,
            rule = rule
        )
    }
    return(fun)
}

#' @title Lightweight extension of the base R plotting function
#'
#' @description
#' Enhances the base \code{\link[graphics]{plot}} function. Supported features
#' include automatic legends and facets for grouped data, additional plot types,
#' theme customization, and so on. Users can call either `tinyplot()`, or its
#' shorthand alias `plt()`.
#'
#' @md
#' @param x,y the x and y arguments provide the x and y coordinates for the
#'   plot. Any reasonable way of defining the coordinates is acceptable; most
#'   likely the names of existing vectors or columns of data frames. See the
#'   'Examples' section below, or the function
#'   \code{\link[grDevices]{xy.coords}} for details. If supplied separately, `x`
#'   and `y` must be of the same length.
#' @param xmin,xmax,ymin,ymax minimum and maximum coordinates of relevant area
#'   or interval plot types. Only used when the `type` argument is one of
#'   `"rect"` or `"segments"` (where all four min-max coordinates are required),
#'   or `"pointrange"`, `"errorbar"`, or `"ribbon"` (where only `ymin` and
#'   `ymax` required alongside `x`). Can be passed as unquoted (NSE) variable
#'   names when called from the `tinyplot.formula` method, e.g. `ymin = varname`
#'   where `varname` is a variable in `data`.
#' @param by grouping variable(s). The default behaviour is for groups to be
#'   represented in the form of distinct colours, which will also trigger an
#'   automatic legend. (See `legend` below for customization options.) However,
#'   groups can also be presented through other plot parameters (e.g., `pch`,
#'   `lty`, or `cex`) by passing an appropriate `"by"` keyword; see Examples.
#'   Note that continuous (i.e., gradient) colour legends are also supported if
#'   the user passes a numeric or integer to `by`. To group by multiple
#'   variables, wrap them with \code{\link[base]{interaction}}.
#' @param facet the faceting variable(s) that you want arrange separate plot
#'   windows by. Can be specified in various ways:
#'   - In "atomic" form, e.g. `facet = fvar`. To facet by multiple variables in
#'   atomic form, simply interact them, e.g.
#'   `interaction(fvar1, fvar2)` or `factor(fvar1):factor(fvar2)`.
#'   - As a one-sided formula, e.g. `facet = ~fvar`. Multiple variables can be
#'   specified in the formula RHS, e.g. `~fvar1 + fvar2` or `~fvar1:fvar2`. Note
#'   that these multi-variable cases are _all_ treated equivalently and
#'   converted to `interaction(fvar1, fvar2, ...)` internally. (No distinction
#'   is made between different types of binary operators, for example, and so
#'   `f1+f2` is treated the same as `f1:f2`, is treated the same as `f1*f2`,
#'   etc.)
#'   - As a two-side formula, e.g. `facet = fvar1 ~ fvar2`. In this case, the
#'   facet windows are arranged in a fixed grid layout, with the formula LHS
#'   defining the facet rows and the RHS defining the facet columns. At present
#'   only single variables on each side of the formula are well supported. (We
#'   don't recommend trying to use multiple variables on either the LHS or RHS
#'   of the two-sided formula case.)
#'   - As a special `"by"` convenience keyword, in which case facets will match
#'   the grouping variable(s) passed to `by` above.
#' @param facet.args an optional list of arguments for controlling faceting
#'   behaviour. (Ignored if `facet` is NULL.) Supported arguments are as
#'   follows:
#'   - `nrow`, `ncol` for overriding the default "square" facet window
#'   arrangement. Only one of these should be specified, but `nrow` will take
#'   precedence if both are specified together. Ignored if a two-sided formula
#'   is passed to the main `facet` argument, since the layout is arranged in a
#'   fixed grid.
#'   - `free` a logical value indicating whether the axis limits (scales) for
#'   each individual facet should adjust independently to match the range of
#'   the data within that facet. Default is `FALSE`. Separate free scaling of
#'   the x- or y-axis (i.e., whilst holding the other axis fixed) is not
#'   currently supported.
#'   - `fmar` a vector of form `c(b,l,t,r)` for controlling the base margin
#'   between facets in terms of lines. Defaults to the value of `tpar("fmar")`,
#'   which should be `c(1,1,1,1)`, i.e. a single line of padding around each
#'   individual facet, assuming it hasn't been overridden by the user as part
#'   their global \code{\link[tinyplot]{tpar}} settings. Note some automatic
#'   adjustments are made for certain layouts, and depending on whether the plot
#'   is framed or not, to reduce excess whitespace. See
#'   \code{\link[tinyplot]{tpar}} for more details.
#'   - `cex`, `font`, `col`, `bg`, `border` for adjusting the facet title text
#'   and background. Default values for these arguments are inherited from
#'   \code{\link[tinyplot]{tpar}} (where they take a "facet." prefix, e.g.
#'   `tpar("facet.cex")`). The latter function can also be used to set these
#'   features globally for all `tinyplot` plots.
#' @param formula a \code{\link[stats]{formula}} that optionally includes
#'   grouping variable(s) after a vertical bar, e.g. `y ~ x | z`. One-sided
#'   formulae are also permitted, e.g. `~ y | z`. Only a single `y` and `x`
#'   variable (if any) must be specified but multiple grouping variables
#'   can be included in different ways, e.g. `y ~ x | z1:z2` or
#'   `y ~ x | z1 + z2`. (These two representations are treated as equivalent;
#'   both are parsed as `interaction(z1, z2)` internally.) If arithmetic
#'   operators are used for transforming variables, they should be wrapped in
#'   `I()`, e.g., `I(y1/y2) ~ x`. Note that the `formula` and `x` arguments
#'   should not be specified in the same call.
#' @param data a data.frame (or list) from which the variables in formula
#'   should be taken. A matrix is converted to a data frame.
#' @param type character string or call to a `type_*()` function giving the
#'   type of plot desired.
#'   - NULL (default): Choose a sensible type for the type of `x` and `y` inputs
#'     (i.e., usually `"p"`).
#'   - 1-character values supported by \code{\link[graphics]{plot}}:
#'     - `"p"` Points
#'     - `"l"` Lines
#'     - `"b"` Both points and lines
#'     - `"c"` Empty points joined by lines
#'     - `"o"` Overplotted points and lines
#'     - `"s"` Stair steps
#'     - `"S"` Stair steps
#'     - `"h"` Histogram-like vertical lines
#'     - `"n"` Empty plot over the extent of the data
#'   - `tinyplot`-specific types. These fall into several categories:
#'     - Shapes:
#'       - `"area"` / [`type_area()`]: Plots the area under the curve from `y` = 0 to `y` = f(`x`).
#'       - `"errorbar"` / [`type_errorbar()`]: Adds error bars to points; requires `ymin` and `ymax`.
#'       - `"pointrange"` / [`type_pointrange()`]: Combines points with error bars.
#'       - `"polygon"` / [`type_polygon()`]: Draws polygons.
#'       - `"polypath"` / [`type_polypath()`]: Draws a path whose vertices are given in `x` and `y`.
#'       - `"rect"` / [`type_rect()`]: Draws rectangles; requires `xmin`, `xmax`, `ymin`, and `ymax`.
#'       - `"ribbon"` / [`type_ribbon()`]: Creates a filled area between `ymin` and `ymax`.
#'       - `"segments"` / [`type_segments()`]: Draws line segments between pairs of points.
#'       - `"text"` / [`type_text()`]: Add text annotations.
#'     - Visualizations:
#'       - `"barplot"` / [`type_barplot()`]: Creates a bar plot.
#'       - `"boxplot"` / [`type_boxplot()`]: Creates a box-and-whisker plot.
#'       - `"chull"` / [`type_chull()`]: Draws convex hull(s) around grouped points.
#'       - `"density"` / [`type_density()`]: Plots the density estimate of a variable.
#'       - `"histogram"` / [`type_histogram()`]: Creates a histogram of a single variable.
#'       - `"jitter"` / [`type_jitter()`]: Jittered points.
#'       - `"qq"` / [`type_qq()`]: Creates a quantile-quantile plot.
#'       - `"ridge"` / [`type_ridge()`]: Creates a ridgeline (aka joy) plot.
#'       - `"rug"` / [`type_rug()`]: Adds a rug to an existing plot.
#'       - `"spineplot"` / [`type_spineplot()`]: Creates a spineplot or spinogram.
#'       - `"violin"` / [`type_violin()`]: Creates a violin plot.
#'     - Models:
#'       - `"loess"` / [`type_loess()`]: Local regression curve.
#'       - `"lm"` / [`type_lm()`]: Linear regression line.
#'       - `"glm"` / [`type_glm()`]: Generalized linear model fit.
#'       - `"spline"` / [`type_spline()`]: Cubic (or Hermite) spline interpolation.
#'     - Functions:
#'       - [`type_abline()`]: line(s) with intercept and slope.
#'       - [`type_hline()`]: horizontal line(s).
#'       - [`type_vline()`]: vertical line(s).
#'       - [`type_function()`]: arbitrary function.
#'       - [`type_summary()`]: summarize `y` by unique values of `x`.
#' @param legend one of the following options:
#'    - NULL (default), in which case the legend will be determined by the
#'    grouping variable. If there is no group variable (i.e., `by` is NULL) then
#'    no legend is drawn. If a grouping variable is detected, then an automatic
#'    legend is drawn to the _outer_ right of the plotting area. Note that the
#'    legend title and categories will automatically be inferred from the `by`
#'    argument and underlying data.
#'    - A convenience string indicating the legend position. Supported keywords:
#'      - Standard position keywords from base `legend()`, e.g. `"right"`,
#'        `"topleft"`, `"bottom"`, etc.
#'      - Outer positions via a trailing `"!"`, e.g. `"right!"`, `"topleft!"`,
#'        or `"bottom!"`. This places the legend _outside_ the plotting area and
#'        adjusts the margins accordingly.
#'      - `"direct"`: places text labels at the last point of each group's data,
#'        coloured to match. Best suited to line-based plots with x-sorted data,
#'        where "last" corresponds to "rightmost". The right margin is
#'        automatically expanded to prevent clipping. Requires discrete groups
#'        via `by`. For faceted plots, labels are drawn in each panel for the
#'        groups present there. Supports additional arguments when passed
#'        via `legend(...)`:
#'        - `nudge_x`, `nudge_y`: numeric vectors of per-group offsets in data
#'          units. Recycled to the number of groups. Supports named vectors
#'          for targeted adjustment, e.g.
#'          `nudge_y = c("Group A" = -10, "Group B" = 20)`.
#'        - `repel`: if `TRUE`, automatically separates overlapping labels
#'          vertically. If a positive number, sets the minimum gap between
#'          labels in data units.
#'      - `"none"`: turns off legend printing.
#'    - Logical value, where TRUE corresponds to the default case above (same
#'    effect as specifying NULL) and FALSE turns the legend off (same effect as
#'    specifying "none").
#'    - A list or, equivalently, a dedicated `legend()` function with supported
#'    legend arguments, e.g. "bty", "horiz", and so forth. To suppress the
#'    legend title, set `title = NULL` or `title = FALSE`, e.g.
#'    `legend = list(title = FALSE)`.
#' @param main a main title for the plot, see also `title`.
#' @param sub a subtitle for the plot.
#' @param cap a caption for the plot, drawn at the bottom. Useful for
#'   annotations like data sources. Best paired with a dynamic
#'   \code{\link[tinyplot]{tinytheme}}. For the default theme, should be seen as
#'   a substitute for `sub`, since these two will otherwise overlap. Appearance
#'   can be customized via \code{\link[tinyplot]{tpar}} parameters `adj.cap`,
#'   `cex.cap`, `col.cap`, `font.cap`, and `line.cap`.
#' @param xlab a label for the x axis, defaults to a description of x.
#' @param ylab a label for the y axis, defaults to a description of y.
#' @param ann a logical value indicating whether the default annotation (title
#'   and x and y axis labels) should appear on the plot.
#' @param xlim the x limits (x1, x2) of the plot. Note that x1 > x2 is allowed
#'   and leads to a reversed axis (although see the `"rev(erse)"` keyword option
#'   below). The default value, `NULL`, indicates that the range of the `finite`
#'   values to be plotted should be used. Alongside the standard length-2 vector
#'   (e.g., `xlim = c(0, 1)`), `tinyplot` supports three further convenience
#'   forms:
#'   - a single scalar (e.g. `xlim = 0`) ensures that the provided value is
#'     covered by the axis range, irrespective of the data extent.
#'   - a length-2 vector with one `NA` (e.g. `xlim = c(0, NA)`) pins the
#'     non-`NA` limit and lets the data determine the other limit.
#'   - the convenience string `"rev"` (or the longer `"reverse"`) reverses the
#'     automatically-computed axis range. This is equivalent to passing a
#'     descending vector, but without having to know the data extent in advance.
#' @param ylim the y limits of the plot. Accepts the same input forms as `xlim`.
#' @param axes logical or character. Should axes be drawn (`TRUE` or `FALSE`)?
#'   Or alternatively what type of axes should be drawn: `"standard"` (with
#'   axis, ticks, and labels; equivalent to `TRUE`), `"none"` (no axes;
#'   equivalent to `FALSE`), `"ticks"` (only ticks and labels without axis line),
#'   `"labels"` (only labels without ticks and axis line), `"axis"` (only axis
#'   line and labels but no ticks). To control this separately for the two
#'   axes, use the character specifications for `xaxt` and/or `yaxt`.
#' @param xaxt,yaxt character specifying the type of x-axis and y-axis,
#'   respectively. See `axes` for the possible values.
#' @param xaxs,yaxs character specifying the style of the interval calculation
#'   used for the x-axis and y-axis, respectively. See
#'   \code{\link[graphics]{par}} for the possible values.
#' @param xaxb,yaxb numeric vector (or character vector, if appropriate) giving
#'   the break points at which the axis tick-marks are to be drawn. Break points
#'   outside the range of the data will be ignored if the associated axis
#'   variable is categorical, or an explicit `x/ylim` range is given.
#' @param xaxl,yaxl a function or a character keyword specifying the format of
#'   the x- or y-axis tick labels. Note that this is a post-processing step that
#'   affects the _appearance_ of the tick labels only; use in conjunction with
#'   `x/yaxb` if you would like to adjust the position of the tick marks too. In
#'   addition to user-supplied formatting functions (e.g., [`format`],
#'   [`toupper`], [`abs`], or other custom function), several convenience
#'   keywords (or their symbol equivalents) are available for common formatting
#'   transformations: `"percent"` (`"%"`), `"comma"` (`","`), `"log"` (`"l"`),
#'   `"dollar"` (`"$"`), `"euro"` (`"€"`), or `"sterling"` (`"£"`). See the
#'   [`tinylabel`] documentation for examples.
#' @param log a character string which contains `"x"` if the x axis is to be
#'   logarithmic, `"y"` if the y axis is to be logarithmic and `"xy"` or `"yx"`
#'   if both axes are to be logarithmic.
#' @param flip logical. Should the plot orientation be flipped, so that the
#'   y-axis is on the horizontal plane and the x-axis is on the vertical plane?
#'   Default is FALSE.
#' @param frame.plot a logical indicating whether a box should be drawn around
#'   the plot. Can also use `frame` as an acceptable argument alias.
#'   The default is to draw a frame if both axis types (set via `axes`, `xaxt`,
#'   or `yaxt`) include axis lines.
#' @param grid argument for plotting a background panel grid, one of:
#'    - a logical (i.e., `TRUE` to draw the grid on both axes).
#'    - a character string controlling which axes get grid lines and at what
#'      resolution. Uppercase letters (`"X"`, `"Y"`, `"XY"`) draw grid lines at
#'      the standard axis tick positions (with the latter equivalent to `TRUE`),
#'      while lowercase letters (`"x"`, `"y"`, `"xy"`) draw a finer grid with
#'      additional lines at the midpoints between ticks. (Note: the finer grid
#'      has no effect on log-scale axes, since tick positions already include
#'      intermediate values.)
#'    - a panel grid plotting function like `grid()`.
#'   Note that this argument replaces the `panel.first` and `panel.last`
#'   arguments from base `plot()` and tries to make the process more seamless
#'   with better default behaviour. The default behaviour is determined by (and
#'   can be set globally through) the value of `tpar("grid")`.
#' @param palette one of the following options:
#'    - NULL (default), in which case the palette will be chosen according to
#'    the class and cardinality of the "by" grouping variable. For non-ordered
#'    factors or strings with a reasonable number of groups, this will inherit
#'    directly from the user's default \code{\link[grDevices]{palette}} (e.g.,
#'    "R4"). In other cases, including ordered factors and high cardinality, the
#'    "Viridis" palette will be used instead. Note that a slightly restricted
#'    version of the "Viridis" palette---where extreme color values have been
#'    trimmed to improve visual perception---will be used for ordered factors
#'    and continuous variables. In the latter case of a continuous grouping
#'    variable, we also generate a gradient legend swatch.
#'    - A convenience string corresponding to one of the many palettes listed by
#'    either `palette.pals()` or `hcl.pals()`. Note that the string can be
#'    case-insensitive (e.g., "Okabe-Ito" and "okabe-ito" are both valid).
#'    - A palette-generating function. This can be "bare" (e.g.,
#'    `palette.colors`) or "closed" with a set of named arguments (e.g.,
#'    `palette.colors(palette = "Okabe-Ito", alpha = 0.5)`). Note that any
#'    unnamed arguments will be ignored and the key `n` argument, denoting the
#'    number of colours, will automatically be spliced in as the number of
#'    groups.
#'    - A vector or list of colours, e.g. `c("darkorange", "purple", "cyan4")`.
#'    If too few colours are provided for a discrete (qualitative) set of
#'    groups, then the colours will be recycled with a warning. For continuous
#'    (sequential) groups, a gradient palette will be interpolated. 
#' @param col plotting color. Character, integer, or vector of length equal to
#'   the number of categories in the `by` variable. See `col`. Note that the
#'   default behaviour in `tinyplot` is to vary group colors along any variables
#'   declared in the `by` argument. Thus, specifying colors manually should not
#'   be necessary unless users wish to override the automatic colors produced by
#'   this grouping process. Typically, this would only be done if grouping
#'   features are deferred to some other graphical parameter (i.e., passing the
#'   "by" keyword to one of `pch`, `lty`, `lwd`, or `bg`; see below.)
#' @param pch plotting "character", i.e., symbol to use. Character, integer, or
#'   vector of length equal to the number of categories in the `by` variable.
#'   See `pch`. In addition, users can supply a special `pch = "by"` convenience
#'   argument, in which case the characters will automatically loop over the
#'   number groups. This automatic looping will begin at the global character
#'   value (i.e., `par("pch")`) and recycle as necessary.
#' @param lty line type. Character, integer, or vector of length equal to the
#'   number of categories in the `by` variable. See `lty`. In addition, users
#'   can supply a special `lty = "by"` convenience argument, in which case the
#'   line type will automatically loop over the number groups. This automatic
#'   looping will begin at the global line type value (i.e., `par("lty")`) and
#'   recycle as necessary.
#' @param lwd line width. Numeric scalar or vector of length equal to the
#'   number of categories in the `by` variable. See `lwd`. In addition, users
#'   can supply a special `lwd = "by"` convenience argument, in which case the
#'   line width will automatically loop over the number of groups. This
#'   automatic looping will be centered at the global line width value (i.e.,
#'   `par("lwd")`) and pad on either side of that.
#' @param bg background fill color for the open plot symbols 21:25 (see
#'   `points.default`), as well as ribbon and area plot types.
#'   Users can also supply either one of two special convenience arguments that
#'   will cause the background fill to inherit the automatic grouped coloring
#'   behaviour of `col`:
#'
#'   - `bg = "by"` will insert a background fill that inherits the main color
#'   mappings from `col`.
#'   - `by = <numeric[0,1]>` (i.e., a numeric in the range `[0,1]`) will insert
#'   a background fill that inherits the main color mapping(s) from `col`, but
#'   with added alpha-transparency.
#'
#'   For both of these convenience arguments, note that the (grouped) `bg`
#'   mappings will persist even if the (grouped) `col` defaults are themselves
#'   overridden. This can be useful if you want to preserve the grouped palette
#'   mappings by background fill but not boundary color, e.g. filled points. See
#'   examples.
#' @param fill alias for `bg`. If non-NULL values for both `bg` and `fill` are
#'   provided, then the latter will be ignored in favour of the former.
#' @param alpha a numeric in the range `[0,1]` for adjusting the alpha channel
#'   of the color palette, where 0 means transparent and 1 means opaque. Use
#'   fractional values, e.g. `0.5` for semi-transparency.
#' @param cex character expansion. A numerical vector (can be a single value)
#'   giving the amount by which plotting characters and symbols should be scaled
#'   relative to the default. Note that `NULL` is equivalent to 1.0, while `NA`
#'   renders the characters invisible. There are two additional considerations,
#'   specifically for points-alike plot types (e.g. `"p"`):
#'   
#'   - users can also supply a special `cex = "by"` convenience argument, in
#'     which case the character expansion will automatically adjust by group
#'     too. The range of this character expansion is controlled by the `clim`
#'     argument in the respective types; see [`type_points()`] for example.
#'   - passing a `cex` vector of equal length to the main `x` and `y` variables
#'     (e.g., another column in the same dataset) will yield a "bubble"plot with
#'     its own dedicated legend. This can provide a useful way to visualize an
#'     extra dimension of the data; see Examples.
#' @param weights an optional numeric vector of observation weights, of the same
#'   length as the number of `x`,`y` coordinates. Can be passed as an unquoted
#'   (NSE) variable name when called from the `tinyplot.formula` method, e.g.
#'   `weights = varname` where `varname` is a variable in `data`. Weights are
#'   only consumed by the plot types that support them; specifically, the model
#'   fit and distribution types like [`type_lm()`] and [`type_spineplot()`].
#'   Because weights trigger a statistical transformation, a warning is emitted
#'   if this argument is supplied to a non-supported type (even though the
#'   actual weights will be ignored, regardless).
#' @param labels an optional vector of text labels for use with [`type_text()`],
#'   of length 1 or the same length as the number of `x`,`y` coordinates. Can be
#'   passed as an unquoted (NSE) variable name when called from the
#'   `tinyplot.formula` method, e.g. `labels = varname` where `varname` is a
#'   variable in `data`. Ignored by other plot types. Overrides the `labels`
#'   argument of [`type_text()`] if both are supplied.
#' @param subset,na.action,drop.unused.levels arguments passed to `model.frame`
#'   when extracting the data from `formula` and `data`.
#' @param add logical. If TRUE, then elements are added to the current plot rather
#'   than drawing a new plot window. Note that the automatic legend for the
#'   added elements will be turned off. See also [tinyplot_add], which provides
#'   a convenient wrapper around this functionality for layering on top of an
#'   existing plot without having to repeat arguments.
#' @param draw a function that draws directly on the plot canvas (before `x` and
#'   `y` are plotted). The `draw` argument is primarily useful for adding common
#'   elements to each facet of a faceted plot, e.g.
#'   \code{\link[graphics]{abline}} or \code{\link[graphics]{text}}. Note that
#'   this argument is somewhat experimental and that _no_ internal checking is
#'   done for correctness; the provided argument is simply captured and
#'   evaluated as-is within `tinyplot()` and thus has access to the local
#'   definition of all variables such as `x`, `y`, etc. See Examples.
#' @param restore.par a logical value indicating whether the
#'   \code{\link[graphics]{par}} settings prior to calling `tinyplot` should be
#'   restored on exit. Defaults to FALSE, which makes it possible to add
#'   elements to the plot after it has been drawn. However, note the the outer
#'   margins of the graphics device may have been altered to make space for the
#'   `tinyplot` legend. Users can opt out of this persistent behaviour by
#'   setting to TRUE instead. See also [get_saved_par] for another option to
#'   recover the original \code{\link[graphics]{par}} settings, as well as
#'   longer discussion about the trade-offs involved.
#' @param empty logical indicating whether the interior plot region should be
#'  left empty. The default is `FALSE`. Setting to `TRUE` has a similar effect
#'  to invoking `type = "n"` above, except that any legend artifacts owing to a
#'  particular plot type (e.g., lines for `type = "l"` or squares for
#'  `type = "area"`) will still be drawn correctly alongside the empty plot. In
#'  contrast,`type = "n"` implicitly assumes a scatterplot and so any legend
#'  will only depict points.
#' @param file character string giving the file path for writing a plot to disk.
#'   If specified, the plot will not be displayed interactively, but rather sent
#'   to the appropriate external graphics device (i.e.,
#'   \code{\link[grDevices]{png}}, \code{\link[grDevices]{jpeg}},
#'   \code{\link[grDevices]{pdf}}, or \code{\link[grDevices]{svg}}). As a point
#'   of convenience, note that any global parameters held in `(t)par` are
#'   automatically carried over to the external device and don't need to be
#'   reset (in contrast to the conventional base R approach that requires
#'   manually opening and closing the device). The device type is determined by
#'   the file extension at the end of the provided path, and must be one of
#'   ".png", ".jpg" (".jpeg"), ".pdf", or ".svg". (Other file types may be
#'   supported in the future.) The file dimensions can be controlled by the
#'   corresponding `width` and `height` arguments below, otherwise will fall
#'   back to the `"file.width"` and `"file.height"` values held in
#'   \code{\link[tinyplot]{tpar}} (i.e., both defaulting to 7 inches, and where
#'   the default resolution for bitmap files is also specified as 300
#'   DPI).
#' @param width numeric giving the plot width in inches. Together with `height`,
#'  typically used in conjunction with the `file` argument above, overriding the
#'  default values held in `tpar("file.width", "file.height")`. If either `width`
#'  or `height` is specified, but a corresponding `file` argument is not
#'  provided as well, then a new interactive graphics device dimensions will be
#'  opened along the given dimensions. Note that this interactive resizing may
#'  not work consistently from within an IDE like RStudio that has an integrated
#'  graphics windows.
#' @param height numeric giving the plot height in inches. Same considerations as
#'  `width` (above) apply, e.g. will default to `tpar("file.height")` if not
#'  specified.
#' @param asp the y/xy/x aspect ratio, see `plot.window`.
#' @param theme keyword string (e.g. `"clean"`) or list defining a theme. Passed
#'  on to [`tinytheme`], but reset upon exit so that the theme effect is only
#'  temporary. Useful for invoking ephemeral themes. 
#' @param ... other graphical parameters. If `type` is a character specification
#'   (such as `"hist"`) then any argument names that match those from the corresponding
#'   `type_*()` function (such as \code{\link{type_hist}}) are passed on to that.
#'   All remaining arguments from `...` can be further graphical parameters, see
#'   \code{\link[graphics]{par}}).
#'
#' @returns No return value, called for side effect of producing a plot.
#'
#' @details
#' Disregarding the enhancements that it supports, `tinyplot` tries as far as
#' possible to mimic the behaviour and syntax logic of the original base
#' \code{\link[graphics]{plot}} function. Users should therefore be able to swap
#' out existing `plot` calls for `tinyplot` (or its shorthand alias `plt`),
#' without causing unexpected changes to the output.
#'
#' @importFrom grDevices axisTicks adjustcolor cairo_pdf chull colorRampPalette dev.cur dev.list dev.off dev.new extendrange hcl.colors hcl.pals jpeg palette palette.colors palette.pals pdf png svg xy.coords
#' @importFrom graphics abline arrows axis Axis axTicks box boxplot grconvertX grconvertY hist lines mtext par plot.default plot.new plot.window points polygon polypath segments rect text title
#' @importFrom utils modifyList head tail
#' @importFrom stats na.omit setNames
#' @importFrom tools file_ext
#'
#' @examples
#' aq = transform(
#'   airquality,
#'   Month = factor(Month, labels = month.abb[unique(Month)])
#' )
#'
#' # In most cases, `tinyplot` should be a drop-in replacement for regular
#' # `plot` calls. For example:
#'
#' op = tpar(mfrow = c(1, 2))
#' plot(0:10, main = "plot")
#' tinyplot(0:10, main = "tinyplot")
#' tpar(op) # restore original layout
#'
#' # Aside: `tinyplot::tpar()` is a (near) drop-in replacement for `par()`
#'
#' # Unlike vanilla plot, however, tinyplot allows you to characterize groups
#' # using either the `by` argument or equivalent `|` formula syntax.
#'
#' with(aq, tinyplot(Day, Temp, by = Month)) ## atomic method
#' tinyplot(Temp ~ Day | Month, data = aq) ## formula method
#'
#' # (Notice that we also get an automatic legend.)
#'
#' # You can also use the equivalent shorthand `plt()` alias if you'd like to
#' # save on a few keystrokes
#'
#' plt(Temp ~ Day | Month, data = aq) ## shorthand alias
#'
#' # Use standard base plotting arguments to adjust features of your plot.
#' # For example, change `pch` (plot character) to get filled points and `cex`
#' # (character expansion) to increase their size.
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   pch = 16,
#'   cex = 2
#' )
#' 
#' # Use the special "by" convenience keyword if you would like to map these
#' # aesthetic features over groups too (i.e., in addition to the default
#' # colour grouping)
#' 
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   pch = "by",
#'   cex = "by"
#' )
#'
#' # We can add alpha transparency for overlapping points
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   pch = 16,
#'   cex = 2,
#'   alpha = 0.3
#' )
#'
#' # To get filled points with a common solid background color, use an
#' # appropriate plotting character (21:25) and combine with one of the special
#' # `bg`/`fill` convenience arguments.
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   pch = 21, # use filled circles
#'   cex = 2,
#'   bg = 0.3, # numeric in [0,1] adds a grouped background fill with transparency
#'   col = "black" # override default color mapping; give all points a black border
#' )
#' 
#' # Aside: For "bubble" plots, pass an appropriate vector to the `cex` arg.
#' # This can be useful for depicting an additional dimension of the data (here:
#' # Wind).
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   pch = 21,
#'   cex = aq$Wind, # map character size to another feature in the data
#'   bg = 0.3,
#'   col = "black"
#' )
#'
#' # Converting to a grouped line plot is a simple matter of adjusting the
#' # `type` argument.
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "l"
#' )
#'
#' # Similarly for other plot types, including some additional ones provided
#' # directly by tinyplot, e.g. density plots or internal plots (ribbons,
#' # pointranges, etc.)
#'
#' tinyplot(
#'   ~ Temp | Month,
#'   data = aq,
#'   type = "density",
#'   fill = "by"
#' )
#'
#' # Facet plots are supported too. Facets can be drawn on their own...
#'
#' tinyplot(
#'   Temp ~ Day,
#'   facet = ~Month,
#'   data = aq,
#'   type = "area",
#'   main = "Temperatures by month"
#' )
#'
#' # ... or combined/contrasted with the by (colour) grouping.
#'
#' aq = transform(aq, Summer = Month %in% c("Jun", "Jul", "Aug"))
#' tinyplot(
#'   Temp ~ Day | Summer,
#'   facet = ~Month,
#'   data = aq,
#'   type = "area",
#'   palette = "dark2",
#'   main = "Temperatures by month and season"
#' )
#'
#' # Users can override the default square window arrangement by passing `nrow`
#' # or `ncol` to the helper facet.args argument. Note that we can also reduce
#' # axis label repetition across facets by turning the plot frame off.
#'
#' tinyplot(
#'   Temp ~ Day | Summer,
#'   facet = ~Month, facet.args = list(nrow = 1),
#'   data = aq,
#'   type = "area",
#'   palette = "dark2",
#'   frame = FALSE,
#'   main = "Temperatures by month and season"
#' )
#'
#' # Use a two-sided formula to arrange the facet windows in a fixed grid.
#' # LHS -> facet rows; RHS -> facet columns
#'
#' aq$hot = ifelse(aq$Temp >= 75, "hot", "cold")
#' aq$windy = ifelse(aq$Wind >= 15, "windy", "calm")
#' tinyplot(
#'   Temp ~ Day,
#'   facet = windy ~ hot,
#'   data = aq
#' )
#'
#' # To add common elements to each facet, use the `draw` argument
#'
#' tinyplot(
#'   Temp ~ Day,
#'   facet = windy ~ hot,
#'   data = aq,
#'   draw = abline(h = 75, lty = 2, col = "hotpink")
#' )
#'
#' # The (automatic) legend position and look can be customized using
#' # appropriate arguments. Note the trailing "!" in the `legend` position
#' # argument below. This tells `tinyplot` to place the legend _outside_ the plot
#' # area.
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "l",
#'   legend = legend("bottom!", title = "Month of the year", bty = "o")
#' )
#'
#' # Use legend = "direct" to place text labels at the last point of each
#' # group's data, coloured to match. Best suited to line-based plots with
#' # x-sorted data, where "last" corresponds to "rightmost". The right
#' # margin is automatically expanded to fit the labels. Pairs well with
#' # dynamic themes for tighter margins overall.
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "l",
#'   legend = "direct",
#'   theme = "clean2"
#' )
#'
#' # Use `nudge_x/y`` to manually adjust label positions, or `repel` to
#' # auto-separate overlapping labels:
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "l",
#'   legend = legend("direct", nudge_y = c(May = -1, Jun = 2)),
#'   # legend = legend("direct", repel = TRUE), # another option
#'   theme = "clean2"
#' )
#'
#' # The default group colours are inherited from either the "R4" or "Viridis"
#' # palettes, depending on the number of groups. However, all palettes listed
#' # by `palette.pals()` and `hcl.pals()` are supported as convenience strings,
#' # or users can supply a valid palette-generating function for finer control
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "l",
#'   palette = "tableau"
#' )
#'
#' # It's possible to customize the look of your plots by setting graphical
#' # parameters (e.g., via `(t)par`)... But a more convenient way is to just use
#' # built-in themes (see `?tinytheme`).
#'
#' tinyplot(
#'   Temp ~ Day | Month,
#'   data = aq,
#'   type = "b",
#'   alpha = 0.5,
#'   main = "Daily temperatures by month",
#'   sub = "Brought to you by tinyplot",
#'   cap = "Source: Base R airquality dataset",
#'   theme = "clean2"
#' )
#'
#' # For more examples and a detailed walkthrough, please see the introductory
#' # tinyplot tutorial available online:
#' # https://grantmcdermott.com/tinyplot/vignettes/introduction.html
#'
#' @rdname tinyplot
#' @export
tinyplot =
  function(x, ...) {
    UseMethod("tinyplot")
  }

#' @rdname tinyplot
#' @export
tinyplot.default = function(
    x = NULL,
    y = NULL,
    xmin = NULL,
    xmax = NULL,
    ymin = NULL,
    ymax = NULL,
    weights = NULL,
    labels = NULL,
    by = NULL,
    facet = NULL,
    facet.args = NULL,
    data = NULL,
    type = NULL,
    legend = NULL,
    main = NULL,
    sub = NULL,
    cap = NULL,
    xlab = NULL,
    ylab = NULL,
    ann = par("ann"),
    xlim = NULL,
    ylim = NULL,
    axes = TRUE,
    xaxt = NULL,
    yaxt = NULL,
    xaxs = NULL,
    yaxs = NULL,
    xaxb = NULL,
    yaxb = NULL,
    xaxl = NULL,
    yaxl = NULL,
    log = "",
    flip = FALSE,
    frame.plot = NULL,
    grid = NULL,
    palette = NULL,
    pch = NULL,
    lty = NULL,
    lwd = NULL,
    col = NULL,
    bg = NULL,
    fill = NULL,
    alpha = NULL,
    cex = NULL,
    add = FALSE,
    draw = NULL,
    empty = FALSE,
    restore.par = FALSE,
    file = NULL,
    width = NULL,
    height = NULL,
    asp = NA,
    theme = NULL,
    ...) {

  # Force evaluation of legend if it's a symbol to avoid downstream promise
  # issues. Let sanitize_legend handle it
  if (!missing(legend) && is.symbol(substitute(legend))) {
    legend = legend
  }

  #
  ## save parameters and calls -----
  #
  par_first = get_saved_par("first")
  if (is.null(par_first)) set_saved_par("first", par())
  
  # Validate grid only for simple values; skip for unevaluated calls like grid()
  # which are passed as language objects from tinyplot.formula via substitute(). (#193)
  if (!is.null(grid) && !is.call(grid)) {
    assert_grid(grid, null.ok = TRUE, name = "grid")
  }

  # save for tinyplot_add()
  assert_logical(add)
  if (!add) {
    calls = sys.calls()
    is_tinyplot_call = function(x) identical(tinyplot, try(eval(x[[1L]]), silent = TRUE))
    idx = which(vapply(calls, is_tinyplot_call, FALSE))
    if (length(idx) > 0) {
      set_environment_variable(.last_call = calls[[idx[1L]]])
    }
  }

  # Save current graphical parameters
  opar = par(no.readonly = TRUE)
  if (restore.par || !is.null(facet)) {
    if (!is.null(file) || !is.null(width) || !is.null(height)) {
      opar$new = FALSE # catch for some interfaces
    }
    on.exit(par(opar), add = TRUE)
  }
  set_saved_par(when = "before", opar)

  # Catch for adding to existing facet plot
  if (!is.null(facet) && add) {
    recordGraphics(
      par(get_saved_par(when = "after")),
      list = list(),
      env = getNamespace('tinyplot')
    )
  }

  # Ephemeral theme
  if (!is.null(theme)) {
    if (is.character(theme) && length(theme) == 1) {
      tinytheme(theme)
    } else if (is.list(theme)) {
      do.call(tinytheme, theme)
    } else {
      warning('Argument `theme` must be a character of length 1 (e.g. "clean"), or a list. Ignoring.')
    }
    if (is.character(theme) && theme == "default") {
      # Reset mar to pre-theme value so legend margin adjustment isn't
      # clobbered. Only needed for "default" theme which uses hook = FALSE
      # and thus sets par(mar) immediately. (#557)
      par(mar = opar$mar)
      on.exit(init_tpar(rm_hook = TRUE), add = TRUE)
    } else {
      dtheme = theme_default
      otheme = opar[names(dtheme)]
      on.exit(do.call(tinytheme, otheme), add = TRUE)
    }
  }

  #
  ## settings container -----
  #

  dots = list(...)

  # resolve any axis-reversal keyword (e.g. xlim = "reverse") into a flag, and
  # reset the limit to NULL so the normal auto-range path runs (see lim.R)
  .revx = sanitize_lim_rev(xlim, "xlim")
  xlim = .revx[["lim"]]; rev_x = .revx[["rev"]]
  .revy = sanitize_lim_rev(ylim, "ylim")
  ylim = .revy[["lim"]]; rev_y = .revy[["rev"]]

  settings_list = list(
    # save call to check user input later
    call          = match.call(),

    # save to file & device dimensions
    file          = file,
    width         = width,
    height        = height,

    # deparsed input for use in labels
    by_dep        = deparse1(substitute(by)),
    cex_dep       = if (!is.null(cex)) deparse1(substitute(cex)) else NULL,
    facet_dep     = deparse1(substitute(facet)),
    x_dep         = if (is.null(x))    NULL else deparse1(substitute(x)),
    xmax_dep      = if (is.null(xmax)) NULL else deparse1(substitute(xmax)),
    xmin_dep      = if (is.null(xmin)) NULL else deparse1(substitute(xmin)),
    y_dep         = if (is.null(y))    NULL else deparse1(substitute(y)),
    ymax_dep      = if (is.null(ymax)) NULL else deparse1(substitute(ymax)),
    ymin_dep      = if (is.null(ymin)) NULL else deparse1(substitute(ymin)),

    # types
    type          = type,
    type_data     = NULL,
    type_draw     = NULL,
    type_name     = NULL,

    # type-specific settings
    bubble        = FALSE,
    bubble_pch    = NULL,
    bubble_alpha  = NULL,
    bubble_bg_alpha = NULL,
    ygroup        = NULL,  # for type_ridge()

    # data points and labels
    x             = x,
    xmax          = xmax,
    xmin          = xmin,
    xlab          = xlab,
    xlabs         = NULL,
    y             = y,
    ymax          = ymax,
    ymin          = ymin,
    ylab          = ylab,
    ylabs         = NULL,
    weights       = weights,
    labels        = labels,
    weights_used  = FALSE, # set TRUE by type_data() functions that consume weights

    # axes
    axes          = axes,
    xaxt          = xaxt,
    xaxb          = xaxb,
    xaxl          = xaxl,
    xaxs          = xaxs,
    yaxt          = yaxt,
    yaxb          = yaxb,
    yaxl          = yaxl,
    yaxs          = yaxs,
    frame.plot    = frame.plot,
    xlim          = xlim,
    ylim          = ylim,
    rev_x         = rev_x,
    rev_y         = rev_y,

    # flags to check user input (useful later on)
    null_by       = is.null(by),
    null_xlim     = is.null(xlim),
    null_ylim     = is.null(ylim),
    # when palette functions need pre-processing this check raises error
    null_palette  = tryCatch(is.null(palette), error = function(e) FALSE),
    x_by          = identical(x, by), # for "boxplot", "spineplot" and "ridge"

    # unevaluated expressions with side effects
    draw          = substitute(draw),
    facet         = facet,
    facet.args    = facet.args,
    palette       = substitute(palette),
    legend        = if (add) FALSE else substitute(legend),

    # aesthetics
    lty           = lty,
    lwd           = lwd,
    col           = col,
    bg            = bg,
    log           = log,
    fill          = fill,
    alpha         = alpha,
    cex           = cex,
    pch           = if (is.null(pch)) get_tpar("pch", default = NULL) else pch,

    # ribbon.alpha overwritten by some type_data() functions
    # sanitize_ribbon_alpha: returns default alpha transparency for ribbon-type plots
    ribbon.alpha  = sanitize_ribbon_alpha(NULL),

    # misc
    add           = add,
    by            = by,
    cap           = cap,
    dodge         = NULL,
    dots          = dots,
    flip          = flip,
    lighten       = TRUE, # area types may override via their type_data() fn
    group_offsets = NULL,
    offsets_axis  = NULL,
    sub           = sub,
    type_info     = list() # pass type-specific info from type_data to type_draw
  )

  settings = new.env()
  list2env(settings_list, settings)

  #
  ## devices and files -----
  #
  # Write plot to output file or window with fixed dimensions
  setup_device(settings)
  if (!is.null(settings$file)) on.exit(dev.off(), add = TRUE)


  #
  ## sanitize arguments -----
  #

  # extract legend_args from dots
  if ("legend_args" %in% names(dots)) {
    settings$legend_args = settings$dots[["legend_args"]]
    settings$dots$legend_args = NULL # avoid passing both directly and via ...
  } else {
    settings$legend_args = list(x = NULL)
  }
  # normalize legend position up front so downstream code can read
  # legend_args[["x"]] directly (idempotent: guarded inside sanitize_legend).
  # Use settings$legend (captured via substitute()) rather than the raw
  # `legend` promise, since the latter may be an unevaluated call like
  # `legend("bottom!", ...)` that would error if forced by is.null() etc.
  # Skip when add=TRUE: no new legend is drawn in add-mode, and
  # settings$legend is coerced to FALSE which sanitize_legend would
  # spuriously normalize to the "right!" default.
  if (!isTRUE(add)) {
    settings$legend_args = sanitize_legend(settings$legend, settings$legend_args)
  }

  # alias: bg = fill
  if (is.null(bg) && !is.null(fill)) settings$bg = fill

  # validate types and returns a list with name, data, and draw components
  sanitize_type(settings)
  
  # standardize axis arguments and returns consistent axes, xaxt, yaxt, frame.plot
  sanitize_axes(settings)

  # generate appropriate axis labels based on input data and plot type
  sanitize_xylab(settings)

  # palette default
  if (is.null(settings$palette)) {
    settings$palette = get_tpar("palette", default = NULL)
  }

  # by: coerce character groups to factor
  if (!settings$null_by && is.character(settings$by)) {
    settings$by = factor(settings$by)
  }

  # flag if x==by, currently only used for 
  # 


  # facet: parse facet formula and prepares variables when facet==by
  sanitize_facet(settings)

  # x / y processing
  # convert characters to factors
  # set x automatically when omitted (ex: rect)
  # set y automatically when omitted (ex: boxplots)
  # combine x, y, xmax, by, facet etc. into a single `datapoints` data.frame
  sanitize_datapoints(settings)


  #
  ## transform datapoints using type_data() -----
  #

  if (!is.null(settings$type_data)) {
    settings$type_data(settings, ...)
  }

  # Warn if weights were supplied but the plot type ignored them. This is a
  # deliberate exception to how other unconsumed top-level args (e.g.
  # xmin/xmax/ymin/ymax) are silently dropped: a missing ribbon is visually
  # obvious, but ignored weights produce a normal-looking yet statistically
  # wrong plot. Consuming types opt in by setting `weights_used`, so new
  # weight-aware types are covered automatically (#332).
  if (!is.null(settings$weights) && !isTRUE(settings$weights_used)) {
    warning(
      "`weights` were supplied but are ignored by this plot type.",
      call. = FALSE
    )
  }

  # ensure axis aligment of any added layers
  if (!add) {
    assign("xlabs_orig", settings[["xlabs"]], envir = get(".tinyplot_env", envir = parent.env(environment())))
    assign(".group_offsets", settings[["group_offsets"]], envir = get(".tinyplot_env", envir = parent.env(environment())))
    assign(".offsets_axis", settings[["offsets_axis"]], envir = get(".tinyplot_env", envir = parent.env(environment())))
  } else {
    align_layer(settings)
  }

  # flip -> swap x and y after type_data, except for boxplots (which has its own bespoke flip logic)
  flip_datapoints(settings)


  #
  ## bubble plot -----
  #
  
  # Transform cex values for bubble charts. Handles size transformation, legend
  # gotchas, and aesthetic sanitization.
  # Currently limited to "p" and "text" types, but could expand to others.
  bubble(settings)


  #
  ## axis breaks and limits -----
  #
  
  # do this after computing yaxb because limits will depend on the previous calculations
  if (!add) {
    lim_args(settings)
  }

  #
  ## facets: count -----
  #

  # facet_layout processes facet simplification, attribute restoration, and layout
  facet_layout(settings)


  #
  ## aesthetics by group -----
  #
  
  by_aesthetics(settings)


  #
  ## legends -----
  #
  
  prepare_legend(settings)

  #
  ## make settings available in the environment directly -----
  #

  env2env(settings, environment())

  #
  ## dynmar: compute margins up front -----
  #
  # Under dynmar themes, compute the full margin once before any drawing.
  # theme_mar (from the theme) is the baseline padding; dynmar_side() adds
  # space for ticks, axis labels, main, and sub. whtsbp adds tick-label
  # width/height for horizontal y-labels or vertical x-labels.
  # For outer legends ("bottom!", "left!", etc.), skip dynmar_side on the
  # legend's side — the legend code owns that side's mar via oma.
  #
  dynmar_computed = NULL
  .whtsbp = c(0, 0, 0, 0)
  if (!add && isTRUE(get_tpar("dynmar"))) {
    .side.sub = get_tpar("side.sub", default = 3)
    # Read the theme's intended mar. Also build a tpars list from the theme
    # definition so dynmar_side uses theme mgp/tcl/las (which aren't in
    # par() yet since the before.plot.new hook hasn't fired).
    .tinytheme = get_tpar("tinytheme", default = "default")
    .theme_def = get_theme_def(.tinytheme)
    if (identical(.theme_def, theme_default)) .theme_def = NULL
    .theme_mar = if (!is.null(.theme_def[["mar"]])) .theme_def[["mar"]] else par("mar")
    .tpars = if (!is.null(.theme_def)) modifyList(.theme_def, tpar()) else tpar()
    # Merge pending before.plot.new hook values into .tpars so user
    # overrides passed via tinytheme(..., las = 2) (or tpar(...)) are
    # visible to dynmar_side()/whtsbp before plot.new fires. Without this,
    # user overrides for par-level values (las, cex.lab, mgp, tcl, etc.)
    # are queued in hook closures and unreachable from par() at this point.
    .pending_hooks = get_environment_variable(".tpar_hooks")
    for (.h in .pending_hooks) {
      .bp = environment(.h)[["base_par"]]
      if (is.list(.bp)) .tpars = modifyList(.tpars, .bp)
    }
    if (!is.null(.tpars[["mar"]])) .theme_mar = .tpars[["mar"]]

    .cex_axis = get_tpar("cex.axis", tpar_list = .tpars, default = 1)
    .cex_lab = get_tpar(c("cex.ylab", "cex.lab"), tpar_list = .tpars, default = 1)
    .las = get_tpar("las", tpar_list = .tpars, default = par("las"))
    .ymgp_shift = if (.las %in% c(0L, 1L)) 0.5 * (.cex_axis - 1) else 0
    .ylab_cex_shift = 0.5 * (.cex_lab - 1)

    # Detect outer-legend sides (order: bottom, left, top, right).
    .lgnd_pos = settings$legend_args[["x"]]
    .outer_sides = c(
      grepl("bottom!$", .lgnd_pos),
      grepl("left!$",   .lgnd_pos),
      grepl("top!$",    .lgnd_pos),
      grepl("right!$",  .lgnd_pos)
    )

    # Types that suppress the standard axes (xaxt/yaxt = "n") only because they
    # draw their own in the same place (e.g. spineplot category + numeric labels
    # via spine_axis(); ridge y-axis category labels via tinyAxis()). Their
    # tick-row margin must still be reserved, else the self-drawn labels clip --
    # or, under las 0/1 where mar is further shifted, error -- when xlab/ylab = NA
    # makes label_extent = 0 (#635, #650).
    .self_axis = type %in% c("spineplot", "ridge")

    .dyn = c(
      dynmar_side(1, xlab, main = main, sub = sub,
                  cap = if (.outer_sides[1]) NULL else cap,
                  side.sub = .side.sub,
                  axis_on = .self_axis ||
                    (!identical(xaxt, "none") && !identical(xaxt, "n")),
                  tpars = .tpars),
      dynmar_side(2, ylab,
                  axis_on = .self_axis ||
                    (!identical(yaxt, "none") && !identical(yaxt, "n")),
                  tpars = .tpars),
      dynmar_side(3, NULL, main = main, sub = sub, side.sub = .side.sub,
                  tpars = .tpars),
      dynmar_side(4, NULL, tpars = .tpars)
    )
    if (.ymgp_shift > 0) .dyn[2] = .dyn[2] - .ymgp_shift
    # Drop the theme's baseline padding on outer-legend sides so the plot
    # region meets the legend's oma flush. Only .theme_mar is zeroed — the
    # axis-driven bumps in .dyn (tick rows, axis labels, main/sub) are kept
    # so that "left!" and "bottom!" legends don't collide with axis content.
    # Exception: "top!" lives in mar[3] (not oma), so keep the baseline padding.
    .outer_sides_oma = .outer_sides & c(TRUE, TRUE, FALSE, TRUE)
    .theme_mar[.outer_sides_oma] = 0

    # whtsbp uses strwidth(units="figure") + grconvertX("nfc" → "lines"),
    # both of which give device-default font metrics without requiring
    # plot.new()/plot.window() first. A preparatory plot.new() here would
    # advance par("mfg") (breaking mfrow layouts) and create a blank page
    # in IDE plot panes (Positron). Left/bottom/top margin sizing for
    # title alignment is handled later by draw_legend or the no-legend
    # path's own plot.new(), after which the margins are reinstated
    # via dynmar_computed + .whtsbp before draw_title runs.

    # Compute whtsbp (tick-label width/height bump). Read `las` from .tpars
    # (the theme definition) rather than par() — par("las") isn't set to the
    # theme's intended value until the before.plot.new hook fires, but this
    # block runs before that. Pass .cex_axis to strwidth so measurements
    # reflect the intended text size (par("cex.axis") isn't set yet either).
    .whtsbp = c(0, 0, 0, 0)
    .whtsbp_y_raw = 0
    .whtsbp_x_raw = 0
    .las = get_tpar("las", tpar_list = .tpars, default = par("las"))
    if (.las %in% 1:2) {
      if (type == "ridge") {
        yaxlabs = levels(y)
      } else if (!is.null(ylabs)) {
        yaxlabs = if (!is.null(names(ylabs))) names(ylabs) else ylabs
      } else if (type == "boxplot" && isTRUE(flip) && !is.null(xlabs)) {
        yaxlabs = if (!is.null(names(xlabs))) names(xlabs) else xlabs
      } else {
        ylim_usr = if (diff(ylim) == 0 && is.null(yaxb)) ylim + c(-0.5, 0.5) else extendrange(ylim, f = 0.04)
        yaxlabs = axisTicks(usr = ylim_usr, log = par("ylog"))
      }
      if (!is.null(yaxl)) yaxlabs = tinylabel(yaxlabs, yaxl)
      .whtsbp_y_raw = grconvertX(max(strwidth(yaxlabs, "figure", cex = .cex_axis)), from = "nfc", to = "lines") -
                      grconvertX(0, from = "nfc", to = "lines") - 0.5
      if (is.finite(.whtsbp_y_raw)) .whtsbp[2] = .whtsbp_y_raw
    }
    if (.las %in% 2:3) {
      xlim_usr = if (diff(xlim) == 0 && is.null(xaxb)) xlim + c(-0.5, 0.5) else extendrange(xlim, f = 0.04)
      xaxlabs = if (is.null(xlabs)) axisTicks(usr = xlim_usr, log = par("xlog")) else
        if (!is.null(names(xlabs))) names(xlabs) else xlabs
      if (!is.null(xaxl)) xaxlabs = tinylabel(xaxlabs, xaxl)
      .whtsbp_x_raw = grconvertX(max(strwidth(xaxlabs, "figure", cex = .cex_axis)), from = "nfc", to = "lines") - 0.5
      if (is.finite(.whtsbp_x_raw)) .whtsbp[1] = .whtsbp_x_raw
    }

    # Under facets, per-facet tick labels render smaller (scaled by
    # cex_fct_adj), so whtsbp — which is computed from device font metrics
    # — needs the same scaling to match the actual rendered margin used by
    # draw_facet_window. Without this, draw_title's mar reserves too much
    # space on the LHS and anchors the title too far right.
    if (cex_fct_adj != 1) {
      .whtsbp = .whtsbp * cex_fct_adj
      .whtsbp_y_raw = .whtsbp_y_raw * cex_fct_adj
      .whtsbp_x_raw = .whtsbp_x_raw * cex_fct_adj
    }

    dynmar_computed = .theme_mar + .dyn
    par(mar = dynmar_computed + .whtsbp)
  }

  if (legend_draw_flag && !identical(legend_args[["x"]], "direct")) {
    if (!multi_legend) {
      ## simple case: single legend only
      if (is.null(lgnd_cex)) lgnd_cex = cex * cex_fct_adj
      draw_legend(
        legend = legend,
        legend_args = legend_args,
        by_dep = by_dep,
        lgnd_labs = lgnd_labs,
        type = type,
        pch = pch,
        lty = lty,
        lwd = lwd,
        col = col,
        bg = bg,
        gradient = by_continuous,
        cex = lgnd_cex,
        has_sub = has_sub,
        has_cap = has_cap,
        cap_text = cap,
        dynmar_title_mar = if (!is.null(dynmar_computed) && .outer_sides[3] && text_line_count(main) >= 1L) dynmar_computed[3] else NULL
      )
    } else {
      ## multi-legend case...
      prepare_legend_multi(settings)
      env2env(settings, environment(), c("legend_args", "lgby", "lgbub"))
      # draw multi-legend
      draw_multi_legend(list(lgby, lgbub), position = legend_args[["x"]])

    }

    has_legend = TRUE
    } else if (legend_args[["x"]] %in% c("none", "direct") && !isTRUE(add)) {
    omar = par("mar")
    ooma = par("oma")
    topmar_epsilon = 0.1

    # Catch to avoid recursive offsets, e.g. repeated tinyplot calls with
    # "bottom!" legend position.
    restore_margin_inner(ooma)
    # clean up for now
    rm(omar, ooma, topmar_epsilon)

    # Draw new plot
    plot.new()
  }


  #
  ## title and subtitle -----
  #

  direct_labels_flag = !isTRUE(add) && identical(legend_args[["x"]], "direct") &&
    !isTRUE(by_continuous) && !null_by
  if (identical(legend_args[["x"]], "direct") && !direct_labels_flag && !isTRUE(add)) {
    warning("legend=\"direct\" requires discrete groups via `by`. Falling back to no legend.")
  }

  if (!add) {
    # Reinstate dynmar margins and user coordinates after draw_legend
    # (which may have called plot.new and reset par via hooks).
    if (!is.null(dynmar_computed)) {
      if (has_legend && .outer_sides[3]) {
        dynmar_computed[3] = par("mar")[3] - .whtsbp[3]
      }
      par(mar = dynmar_computed + .whtsbp)
      if (!is.null(xlim) && !is.null(ylim)) {
        plot.window(xlim = xlim, ylim = ylim)
      }
    } else if (direct_labels_flag && !is.null(xlim) && !is.null(ylim)) {
      plot.window(xlim = xlim, ylim = ylim)
    }

    # Expand right margin for direct labels based on actual label overshoot
    if (direct_labels_flag && !is.null(xlim) && !is.null(ylim)) {
      usr_right = par("usr")[2]
      last_x = tapply(datapoints$x, datapoints$by, function(z) tail(z, 1))
      dl_nudge_x = legend_args[["nudge_x"]] %||% rep(0, length(last_x))
      dl_nudge_x = rep_len(dl_nudge_x, length(last_x))
      offset_usr = strwidth("m", units = "user") * 0.3
      label_widths = strwidth(lgnd_labs, units = "user")
      overshoots = (last_x + dl_nudge_x + offset_usr + label_widths) - usr_right
      max_overshoot = max(0, overshoots, na.rm = TRUE)
      if (max_overshoot > 0) {
        overshoot_lines = max_overshoot * par("pin")[1] / diff(par("usr")[1:2]) / par("csi")
        if (!is.null(dynmar_computed)) {
          dynmar_computed[4] = dynmar_computed[4] + overshoot_lines
          par(mar = dynmar_computed + .whtsbp)
        } else {
          cur_mar = par("mar")
          cur_mar[4] = cur_mar[4] + overshoot_lines
          par(mar = cur_mar)
        }
        plot.window(xlim = xlim, ylim = ylim)
      }
    }

    ann = as.logical(ann)
    if (ann) draw_title(
      main, sub, cap, xlab, ylab, legend, legend_args, opar,
      xlab_line_offset = if (!is.null(dynmar_computed)) .whtsbp_x_raw else 0,
      ylab_line_offset = if (!is.null(dynmar_computed)) .whtsbp_y_raw - max(0, .ymgp_shift) - .ylab_cex_shift else 0
    )
  }


  #
  ## facets: draw -----
  #

  # Two-phase plotting logic: First determine and draw all exterior elements 
  # (facet windows, axes, grid, etc.), then circle back to each facet and 
  # draw the interior elements (grouped points, lines, etc.)

  omar = NULL # Placeholder variable for now, which we re-assign as part of facet margins

  # placeholders for facet_window_args() call
  facet_newlines = facet_text = facet_rect = facet_font = facet_col = facet_bg = facet_border = NULL

  if (!is.null(facet) && !add) {
    if (is.null(omar)) omar = par("mar")

    # Grab some of the customizable facet args that we'll be using later
    facet_rect = FALSE
    facet_text = .tpar[["facet.cex"]]
    facet_font = .tpar[["facet.font"]]
    facet_col = .tpar[["facet.col"]]
    facet_bg = .tpar[["facet.bg"]]
    facet_border = .tpar[["facet.border"]]
    if (!is.null(facet.args)) {
      if (!is.null(facet.args[["cex"]])) facet_text = facet.args[["cex"]]
      if (!is.null(facet.args[["col"]])) facet_col = facet.args[["col"]]
      if (!is.null(facet.args[["font"]])) facet_font = facet.args[["font"]]
      if (!is.null(facet.args[["bg"]])) facet_bg = facet.args[["bg"]]
      if (!is.null(facet.args[["border"]])) facet_border = facet.args[["border"]]
    }
    if (!is.null(facet_bg) || !is.null(facet_border)) facet_rect = TRUE

    # Need extra adjustment to top margin if facet titles have "\n" newline
    # separator. (Note that we'll also need to take account for this in the
    # individual facet margins / gaps further below.)
    facet_newlines = lengths(gregexpr("\n", grep("\\n", facets, value = TRUE)))
    # if (length(facet_newlines)==0) facet_newlines = 0
    # omar[3] = omar[3] + max(facet_newlines)
    facet_newlines = ifelse(length(facet_newlines) == 0, 0, max(facet_newlines))
    omar[3] = omar[3] + facet_newlines * facet_text
    # apply the changes
    par(mar = omar)
  }

  # Now draw the individual facet windows (incl. axes, grid lines, and facet titles)
  # Will be skipped if adding to an existing plot; see ?facet

  dl_overshoot = 0
  if (direct_labels_flag && nfacets > 1) {
    dl_label_lines = max(strwidth(lgnd_labs, "inches")) / par("csi")
    dl_overshoot = dl_label_lines * cex_fct_adj * 0.5
  }

  facet_window_args = recordGraphics(
    draw_facet_window(
      add = add,
      # facet-specific args
      cex_fct_adj = cex_fct_adj,
      facet.args = facet.args,
      facet_newlines = facet_newlines, facet_font = facet_font,
      facet_rect = facet_rect, facet_text = facet_text,
      facet_col = facet_col, facet_bg = facet_bg, facet_border = facet_border,
      facet = facet,
      facets = facets, ifacet = ifacet,
      nfacets = nfacets, nfacet_cols = nfacet_cols, nfacet_rows = nfacet_rows,
      # axes args
      axes = axes, flip = flip, frame.plot = frame.plot,
      oxaxis = oxaxis, oyaxis = oyaxis,
      xlabs = xlabs, xlim = xlim, null_xlim = null_xlim, xaxt = xaxt, xaxs = xaxs, xaxb = xaxb, xaxl = xaxl,
      ylabs = ylabs, ylim = ylim, null_ylim = null_ylim, yaxt = yaxt, yaxs = yaxs, yaxb = yaxb, yaxl = yaxl,
      rev_x = rev_x, rev_y = rev_y,
      asp = asp, log = log,
      # other args (in approx. alphabetical + group ordering)
      dots = dots,
      draw = draw,
      grid = grid,
      has_legend = has_legend,
      main = main,
      sub = sub,
      cap = cap,
      type = type,
      xlab = xlab,
      x = x, xmax = xmax, xmin = xmin,
      ylab = ylab,
      y = y, ymax = ymax, ymin = ymin,
      tpars = tpars,
      dynmar_computed = dynmar_computed,
      dl_overshoot = dl_overshoot
    ),
    list = list(
      add = add,
      cex_fct_adj = cex_fct_adj,
      facet.args = facet.args,
      facet_newlines = facet_newlines, facet_font = facet_font,
      facet_rect = facet_rect, facet_text = facet_text,
      facet_col = facet_col, facet_bg = facet_bg, facet_border = facet_border,
      facet = datapoints$facet,
      facets = facets, ifacet = ifacet,
      nfacets = nfacets, nfacet_cols = nfacet_cols, nfacet_rows = nfacet_rows,
      axes = axes, flip = flip, frame.plot = frame.plot,
      oxaxis = oxaxis, oyaxis = oyaxis,
      xlabs = xlabs, xlim = xlim, null_xlim = null_xlim, xaxt = xaxt, xaxs = xaxs, xaxb = xaxb, xaxl = xaxl,
      ylabs = ylabs, ylim = ylim, null_ylim = null_ylim, yaxt = yaxt, yaxs = yaxs, yaxb = yaxb, yaxl = yaxl,
      rev_x = rev_x, rev_y = rev_y,
      asp = asp, log = log,
      dots = dots,
      draw = draw,
      grid = grid,
      has_legend = has_legend,
      main = main,
      sub = sub,
      cap = cap,
      type = type,
      xlab = xlab,
      x = datapoints$x, xmax = datapoints$xmax, xmin = datapoints$xmin,
      ylab = ylab,
      y = datapoints$y, ymax = datapoints$ymax, ymin = datapoints$ymin,
      tpars = tpar(), # https://github.com/grantmcdermott/tinyplot/issues/474
      dynmar_computed = dynmar_computed,
      dl_overshoot = dl_overshoot
    ),
    getNamespace("tinyplot")
  )
  list2env(facet_window_args, environment())

  #
  ## split and draw datapoints -----
  #

  # Finally, we can draw all of the plot elements (points, lines, etc.)
  # We'll do this via a nested loops:
  #  1) Outer loop over facets
  #  2) Inner loop over groups
  if (!is.null(datapoints$facet)) {
    split_data = split(datapoints, datapoints$facet)
    split_data = lapply(split_data, as.list)
  } else {
    split_data = list(as.list(datapoints))
  }

  if (direct_labels_flag) .dl_info = lapply(seq_along(split_data), function(x) vector("list", ngrps))

  ## Outer loop over the facets
  for (i in seq_along(split_data)) {
    # Split group-level data again to grab any "by" groups
    idata = split_data[[i]]
    iby = idata[["by"]]
    if (!null_by) { ## maybe all(iby=="")
      if (isTRUE(by_continuous)) {
        idata[["col"]] = col[round(rescale_num(idata$by, from = range(datapoints$by), to = c(1, 100)))]
        idata[["bg"]] = bg[round(rescale_num(idata$by, from = range(datapoints$by), to = c(1, 100)))]
        idata = list(idata)
      } else {
        idata = lapply(idata, split, iby)
        idata = do.call(function(...) Map("list", ...), idata)
      }
    } else {
      idata = list(idata)
      if (isTRUE(by_continuous)) {
        if (length(col) != 1) {
          idata[[1]][["col"]] = col[round(rescale_num(by, to = c(1, 100)))]
        } else {
          idata[[1]][["col"]] = col
        }
        if (length(bg) != 1) {
          idata[[1]][["bg"]] = bg[round(rescale_num(by, to = c(1, 100)))]
        } else {
          idata[[1]][["bg"]] = bg
        }
      }
    }
    
    # Set the facet "window" manually
    # See: https://github.com/grantmcdermott/tinyplot/issues/65
    if (nfacets > 1) {
      mfgi = ceiling(i / nfacet_cols)
      mfgj = i %% nfacet_cols
      if (mfgj == 0) mfgj = nfacet_cols
      par(mfg = c(mfgi, mfgj))

      # For free facets, we need to reset par(usr) based extent of that
      # particular facet... which we calculated and saved to the .fusr env var
      # (list) back in draw_facet_window()
      if (isTRUE(facet.args[["free"]])) {
        fusr = get(".fusr", envir = get(".tinyplot_env", envir = parent.env(environment())))
        par(usr = fusr[[i]])
      }
    }


    ## Inner loop over the "by" groupings
    for (ii in seq_along(idata)) {
      icol = col[ii]
      ibg = bg[ii]
      ipch = pch[ii]
      ilty = lty[ii]
      ilwd = lwd[ii]
      icex = if (bubble) idata[[ii]][["cex"]] else cex[[ii]]
      
      ix = idata[[ii]][["x"]]
      iy = idata[[ii]][["y"]]
      iz = idata[[ii]][["z"]]
      ixmin = idata[[ii]]$xmin
      ixmax = idata[[ii]]$xmax
      iymin = idata[[ii]]$ymin
      iymax = idata[[ii]]$ymax
      ilabels = idata[[ii]][["labels"]]

      if (isTRUE(by_continuous)) {
        icol = idata[[ii]]$col
        ibg = idata[[ii]]$bg
      }

      # empty plot flag
      empty_plot = FALSE
      if (isTRUE(empty) || isTRUE(type == "n") || ((length(ix) == 0) && !(type %in% c("histogram", "hist", "rect", "segments", "spineplot")))) {
        empty_plot = TRUE
      }

      # Draw the individual plot elements...
      if (!isTRUE(empty_plot)) {
        if (is.null(type_draw)) {
          type_draw = switch(type,
            "ribbon" = type_ribbon()$draw,
            "polygon" = type_polygon()$draw,
            "rect" = type_rect()$draw,
            "p" = ,
            "points" = type_points()$draw,
            "l" = ,
            "o" = ,
            "b" = ,
            "c" = ,
            "h" = ,
            "s" = ,
            "S" = type_lines(type = type)$draw
          )
        }
        type_draw(
          ibg = ibg,
          icol = icol,
          ilty = ilty,
          ilwd = ilwd,
          ipch = ipch,
          ix = ix,
          ixmax = ixmax,
          ixmin = ixmin,
          iy = iy,
          iymax = iymax,
          iymin = iymin,
          ilabels = ilabels,
          iz = iz,
          # cex = cex,
          icex = icex,
          dots = dots,
          type = type,
          x_by = x_by,
          by_continuous = by_continuous,
          iby = ii,
          ifacet = i,
          facet_by = facet_by,
          data_facet = idata,
          ngrps = ngrps,
          nfacets = nfacets,
          flip = flip,
          type_info = type_info,
          facet_window_args = facet_window_args
        )
      }
      if (direct_labels_flag && !empty_plot && length(ix) > 0) {
        .dl_info[[i]][[ii]] = list(x = tail(ix, 1), y = tail(iy, 1), col = icol)
      }
    }
  }

  if (direct_labels_flag) {
    dl_labs = lgnd_labs
    repel = legend_args[["repel"]]
    has_nudge = !is.null(legend_args[["nudge_x"]]) || !is.null(legend_args[["nudge_y"]])
    if (has_nudge && !is.null(repel) && !isFALSE(repel)) {
      warning("Direct labels: both `nudge` and `repel` specified. Using `nudge`.", call. = FALSE)
      repel = FALSE
    }
    nudge_x = rep(0, ngrps)
    nudge_y = rep(0, ngrps)
    if (has_nudge) {
      nudge_x = legend_args[["nudge_x"]] %||% rep(0, ngrps)
      nudge_y = legend_args[["nudge_y"]] %||% rep(0, ngrps)
      if (!is.null(names(nudge_x))) {
        idx = match(lgnd_labs, names(nudge_x))
        nudge_x = ifelse(is.na(idx), 0, nudge_x[idx])
      }
      if (!is.null(names(nudge_y))) {
        idx = match(lgnd_labs, names(nudge_y))
        nudge_y = ifelse(is.na(idx), 0, nudge_y[idx])
      }
      nudge_x = rep_len(nudge_x, ngrps)
      nudge_y = rep_len(nudge_y, ngrps)
    }

    for (fi in seq_along(.dl_info)) {
      if (nfacets > 1) {
        mfgi = ceiling(fi / nfacet_cols)
        mfgj = fi %% nfacet_cols
        if (mfgj == 0) mfgj = nfacet_cols
        par(mfg = c(mfgi, mfgj))
        if (isTRUE(facet.args[["free"]])) {
          fusr = get(".fusr", envir = get(".tinyplot_env", envir = parent.env(environment())))
          par(usr = fusr[[fi]])
        }
      }

      fi_info = .dl_info[[fi]]
      dl_x = vapply(fi_info, function(d) if (!is.null(d)) d$x else NA_real_, numeric(1))
      dl_y = vapply(fi_info, function(d) if (!is.null(d)) d$y else NA_real_, numeric(1))
      dl_x = dl_x + nudge_x
      dl_y = dl_y + nudge_y

      if (!is.null(repel) && !isFALSE(repel)) {
        min_gap = if (isTRUE(repel)) 0 else as.numeric(repel)
        dl_y = repel_text(
          x = rep(0, length(dl_y)), y = dl_y,
          widths = rep(0, length(dl_y)),
          heights = strheight(dl_labs, units = "user"),
          min_gap = min_gap, axis = "y"
        )[["y"]]
      }

      for (k in seq_along(fi_info)) {
        if (!is.null(fi_info[[k]])) {
          text(dl_x[k], dl_y[k], labels = dl_labs[k],
               col = fi_info[[k]]$col, pos = 4, offset = 0.3, xpd = NA)
        }
      }
    }
  }

  #
  ## save end pars for possible recall later -----
  #

  if (!add) {
    # Capture device and usr before recordGraphics (in current plot context)
    current_dev = dev.cur()
    current_usr = if (isTRUE(settings$flip)) par("usr")[c(3,4,1,2)] else par("usr")
    
    # Store usr and dev for validating layer alignment
    assign("usr_orig", current_usr, envir = get(".tinyplot_env", envir = parent.env(environment())))
    assign("dev_orig", current_dev, envir = get(".tinyplot_env", envir = parent.env(environment())))
    
    recordGraphics(
      {
        apar = par(no.readonly = TRUE)
        set_saved_par(when = "after", apar)
      },
      list = list(),
      env = getNamespace('tinyplot')
    )
  }

}




#' @rdname tinyplot
#' @importFrom stats as.formula model.frame terms
#' @export
tinyplot.formula = function(
    x = NULL,
    data = parent.frame(),
    facet = NULL,
    facet.args = NULL,
    type = NULL,
    xmin = NULL,
    xmax = NULL,
    ymin = NULL,
    ymax = NULL,
    weights = NULL,
    labels = NULL,
    xlim = NULL,
    ylim = NULL,
    # log = "",
    main = NULL,
    sub = NULL,
    cap = NULL,
    xlab = NULL,
    ylab = NULL,
    ann = par("ann"),
    axes = TRUE,
    frame.plot = NULL,
    asp = NA,
    grid = NULL,
    pch = NULL,
    col = NULL,
    lty = NULL,
    lwd = NULL,
    restore.par = FALSE,
    formula = NULL,
    subset = NULL,
    na.action = NULL,
    drop.unused.levels = TRUE,
    ...) {
  ## formula for variables must be specified through 'x' or 'formula' but not both
  if (is.null(x)) {
    if (missing(formula)) {
      stop("plot formula must be specified by either 'x' or 'formula' argument")
    }
  } else {
    if (missing(formula)) {
      formula = x
    } else {
      warning("only one of the arguments 'x' and 'formula' should be specified, defaulting to the 'formula' argument")
    }
  }

  ## placeholder for legend title
  legend_args = list(x = NULL)

  ## turn facet into a formula if it does not evaluate successfully
  if (inherits(try(facet, silent = TRUE), "try-error")) {
    facet = as.formula(paste("~", deparse(substitute(facet))))
    environment(facet) = environment(formula)
  }

  ## process all formulas
  tf = tinyformula(formula, facet)

  ## set up model frame
  m = match.call(expand.dots = FALSE)
  m = m[c(1L, match(c("formula", "data", "subset", "na.action", "drop.unused.levels", "xmin", "xmax", "ymin", "ymax", "weights", "labels"), names(m), 0L))]
  m$formula = tf$full
  ## need stats:: for non-standard evaluation
  m[[1L]] = quote(stats::model.frame)
  mf = eval.parent(m)

  ## extract x (if any)
  x = tinyframe(tf$x, mf)
  if (!is.null(x) && ncol(x) > 0L) {
    xnam = names(x)[[1L]]
    if (length(names(x)) > 1L) warning(paste("formula should specify at most one x-variable, using:", xnam),
    "\nif you want to use arithmetic operators, make sure to wrap them inside I()")
    x = x[[xnam]]
  } else {
    x = NULL
    xnam = NULL
  }

  ## extract y (if any)
  y = tinyframe(tf$y, mf)
  if (!is.null(y) && ncol(y) > 0L) {
    ynam = names(y)[[1L]]
    if (length(names(y)) > 1L) warning(paste("formula should specify at most one y-variable, using:", ynam),
    "\nif you want to use arithmetic operators, make sure to wrap them inside I()")
    y = y[[ynam]]
  } else {
    y = NULL
    ynam = NULL
  }

  ## extract by (if any)
  by = tinyframe(tf$by, mf)
  if (!is.null(by) && ncol(by) > 0L) {
    bynam = names(by)
    by = if (length(bynam) == 1L) by[[bynam]] else interaction(by, sep = ":")
  } else {
    by = NULL
  }

  ## extract x/y facet (if formula)
  if (!is.null(tf$xfacet) || !is.null(tf$yfacet)) {
    ## xfacet/yfacet can be either:
    ## - no formula: NULL
    ## - formula without variables: empty (zero-column) data
    ## - formula with variable(s): data frame
    xfacet = tinyframe(tf$xfacet, mf)
    yfacet = tinyframe(tf$yfacet, mf)
    xtype = if (is.null(xfacet)) "none" else if (ncol(xfacet) == 0L) "empty" else "data"
    ytype = if (is.null(yfacet)) "none" else if (ncol(yfacet) == 0L) "empty" else "data"
    
    ## turn data frame (if specified) into a single factor
    if (xtype == "data") xfacet = if (ncol(xfacet) == 1L) xfacet[[1L]] else interaction(xfacet, sep = ":")
    if (ytype == "data") yfacet = if (ncol(yfacet) == 1L) yfacet[[1L]] else interaction(yfacet, sep = ":")

    ## set up actual facet, potentially with grid information
    if (xtype %in% c("none", "empty") && ytype %in% c("none", "empty")) {
      facet = NULL
    } else {
      if (xtype %in% c("none", "empty")) {
        facet = yfacet
        if (xtype == "empty") {
          if (is.null(facet.args)) facet.args = list()
          if (is.null(facet.args[["nrow"]])) facet.args[["nrow"]] = length(unique(yfacet))
        }
      } else if (ytype %in% c("none", "empty")) {
        facet = xfacet
        if (ytype == "empty") {
          if (is.null(facet.args)) facet.args = list()
          if (is.null(facet.args[["nrow"]])) facet.args[["nrow"]] = 1L
        }
      } else {
        facet = interaction(xfacet, yfacet, sep = "~")
        attr(facet, "facet_grid") = TRUE
        attr(facet, "facet_nrow") = length(unique(yfacet))
      }
    }
  }

  ## nice axis and legend labels
  dens_type = !is.null(type) && (is.atomic(type) && identical(type, "density")) || (!is.atomic(type) && identical(type$name, "density"))
  hist_type = !is.null(type) && (is.atomic(type) && type %in% c("hist", "histogram")) || (!is.atomic(type) && identical(type$name, "histogram"))
  barp_type = !is.null(type) &&  (is.atomic(type) && identical(type, "barplot")) || (!is.atomic(type) && identical(type$name, "barplot"))
  if (is.null(x) && is.null(y)) {
    # Exception: both x and y NULL (e.g., ~ 0 with type = "segments").
    # Build labels from xmin/xmax/ymin/ymax names in the original call (m),
    # since deparse(substitute()) in the default method would see mf[["..."]].
    if (is.null(xlab) && !is.null(m[["xmin"]]) && !is.null(m[["xmax"]])) {
      xlab = sprintf("[%s, %s]", deparse1(m[["xmin"]]), deparse1(m[["xmax"]]))
    }
    if (is.null(ylab) && !is.null(m[["ymin"]]) && !is.null(m[["ymax"]])) {
      ylab = sprintf("[%s, %s]", deparse1(m[["ymin"]]), deparse1(m[["ymax"]]))
    }
  } else if (is.null(x) && !is.null(y)) {
    # Exception: univariate y ~ 1 formulas. sanitize_type() will swap x/y and
    # infer the type (histogram or barplot). Set xlab from the variable name
    # and let sanitize_xylab() determine ylab after the type is known.
    if (is.null(xlab)) xlab = ynam
  } else if (dens_type) {
    # if (is.null(ylab)) ylab = "Density" ## rather assign ylab as part of internal type_density() logic
    if (is.null(xlab)) xlab = xnam
  } else if (hist_type) {
    # if (is.null(ylab)) ylab = "Frequency" ## rather assign ylab as part of internal type_histogram() logic
    if (is.null(xlab)) xlab = xnam
  } else if (is.null(y)) {
    if (is.factor(x) || is.character(x) || barp_type) {
      if (is.null(xlab)) xlab = xnam
    } else {
      if (is.null(ylab)) ylab = xnam
      if (is.null(xlab)) xlab = "Index"
    }
  } else {
    if (is.null(ylab)) ylab = ynam
    if (is.null(xlab)) xlab = xnam
  }
  if (!is.null(by)) {
    legend_args[["title"]] = if (length(bynam) == 1L) bynam else sprintf("interaction(%s)", paste(bynam, collapse = ", "))
  }

  tinyplot.default(
    x = x, y = y, by = by,
    facet = facet, facet.args = facet.args,
    data = data,
    type = type,
    xmin = mf[["(xmin)"]],
    xmax = mf[["(xmax)"]],
    ymin = mf[["(ymin)"]],
    ymax = mf[["(ymax)"]],
    weights = mf[["(weights)"]],
    labels = mf[["(labels)"]],
    xlim = xlim,
    ylim = ylim,
    # log = "",
    main = main,
    sub = sub,
    cap = cap,
    xlab = xlab,
    ylab = ylab,
    ann = ann,
    axes = axes,
    frame.plot = frame.plot,
    asp = asp,
    grid = substitute(grid), # issue #193
    legend_args = legend_args,
    pch = pch,
    col = col,
    lty = lty,
    lwd = lwd,
    restore.par = restore.par,
    ...
  )
}

#' @rdname tinyplot
#' @export
tinyplot.density = function(
    x = NULL,
    type = c("l", "area"),
    ...) {
  
  dots = list(...)
  
  if (!is.null(dots[["by"]]) || !is.null(dots[["facet"]])) {
    stop(
      '\nGrouped and/or faceted plots are no longer supported with the tinyplot.density() method. ',
      '\nPlease use the dedicated type argument instead, e.g. `tinyplot(..., type = "density")`. See `?type_density` for details.',
      '\n\nThis breaking change was introduced in tinyplot v0.3.0.'
    )
  }
  
  type = match.arg(type)
  
  ## override if bg = "by"
  if (!is.null(dots[["bg"]]) || !is.null(dots[["fill"]])) type = "area"
  
  if (inherits(x, "density")) {
    object = x
    # legend_args = list(x = NULL)
    # # Grab by label to pass on legend title to tinyplot.default
    # legend_args[["title"]] = deparse(substitute(by))
  } else {
    ## An internal catch for non-density objects that were forcibly
    ## passed to tinyplot.density (e.g., via a one-side formula)
    if (anyNA(x)) {
      x = na.omit(x)
      x = as.numeric(x)
    }
    object = density(x)
  }
  
  x = object$x
  y = object$y
  
  if (type == "area") {
    ymin = rep(0, length(y))
    ymax = y
    # # set extra legend params to get bordered boxes with fill
    # legend_args[["x.intersp"]] = 1.25
    # legend_args[["lty"]] = 0
    # legend_args[["pt.lwd"]] = 1
  }
  
  # splice in change arguments
  dots[["x"]] = x
  dots[["y"]] = y
  dots[["type"]] = type
  
  ## axes range
  if (is.null(dots[["xlim"]])) dots[["xlim"]] = range(x)
  if (is.null(dots[["ylim"]])) dots[["ylim"]] = range(y)
  
  ## nice labels and titles
  if (is.null(dots[["ylab"]])) dots[["ylab"]] = "Density"
  if (is.null(dots[["xlab"]])) dots[["xlab"]] = paste0("N = ", object$n, "   Bandwidth = ", sprintf("%.4g", object$bw))
  if (is.null(dots[["main"]])) dots[["main"]] = paste0(paste(object$call, collapse = "(x = "), ")")
  
  do.call(tinyplot.default, args = dots)
  
}


#' @export
#' @name plt
#' @rdname tinyplot
plt = tinyplot

#' Draw facet windows
#' 
#' @description Internal functions called from `tinyplot` in order to draw the
#' plot window with different facets, grids, axes, etc. 
#' 
#' `facet_layout` determines the layout of the facets, based on a set of inputs.
#' 
#' `draw_facet_window` is the main workhorse function for setting the exterior
#' plot elements as part of a `tinyplot` call, including adjustment of margins
#' for dynamic themes, etc.
#' 
#' @keywords internal
#' @rdname facet
draw_facet_window = function(
    # add arg first, since that determines what happens (if at all)
    add,
    # facet-specific args
    cex_fct_adj,
    facet.args,
    facet_newlines, facet_font, facet_rect, facet_text,
    facet_col, facet_bg, facet_border,
    facet, facets, ifacet,
    nfacets, nfacet_cols, nfacet_rows,
    # axes args
    axes, flip, frame.plot, oxaxis, oyaxis,
    xlabs, xlim, null_xlim, xaxt, xaxs, xaxb, xaxl,
    ylabs, ylim, null_ylim, yaxt, yaxs, yaxb, yaxl,
    rev_x = FALSE, rev_y = FALSE,
    asp, log,
    # other args (in approx. alphabetical + group ordering)
    dots,
    draw,
    grid,
    has_legend,
    main,
    sub,
    cap,
    type,
    xlab,
    x, xmax, xmin,
    ylab,
    y, ymax, ymin,
    tpars = NULL,
    dynmar_computed = NULL,
    dl_overshoot = 0
    ) {
  
  if (is.null(tpars)) tpars = tpar()
  
  # if add is TRUE, just return inputs without any calculations
  if (isTRUE(add)) {
    return(as.list(environment()))
  }
  
  # if breaks are provided use these (but only if x/ylabs are null)
  if (!is.null(xaxb) && !is.null(xlabs)) xlabs = xaxb
  if (!is.null(yaxb) && !is.null(ylabs)) ylabs = yaxb
  
  # draw background color only in the grid rectangle
  grid.bg = get_tpar("grid.bg", tpar_list = tpars)
  if (!is.null(grid.bg)) {
    corners = par("usr")
    rect(corners[1], corners[3], corners[2], corners[4], col = grid.bg, border = NA)
  }

  ## dynamic margins flag
  dynmar = isTRUE(get_tpar("dynmar", tpar_list = tpars))
  
  ## optionally allow to modify and restore the style of axis interval calculation
  if (!is.null(xaxs) || !is.null(yaxs)) {
    op = par()
    if (!is.null(xaxs)) {
      par(xaxs = xaxs)
      on.exit(par(xaxs = op$xaxs), add = TRUE)
    }
    if (!is.null(yaxs)) {
      par(yaxs = yaxs)
      on.exit(par(yaxs = op$yaxs), add = TRUE)
    }
  }

  if (nfacets > 1) {
    # Set facet margins (i.e., gaps between facets)
    if (is.null(facet.args[["fmar"]])) {
      fmar = tpar("fmar")
    } else {
      if (length(facet.args[["fmar"]]) != 4) {
        warning(
          "`fmar` has to be a vector of length four, e.g.",
          "`facet.args = list(fmar = c(b,l,t,r))`.",
          "\n",
          "Resetting to fmar = c(1,1,1,1) default.",
          "\n"
        )
        fmar = tpar("fmar")
      } else {
        fmar = facet.args[["fmar"]]
      }
    }
    # We need to adjust for n>=3 facet cases for correct spacing...
    if (nfacets >= 3) {
      ## ... exception for 2x2 cases
      if (!(nfacet_rows == 2 && nfacet_cols == 2)) fmar = fmar * .75
    }
    # Extra reduction if no plot frame to reduce whitespace
    if (isFALSE(frame.plot) && !isTRUE(facet.args[["free"]])) {
      fmar = fmar - 0.5
    }

    ooma = par("oma")

    # Bump top margin for facet strip. Use facet_text (not / cex_fct_adj)
    # because nmar = (fmar + 0.1) / cex_fct_adj already divides — using
    # facet_text directly keeps the inter-panel gap constant as newlines grow.
    fmar[3] = fmar[3] + facet_text
    if (isTRUE(attr(facet, "facet_grid"))) {
      fmar[3] = max(0, fmar[3] - facet_text)
      # Indent for RHS facet_grid title strip if "right!" legend
      if (has_legend && ooma[4] > 0) ooma[4] = ooma[4] + 1
    }
    fmar[3] = fmar[3] + facet_newlines * facet_text

    omar = par("mar")
    
    ## Dynamic plot margin adjustments
    if (dynmar) {
      # Margins were pre-computed in tinyplot.default (dynmar_computed).
      # Use that as the base instead of par("mar") which may have been
      # reset by the before.plot.new hook.
      side.sub = get_tpar("side.sub", tpar_list = tpars, default = 3)
      omar = dynmar_computed
      omar[3] = dynmar_computed[3] + (1 + facet_newlines + 0.1) * facet_text
      # Ensure fmar[3] doesn't exceed omar[3] - 0.1, which would make
      # noma[3] negative and get clamped to 0, creating excess top space.
      if (fmar[3] + 0.1 > omar[3]) fmar[3] = omar[3] - 0.1
      if (par("las") %in% 1:2) {
        # extra whitespace bump on the y axis
        ## overrides for ridge and some types that use integer spacing with (named) axis labels ## FXIME
        if (type == "ridge") {
          yaxlabs = levels(y)
        } else if (!is.null(ylabs)) {
          yaxlabs = if (!is.null(names(ylabs))) names(ylabs) else ylabs 
        } else if (type == "boxplot" && isTRUE(flip) && !is.null(xlabs)) {
          yaxlabs = if (!is.null(names(xlabs))) names(xlabs) else xlabs 
        } else if (isTRUE(facet.args[["free"]]) && null_ylim && !is.null(facet)) {
          yfree_split = split(c(y, ymin, ymax), facet)
          yaxlabs_all = lapply(yfree_split, function(yf) {
            axisTicks(usr = extendrange(range(yf, na.rm = TRUE), f = 0.04), log = par("ylog"))
          })
          widths = vapply(yaxlabs_all, function(labs) max(strwidth(labs, "inches", cex = par("cex.axis"))), numeric(1L))
          yaxlabs = yaxlabs_all[[which.max(widths)]]
        } else {
          yaxlabs = axisTicks(usr = extendrange(ylim, f = 0.04), log = par("ylog"))
        }
        if (!is.null(yaxl)) yaxlabs = tinylabel(yaxlabs, yaxl)
        # whtsbp = grconvertX(max(strwidth(yaxl, "figure")), from = "nfc", to = "lines") - 1
        whtsbp = grconvertX(max(strwidth(yaxlabs, "figure", cex = par("cex.axis"))), from = "nfc", to = "lines") - grconvertX(0, from = "nfc", to = "lines") - 0.5
        if (whtsbp > 0) {
          omar = omar + c(0, whtsbp, 0, 0) * cex_fct_adj
          fmar[2] = fmar[2] + whtsbp * cex_fct_adj
        }
        # Extra reduction if no plot frame to reduce whitespace
        if (isFALSE(frame.plot) && !isTRUE(facet.args[["free"]])) {
          fmar[2] = fmar[2] - (whtsbp * cex_fct_adj)
        }
      }
      if (par("las") %in% 2:3) {
        # extra whitespace bump on the x axis
        if (is.null(xlabs) && isTRUE(facet.args[["free"]]) && null_xlim && !is.null(facet)) {
          xfree_split = split(c(x, xmin, xmax), facet)
          xaxlabs_all = lapply(xfree_split, function(xf) {
            axisTicks(usr = extendrange(range(xf, na.rm = TRUE), f = 0.04), log = par("xlog"))
          })
          widths = vapply(xaxlabs_all, function(labs) max(strwidth(labs, "inches", cex = par("cex.axis"))), numeric(1L))
          xaxlabs = xaxlabs_all[[which.max(widths)]]
        } else {
          xaxlabs = if (is.null(xlabs)) axisTicks(usr = extendrange(xlim, f = 0.04), log = par("xlog")) else
            if (!is.null(names(xlabs))) names(xlabs) else xlabs
        }
        if (!is.null(xaxl)) xaxlabs = tinylabel(xaxlabs, xaxl)
        whtsbp = grconvertX(max(strwidth(xaxlabs, "figure", cex = par("cex.axis"))), from = "nfc", to = "lines") - 0.5
        if (whtsbp > 0) {
          omar = omar + c(whtsbp, 0, 0, 0) * cex_fct_adj
          fmar[1] = fmar[1] + whtsbp * cex_fct_adj
        }
        # Extra reduction if no plot frame to reduce whitespace
        if (isFALSE(frame.plot) && !isTRUE(facet.args[["free"]])) {
          fmar[1] = fmar[1] - (whtsbp * cex_fct_adj)
        }
      }

      if (type == "spineplot") omar[4] = 2.1 # FIXME catch for spineplot RHS axis labs

      # FIXME: Is this causing issues for lhs legends with facet_grid?
      # catch for missing rhs legend
      if (isTRUE(attr(facet, "facet_grid")) && !has_legend) {
        omar[4] = omar[4] + 1
      }
    }

    if (dl_overshoot > 0) {
      fmar[4] = fmar[4] + dl_overshoot
    }

    # Now we set the margins. The trick here is that we simultaneously adjust
    # inner (mar) and outer (oma) margins by the same amount, but in opposite
    # directions, to preserve the overall facet and plot centroids.
    nmar = (fmar + .1) / cex_fct_adj
    noma = (ooma + omar - fmar - .1) / cex_fct_adj
    # Catch in case of negative oma values. (Probably only occurs with some
    # user-supplied tpar(lmar) values and a "left!" positioned legend.)
    if (any(noma < 0)) {
      noma_orig = noma
      noma[noma < 0] = 0
      # noma_diff = noma-noma_orig
      # nmar = nmar + noma_diff
    }
    # apply changes
    par(oma = noma)
    par(mar = nmar)

    # Now that the margins have been set, arrange facet rows and columns based
    # on our earlier calculations.
    par(mfrow = c(nfacet_rows, nfacet_cols))
  } else if (dynmar) {
    # Dynamic plot margin adjustments (no facets). Margins were pre-computed
    # in tinyplot.default and passed via dynmar_computed; use them directly.
    # Tick-label *width/height* (whtsbp) is added further below.
    side.sub = get_tpar("side.sub", tpar_list = tpars, default = 3)
    omar = dynmar_computed
    if (type == "spineplot") omar[4] = 2.1 # FIXME catch for spineplot RHS axis labs
    if (par("las") %in% 1:2) {
      # extra whitespace bump on the y axis
      ## overrides for ridge and some types that use integer spacing with (named) axis labels ## FXIME
      if (type == "ridge") {
        yaxlabs = levels(y)
      } else if (!is.null(ylabs)) {
        yaxlabs = if (!is.null(names(ylabs))) names(ylabs) else ylabs 
      } else if (type == "boxplot" && isTRUE(flip) && !is.null(xlabs)) {
        yaxlabs = if (!is.null(names(xlabs))) names(xlabs) else xlabs 
      } else {
        # yaxl = axTicks(2)
        ylim_usr = if (diff(ylim) == 0 && is.null(yaxb)) ylim + c(-0.5, 0.5) else extendrange(ylim, f = 0.04)
        yaxlabs = axisTicks(usr = ylim_usr, log = par("ylog"))
      }
      if (!is.null(yaxl)) yaxlabs = tinylabel(yaxlabs, yaxl)
      # whtsbp = grconvertX(max(strwidth(yaxlabs, "figure", cex = par("cex.axis"))), from = "nfc", to = "lines") - 1
      whtsbp = grconvertX(max(strwidth(yaxlabs, "figure", cex = par("cex.axis"))), from = "nfc", to = "lines") - grconvertX(0, from = "nfc", to = "lines") - 0.5
      omar[2] = omar[2] + whtsbp
    }
    if (par("las") %in% 2:3) {
      # extra whitespace bump on the x axis
      # xaxl = axTicks(1)
      xlim_usr = if (diff(xlim) == 0 && is.null(xaxb)) xlim + c(-0.5, 0.5) else extendrange(xlim, f = 0.04)
      xaxlabs = if (is.null(xlabs)) axisTicks(usr = xlim_usr, log = par("xlog")) else
        if (!is.null(names(xlabs))) names(xlabs) else xlabs
      if (!is.null(xaxl)) xaxlabs = tinylabel(xaxlabs, xaxl)
      whtsbp = grconvertX(max(strwidth(xaxlabs, "figure", cex = par("cex.axis"))), from = "nfc", to = "lines") - 0.5
      omar[1] = omar[1] + whtsbp
    }

     par(mar = omar)
  }

  ## Loop over the individual facet windows and draw the plot region
  ## components (axes, titles, box, grid, etc.)
  for (ii in ifacet) {
    # See: https://github.com/grantmcdermott/tinyplot/issues/65
    if (nfacets > 1) {
      mfgi = ceiling(ii / nfacet_cols)
      mfgj = ii %% nfacet_cols
      if (mfgj == 0) mfgj = nfacet_cols
      par(mfg = c(mfgi, mfgj))
    }

    ## Set the plot window
    ## Problem: Passing extra args through ... (e.g., legend_args) to plot.window
    ## triggers an annoying warning about unrecognized graphical params.
    # plot.window(
    #   xlim = xlim, ylim = ylim,
    #   asp = asp, log = log,
    #   # ...
    # )
    ## Solution: Only pass on relevant args using name checking and do.call.
    ## Idea borrowed from here: https://stackoverflow.com/a/4128401/4115816
    pdots = dots[names(dots) %in% names(formals(plot.default))]
    ## catch for flipped boxplots...
    if (type == "boxplot" && isTRUE(flip)) {
      log_flip = log
      if (!is.null(log)) {
        if (log == "x") log_flip = "y"
        if (log == "y") log_flip = "x"
      }
      do.call(
        "plot.window",
        c(list(xlim = ylim, ylim = xlim, asp = asp, log = log_flip), pdots)
      )
      xside = 2
      yside = 1
    } else {
      ## ... standard plot window for all other cases
      do.call(
        "plot.window",
        c(list(xlim = xlim, ylim = ylim, asp = asp, log = log), pdots)
      )
      xside = 1
      yside = 2
    }


    # axes, frame.plot and grid
    if (isTRUE(axes) || isTRUE(facet.args[["free"]])) {
      args_x = list(x,
        side = xside,
        type = xaxt,
        labeller = xaxl,
        cex = get_tpar(c("cex.xaxs", "cex.axis"), 0.8, tpar_list = tpars),
        lwd = get_tpar(c("lwd.xaxs", "lwd.axis"), 1, tpar_list = tpars),
        lty = get_tpar(c("lty.xaxs", "lty.axis"), 1, tpar_list = tpars)
      )
      .ca = get_tpar(c("cex.yaxs", "cex.axis"), 0.8, tpar_list = tpars)
      .ymgp_shift = if (par("las") %in% c(0L, 1L)) 0.5 * (.ca - 1) else 0
      args_y = list(y,
        side = yside,
        type = yaxt,
        labeller = yaxl,
        cex = .ca,
        lwd = get_tpar(c("lwd.yaxs", "lwd.axis"), 1, tpar_list = tpars),
        lty = get_tpar(c("lty.yaxs", "lty.axis"), 1, tpar_list = tpars)
      )
      if (!is.null(xaxb)) args_x$at = xaxb
      if (!is.null(yaxb)) args_y$at = yaxb
      # `xlabs` is only non-NULL when a type has placed categorical data on the
      # x-axis, so its presence is the signal to draw labelled ticks.
      type_range_x = !is.null(xlabs)
      type_range_y = !is.null(ylabs) && (type == "p" || (isTRUE(flip) && type %in% c("barplot", "pointrange", "errorbar", "ribbon", "boxplot", "violin")))
      if (type_range_x) {
        args_x = modifyList(args_x, list(at = xlabs, labels = names(xlabs)))
      }
      if (type_range_y) {
        args_y = modifyList(args_y, list(at = ylabs, labels = names(ylabs)))
      }

      if (isTRUE(facet.args[["free"]]) && (par("xlog") || par("ylog"))) {
        warning(
          "\nFree scale axes for faceted plots are currently not supported if the axes are logged. Reverting back to fixed scales.",
          "\nIf support for this feature is important to you, please raise an issue on our GitHub repo:",
          "\nhttps://github.com/grantmcdermott/tinyplot/issues\n"
        )
        facet.args[["free"]] = FALSE
      }

      # Special logic if facets are free...
      if (isTRUE(facet.args[["free"]])) {
        # First, we need to calculate the plot extent and axes range of each
        # individual facet.
        xfree = if (!is.null(facet)) split(c(x, xmin, xmax), facet)[[ii]] else c(x, xmin, xmax)
        yfree = if (!is.null(facet)) split(c(y, ymin, ymax), facet)[[ii]] else c(y, ymin, ymax)
        if (null_xlim) xlim = range(xfree, na.rm = TRUE)
        if (null_ylim) ylim = range(yfree, na.rm = TRUE)
        # An axis is reversed either via the `rev_x`/`rev_y` flag (e.g. the
        # "reverse" keyword) or when the user supplies descending fixed limits
        # (e.g. xlim = c(10, 0)). The latter must be detected before extendrange()
        # below, which always returns an ascending pair and would otherwise drop
        # the descending order. (#644)
        rev_xext = isTRUE(rev_x) || (!null_xlim && length(xlim) == 2L && xlim[2L] < xlim[1L])
        rev_yext = isTRUE(rev_y) || (!null_ylim && length(ylim) == 2L && ylim[2L] < ylim[1L])
        # extendrange() returns an ascending pair, so reverse afterwards
        xext = extendrange(sort(xlim), f = 0.04)
        yext = extendrange(sort(ylim), f = 0.04)
        # base axTicks() misbehaves on a reversed usr (it collapses to a single
        # tick), so precompute ticks from the ascending extent and pass them as
        # an explicit `at` below; placement against the reversed usr is fine.
        xat = if (rev_xext) axisTicks(usr = xext, log = par("xlog")) else NULL
        yat = if (rev_yext) axisTicks(usr = yext, log = par("ylog")) else NULL
        if (rev_xext) xext = rev(xext)
        if (rev_yext) yext = rev(yext)
        # We'll save this in a special .fusr env var (list) that we'll re-use
        # when it comes to plotting the actual elements later
        if (ii == 1) {
          fusr = replicate(4, vector("double", length = nfacets), simplify = FALSE)
          assign(".fusr", fusr, envir = get(".tinyplot_env", envir = parent.env(environment())))
        }
        fusr = get(".fusr", envir = get(".tinyplot_env", envir = parent.env(environment())))
        fusr[[ii]] = c(xext, yext)
        assign(".fusr", fusr, envir = get(".tinyplot_env", envir = parent.env(environment())))
        # Explicitly set (override) the current facet extent
        par(usr = fusr[[ii]])
        # if plot frame is true then print axes per normal...
        if (!is.null(xlabs)) {
          tinyAxis(xfree, side = xside, at = xlabs, labels = names(xlabs), type = xaxt, labeller = xaxl)
        } else if (!is.null(xat)) {
          tinyAxis(xfree, side = xside, at = xat, type = xaxt, labeller = xaxl)
        } else {
          tinyAxis(xfree, side = xside, type = xaxt, labeller = xaxl)
        }
        if (.ymgp_shift > 0) par(mgp = par("mgp") - c(0, .ymgp_shift, 0))
        if (isTRUE(flip) && type %in% c("barplot", "pointrange", "errorbar", "ribbon", "boxplot", "p", "violin") && !is.null(ylabs)) {
          tinyAxis(yfree, side = yside, at = ylabs, labels = names(ylabs), type = yaxt, labeller = yaxl)
        } else if (!is.null(yat)) {
          tinyAxis(yfree, side = yside, at = yat, type = yaxt, labeller = yaxl)
        } else {
          tinyAxis(yfree, side = yside, type = yaxt, labeller = yaxl)
        }
        if (.ymgp_shift > 0) par(mgp = par("mgp") + c(0, .ymgp_shift, 0))

        # For fixed facets we can just reuse the same plot extent and axes limits
      } else if (isTRUE(frame.plot)) {
        # if plot frame is true then print axes per normal...
        do.call(tinyAxis, args_x)
        if (.ymgp_shift > 0) par(mgp = par("mgp") - c(0, .ymgp_shift, 0))
        do.call(tinyAxis, args_y)
        if (.ymgp_shift > 0) par(mgp = par("mgp") + c(0, .ymgp_shift, 0))
      } else {
        # ... else only print the "outside" axes.
        if (ii %in% oxaxis) do.call(tinyAxis, args_x)
        if (.ymgp_shift > 0) par(mgp = par("mgp") - c(0, .ymgp_shift, 0))
        if (ii %in% oyaxis) do.call(tinyAxis, args_y)
        if (.ymgp_shift > 0) par(mgp = par("mgp") + c(0, .ymgp_shift, 0))
      }
    }

    # facet titles
    ## Note: facet titles could be done more simply with mtext... but then we
    ## couldn't adjust background features (e.g., fill), or rotate the rhs
    ## facet grid text. So we're rolling our own "manual" versions with text
    ## and rect.
    if (!is.null(facet)) {
      # Get the four corners of plot area (x1, x2, y1, y2)
      corners = par("usr")
      # catch for logged axes
      xlog = isTRUE(par("xlog"))
      ylog = isTRUE(par("ylog"))
      if (xlog) corners[1:2] = 10^(corners[1:2])
      if (ylog) corners[3:4] = 10^(corners[3:4])
      # special logic for facet grids
      if (is.null(facet_newlines) || facet_newlines == 0) {
        facet_title_lines = 1
      } else {
        facet_title_lines = 1 + facet_newlines
      }
      # different logic for facet grids versus regular facets
      if (isTRUE(attr(facet, "facet_grid"))) {
        ## top facet strips
        if (ii %in% 1:nfacet_cols) {
          line_height_lines = (facet_title_lines + .1) * facet_text / cex_fct_adj
          if (isTRUE(facet_rect)) {
            if (ylog) {
              line_height = grconvertY(line_height_lines, from = "lines", to = "user") / grconvertY(0, from = "lines", to = "user")
              rect_height = corners[4] * line_height
            } else {
              line_height = grconvertY(line_height_lines, from = "lines", to = "user") - grconvertY(0, from = "lines", to = "user")
              rect_height = corners[4] + line_height
            }
            rect(
              corners[1], corners[4], corners[2], rect_height,
              col = facet_bg, border = facet_border,
              xpd = NA
            )
          }
          xpos = if (xlog) 10^(mean(log10(corners[1:2]))) else mean(corners[1:2])
          if (ylog) {
            ypos = grconvertY(line_height_lines / 2, from = "lines", to = "user") / grconvertY(0, from = "lines", to = "user")
            ypos = corners[4] * ypos
          } else {
            ypos = grconvertY(line_height_lines / 2, from = "lines", to = "user") - grconvertY(0, from = "lines", to = "user")
            ypos = corners[4] + ypos
          }
          text(
            x = xpos,
            y = ypos,
            labels = sub("^(.*?)~.*", "\\1", facets[[ii]]),
            adj = c(0.5, 0.5),
            cex = facet_text / cex_fct_adj,
            col = facet_col,
            font = facet_font,
            xpd = NA,
          )
        }
        ## right facet strips
        if (ii %% nfacet_cols == 0 || ii == nfacets) {
          line_height_lines = (facet_title_lines + .1) * facet_text / cex_fct_adj
          if (isTRUE(facet_rect)) {
            if (xlog) {
              line_height = grconvertX(line_height_lines, from = "lines", to = "user") / grconvertX(0, from = "lines", to = "user")
              rect_width = corners[2] * line_height
            } else {
              line_height = grconvertX(line_height_lines, from = "lines", to = "user") - grconvertX(0, from = "lines", to = "user")
              rect_width = corners[2] + line_height
            }
            rect(
              corners[2], corners[3], rect_width, corners[4],
              col = facet_bg, border = facet_border,
              xpd = NA
            )
          }
          if (xlog) {
            xpos = grconvertX(line_height_lines / 2, from = "lines", to = "user") / grconvertX(0, from = "lines", to = "user")
            xpos = corners[2] * xpos
          } else {
            xpos = grconvertX(line_height_lines / 2, from = "lines", to = "user") - grconvertX(0, from = "lines", to = "user")
            xpos = corners[2] + xpos
          }
          ypos = if (ylog) 10^(mean(log10(corners[3:4]))) else mean(corners[3:4])
          text(
            x = xpos,
            y = ypos,
            labels = sub("^.*?~(.*)", "\\1", facets[[ii]]),
            srt = 270,
            adj = c(0.5, 0.5),
            cex = facet_text / cex_fct_adj,
            col = facet_col,
            font = facet_font,
            xpd = NA
          )
        }
      } else {
        line_height_lines = (facet_title_lines + .1) * facet_text / cex_fct_adj
        if (isTRUE(facet_rect)) {
          if (ylog) {
            line_height = grconvertY(line_height_lines, from = "lines", to = "user") / grconvertY(0, from = "lines", to = "user")
            rect_height = corners[4] * line_height
          } else {
            line_height = grconvertY(line_height_lines, from = "lines", to = "user") - grconvertY(0, from = "lines", to = "user")
            rect_height = corners[4] + line_height
          }
          rect(
            corners[1], corners[4], corners[2], rect_height,
            col = facet_bg, border = facet_border,
            xpd = NA
          )
        }
        xpos = if (xlog) 10^(mean(log10(corners[1:2]))) else mean(corners[1:2])
        if (ylog) {
          ypos = grconvertY(line_height_lines / 2, from = "lines", to = "user") / grconvertY(0, from = "lines", to = "user")
          ypos = corners[4] * ypos
        } else {
          ypos = grconvertY(line_height_lines / 2, from = "lines", to = "user") - grconvertY(0, from = "lines", to = "user")
          ypos = corners[4] + ypos
        }
        text(
          x = xpos,
          y = ypos,
          labels = paste(facets[[ii]]),
          adj = c(0.5, 0.5),
          cex = facet_text / cex_fct_adj,
          col = facet_col,
          font = facet_font,
          xpd = NA
        )
      }
    }

    # plot frame
    if (frame.plot) box()

    # panel grid lines
    if (is.null(grid)) grid = get_tpar("grid", tpar_list = tpars)
    if (!is.null(grid) && !isFALSE(grid)) {
      gcol = get_tpar("grid.col", tpar_list = tpars)
      glty = get_tpar("grid.lty", tpar_list = tpars)
      glwd = get_tpar("grid.lwd", tpar_list = tpars)

      if (isTRUE(grid)) {
        draw_x = draw_y = TRUE
        fine_x = fine_y = FALSE
      } else if (is.character(grid)) {
        draw_x = grepl("x", grid, fixed = TRUE) || grepl("X", grid, fixed = TRUE)
        draw_y = grepl("y", grid, fixed = TRUE) || grepl("Y", grid, fixed = TRUE)
        fine_x = grepl("x", grid, fixed = TRUE)
        fine_y = grepl("y", grid, fixed = TRUE)
      } else {
        eval(grid) # issue #193
        draw_x = draw_y = FALSE
      }

      if (draw_x || draw_y) {

        if (draw_x) {
          if (!is.null(xaxb)) {
            xg = xaxb
          } else {
            xg = if (!inherits(x, c("POSIXt", "Date"))) axTicks(side = 1) else axTicksDateTime(side = 1, x = x)
          }
          if (fine_x && !par("xlog") && length(xg) >= 2L) {
            xg = as.numeric(xg)
            half = (xg[2L] - xg[1L]) / 2
            xg = seq(xg[1L] - half, xg[length(xg)] + half, by = half)
          }
          abline(v = xg, col = gcol, lty = glty, lwd = glwd)
        }

        if (draw_y) {
          if (!is.null(yaxb)) {
            yg = yaxb
          } else {
            yg = if (!inherits(y, c("POSIXt", "Date"))) axTicks(side = 2) else axTicksDateTime(side = 2, x = x)
          }
          if (fine_y && !par("ylog") && length(yg) >= 2L) {
            yg = as.numeric(yg)
            half = (yg[2L] - yg[1L]) / 2
            yg = seq(yg[1L] - half, yg[length(yg)] + half, by = half)
          }
          abline(h = yg, col = gcol, lty = glty, lwd = glwd)
        }

      }
    }

    # add any drawn elements
    if (!is.null(draw)) eval(draw)
  } # end of ii facet loop

  return(as.list(environment()))
}


#' @rdname facet
#' @keywords internal
#' @param settings A list of settings as created by `tinyplot()`.
facet_layout = function(settings) {
  # Extract needed variables from settings
  add = settings$add
  facet.args = settings$facet.args
  datapoints = settings$datapoints
  facet_attr = settings$facet_attr

  # Simplify facet if only one unique value
  facet = datapoints$facet
  if (!is.null(facet) && length(unique(facet)) == 1) {
    facet = NULL
    datapoints$facet = NULL
  }

  # Restore facet attributes
  if (!is.null(facet)) {
    attributes(facet) = facet_attr
    attributes(datapoints$facet) = facet_attr
  }

  nfacet_rows = 1
  nfacet_cols = 1
  if (!is.null(facet)) {
    facets = if (is.factor(facet)) levels(facet) else sort(unique(facet))
    ifacet = seq_along(facets)
    nfacets = length(facets)
    if (isTRUE(add)) {
      omfrow = par("mfrow")
      nfacet_rows = omfrow[1]
      nfacet_cols = omfrow[2]
    } else {
      if (isTRUE(attr(facet, "facet_grid"))) {
        facet.args[["nrow"]] = attr(facet, "facet_nrow")
      }
      if (!is.null(facet.args[["nrow"]])) {
        nfacet_rows = facet.args[["nrow"]]
        nfacet_cols = ceiling(nfacets / nfacet_rows)
      } else if (!is.null(facet.args[["ncol"]])) {
        nfacet_cols = facet.args[["ncol"]]
        nfacet_rows = ceiling(nfacets / nfacet_cols)
      } else {
        if (nfacets > 3) {
          nfacet_cols = ceiling(sqrt(nfacets))
          nfacet_rows = ceiling(nfacets / nfacet_cols)
        } else {
          nfacet_rows = 1L
          nfacet_cols = nfacets
        }
      }
    }

    oxaxis = tail(ifacet, nfacet_cols)
    oyaxis = seq(1, nfacets, by = nfacet_cols)

    if (nfacet_rows >= 3 || nfacet_cols >= 3) {
      cex_fct_adj = 0.66
    } else if (nfacet_rows == 2 && nfacet_cols == 2) {
      cex_fct_adj = 0.83
    } else {
      cex_fct_adj = 1
    }
  } else {
    facets = ifacet = nfacets = oxaxis = oyaxis = 1
    cex_fct_adj = 1
  }

  # update settings
  env2env(
    environment(),
    settings,
    c("datapoints", "facets", "ifacet", "nfacets", "nfacet_rows", "nfacet_cols", "oxaxis", "oyaxis", "cex_fct_adj")
  )
}



#
# helper functions
#


# utility function for converting facet formulas into variables
get_facet_fml = function(formula, data = NULL) {
  xfacet = yfacet = NULL

  ## catch one-sided formula ~ x or ~ x | z with no "y" variable
  if (!inherits(formula, "formula")) formula = as.formula(formula)
  no_yfacet = length(formula) == 2L
  fml_rhs = if (no_yfacet) 2L else 3L

  ## set up model frame
  m = match.call(expand.dots = FALSE)

  if (!is.null(data)) {
    m = m[c(1L, match(c("formula", "data", "subset", "na.action", "drop.unused.levels"), names(m), 0L))]
  }

  m$formula = formula
  ## need stats:: for non-standard evaluation
  m[[1L]] = quote(stats::model.frame)
  mf = eval.parent(m)

  ## extract variables: x, y (if any)
  if (no_yfacet) {
    yfacet_loc = NULL
    xfacet_loc = 1L
  } else {
    yfacet_loc = 1L
    xfacet_loc = 2L
  }
  if (NCOL(mf) < xfacet_loc) stop("formula should specify at least one variable on the right-hand side")
  yfacet = if (no_yfacet) NULL else mf[, yfacet_loc]
  xfacet = mf[, xfacet_loc:NCOL(mf)]

  ## return object
  xfacet = interaction(xfacet, sep = ":")
  if (no_yfacet) {
    ret = xfacet
  } else {
    # yfacet = interaction(yfacet, sep = ":")
    ## NOTE: We "swap" the formula LHS and RHS since mfrow plots rowwise
    ret = interaction(xfacet, yfacet, sep = "~")
    attr(ret, "facet_grid") = TRUE
    attr(ret, "facet_nrow") = length(unique(yfacet))
  }

  return(ret)
}


## internal convenience function to determine whether the current facet panel
## has the position "left", "right", "top", or "bottom" in the facet grid
is_facet_position = function(position, ifacet, facet_window_args) {
  id = facet_window_args$ifacet
  nc = facet_window_args$nfacet_cols
  ni = tail(id, 1L)
  switch(position,
    "left"   = ifacet %in% seq(1L, ni, by = nc),
    "right"  = ifacet %in% pmin(ni, seq(1L, ni, by = nc) + nc - 1L),
    "top"    = ifacet %in% head(id, nc),
    "bottom" = ifacet %in% tail(id, nc),
    NA
  )
}

#' Points plot type
#'
#' @description Type function for plotting points, i.e. a scatter plot.
#' @param clim Numeric giving the lower and upper limits of the character
#'   expansion (`cex`) normalization for bubble charts.
#' @inheritParams dodge_positions
#'
#' @examples
#' # "p" type convenience character string
#' tinyplot(Sepal.Length ~ Petal.Length, data = iris, type = "p")
#'
#' # Same result with type_points()
#' tinyplot(Sepal.Length ~ Petal.Length, data = iris, type = type_points())
#'
#' # Note: Specifying the type here is redundant. Like base plot, tinyplot
#' # automatically produces a scatter plot if x and y are numeric
#' tinyplot(Sepal.Length ~ Petal.Length, data = iris)
#'
#' # Grouped scatter plot example
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris)
#'
#' # Continuous grouping (with gradient legend)
#' tinyplot(Sepal.Length ~ Petal.Length | Sepal.Width, data = iris, pch = 19)
#'
#' # Bubble chart version
#' tinyplot(Sepal.Length ~ Petal.Length, data = iris, cex = iris$Sepal.Width)
#'
#' # Fancier version with dual legends and extra customization
#' tinyplot(Sepal.Length ~ Petal.Length | Species,
#'   data = iris,
#'   cex = iris$Sepal.Width, clim = c(1, 5),
#'   pch = 21, fill = 0.3)
#'
#' @export
type_points = function(clim = c(0.5, 2.5), dodge = 0, fixed.dodge = FALSE) {
  out = list(
    data = data_points(clim = clim, dodge = dodge, fixed.dodge = fixed.dodge),
    draw = draw_points(),
    name = "p"
  )
  class(out) = "tinyplot_type"
  return(out)
}

data_points = function(clim = c(0.5, 2.5), dodge = 0, fixed.dodge = FALSE) {
  fun = function(settings, ...) {
    env2env(settings, environment(), "datapoints")

    # Store clim for bubble() function
    settings$clim = clim

    # catch for factors (we should still be able to "force" plot these with points)
    if (is.factor(datapoints$x)) {
      xlvls = levels(datapoints$x)
      xlabs = seq_along(xlvls)
      names(xlabs) = xlvls
      datapoints$x = as.integer(datapoints$x)
    } else {
      xlabs = NULL
    }
    if (is.factor(datapoints$y)) {
      ylvls = levels(datapoints$y)
      ylabs = seq_along(ylvls)
      names(ylabs) = ylvls
      datapoints$y = as.integer(datapoints$y)
    } else {
      ylabs = NULL
    }

    # dodge
    if (dodge != 0) {
      datapoints = dodge_positions(datapoints, dodge, fixed.dodge)
    }

    # legend customizations
    settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% settings$lwd

    env2env(environment(), settings, c(
      "datapoints",
      "xlabs",
      "ylabs"
    ))
  }
}

draw_points = function() {
  fun = function(ix, iy, icol, ibg, ipch, ilwd, icex, ...) {
    # browser()
    points(
      x = ix,
      y = iy,
      col = icol,
      bg = ibg,
      type = "p",
      pch = ipch,
      lwd = ilwd,
      cex = icex
    )
  }
  return(fun)
}

#' Force-directed text repelling
#'
#' Resolves overlapping text labels using a two-phase algorithm: (1) push
#' overlapping labels apart symmetrically, (2) spring-pull labels back toward
#' their anchors without reintroducing overlaps.
#'
#' @param x,y Numeric vectors of current label positions.
#' @param widths,heights Numeric vectors of label bounding box dimensions (in
#'   the same units as `x` and `y`).
#' @param anchor_x,anchor_y Numeric vectors of original target positions that
#'   labels are pulled back toward during the spring pass. Defaults to `x`
#'   and `y`.
#' @param min_gap Minimum padding between labels. Default `0`.
#' @param iterations Maximum number of repulsion iterations. Default `20`.
#' @param axis Movement constraint: `"both"` (default), `"x"`, or `"y"`.
#' @return A list with elements `x` and `y` giving adjusted positions.
#' @keywords internal
repel_text = function(x, y, widths, heights,
                        anchor_x = x, anchor_y = y,
                        min_gap = 0, iterations = 20,
                        axis = "both") {
  n = length(x)
  if (n <= 1) return(list(x = x, y = y))

  valid = !is.na(x) & !is.na(y)
  vx = x[valid]
  vy = y[valid]
  ax = anchor_x[valid]
  ay = anchor_y[valid]
  vw = widths[valid]
  vh = heights[valid]
  nv = sum(valid)

  move_x = axis %in% c("both", "x")
  move_y = axis %in% c("both", "y")

  # Phase 1: resolve overlaps by pushing apart (full overlap each step)
  for (iter in seq_len(iterations)) {
    any_overlap = FALSE
    for (i in seq_len(nv - 1)) {
      for (j in (i + 1):nv) {
        ox = (vw[i] + vw[j]) / 2 + min_gap - abs(vx[i] - vx[j])
        oy = (vh[i] + vh[j]) / 2 + min_gap - abs(vy[i] - vy[j])

        overlapping = if (axis == "y") oy > 0
          else if (axis == "x") ox > 0
          else ox > 0 && oy > 0
        if (overlapping) {
          any_overlap = TRUE
          push_y = move_y && (oy <= ox || !move_x)
          if (push_y) {
            push = oy / 2
            if (vy[i] <= vy[j]) {
              vy[i] = vy[i] - push
              vy[j] = vy[j] + push
            } else {
              vy[i] = vy[i] + push
              vy[j] = vy[j] - push
            }
          } else {
            push = ox / 2
            if (vx[i] <= vx[j]) {
              vx[i] = vx[i] - push
              vx[j] = vx[j] + push
            } else {
              vx[i] = vx[i] + push
              vx[j] = vx[j] - push
            }
          }
        }
      }
    }
    if (!any_overlap) break
  }

  # Phase 2: spring pass — pull toward anchors without reintroducing overlaps
  displacements = (vx - ax)^2 + (vy - ay)^2
  for (idx in order(displacements, decreasing = TRUE)) {
    step_x = if (move_x) (ax[idx] - vx[idx]) * 0.5 else 0
    step_y = if (move_y) (ay[idx] - vy[idx]) * 0.5 else 0
    new_x = vx[idx] + step_x
    new_y = vy[idx] + step_y

    overlaps = FALSE
    for (k in seq_len(nv)) {
      if (k == idx) next
      ox = (vw[idx] + vw[k]) / 2 + min_gap - abs(new_x - vx[k])
      oy = (vh[idx] + vh[k]) / 2 + min_gap - abs(new_y - vy[k])
      hit = if (axis == "y") oy > 0
        else if (axis == "x") ox > 0
        else ox > 0 && oy > 0
      if (hit) {
        overlaps = TRUE
        break
      }
    }
    if (!overlaps) {
      vx[idx] = new_x
      vy[idx] = new_y
    }
  }

  x[valid] = vx
  y[valid] = vy
  list(x = x, y = y)
}

#' Text annotations plot type
#'
#' @description Type function for adding text annotations to a plot at the
#' specified (`x`,`y`) coordinates.
#'
#' @param labels Character vector of length `1` or the same length as the
#'   number of `x`,`y` coordinates. If left as `NULL`, then the labels will
#'   automatically inherit the corresponding `y` values. Can also be supplied as
#'   a top-level [`tinyplot`] argument, which additionally supports non-standard
#'   evaluation against `data` and takes precedence if both are given. See
#'   Examples.
#' @param labeller A formatting function (or convenience string) passed to
#'   [`tinylabel`] for formatting the `labels`. Useful for ensuring that the
#'   text labels match the formatting of an axis, e.g. `labeller = "%"` to
#'   display the labels as percentages. Default is `NULL`, i.e. no formatting.
#'   See Examples.
#' @param family The name of a font family. Default of `NULL` means that the
#' family will be the same as the main plot text, following
#' \code{\link[graphics]{par}}. Note that if a `family` argument is provided,
#' then `vfont` (below) will automatically be ignored.
#' @param font Integer giving the font face to be used,
#' following \code{\link[graphics]{par}}. On most devices, the mapping is: `1` =
#' regular, `2` = bold, `3` = italic, `4` = bold italic, and `5` = symbol.
#' @param xpd Logical value or `NA` denoting text clipping behaviour, following
#'  \code{\link[graphics]{par}}.
#' @param srt Numeric giving the desired string rotation in degrees.
#' @param repel Logical or numeric controlling automatic repulsion of
#'   overlapping text labels. The default `FALSE` draws the labels at their
#'   exact (`x`,`y`) coordinates. `TRUE` nudges overlapping labels apart using a
#'   force-directed algorithm (labels are pushed off each other and then sprung
#'   back toward their original positions). A numeric value does the same but
#'   additionally enforces that much minimum padding (in user coordinates)
#'   between labels. Works best with the default centered text placement (i.e.
#'   without `pos`). **Caveat:** The repulsion logic currently operates
#'   within each group rather than across groups. So the text of different
#'   groups may still overlap. See Examples.
#' @param clim Numeric giving the lower and upper limits of the character
#'   expansion (`cex`) normalization for bubble charts.
#' @inheritParams graphics::text
#' @examples
#' # simplest case (no labels), will auto revert to y labels
#' tinyplot(1:12, type = "text")
#'
#' # pass explicit `labels` arg if you want specific text
#' tinyplot(1:12, type = "text", labels = month.abb)
#'
#' # you can also use a labeller function (passed to `tinylabel`) to
#' # customize
#' tinyplot(1:12, type = "text", labels = month.abb, labeller = toupper)
#' 
#' # tip: use `xpd = NA` to avoid clipping text at the plot region
#' tinyplot(1:12, type = "text", labels = month.name, xpd = NA)
#'
#' # for advanced customization, it's safer to pass args through `type_text()`
#' tinyplot(
#'   1:12,
#'   type = type_text(
#'   labels = month.abb,
#'   family = "HersheyScript",
#'   srt = -20)
#' )
#' 
#' # some formula + data.frame examples...
#' 
#' mt = transform(mtcars[1:20,], cyl = factor(cyl), model = rownames(mtcars[1:20,]))
#' 
#' # as a top-level argument, `labels` supports non-standard evaluation, so a
#' # bare column name is resolved against `data` (just like the plot variables)
#'
#' # tinyplot(mpg ~ wt, data = mt, type = type_text(labels = mt$model))
#' tinyplot(mpg ~ wt, data = mt, type = "text", labels = model) # same
#'
#' # tip: use `repel = TRUE` to automatically nudge overlapping labels apart
#' tinyplot(
#'   mpg ~ wt, data = mt,
#'   type = "text", labels = model,
#'   repel = TRUE, xpd = NA
#' )
#'
#' # limitation: `repel` logic currently works per group, so grouped text data
#' # may still overlap
#' tinyplot(
#'   mpg ~ wt | cyl, data = mt,
#'   type = "text", labels = model,
#'   repel = TRUE, xpd = NA
#' )
#' 
#' # an advantage of the top-level `labels` approach is that data subsetting and 
#' # sanitizing ops pass through automatically
#' tinyplot(hp ~ wt, data = mt)
#' tinyplot_add(subset = carb==2, col = "red", pch = 16)
#' tinyplot_add(subset = carb==2, col = "red", type = "text", labels = model, pos = 4)
#'
#' # fancier plot that combines a theme and some other customizations
#' tinyplot(
#'   mpg ~ hp | cyl,
#'   data = mt,
#'   legend = list(title = "Cylinders"),
#'   theme = "web"
#' )
#' tinyplot_add(
#'   labels = model,
#'   type = type_text(
#'     family = "HersheySans",
#'     font = 2,
#'     pos = 4,
#'     repel = TRUE,
#'     xpd = NA
#'   )
#' )
#'
#' @export
type_text = function(
  labels = NULL,
  labeller = NULL,
  adj = NULL,
  pos = NULL,
  offset = 0.5,
  family = NULL,
  font = NULL,
  vfont = NULL,
  xpd = NULL,
  srt = 0,
  repel = FALSE,
  clim = c(0.5, 2.5)
) {
  out = list(
    draw = draw_text(
      adj = adj,
      pos = pos,
      offset = offset,
      vfont = vfont,
      family = family,
      font = font,
      xpd = xpd,
      srt = srt,
      repel = repel
    ),
    data = data_text(labels = labels, labeller = labeller, clim = clim),
    name = "text"
  )
  class(out) = "tinyplot_type"
  return(out)
}

data_text = function(labels = NULL, labeller = NULL, clim = c(0.5, 2.5)) {
  fun = function(settings, ...) {
    env2env(settings, environment(), "datapoints")

    # Store clim for bubble() function
    settings$clim = clim

    # Precedence: a top-level `labels` column (carried on datapoints via NSE)
    # wins over the type_text(labels=) constructor arg, which in turn falls
    # back to the y values.
    if (!is.null(datapoints[["labels"]])) {
      labels = datapoints[["labels"]]
    } else if (is.null(labels)) {
      labels = datapoints$y
    }
    if (length(labels) != 1 && length(labels) != nrow(datapoints)) {
      msg = sprintf("`labels` must be of length 1 or %s.", nrow(datapoints))
      stop(msg, call. = FALSE)
    }
    if (!is.null(labeller)) {
      labels = tinylabel(labels, labeller)
    }
    datapoints$labels = labels
    if (is.factor(datapoints$x)) {
      datapoints$x = as.numeric(datapoints$x)
    }
    if (is.factor(datapoints$y)) {
      datapoints$y = as.numeric(datapoints$y)
    }

    env2env(environment(), settings, "datapoints")
  }
  return(fun)
}

draw_text = function(
  adj = NULL,
  pos = NULL,
  offset = 0.5,
  vfont = NULL,
  family = NULL,
  font = NULL,
  xpd = NULL,
  srt = 0,
  repel = FALSE
) {
  if (is.null(xpd)) {
    xpd = par("xpd")
  }
  if (!is.null(family)) {
    vfont = NULL
  }
  fun = function(ix, iy, ilabels, icol, icex, ...) {
    # Optionally repel overlapping labels. Measurement requires user
    # coordinates, which are only correct here (post `plot.window()`), not in
    # the type's data function. See `repel_text()` in R/repel.R.
    if (!isFALSE(repel) && length(ix) > 1) {
      min_gap = if (isTRUE(repel)) 0 else as.numeric(repel)
      w = strwidth(ilabels, units = "user", cex = icex, font = font, family = family)
      h = strheight(ilabels, units = "user", cex = icex, font = font, family = family)
      rp = repel_text(
        x = ix, y = iy, widths = w, heights = h,
        min_gap = min_gap, axis = "both"
      )
      ix = rp[["x"]]
      iy = rp[["y"]]
    }
    text(
      x = ix,
      y = iy,
      labels = ilabels,
      col = icol,
      adj = adj,
      pos = pos,
      offset = offset,
      family = family,
      font = font,
      vfont = vfont,
      xpd = xpd,
      srt = srt,
      cex = icex
    )
  }
}

## Null coalescing operator
if (getRversion() <= "4.4.0") {
  `%||%` = function(x, y) if (is.null(x)) y else x
}

# Count text lines. Returns 0 for absent text and 1 for expression objects.
text_line_count = function(x) {
  if (is.null(x)) return(0L)
  if (identical(x, NA) || identical(x, NA_character_)) return(0L)
  if (!is.character(x)) return(1L)
  if (!length(x)) return(0L)
  keep = !is.na(x) & nzchar(x)
  if (!any(keep)) return(0L)
  x = x[which(keep)[1L]]
  as.integer(1L + nchar(gsub("[^\n]", "", x)))
}

# Compute additive margin "build" for a given side under dynmar.
# Starts from zero and adds only what the plot actually needs. The tick
# row and the axis-label row occupy overlapping vertical space (tick row
# ends at ~|tcl| + mgp[2] + descent, axis label baseline sits at mgp[1]),
# so the margin is the max of:
#   - tick-row height (|tcl| + mgp[2] + descent), when the axis is drawn
#   - axis-label extent (mgp[1] + (N - 1) * cex + descent), when present
# The descent term = 0.4*cex + 0.6: text at the mgp reference line extends
# ~0.4*cex lines below baseline (empirically measured character cell descent),
# plus 0.6 lines fixed buffer for descender characters and spacing.
# Main/sub sit above/below the plot box on the top/bottom side and add to the
# margin additively.
# Tick-label *width* for sides 2/4 (and *height* for 1/3 under las 2:3) is
# handled separately by the existing whtsbp logic in draw_facet_window().
# The caller is expected to take max(theme_mar[side], dynmar_side(...)) so
# that the theme's starting `mar` acts as a baseline padding.
dynmar_side = function(side, label, main = NULL, sub = NULL, cap = NULL,
                      side.sub = 3, axis_on = TRUE, tpars = NULL) {
  mgp = get_tpar("mgp", tpar_list = tpars)
  tcl = get_tpar("tcl", tpar_list = tpars, default = par("tcl"))
  tick_extent = if (side %in% 1:2 && isTRUE(axis_on)) {
    cex_axis = get_tpar("cex.axis", tpar_list = tpars, default = 1)
    max(0, -tcl) + mgp[2] + 0.4 * cex_axis + 0.6
  } else 0
  label_extent = 0
  lines = text_line_count(label)
  if (side %in% 1:2 && lines >= 1L) {
    cex_lab = get_tpar(
      if (side == 1L) c("cex.xlab", "cex.lab") else c("cex.ylab", "cex.lab"),
      tpar_list = tpars, default = 1
    )
    # Side 2: rotated ylab baseline is shifted by ylab_cex_shift, so far edge
    # = mgp[1] + 0.1*cex_lab + 0.5; descent = 0.1*cex + 0.9 gives constant
    # 0.4-line buffer (and exactly 1.0 at cex=1, preserving existing behavior).
    # Side 1: horizontal xlab ink extends ~0.2*cex below center; +0.8 buffer.
    descent = if (side == 2L) 0.1 * cex_lab + 0.9 else 0.2 * cex_lab + 0.8
    label_extent = mgp[1] + (lines - 1) * cex_lab + descent
    # Expressions (e.g., ylab = expression(mm^{1/2})) can be taller than a
    # plain text line due to superscripts, subscripts, fractions, etc. Measure
    # the actual rendered height and add the excess over a normal text line.
    if (is.language(label)) {
      expr_lines = strheight(label, units = "inches", cex = cex_lab) / par("csi")
      text_lines = strheight("X", units = "inches", cex = cex_lab) / par("csi")
      excess = expr_lines - text_lines
      if (excess > 0) label_extent = label_extent + excess
    }
  }
  mar = max(tick_extent, label_extent)
  if (side == 3L) {
    mlines = text_line_count(main)
    if (mlines >= 1L) {
      cex_main = get_tpar("cex.main", tpar_list = tpars, default = 1.2)
      gap_main = get_tpar("gap.main", tpar_list = tpars, default = 0.7)
      mar = mar + gap_main + (mlines - 1 + 0.6) * cex_main
    }
  }
  slines = text_line_count(sub)
  if (slines >= 1L && side == side.sub && side %in% c(1L, 3L)) {
    cex_sub = get_tpar("cex.sub", tpar_list = tpars, default = 1.2)
    gap_sub = get_tpar("gap.sub", tpar_list = tpars, default = 0.7)
    has_main_here = side == 3L && text_line_count(main) >= 1L
    asc = if (has_main_here) 0 else 0.6 * cex_sub
    mar = mar + gap_sub + (slines - 1 + 0.6) * cex_sub + asc
  }
  clines = text_line_count(cap)
  if (clines >= 1L && side == 1L) {
    cex_cap = get_tpar("cex.cap", tpar_list = tpars, default = 1)
    mar = mar + (cex_cap + 0.2) + (clines - 1) * cex_cap
  }
  mar
}


## Function that computes an appropriate bandwidth kernel based on a string
## input
bw_fun = function(kernel, x) {
  kernel = tolower(kernel)
  switch(kernel,
    nrd0 = bw.nrd0(x),
    nrd  = bw.nrd(x),
    ucv  = bw.ucv(x),
    bcv  = bw.bcv(x),
    sj   = bw.SJ(x),
    stop("Invalid `bw` string. Choose from 'nrd0', 'nrd', 'ucv', 'bcv', or 'SJ'.")
  )
}


# Assign (inject) elements from one environment into another
env2env = function(source_env, target_env, keys = NULL) {
  if (is.null(keys)) {
    keys = ls(source_env, all.names = TRUE)
  }
  for (nm in keys) {
    assign(nm, source_env[[nm]], envir = target_env)
  }
}


## Function for efficiently checking whether a vector has more than n unique
## values (uses a hash set approach for large vectors to check sequentially)
more_than_n_unique = function(x, n, small_vec_len = 1e3L) {
  len_x = length(x)
  # For "small" vectors, just use direct length(unique(x)) approach
  if (len_x <= small_vec_len) {
    return(length(unique(x)) > n)
  } else {
    # For larger vectors, use the hash set approach
    env = new.env(hash = TRUE, size = n)
    count = 0
    for (val in x) {
      if (!exists(as.character(val), env)) {
        assign(as.character(val), TRUE, env)
        count = count + 1
        if (count > n) {
          return(TRUE)
        }
      }
    }
    return(FALSE)
  }
}


# Rescale numeric (used for continuous legends, etc.)
rescale_num = function(x, from = NULL, to = NULL) {
  if (is.null(from)) from = range(x)
  if (is.null(to)) to = c(0, 1)
  (x - from[1]) / diff(from) * diff(to) + to[1]
}


# Convenience function for swapping variables (e.g., use in flipped plots)
swap_variables = function(env, ...) {
  pairs = list(...)
  for (p in pairs) {
    tmp = get(p[1], envir = env)
    assign(p[1], get(p[2], envir = env), envir = env)
    assign(p[2], tmp, envir = env)
  }
}

# Convenience function for swapping columns (e.g., use in flipped plots)
swap_columns = function(dp, a, b) {
  va = dp[[a]]
  vb = dp[[b]]
  dp[[a]] = if (!is.null(vb)) vb else NULL
  dp[[b]] = if (!is.null(va)) va else NULL
  dp
}


#' Restore outer margin defaults
#'
#' @description Resets the outer margin display (omd) to default full device.
#'   Used to clean up after legend drawing that may have adjusted margins.
#'
#' @returns NULL (called for side effect of resetting par("omd"))
#'
#' @keywords internal
restore_margin_outer = function() {
  par(omd = c(0, 1, 0, 1))
}


#' Restore inner margin defaults
#'
#' @description Resets inner margins that may have been adjusted for legend
#'   placement. Handles special cases for each margin side and checks for
#'   custom mfrow layouts.
#'
#' @param ooma Outer margins (from par("oma"))
#' @param topmar_epsilon Small epsilon value for top margin adjustment (default 0.1)
#'
#' @returns NULL (called for side effect of resetting par("mar"))
#'
#' @keywords internal
restore_margin_inner = function(ooma, topmar_epsilon = 0.1) {
  ooma = par("oma")
  omar = par("mar")

  # omd reset (from restore_margin_outer) clears oma as a side-effect, so
  # detect dirty margins directly from mar values rather than relying on oma.
  top_dirty = omar[3] > 4.1
  right_dirty = omar[4] == 0
  left_dirty = ooma[2] != 0 && omar[2] == par("mgp")[1] + 1 * par("cex.lab")
  bottom_dirty = ooma[1] != 0 && omar[1] == par("mgp")[1] + 1 * par("cex.lab")

  if (!any(ooma != 0) && !top_dirty && !right_dirty) return(invisible(NULL))

  # Restore inner margin defaults (in case affected by preceding tinyplot call)
  if (bottom_dirty) omar[1] = 5.1
  if (left_dirty) omar[2] = 4.1
  if (top_dirty) omar[3] = 4.1
  if (right_dirty) omar[4] = 2.1
  par(mar = omar)

  # Restore outer margin defaults (with a catch for custom mfrow plots)
  if (all(par("mfrow") == c(1, 1))) {
    par(omd = c(0, 1, 0, 1))
  }
}


# Convert colour(s) to HCL-like (Luv) coordinates, preserving alpha. Helper for
# seq_palette(). (Originally lived in type_spineplot.R.)
#' @importFrom grDevices col2rgb convertColor hcl
to_hcl = function(x) {
    x = t(col2rgb(x, alpha = TRUE)/255)
    alpha = x[, 4]
    x = x[, 1:3]
    x = convertColor(x, from = "sRGB", to = "Luv")
    x = cbind(H = atan2(x[, 3L], x[, 2L]) * 180/pi, C = sqrt(x[, 2L]^2 + x[, 3L]^2), L = x[, 1L])
    x[is.na(x[, 1L]), 1L] = 0
    x[x[, 1L] < 0, 1L] = x[x[, 1L] < 0, 1L] + 360
    attr(x, "alpha") = alpha
    return(x)
}

# Is a colour (effectively) neutral grey, i.e. has near-zero chroma? Used to
# decide when a single-group fill should fall back to a plain grey rather than a
# coloured tint (see by_bg() and draw_spineplot()). Black, white, and greys are
# all achromatic; any palette hue is not.
is_achromatic = function(x, tol = 1) {
  drop(to_hcl(x))[2L] < tol
}

# Build an n-step sequential ramp from colour `x` toward near-white, in HCL
# space (reduces chroma, increases lightness). Used for single-group fills
# (boxplot/violin/barplot/histogram), ridge fills, legend swatches, and
# spineplot shading. `seq_palette(col, 3)[3]` is thus a lighter-but-*opaque*
# tint of `col`. When `grayscale = TRUE`, returns a neutral grey ramp via
# gray.colors() instead (used by spineplot when no colour grouping is active).
#' @importFrom grDevices gray.colors
seq_palette = function(x, n, power = 1.5, grayscale = FALSE) {
    if (isTRUE(grayscale)) return(gray.colors(n))
    x = drop(to_hcl(x[1L]))
    alpha = attr(x, "alpha")
    hcl(
      h = x[1L],
      c = seq.int(from = x[2L]^(1/power), to = 0, length.out = n + 1)[1L:n]^power,
      l = 100 - seq.int(from = (100 - x[3L])^(1/power), to = pmin(8, (100 - x[3L])/2)^(1/power), length.out = n)^power,
      alpha = alpha
    )[1L:n]
}

#' @param v x-value(s) for vertical line(s). Numeric of length 1, or equal to
#'   the number of groups or number of facets (or the product thereof).
#' @rdname type_abline
#' @export
type_vline = function(v = 0) {
  assert_numeric(v)
  data_vline = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "lwd", "lty", "col"))
    if (nrow(datapoints) == 0) {
      msg = "`type_vline() only works on existing plots with x and y data points."
      stop(msg, call. = FALSE)
    }
    # keep track of unique lty and lwd (needed for group catch / escape hatch
    # later in draw_hline)
    ul_lwd = length(unique(lwd))
    ul_lty = length(unique(lty))
    ul_col = length(unique(col))

    type_info = list(ul_lty = ul_lty, ul_lwd = ul_lwd, ul_col = ul_col)
    env2env(environment(), settings, "type_info")
  }
  draw_vline = function() {
    fun = function(ifacet, iby, data_facet, icol, ilty, ilwd,
                   ngrps, nfacets, by_continuous, facet_by,
                   type_info,
                   ...) {
      # flag for aesthetics by groups
      grp_aes = type_info[["ul_col"]] == 1 || type_info[["ul_lty"]] == ngrps || type_info[["ul_lwd"]] == ngrps

      if (length(v) != 1) {
        if (!length(v) %in% c(ngrps, nfacets, ngrps * nfacets)) {
          msg = "Length of 'v' must be 1, or equal to the number of facets or number of groups (or product thereof)."
          stop(msg, call. = FALSE)
        }
        if (!facet_by && length(v) == nfacets) {
          v = v[ifacet]
          if (!grp_aes && type_info[["ul_col"]] != ngrps) {
            icol = 1
          } else if (by_continuous) {
            icol = 1
          }
        } else if (!by_continuous && length(v) == ngrps * nfacets) {
          v = v[ifacet * ngrps - c(ngrps - iby)]
        } else if (!by_continuous) {
          v = v[iby]
        }
      } else if (!grp_aes) {
        icol = 1
      }

      abline(v = v, col = icol, lty = ilty, lwd = ilwd)
    }
    return(fun)
  }
  out = list(
    draw = draw_vline(),
    data = data_vline,
    name = "vline"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' Local polynomial regression plot type
#'
#' @description Type function for plotting a LOESS (LOcal regrESSion) fit.
#' Arguments are passed to \code{\link[stats]{loess}}.
#'
#' @inheritParams type_glm
#' @inheritParams stats::loess
#' @param se logical. If `TRUE` (the default), confidence intervals are drawn.
#' @param level the confidence level required if `se = TRUE`. Default is 0.95.
#' @importFrom stats loess loess.control predict
#' @examples
#' # "loess" type convenience string
#' tinyplot(dist ~ speed, data = cars, type = "loess")
#'
#' # Use `type_loess()` to pass extra arguments for customization
#' tinyplot(dist ~ speed, data = cars, type = type_loess(span = 0.5, degree = 1))
#' @export
type_loess = function(
    span = 0.75,
    degree = 2,
    family = "gaussian",
    control = loess.control(),
    se = TRUE,
    level = 0.95,
    weights = NULL) {
    out = list(
        draw = draw_ribbon(),
        data = data_loess(span = span, degree = degree, family = family, control = control, se = se, level = level, weights = weights),
        name = if (isTRUE(se)) "ribbon" else "l"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_loess = function(span, degree, family, control, se, level, weights = NULL, ...) {
    fun = function(settings, ...) {
        env2env(settings, environment(), "datapoints")
        # top-level `weights` (carried on datapoints via NSE) take precedence
        # over the constructor-level `weights` argument
        if (is.null(datapoints[["weights"]]) && !is.null(weights)) {
            datapoints[["weights"]] = weights
        }
        has_weights = !is.null(datapoints[["weights"]])
        if (has_weights) settings$weights_used = TRUE
        datapoints = split(datapoints, list(datapoints$facet, datapoints$by))
        datapoints = Filter(function(k) nrow(k) > 0, datapoints)
        datapoints = lapply(datapoints, function(dat) {
            # .w is NULL when no weights column is present, which loess() treats
            # the same as omitting the argument
            .w = dat[["weights"]]
            fit = loess(y ~ x, data = dat, span = span, degree = degree, family = family, control = control, weights = .w)
            if (se == TRUE) {
                p = predict(fit, newdata = dat, se = TRUE)
                p = ci(p$fit, p$se.fit, conf.level = level, p$df)
                dat$y = p$estimate
                dat$ymax = p$conf.high
                dat$ymin = p$conf.low
            } else {
                dat$y = predict(fit, dat)
            }
            dat
        })
        datapoints = do.call(rbind, datapoints)
        datapoints = datapoints[order(datapoints$facet, datapoints$by, datapoints$x), ]
        
        # legend customizations - same as ribbon but add line through square
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% 0
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% par("lty")
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, "datapoints")
    }
    return(fun)
}

sanitize_ribbon_alpha = function(ribbon.alpha) {
  assert_numeric(ribbon.alpha, len = 1, lower = 0, upper = 1, null.ok = TRUE)
  if (is.null(ribbon.alpha)) ribbon.alpha = .tpar[["ribbon.alpha"]]
  return(ribbon.alpha)
}

#' Add straight lines to a plot
#' @description
#' These functions add straight line(s) through the current plot.
#' @details
#' While `type_abline`, `type_hline`, and `type_vline` can be called in a base
#' plot layer, we expect that they will typically be called as subsequent
#' layers via [`tinyplot_add`].
#' @section Recycling logic:
#' The recycling behaviour of the line parameters (i.e., `a`, `b`, `h`, or `v`)
#' is adaptive, depending on whether `by` or `facet` grouping is detected. While
#' this leads to different recycling scenarios, the underlying code logic
#' follows sensible heuristics designed to match user expectations.
#'
#' Parameter lengths must equal one of four options:
#'
#' 1. Single value (i.e., length = 1), i.e. simplest case where the same line is
#' applied uniformly across all groups and facets. Uses the default user colour
#' (e.g. `"black"`, or `tpar("palette.qualitative")[1]` if a theme is set).
#' 2. Number of `by` groups, i.e. one parameter per group. For example,
#' `tinyplot(mpg ~ wt | factor(cyl), data = mtcars, type = type_hline(h = 21:23))`
#' will give three horizontal lines, with colours matching the user's qualitative
#' palette.
#' 3. Number of `facet` groups, i.e. one parameter per facet panel. For example:
#' `tinyplot(mpg ~ wt, facet = ~am, data = mtcars, type = type_hline(h = c(20,30)))`
#' would give separate horizontal lines per facet, but both using the same
#' default color.
#' 4. Product of `by` and `facet` groups, i.e. one parameter for each unique
#' by-facet combination. Orders over facets first and then, within that, by
#' group. For example:
#' `tinyplot(mpg ~ wt | factor(cyl), facet = ~am, data = mtcars, type = type_hline(h = 21:26))`
#' will give six separate lines, with the first three (`21:23`) coloured by
#' group in the first facet, and second three (`24:26`) coloured by by group
#' in the second facet.
#'
#' Alongside these general rules, we also try to accommodate special cases when
#' other aesthetic parameters like `lwd` or `lty` are invoked by the user. See
#' Examples.
#'
#' @param a,b the intercept (default: `a` = 0) and slope (default: `b` = 1)
#'   terms. Numerics of length 1, or equal to the number of groups or number of
#'   facets (or the product thereof).
#' @examples
#' #
#' ## abline
#'
#' tinyplot(x = -10:10, y = rnorm(21) + -10:10, grid = TRUE)
#' tinyplot_add(type = "abline")
#' # same as...
#' # tinyplot_add(type = type_abline(a = 0, b = 1))
#'
#' # customize by passing bespoke intercept and slope values
#' tinyplot_add(type = type_abline(a = -1, b = -0.5))
#'
#' # note that calling as abline & co. as a base plot layer will still lead to
#' # axes limits that respect the range of the data
#' tinyplot(x = -10:10, y = -10:10, grid = TRUE, type = "abline")
#'
#' #
#' ## hline and vline
#'
#' # Base plot layer
#' tinyplot(mpg ~ hp | cyl, facet = "by", data = mtcars, ylim = c(0, 40))
#'
#' # Add horizontal lines at the (default) 0 y-intercept
#' tinyplot_add(type = "hline", col = "grey")
#'
#' # Note that group+facet aesthetics will be inherited. We can use this to
#' # add customized lines (here: the mean `mpg` for each `cyl` group)
#' tinyplot_add(type = type_hline(with(mtcars, tapply(mpg, cyl, mean))), lty = 2)
#'
#' # Similar idea for vline
#' tinyplot_add(type = type_vline(with(mtcars, tapply(hp, cyl, mean))), lty = 2)
#'
#' #
#' ## Recycling logic
#'
#' # length(h) == no. of groups
#' tinyplot(mpg ~ wt | factor(cyl), data = mtcars, type = type_hline(h = 21:23))
#'
#' # length(h) == no. of facets
#' tinyplot(mpg ~ wt, facet = ~am, data = mtcars, type = type_hline(h = c(20, 30)))
#'
#' # length(h) == no. of groups x no. of facets
#' tinyplot(mpg ~ wt | factor(cyl),
#'   facet = ~am, data = mtcars,
#'   type = type_hline(h = 21:26))
#'
#' # special adjustment case (here: lwd by group)
#' tinyplot(mpg ~ wt | factor(cyl),
#'   facet = ~am, data = mtcars,
#'   type = type_hline(c(20, 30)), lwd = c(21, 14, 7))
#'
#' @export
type_abline = function(a = 0, b = 1) {
  data_abline = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "lwd", "lty", "col"))
    if (nrow(datapoints) == 0) {
      msg = "`type_abline() only works on existing plots with x and y data points."
      stop(msg, call. = FALSE)
    }
    # keep track of unique lty and lwd (needed for group catch / escape hatch
    # later in draw_hline)
    ul_lwd = length(unique(lwd))
    ul_lty = length(unique(lty))
    ul_col = length(unique(col))
    type_info = list(ul_lty = ul_lty, ul_lwd = ul_lwd, ul_col = ul_col)
    env2env(environment(), settings, "type_info")
  }
  draw_abline = function() {
    fun = function(ifacet, iby, data_facet, icol, ilty, ilwd,
                   ngrps, nfacets, by_continuous, facet_by,
                   type_info,
                   ...) {
      # flag for aesthetics by groups
      grp_aes = type_info[["ul_col"]] == 1 || type_info[["ul_lty"]] == ngrps || type_info[["ul_lwd"]] == ngrps

      if (length(a) != 1) {
        if (!length(a) %in% c(ngrps, nfacets, ngrps * nfacets)) {
          msg = "Length of 'a' must be 1, or equal to the number of facets or number of groups (or product thereof)."
          stop(msg, call. = FALSE)
        }
        if (!facet_by && length(a) == nfacets) {
          a = a[ifacet]
          if (!grp_aes && type_info[["ul_col"]] != ngrps) {
            icol = 1
          } else if (by_continuous) {
            icol = 1
          }
        } else if (!by_continuous && length(a) == ngrps * nfacets) {
          a = a[ifacet * ngrps - c(ngrps - iby)]
        } else if (!by_continuous) {
          a = a[iby]
        }
      } else if (!grp_aes) {
        icol = 1
      }

      if (length(b) != 1) {
        if (!length(b) %in% c(ngrps, nfacets, ngrps * nfacets)) {
          msg = "Length of 'b' must be 1, or equal to the number of facets or number of groups (or product thereof)."
          stop(msg, call. = FALSE)
        }
        if (!facet_by && length(b) == nfacets) {
          b = b[ifacet]
          if (!grp_aes && type_info[["ul_col"]] != ngrps) {
            icol = 1
          } else if (by_continuous) {
            icol = 1
          }
        } else if (!by_continuous && length(b) == ngrps * nfacets) {
          b = b[ifacet * ngrps - c(ngrps - iby)]
        } else if (!by_continuous) {
          b = b[iby]
        }
      } else if (!grp_aes) {
        icol = 1
      }

      if (type_info[["ul_col"]] != 1 && !(type_info[["ul_lty"]] == ngrps || type_info[["ul_lwd"]] == ngrps)) {
        icol = 1
      }

      abline(a = a, b = b, col = icol, lty = ilty, lwd = ilwd)
    }
    return(fun)
  }
  out = list(
    draw = draw_abline(),
    data = data_abline,
    name = "abline"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' Add a rug to a plot
#' 
#' @description
#' Adds a rug representation (1-d plot) of the data to the plot.
#' 
#' @details
#' This function should only be used as part of [`tinyplot_add()`], i.e. adding
#' to an existing plot.
#' 
#' In most cases, determining which variable receives the rug representation
#' will be based on the `side` argument (i.e., x-variable if side is 1 or 3, and
#' y-variable if side is 2 or 4). An exception is if the preceding plot type was
#' either `"density"` or `"histogram"`; for these latter cases, the x-variable
#' will always be used. See Examples.
#' 
#' @inheritParams graphics::rug
#' @param jitter Logical. Add jittering to separate ties? Default is `FALSE`.
#' @param amount Numeric. Amount of jittering (see \code{\link[base]{jitter}}).
#'   Only used if `jitter` is `TRUE`.
#' @examples
#' tinyplot(~wt | am, data = mtcars, type = "density", facet = "by", fill = "by")
#' tinyplot_add(type = "rug")
#' # use type_rug() to pass extra options
#' tinyplot_add(type = type_rug(side = 3, ticksize = 0.05))
#' 
#' # For ties, use jittering
#' tinyplot(eruptions ~ waiting, data = faithful, type = "lm")
#' tinyplot_add(type = type_rug(jitter = TRUE, amount = 0.3))
#' tinyplot_add(type = type_rug(jitter = TRUE, amount = 0.1, side = 2))
#' # Add original points just for reference
#' tinyplot_add(type = "p")
#' 
#' @importFrom graphics rug
#' @export
type_rug = function(ticksize = 0.03, side = 1, quiet = getOption("warn") < 0, jitter = FALSE, amount = NULL) {
  data_rug = function(settings, ...) {
    env2env(settings, environment(), "datapoints")
    if (nrow(datapoints) == 0) {
      msg = "`type_rug() only works on existing plots with x and y data points."
      stop(msg, call. = FALSE)
    }

    env2env(environment(), settings, "datapoints")
  }
  draw_rug = function(.ticksize = ticksize, .side = side, .quiet = quiet, .jitter = jitter, .amount = amount) {
      fun = function(ix, iy, icol, ilwd, ...) {
        lc = get_environment_variable(".last_call")
        swapy = !is.null(lc$type) && lc$type %in% c("density", "hist", "histogram")
        rugx = if (swapy) iy else if (side %in% c(1, 3)) ix else iy
        if (isTRUE(jitter)) rugx = jitter(rugx, amount = .amount)
        rug(
          x = rugx,
          col = icol,
          lwd = ilwd,
          ticksize = .ticksize,
          side = .side,
          quiet = .quiet
        )
      }
      return(fun)
  }

  out = list(
    draw = draw_rug(),
    data = data_rug,
    name = "rug"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' Confidence ellipse plot type
#'
#' @description Type function for drawing confidence ellipses around grouped
#'   points.
#'
#' @details
#' A confidence ellipse summarizes the spread and correlation of two variables:
#' its axes align with the principal directions of variation, their lengths
#' reflect the standard deviations along those directions, and the overall size
#' is scaled so that (assuming bivariate normality) the specified proportion of
#' observations falls within the boundary.
#'
#' Operationally, the ellipse is computed from the covariance matrix
#' (\code{\link[stats]{cov}}) of each group via eigendecomposition
#' (\code{\link[base]{eigen}}), scaled by a chi-squared quantile
#' (\code{\link[stats]{qchisq}}) for the specified confidence level.
#' Groups with fewer than 3 points are silently dropped.
#'
#' @importFrom stats cov qchisq
#'
#' @param level Numeric confidence level for the ellipse. Default is `0.95`.
#' @param segments Integer number of line segments used to approximate the
#'   ellipse. Default is `100`.
#' @inheritParams type_polygon
#'
#' @examples
#' # "ellipse" type convenience character string
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, type = "ellipse")
#'
#' # layer filled ellipse(s) on top of points
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic")
#' tinyplot_add(type = "ellipse", fill = 0.2)
#' 
#' # to avoid clipping, it's better to reverse the layering (ellipse first)
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic",
#'          type = "ellipse", fill = 0.2)
#' tinyplot_add(type = "p")
#'
#' # custom confidence level
#' tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris,
#'   type = type_ellipse(level = 0.5))
#'
#' @export
type_ellipse = function(level = 0.95, segments = 100, density = NULL, angle = 45) {
  out = list(
    draw = draw_polygon(density = density, angle = angle),
    data = data_ellipse(level = level, segments = segments),
    name = "ellipse"
  )
  class(out) = "tinyplot_type"
  return(out)
}


data_ellipse = function(level = 0.95, segments = 100) {
  fun = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "by", "facet"))

    datapoints = split(datapoints, list(datapoints$by, datapoints$facet))
    datapoints = Filter(function(k) nrow(k) >= 3, datapoints)

    radius = sqrt(qchisq(level, df = 2))
    theta = seq(0, 2 * pi, length.out = segments + 1)
    unit = cbind(cos(theta), sin(theta))

    datapoints = lapply(datapoints, function(dat) {
      center = c(mean(dat$x), mean(dat$y))
      cov_mat = cov(cbind(dat$x, dat$y))
      eig = eigen(cov_mat, symmetric = TRUE)
      scale_mat = eig$vectors %*% diag(sqrt(eig$values) * radius)
      pts = unit %*% t(scale_mat)
      data.frame(
        x = pts[, 1] + center[1],
        y = pts[, 2] + center[2],
        by = dat$by[1L],
        facet = dat$facet[1L]
      )
    })
    datapoints = do.call(rbind, datapoints)

    by = if (length(unique(datapoints$by)) == 1) by else datapoints$by
    facet = if (length(unique(datapoints$facet)) == 1) facet else datapoints$facet

    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25

    env2env(environment(), settings, c("datapoints", "by", "facet"))
  }
  return(fun)
}

#' Polygon plot type
#'
#' @description Type function for plotting polygons.
#' Arguments are passed to \code{\link[graphics]{polygon}}.
#' 
#' @inheritParams graphics::polygon
#' 
#' @examples
#' # "polygon" type convenience character string
#' tinyplot(1:9, c(2,1,2,1,NA,2,1,2,1), type = "polygon")
#' 
#' # Use `type_polygon()` to pass extra arguments for customization
#' tinyplot(1:9, c(2,1,2,1,NA,2,1,2,1), type = type_polygon(density = c(10, 20)))
#' 
#' @export
type_polygon = function(density = NULL, angle = 45) {
  out = list(
    draw = draw_polygon(density = density, angle = angle),
    data = data_polygon(),
    name = "polygon"
  )
  class(out) = "tinyplot_type"
  return(out)
}


data_polygon = function() {
  fun = function(settings, ...) {
    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
  }
  return(fun)
}


draw_polygon = function(density = density, angle = 45) {
    fun = function(ix, iy, icol, ibg, ilty = par("lty"), ilwd = par("lwd"), ...) {
        polygon(
            x = ix,
            y = iy,
            border = icol,
            col = if (!is.null(density)) icol else ibg,
            lty = ilty,
            lwd = ilwd,
            density = density,
            angle = angle
        )
    }
    return(fun)
}

#' tinyplot Method for Plotting Matrices
#'
#' @description Convenience interface for visualizing
#'   \code{\link[base]{matrix}} objects with tinyplot.
#'
#' @details Internally the matrix is converted to long form and visualized as a
#'   scatter (or other `type`) of each column's values against their row index.
#'   Each column is mapped to a separate `by` category, so a matrix with
#'   multiple columns produces a grouped plot. Optionally, it can also be
#'   faceted via `facet = "by"`. This mirrors the base R
#'   \code{\link[graphics]{matplot}} convention of plotting the columns of a
#'   matrix against the row numbers. If the matrix has column names, these are
#'   used as the group (and legend) labels. Single-column matrices are drawn as
#'   a simple index plot with no grouping or legend.
#'
#' @param x an object of class `"matrix"`.
#' @param type plot type passed on to `tinyplot`. Defaults to `"p"` (points).
#' @param legend specification passed on to `tinyplot`. The default is to draw a
#'   legend when the matrix has named columns, and to suppress it otherwise.
#' @param facet specification of `facet` passed on to `tinyplot`. The only
#'   accepted non-`NULL` value is the `"by"` convenience string, which facets
#'   the plot by matrix column.
#' @param xlab,ylab axis labels passed on to `tinyplot`. `ylab` defaults to the
#'   deparsed matrix name. `xlab` defaults to `"Index"` when the matrix has no
#'   row names; when it does, the row names already label the ticks so the
#'   x-axis title is suppressed.
#' @param ... further arguments passed to `tinyplot`.
#'
#' @returns No return value, called for the side effect of producing a plot.
#'
#' @seealso \code{\link[graphics]{matplot}}
#'
#' @examples
#' # basic use
#' tinyplot(VADeaths)
#' tinyplot(VADeaths, type = "b")
#' tinyplot(VADeaths, type = "b", legend = "direct", theme = "socviz")
#' tinyplot(VADeaths, type = "b", legend = FALSE, facet = "by", theme = "socviz")
#' 
#' # equivalent plot to an example in `?matplot`
#' sines = outer(1:20, 1:4, function(x, y) sin(x / 20 * pi * y))
#' tinyplot(sines, type = "o", pch = "by", lty = "by", col = rainbow(ncol(sines)))
#'
#' @export
tinyplot.matrix = function(x, type = NULL, legend = NULL, facet = NULL, xlab = NULL, ylab = NULL, ...) {
  assert_choice(facet, "by", null.ok = TRUE)
  ## Default to points. We set this explicitly (rather than relying on
  ## tinyplot's auto-inference) because the x-axis row labels are passed as a
  ## factor, which would otherwise be inferred as a boxplot.
  if (is.null(type)) type = "p"
  dep_x = deparse1(substitute(x))
  dims = dim(x)
  if (dims[2] == 1L) {
    bby = NULL
    legend = FALSE
  } else {
    nms = colnames(x)
    if (!is.null(nms)) {
      bby = factor(rep(nms, each = dims[1]), levels = nms)
      if (is.null(legend)) legend = list(title = NULL)
    } else {
      bby = factor(rep(seq_len(dims[2]), each = dims[1]))
      legend = FALSE
    }
  }
  ## If the matrix has row names, use them for the x-axis tick labels via an
  ## ordered factor (preserving row order). Otherwise fall back to a plain
  ## numeric index.
  rnms = rownames(x)
  dim(x) = dims[1] * dims[2]
  y = x
  if (is.null(rnms)) {
    x = rep(seq_len(dims[1]), times = dims[2])
    ## no row names: x is a plain numeric index, so label it as such
    if (is.null(xlab)) xlab = "Index"
  } else {
    x = factor(rep(rnms, times = dims[2]), levels = rnms, ordered = TRUE)
    ## row names already label the ticks, so an "Index" title is redundant
    if (is.null(xlab)) xlab = NA
  }
  if (is.null(ylab)) ylab = dep_x
  tinyplot.default(
    x = x, y = y,
    type = type,
    by = bby,
    facet = facet,
    legend = legend,
    xlab = xlab,
    ylab = ylab,
    ...
  )
}

#' Quantile-Quantile plot (QQ)
#'
#' @description Plots the theoretical quantiles of `x` on the horizontal axis
#' against observed values of `x` on the vertical axis.
#'
#' @param distribution Distribution function to use.
#' @examples
#' tinyplot(~mpg, data = mtcars, type = type_qq())
#'
#' # suppress the line
#' tinyplot(~mpg, data = mtcars, lty = 0, type = type_qq())
#' @importFrom stats qnorm ppoints quantile
#' @export
type_qq = function(distribution = qnorm) {
  data_qq = function(distribution) {
    fun = function(settings, ...) {
      env2env(settings, environment(), "datapoints")

      y = sort(datapoints$y)
      x = datapoints$x
      x = distribution(ppoints(x))
      datapoints$x = x
      datapoints$y = y
      env2env(environment(), settings, "datapoints")
    }
  }

  draw_qq = function() {
    fun = function(ix, iy, icol, ibg, ipch, ilwd, ilty, icex, xlab, ...) {
      points(
        x = ix,
        y = iy,
        col = icol,
        bg = ibg,
        type = "p",
        pch = ipch,
        lwd = ilwd,
        cex = icex
      )

      if (!is.null(ilty)) {
        iy = quantile(iy, c(0.25, 0.75))
        ix = quantile(ix, c(0.25, 0.75))
        slope = diff(iy) / diff(ix)
        intercept = iy[1] - slope * ix[1]
        abline(a = intercept, b = slope, lty = ilty, col = icol, lwd = ilwd)
      }
    }
    return(fun)
  }

  out = list(
    draw = draw_qq(),
    data = data_qq(distribution = distribution),
    name = "qq"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' Ribbon and area plot types
#'
#' @param alpha numeric value between 0 and 1 specifying the opacity of ribbon shading
#'   If no `alpha` value is provided, then will default to `tpar("ribbon.alpha")`
#'   (i.e., probably `0.2` unless this has been overridden by the user in their global
#'   settings.)
#' @inheritParams type_errorbar
#'
#' @description Type constructor functions for producing polygon ribbons, which
#' define a `y` interval (usually spanning from `ymin` to `ymax`) for each
#' `x` value. Area plots are a special case of ribbon plot where `ymin` is
#' set to 0 and `ymax` is set to `y`.
#' 
#' @section Dodging ribbon plots:
#' 
#' We support dodging for grouped ribbon plots, enabling similar functionality
#' to dodged errorbar and pointrange plots. However, it is strongly recommended
#' that dodging is only implemented for cases where the x-axis comprises a
#' limited number of discrete cases (e.g., coefficient or event-study plots).
#' See Examples.
#'
#' @examples
#' x = 1:100 / 10
#' y = sin(x)
#'
#' #
#' ## Ribbon plots
#'
#' # "ribbon" convenience string
#' tinyplot(x = x, ymin = y - 1, ymax = y + 1, type = "ribbon")

#' # Same result with type_ribbon()
#' tinyplot(x = x, ymin = y-1, ymax = y+1, type = type_ribbon())
#'
#' # y will be added as a line if it is specified
#' tinyplot(x = x, y = y, ymin = y-1, ymax = y+1, type = "ribbon")
#'
#' #
#' ## Area plots
#'
#' # "area" type convenience string
#' tinyplot(x, y, type = "area")
#'
#' # Same result with type_area()
#' tinyplot(x, y, type = type_area())
#'
#' # Area plots are often used for time series charts
#' tinyplot(AirPassengers, type = "area")
#' 
#' #
#' ## Dodged ribbon/area plots
#' 
#' # Dodged ribbon or area plots can be useful in cases where there is strong
#' # overlap across groups (and a limited number of discrete x-axis values).
#' 
#' dat = data.frame(
#'   x = rep(c("Before", "After"), each = 2),
#'   grp = rep(c("A", "B"), 2),
#'   y = c(10, 10.5, 15, 15.3),
#'   lwr = c(8, 8.5, 13, 13.3),
#'   upr = c(12, 12.5, 17, 17.3)
#' )
#' 
#' tinyplot(
#'   y ~ x | grp,
#'   data = dat,
#'   ymin = lwr, ymax = upr,
#'   type = type_ribbon(),
#'   main = "Overlappling ribbons"
#' )
#' 
#' tinyplot(
#'   y ~ x | grp,
#'   data = dat,
#'   ymin = lwr, ymax = upr,
#'   type = type_ribbon(dodge = 0.1),
#'   main = "Dodged ribbons"
#' )
#' 
#' @export
type_ribbon = function(alpha = NULL, dodge = 0, fixed.dodge = FALSE) {
    out = list(
        draw = draw_ribbon(),
        data = data_ribbon(ribbon.alpha = alpha, dodge = dodge, fixed.dodge = fixed.dodge),
        name = "ribbon"
    )
    class(out) = "tinyplot_type"
    return(out)
}


draw_ribbon = function() {
    fun = function(ix, iy, ixmin, ixmax, iymin, iymax, ibg, ilty, ilwd, icol, ipch, i, flip = FALSE, ...) {
        polyg = type_polygon()$draw
        lin = type_lines()$draw
        if (isFALSE(flip)) {
            polyg(ix = c(ix, rev(ix)), iy = c(iymin, rev(iymax)), icol = NA, ibg = ibg)
        } else {
            polyg(c(ixmin, rev(ixmax)), iy = c(iy, rev(iy)), icol = NA, ibg = ibg)
        }
        lin(ix = ix, iy = iy, icol = icol, ipch = ipch, ibg = ibg, ilty = ilty, ilwd = ilwd, type = "l")
    }
    return(fun)
}


data_ribbon = function(ribbon.alpha = NULL, dodge = 0, fixed.dodge = FALSE) {
    ribbon.alpha = sanitize_ribbon_alpha(ribbon.alpha)
    fun = function(settings, ...) {
        env2env(settings, environment(), c("datapoints", "xlabs", "null_by", "null_facet"))
        # Convert x to factor if it's not already
        if (is.character(datapoints$x)) {
            datapoints$x = as.factor(datapoints$x)
        }

        if (is.factor(datapoints$x)) {
            xlvls = levels(datapoints$x)
            xlabs = seq_along(xlvls)
            names(xlabs) = xlvls
            datapoints$x = as.integer(datapoints$x)
        } else {
            xlabs = NULL
        }
        
        # dodge (auto-detects x, xmin, xmax columns)
        if (dodge != 0) {
            datapoints = dodge_positions(datapoints, dodge, fixed.dodge)
        }

        if (null_by && null_facet) {
            xord = order(datapoints$x)
        } else if (null_facet) {
            xord = order(datapoints$by, datapoints$x)
        } else if (null_by) {
            xord = order(datapoints$facet, datapoints$x)
        } else {
            xord = order(datapoints$by, datapoints$facet, datapoints$x)
        }

        # Reorder x, y, ymin, and ymax based on the order determined
        datapoints = datapoints[xord, ]

        # Catch for missing ymin and ymax
        if (is.null(datapoints$ymin)) datapoints$ymin = datapoints$y
        if (is.null(datapoints$ymax)) datapoints$ymax = datapoints$y

        x = datapoints$x
        y = datapoints$y
        ymin = datapoints$ymin
        ymax = datapoints$ymax
        by = if (length(unique(datapoints$by)) > 1) datapoints$by else NULL
        facet = if (length(unique(datapoints$facet)) > 1) datapoints$facet else NULL

        # ribbon.alpha comes from parent scope, so assign it locally
        ribbon.alpha = ribbon.alpha

        # legend customizations
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% 0
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25

        vars_to_copy = c("x", "y", "ymin", "ymax", "xlabs", "datapoints", "ribbon.alpha")
        if (!is.null(by)) vars_to_copy = c(vars_to_copy, "by")
        if (!is.null(facet)) vars_to_copy = c(vars_to_copy, "facet")

        env2env(environment(), settings, vars_to_copy)
    }
    return(fun)
}

#' Generalized linear model plot type
#'
#' @description Type function for plotting a generalized model fit.
#' Arguments are passed to \code{\link[stats]{glm}}.
#'
#' @param se logical. If TRUE, confidence intervals are drawn.
#' @param weights an optional numeric vector of observation weights for the model
#'   fit, of the same length as the number of data points. Weights can also be
#'   supplied via the top-level `weights` argument of [`tinyplot`] (which is
#'   evaluated with non-standard evaluation in the formula method, and takes
#'   precedence if both are given).
#' @inheritParams stats::glm
#' @inheritParams stats::predict.glm
#' @inheritParams stats::confint
#' @importFrom stats glm predict
#' @examples
#' # "glm" type convenience string
#' tinyplot(am ~ mpg, data = mtcars, type = "glm")
#'
#' # Use `type_glm()` to pass extra arguments for customization
#' tinyplot(am ~ mpg, data = mtcars, type = type_glm(family = "binomial"))
#' @export
type_glm = function(family = "gaussian", se = TRUE, level = 0.95, type = "response", weights = NULL) {
    assert_flag(se)
    out = list(
        draw = draw_ribbon(),
        data = data_glm(family = family, se = se, level = level, type = type, weights = weights),
        name = if (isTRUE(se)) "ribbon" else "l"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_glm = function(family, se, level, type, weights = NULL, ...) {
    fun = function(settings, ...) {
        env2env(settings, environment(), "datapoints")
        # top-level `weights` (carried on datapoints via NSE) take precedence
        # over the constructor-level `weights` argument
        if (is.null(datapoints[["weights"]]) && !is.null(weights)) {
            datapoints[["weights"]] = weights
        }
        has_weights = !is.null(datapoints[["weights"]])
        if (has_weights) settings$weights_used = TRUE
        dat = split(datapoints, list(datapoints$facet, datapoints$by))
        dat = lapply(dat, function(x) {
            if (nrow(x) == 0) {
                return(x)
            }
            if (nrow(x) < 3) {
                x$y = NA
                return(x)
            }
            # .w is NULL when no weights column is present, which glm() treats
            # the same as omitting the argument
            .w = x[["weights"]]
            fit = glm(y ~ x, data = x, family = family, weights = .w)
            nd = data.frame(x = seq(min(x$x, na.rm = TRUE), max(x$x, na.rm = TRUE), length.out = 100))
            nd$by = x$by[1]
            nd$facet = x$facet[1]
            if (se == TRUE) {
                if (identical(type, "response")) {
                    p = predict(fit, newdata = nd, type = "link", se.fit = TRUE)
                    p = ci(p$fit, p$se.fit, conf.level = level, fit$df.residual, backtransform = stats::family(fit)$linkinv)
                    nd$y = p$estimate
                    nd$ymax = p$conf.high
                    nd$ymin = p$conf.low
                } else {
                    nd$y = predict(fit, newdata = nd, type = type)
                    nd = ci(nd$y, nd$se, level, fit$df.residual, backtransform = stats::family(fit)$linkinv)
                }
            } else {
                nd$y = predict(fit, nd, type = type)
            }
            nd
        })
        datapoints = do.call(rbind, dat)
        datapoints = datapoints[order(datapoints$facet, datapoints$by, datapoints$x), ]
        
        # legend customizations - same as ribbon but add line through square
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% 0
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% par("lty")
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, "datapoints")
    }
    return(fun)
}


#' Calculate confidence intervals
#' @importFrom stats qt
#' @keywords internal
ci = function(estimate, std.error, conf.level, df, backtransform = identity) {
    crit = qt(1 - (1 - conf.level) / 2, df)
    out = list(
        estimate = backtransform(estimate),
        conf.low = backtransform(estimate - crit * std.error),
        conf.high = backtransform(estimate + crit * std.error)
    )
    return(out)
}

#' Dodge positions for grouped data
#'
#' Adjusts x-coordinates (and optionally xmin/xmax) to dodge overlapping points
#' or ranges in grouped plots.
#'
#' @param datapoints Data frame containing plot data with at least `x` and `by`
#'   columns.
#' @param dodge Adjustment parameter for dodging overlapping points or ranges in
#'   grouped plots along the x-axis (or y-axis for flipped plots). Either:
#' 
#'   - numeric value in the range `[0,1)`. Note that values are scaled
#'   relative to the spacing of x-axis breaks, e.g. `dodge = 0.1` places the
#'   outermost groups one-tenth of the way to adjacent breaks, `dodge = 0.5`
#'   places them midway between breaks, etc. Values < 0.5 are recommended.
#'   - logical. If `TRUE`, the dodge width is calculated automatically based on
#'   the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
#'   `FALSE` or 0, no dodging is performed.
#' 
#'   Default value is 0 (no dodging). While we do not check, it is _strongly_
#'   recommended that dodging only be used in cases where the x-axis comprises a
#'   limited number of discrete breaks.
#' @param fixed.dodge Logical. If `FALSE` (default), dodge positions are
#'   calculated independently for each `x` value, based only on the groups
#'   present at that position. If `TRUE`, dodge positions are based on all
#'   groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
#'   groups are missing for a particular `x` value).
#' @param cols Character vector of column names to dodge. If `NULL` (default),
#'   automatically detects and dodges `x`, `xmin`, and `xmax` if they exist.
#' @param settings Environment containing plot settings. If `NULL` (default),
#'   retrieved from the calling environment.
#'
#' @return Modified `datapoints` data frame with dodged positions.
#'
#' @details
#' When `fixed.dodge = TRUE`, all groups are dodged by the same amount across all
#' x values, which is useful when x is categorical. When `fixed.dodge = FALSE`,
#' dodging is calculated independently for each x value, which is useful when
#' the number of groups varies across x values.
#'
#' If `cols` is not specified, the function automatically dodges `x` and any
#' `xmin`/`xmax` columns that exist in the data.
#'
#' @keywords internal
dodge_positions = function(
  datapoints,
  dodge,
  fixed.dodge = TRUE,
  cols = NULL,
  settings = NULL
) {
  if (is.null(settings)) {
    settings = get("settings", envir = parent.frame())
  }
  
  if (is.logical(dodge)) {
    if (isTRUE(dodge)) {
      n = nlevels(datapoints$by)
      dodge = if (n == 1) 0 else if (n <= 5) (n - 1) * 0.1 else 0.45
    } else {
      dodge = 0
    }
  }

  assert_numeric(dodge, len = 1, lower = 0, upper = 1)
  if (dodge >= 1) {
  stop("`dodge` must be in the range [0,1).", call. = FALSE)
  }
  assert_logical(fixed.dodge)
  
  if (dodge == 0) {
    return(datapoints)
  } else if (dodge > 0.5) {
    warning(
      "Argument `dodge = ", dodge, "` exceeds 0.5. ",
      "Large dodge values may position outer groups closer to neighboring axis breaks."
    )
  }
  settings$dodge = dodge
  
  # Auto-detect columns to dodge if not specified
  if (is.null(cols)) {
    cols = c("x", "xmin", "xmax")
    cols = cols[cols %in% names(datapoints)]
  }

  if (fixed.dodge) {
    n = nlevels(datapoints$by)
    d = cumsum(rep(dodge, n))
    d = d - mean(d)
    x_adj = d[as.integer(datapoints$by)]
    for (col in cols) {
      datapoints[[col]] = datapoints[[col]] + x_adj
    }
  } else {
    xuniq = unique(datapoints$x)
    for (i in seq_along(xuniq)) {
      idx = which(datapoints$x == xuniq[i])
      n = length(idx)
      d = cumsum(rep(dodge, n))
      d = d - mean(d)
      for (col in cols) {
        datapoints[[col]][idx] = datapoints[[col]][idx] + d
      }
    }
  }

  datapoints
}

#' Error bar and pointrange plot types
#'
#' @description Type function(s) for producing error bar and pointrange plots.
#'
#' @inheritParams dodge_positions
#' @inheritParams graphics::arrows
#' @examples
#' tinytheme("basic")
#' 
#' #
#' ## Basic coefficient plot(s)
#' 
#' mod = lm(mpg ~ wt * factor(am), mtcars)
#' coefs = data.frame(names(coef(mod)), coef(mod), confint(mod))
#' colnames(coefs) = c("term", "est", "lwr", "upr")
#'
#' # "errorbar" and "pointrange" type convenience strings
#' tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "errorbar")
#' tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "pointrange")
#'
#' # Use `type_errorbar()` to pass extra arguments for customization
#' tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs,
#'          type = type_errorbar(length = 0.2))
#'
#' #
#' ## Flipped plots
#' 
#' # For flipped errobar / pointrange plots, it is recommended to use a dynamic
#' # theme that applies horizontal axis tick labels
#'
#' tinytheme("classic")
#' tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "errorbar",
#'          flip = TRUE)
#' tinyplot_add(type = 'vline', lty = 2)
#' 
#' tinytheme("basic") # back to basic theme for the remaining examples
#' 
#' #
#' ## Dodging groups
#'
#' models = list(
#'     "Model A" = lm(mpg ~ wt, data = mtcars),
#'     "Model B" = lm(mpg ~ wt + cyl, data = mtcars),
#'     "Model C" = lm(mpg ~ wt + cyl + hp, data = mtcars)
#' )
#'
#' models = do.call(
#'   rbind,
#'   lapply(names(models), function(m) {
#'     data.frame(
#'       model = m,
#'       term = names(coef(models[[m]])),
#'       estimate = coef(models[[m]]),
#'       setNames(data.frame(confint(models[[m]])), c("conf.low", "conf.high"))
#'     )
#'   })
#' )
#'
#' tinyplot(estimate ~ term | model,
#'          ymin = conf.low, ymax = conf.high,
#'          data = models,
#'          type = type_pointrange(dodge = 0.1))
#'
#' # Aside 1: relative vs fixed dodge
#' #  The default dodge position is based on the unique groups (here: models)
#' #  available to each x value (here: coefficient term). To "fix" the dodge
#' #  position across all x values, use `fixed.dodge = TRUE`.
#'
#' tinyplot(estimate ~ term | model,
#'          ymin = conf.low, ymax = conf.high,
#'          data = models,
#'          type = type_pointrange(dodge = 0.1, fixed.dodge = TRUE))
#' 
#' # Aside 2: layering
#' #  For layering on top of dodged plots, rather pass the dodging arguments
#' #  through the top-level call if you'd like the dodging behaviour to be
#' #  inherited automatically by the added layers.
#'
#' tinyplot(estimate ~ term | model,
#'          ymin = conf.low, ymax = conf.high,
#'          data = models,
#'          type = "pointrange",
#'          dodge = 0.1, fixed.dodge = TRUE)
#' tinyplot_add(type = "l", lty = 2)
#'
#' tinytheme() # reset theme
#'
#' @export
type_errorbar = function(length = 0.05, dodge = 0, fixed.dodge = FALSE) {
    out = list(
        draw = draw_errorbar(length = length),
        data = data_pointrange(dodge = dodge, fixed.dodge = fixed.dodge),
        name = "p"
    )
    class(out) = "tinyplot_type"
    return(out)
}


draw_errorbar = function(length = 0.05) {
    fun = function(
        ix,
        iy,
        ixmin,
        iymin,
        ixmax,
        iymax,
        icol,
        ibg,
        ipch,
        ilwd,
        icex,
        ...
    ) {
        arrows(
            x0 = ixmin,
            y0 = iymin,
            x1 = ixmax,
            y1 = iymax,
            col = icol,
            lwd = ilwd,
            length = length,
            angle = 90,
            code = 3
        )
        draw_points()(
            ix = ix,
            iy = iy,
            icol = icol,
            ibg = ibg,
            ipch = ipch,
            ilwd = ilwd,
            icex = icex
        )
    }
    return(fun)
}

#' Spineplot and spinogram types
#'
#' @description Type function(s) for producing spineplots and spinograms, which
#'   are modified versions of histograms or mosaic plots, and particularly
#'   useful for visualizing factor variables. Note that [`tinyplot`] defaults
#'   to `type_spineplot()` if `y` is a factor variable.
#' @param xlevels,ylevels a character or numeric vector specifying the ordering of the
#'   levels of the `x` and `y` variables (if character) or the corresponding indexes
#'   (if numeric) for the plot.
#' @inheritParams graphics::spineplot
#' @param lighten logical. For grouped spineplots where the `y` variable is
#'   itself the grouping variable (i.e. `y == by`), should the fills use a
#'   lighter, opaque tint of the series colour(s)? Default is `FALSE`, i.e. the
#'   fills use the fully-saturated palette colour(s). (Unlike the other area
#'   types such as [`type_barplot`], where lightening is the default, spineplot
#'   tiles abut one another with no gap, so the darker saturated fills read
#'   better against their matching border colours.) Set to `TRUE` to opt in to
#'   the lighter tint. Note that `lighten` has no effect on other spineplot
#'   displays (single-group or `x == by`), which always use a sequential shading
#'   ramp of the base colour.
#' @examples
#' # "spineplot" type convenience string
#' tinyplot(Species ~ Sepal.Width, data = iris, type = "spineplot")
#' 
#' # Aside: specifying the type is redundant for this example, since tinyplot()
#' # defaults to "spineplot" if y is a factor (just like base plot).
#' tinyplot(Species ~ Sepal.Width, data = iris)
#' 
#' # Use `type_spineplot()` to pass extra arguments for customization
#' tinyplot(
#'   Species ~ Sepal.Width, data = iris,
#'   type = type_spineplot(breaks = 4)
#' )
#' 
#' # Passing custom colors to the y-axis categories
#' tinyplot(
#'   Species ~ Sepal.Width, data = iris,
#'   type = type_spineplot(breaks = 4, col = palette.colors(3, "Pastel 1"))
#' )
#' 
#' # More idiomatic tinyplot way of drawing the previous plot: use y == by
#' tinyplot(
#'   Species ~ Sepal.Width | Species, data = iris,
#'   type = type_spineplot(breaks = 4),
#'   palette = "Pastel 1", legend = FALSE
#' )
#' 
#' ## Grouped and faceted spineplots
#'
#' ttnc = as.data.frame(Titanic)
#' 
#' # Note: The Titanic (ttnc) dataset is pre-tabulated, so we pass its frequency
#' # counts via the top-level `weights` argument (accepted via non-standard
#' # evaluation in the formula method).
#' tinyplot(
#'   Survived ~ Sex, facet = ~ Class, data = ttnc,
#'   # type_spineplot(weights = ttnc$Freq), ## same thing but not NSE
#'   type = "spineplot", weights = Freq
#' )
#' 
#' # Reorder x and y variable categories either by their character levels or
#' # numeric indexes. (Here we combine a top-level `weights` with constructor-
#' # level arguments passed through `type_spineplot()`.)
#' tinyplot(
#'   Survived ~ Sex, facet = ~ Class, data = ttnc,
#'   type = type_spineplot(xlevels = c("Female", "Male"), ylevels = 2:1),
#'   weights = Freq
#' )
#'
#' # For (colour) grouped "by" spineplots, it's visually better to facet too
#' tinyplot(
#'   Survived ~ Sex | Class, data = ttnc,
#'   facet = "by",
#'   type = "spineplot", weights = Freq
#' )
#'
#' # Fancier version. Note the smart inheritance of spacing etc.
#' tinyplot(
#'   Survived ~ Sex | Class, data = ttnc,
#'   facet = "by", facet.args = list(nrow = 1),
#'   type = "spineplot", weights = Freq,
#'   theme = "void", axes = "t", lty = 0, legend = FALSE,
#'   main = "Who survived the Titanic disaster?",
#'   sub = "Frequencies by boarding class and sex"
#' )
#'
#' # Aside: It's possible to use "by" on its own (without faceting), but the
#' # overlaid result isn't great. We will likely overhaul this behaviour in a
#' # future version of tinyplot...
#' tinyplot(Survived ~ Sex | Class, data = ttnc,
#'   type = "spineplot", weights = Freq, alpha = 0.3
#' )
#' 
#' @export
type_spineplot = function(breaks = NULL, tol.ylab = 0.05, off = NULL, xlevels = NULL, ylevels = NULL, col = NULL, xaxlabels = NULL, yaxlabels = NULL, weights = NULL, lighten = FALSE) {
  col = col
  out = list(
    data = data_spineplot(off = off, breaks = breaks, xlevels = xlevels, ylevels = ylevels, xaxlabels = xaxlabels, yaxlabels = yaxlabels, weights = weights, lighten = lighten),
    draw = draw_spineplot(tol.ylab = tol.ylab, off = off, col = col, xaxlabels = xaxlabels, yaxlabels = yaxlabels, lighten = lighten),
    name = "spineplot"
  )
  class(out) = "tinyplot_type"
  return(out)
}

#' @importFrom grDevices nclass.Sturges
data_spineplot = function(off = NULL, breaks = NULL, xlevels = xlevels, ylevels = ylevels, xaxlabels = NULL, yaxlabels = NULL, weights = NULL, lighten = FALSE) {
    fun = function(settings, ...) {
        env2env(settings, environment(), c("datapoints", "xlim", "ylim", "facet", "facet.args", "by", "xaxb", "yaxb", "null_by", "null_facet", "col", "bg", "axes", "frame.plot", "xaxt", "yaxt", "lwd", "lty"))
        settings[["lighten"]] = lighten
      
        ## process weights: a top-level `weights` column (carried on datapoints
        ## via NSE) takes precedence over the constructor-level `weights` arg.
        ## Either way, unify into the `datapoints$weights` column and the local
        ## `weights` vector that the break/range logic below relies on.
        if (is.null(datapoints[["weights"]]) && !is.null(weights)) {
            datapoints$weights = weights
        }
        weights = datapoints[["weights"]]
        if (!is.null(weights)) settings$weights_used = TRUE
        
        ## process x variable
        if (is.factor(datapoints$x)) {
            breaks = NULL
            off = if(is.null(off)) 0.02 else off/100
            if (is.null(xlim)) xlim = c(0, 1 + (nlevels(datapoints$x) - 1L) * off)
        } else {
            off = 0
            if (is.null(xlim)) xlim = c(0, 1)
    	      x = as.numeric(datapoints$x)
            if (is.null(breaks)) {
              breaks = if (!is.null(xaxb)) xaxb else if (is.null(weights)) nclass.Sturges(x) else ceiling(log2(sum(weights)) + 1)
	    }
            breaks = as.numeric(breaks)
            if (length(breaks) == 1L) {
                if (!is.numeric(breaks) || !is.finite(breaks) || breaks < 1L) stop("invalid number of 'breaks'")
                if (breaks > 1e6) {
                    warning(gettextf("'breaks = %g' is too large and set to 1e6", breaks))
                    breaks = 1000000L
                }
                rg = if (is.null(weights)) range(x, na.rm = TRUE) else range(x[weights > 0], na.rm = TRUE)
                breaks = pretty(rg, n = breaks, min.n = 1L)        
            }
        }

        ## process y variable
        if (!is.factor(datapoints$y)) datapoints$y = factor(datapoints$y)
        if (is.null(ylim)) ylim = c(0, 1)

        ## adjust facet margins
        if (!is.null(facet) && is.null(facet.args[["fmar"]])) {
          facet.args[["fmar"]] = c(2, 2, 2, 2)
        }
        
        x_by = identical(datapoints$x, datapoints$by)
        y_by = identical(datapoints$y, datapoints$by)
        
        x.categorical = is.factor(datapoints$x)
        if (!is.null(xlevels) && x.categorical) {
          xlevels = if(is.numeric(xlevels)) levels(datapoints$x)[xlevels] else xlevels
          if (anyNA(xlevels) || !all(xlevels %in% levels(datapoints$x))) warning("not all 'xlevels' correspond to levels of 'x'")
          datapoints$x = factor(datapoints$x, levels = xlevels)
          if (x_by) datapoints$by = datapoints$x
        }
        if (!is.null(ylevels)) {
          ylevels = if(is.numeric(ylevels)) levels(datapoints$y)[ylevels] else ylevels
          if (anyNA(ylevels) || !all(ylevels %in% levels(datapoints$y))) warning("not all 'ylevels' correspond to levels of 'y'")
          datapoints$y = factor(datapoints$y, levels = ylevels)
          if (y_by) datapoints$by = datapoints$y
        }
        
        x = datapoints$x
        y = datapoints$y
        
        # if either x_by or y_by are TRUE, we'll only split by facets and then
        # use some simple logic to assign colouring on the backend
        if (isTRUE(x_by) || isTRUE(y_by)) {
          datapoints = split(datapoints, list(datapoints$facet))
          datapoints = Filter(function(k) nrow(k) > 0, datapoints)
        } else {
          datapoints = split(datapoints, list(datapoints$by, datapoints$facet))
          datapoints = Filter(function(k) nrow(k) > 0, datapoints)
        }
        
        # construct spineplot rectangles and breaks points for each by-facet combo
        datapoints = Map(function(dat, x.categorical, off) {
          ## set up frequency table
          x = dat$x
          if(!x.categorical) {
            x = cut(as.numeric(x), breaks = breaks, include.lowest = TRUE)
          }
          ## TODO: process by grouping via: interaction + spacing + labeling
          ## (for now just do interaction)
          ## FIXME: data_facet only contains the first by group?
          ## if (any(dat$by != "")) x = interaction(dat$by, x)
          if(is.null(dat$weights)) {
            tab = table(x, dat$y)
          } else {
            tab = as.table(tapply(dat$weights, list(x, dat$y), FUN = sum, na.rm = TRUE))
            tab[is.na(tab)] = 0
          }
          nx = nrow(tab)
          ny = ncol(tab)
          
          ## compute coordinates
          ## cumulative proportions of x (plus off) vs. conditional cumulative proportions of y
          yat = rbind(0, apply(proportions(tab[, ny:1L, drop = FALSE], 1L), 1L, cumsum))
          yat[is.na(yat)] = 1
          xat = c(0, cumsum(proportions(marginSums(tab, 1L)) + off))
          
          ybottom = as.vector(yat[-(ny + 1L),])
          ytop = as.vector(yat[-1L,])
          xleft = rep(xat[1L:nx], rep(ny, nx))
          xright = rep(xat[2L:(nx+1L)] - off, rep(ny, nx))
          
          out = data.frame(
            by = dat$by[1], # already split
            facet = dat$facet[1], # already split
            ymin = ybottom,
            ymax = ytop,
            xmin = xleft,
            xmax = xright
          )
          
          attr(out, "nx") = nx
          attr(out, "ny") = ny
          attr(out, "xat") = xat
          attr(out, "yat") = yat
          return(out)
        }, 
        dat = datapoints,
        x.categorical = x.categorical,
        off = off
        )
        
        nx = attr(datapoints[[1]], "nx") ## should be the same for all by/facet groups
        ny = attr(datapoints[[1]], "ny") ## ditto
        xat = lapply(datapoints, attr, "xat")
        yat = lapply(datapoints, attr, "yat")
        datapoints = do.call(rbind, datapoints)
         
        if (is.null(yaxlabels)) yaxlabels = rev(levels(y))
      
        ## axis labels
        yaxlabels = if(is.null(yaxlabels)) levels(y) else rep_len(yaxlabels, ny)
        if (!is.null(yaxb)) {
          # yaxlabels = yaxlabels[yaxlabels %in% yaxb]
          ## rather use the "" assignment workaround below, since otherwise we 
          ## get a mismatch between the label names and ticks 
          yaxlabels[!(yaxlabels %in% yaxb)] = ""
        }
        if(x.categorical) {
          xaxlabels = if(is.null(xaxlabels)) {
            levels(x)
          } else {
            rep_len(xaxlabels, nx)
          }
        } else {
          xaxlabels = if(is.null(xaxlabels)) {
            if(is.numeric(x)) breaks else c(x[1L], x[c(diff(as.numeric(x)) > 0, TRUE)])
          } else {
            rep_len(xaxlabels, nx + 1L)
          }
        }
        
        # catch for x_by / y/by
        if (isTRUE(x_by)) datapoints$by = factor(rep(xaxlabels, each = ny)) # each x label extends over ny rows
        if (isTRUE(y_by)) datapoints$by = factor(rep_len(yaxlabels, nrow(datapoints)))

        x = c(datapoints$xmin, datapoints$xmax)
        y = c(datapoints$ymin, datapoints$ymax)
        ymin = datapoints$ymin
        ymax = datapoints$ymax
        xmin = datapoints$xmin
        xmax = datapoints$xmax
        by = if (null_by) by else datapoints$by
        facet = if (null_facet) facet else datapoints$facet

        # Save original values for type_info before overwriting
        axes_orig = axes
        xaxt_orig = xaxt
        yaxt_orig = yaxt
        # `frame.plot` defaults to TRUE for numeric-x spinograms (the outer box
        # is drawn by draw_spineplot below, after the tiles); NULL/unset counts
        # as TRUE. Preserve the user's choice before overwriting it, so the box
        # honours the top-level `frame.plot` rather than the tile-border `lwd`.
        frameplot_orig = !isFALSE(frame.plot)

        axes = FALSE
        frame.plot = FALSE
        xaxt = "n"
        yaxt = "n"
        xaxs = "i"
        yaxs = "i"
        ylabs = yaxlabels
        type_info = list(
          off = off,
          x.categorical = x.categorical,
          nx = nx,
          ny = ny,
          xat = xat,
          yat = yat,
          xaxlabels = xaxlabels,
          yaxlabels = yaxlabels,
          breaks = breaks,
          axes = axes_orig,
          frame.plot = frameplot_orig,
          xaxt = xaxt_orig,
          yaxt = yaxt_orig,
          null_by = null_by,
          x_by = x_by,
          y_by = y_by
        )
        
        # legend customizations
        # Mirror type_barplot()
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, c(
          "x", "y", "ymin", "ymax", "xmin", "xmax", "col", "bg", "datapoints",
          "by", "facet", "axes", "frame.plot", "xaxt", "yaxt", "xaxs", "yaxs",
          "ylabs", "type_info", "facet.args"
        ))
        
    }
    return(fun)
}

draw_spineplot = function(tol.ylab = 0.05, off = NULL, col = NULL, xaxlabels = NULL, yaxlabels = NULL, lighten = FALSE) {
    fun = function(ixmin, iymin, ixmax, iymax, ilty, ilwd, icol, ibg, 
                   flip,
                   facet_window_args,
                   type_info,
                   ifacet,
                   ...) {
      
      if (is.null(off)) off = type_info[["off"]]
      if (is.null(xaxlabels)) xaxlabels = type_info[["xaxlabels"]]
      if (is.null(yaxlabels)) yaxlabels = type_info[["yaxlabels"]]
      xat = type_info[["xat"]][[ifacet]]
      yat = type_info[["yat"]][[ifacet]]
      nx = type_info[["nx"]]
      ny = type_info[["ny"]]
      x.categorical = type_info[["x.categorical"]]
      null_by = type_info[["null_by"]]
      x_by = type_info[["x_by"]]
      y_by = type_info[["y_by"]]
      
      ## graphical parameters
      if (is.null(col)) {
        if (isFALSE(y_by)) {
          # For single-group displays, use a neutral grey ramp (gray.colors)
          # whenever the resolved seed colour is achromatic (e.g. the black
          # default of the plain default or the "bw"/"ipsum" themes), so these
          # are consistent regardless of whether a palette is declared -- the
          # same principle as the single-group fill logic in by_bg(). For grouped
          # displays we never switch to grayscale: each group (including one
          # whose palette colour is black) follows the same seq_palette ramp so
          # the fills stay in sync with the legend swatches.
          if (is.null(ibg)) ibg = icol
          gs = isTRUE(null_by) && is_achromatic(ibg)
          ibg = seq_palette(ibg, ny, grayscale = gs)
        } else {
          # When the y variable is itself the grouping (`y_by`), each band is a
          # group's palette colour. The fill is resolved once in prepare_legend()
          # and arrives via `ibg` -- lightened to match the other area types
          # (barplot/boxplot/violin) unless `lighten` is off (issue #646). Only
          # fall back to lightening the group colour `icol` here if no fill was
          # supplied (e.g. a standalone draw outside the legend pipeline).
          if (is.null(ibg)) ibg = if (isTRUE(lighten)) lighten_fill(icol) else icol
        }
        ibg = rep_len(ibg, ny)
      } else {
        ibg = col
      }

      rect(
          xleft = ixmin, ybottom = iymin, xright = ixmax, ytop = iymax,
          lty = ilty,
          lwd = ilwd,
          border = par("fg"), #icol,
          col = ibg
      )
      
      ## axes
      ## - standard categorical axes (xaxt/yaxt == "s") _without_ ticks
      ## - never draw additional axis lines, box always for spinogram
      if(type_info[["axes"]]) {
          if (x.categorical) {
              spine_axis(if (flip) 2 else 1, at = (xat[1L:nx] + xat[2L:(nx+1L)] - off)/2, labels = xaxlabels,
                  type = type_info[["xaxt"]], categorical = TRUE)
          } else {
              spine_axis(if (flip) 2 else 1, at = xat, labels = xaxlabels,
                  type = type_info[["xaxt"]], categorical = FALSE)
          }
          yat = yat[, if(flip) ncol(yat) else 1L]
          equidist = any(diff(yat) < tol.ylab)
          yat = if(equidist) seq.int(1/(2*ny), 1-1/(2*ny), by = 1/ny) else (yat[-1L] + yat[-length(yat)])/2
          spine_axis(if (flip) 3 else 2, at = yat, labels = yaxlabels,
              type = type_info[["yaxt"]], categorical = TRUE)
          if (is_facet_position(if(flip) "bottom" else "right", ifacet, facet_window_args)) spine_axis(if (flip) 1 else 4,
              type = type_info[["yaxt"]], categorical = FALSE)
      }
      # Outer box for numeric-x spinograms. This is a structural frame, so it
      # follows the top-level `frame.plot` (via type_info) rather than the
      # tile-border `lwd` -- otherwise `lwd = 0` would wrongly drop the box too.
      if (!x.categorical && isTRUE(type_info[["frame.plot"]])) box()
      
    }
    return(fun)
}




spine_axis = function(side, ..., type = "standard", categorical = TRUE) {
    type = match.arg(type, c("standard", "none", "labels", "ticks", "axis"))
    ## standard: with axis, ticks (unless categorical), and labels
    ## none: no axes
    ## labels: only labels without ticks and axis line
    ## ticks: only ticks and labels without axis line
    ## axis: only axis line and labels but no ticks

    if (type == "none") {
        invisible(numeric(0L))
    } else {
        args = list(side = side, ...)
        if (type == "labels") {
            args$tick = FALSE
        } else if (type == "ticks") {
            args$lwd = 0
            if (!("lwd.ticks" %in% names(args))) args$lwd.ticks = if (categorical) 0 else 1
        } else if (type == "axis") {
            if (categorical) {
                args$tick = FALSE
            } else {
                args$lwd.ticks = 0
            }
        } else {
            args$tick = !categorical
        }
        do.call("axis", args)
    }
}

#' Boxplot type
#' 
#' @description Type function for producing box-and-whisker plots.
#' Arguments are passed to \code{\link[graphics]{boxplot}}, although `tinyplot`
#' scaffolding allows added functionality such as grouping and faceting.
#' Box-and-whisker plots are the default plot type if `x` is a factor and `y` is
#' numeric.
#'
#' @inheritParams graphics::boxplot
#' @param lighten logical. Should the fills use a lighter, opaque tint of the
#'   series colour(s)? Default is `TRUE`, which keeps single- and multi-group
#'   displays consistent and lets the fill read cleanly over grid lines. Set to
#'   `FALSE` to use the fully-saturated palette colour(s) instead.
#' @examples
#' # "boxplot" type convenience string
#' tinyplot(weight ~ feed, data = chickwts, type = "boxplot")
#' 
#' # Note: Specifying the type here is redundant. Like base plot, tinyplot
#' # automatically produces a boxplot if x is a factor and y is numeric
#' tinyplot(weight ~ feed, data = chickwts)
#' 
#' # For flipped boxplots, it's usually better to use a dynamic theme to
#' # accommodate (horizontal) y-axis labels
#' tinyplot(weight ~ feed, data = chickwts, flip = TRUE, theme = "dynamic")
#' 
#' # Grouped boxplot example using a different dataset
#' # (C.f., the final example in `?boxplot`)
#' tinyplot(
#'   len ~ dose | supp, data = ToothGrowth, type = "boxplot",
#'   main = "Guinea Pigs' Tooth Growth",
#'   legend = list(title = "Supplement"),
#'   xlab = "Vitamin C dose mg", ylab = "tooth length"
#' )
#' 
#' # Use `type_boxplot()` to pass extra arguments for customization
#' tinyplot(
#'   len ~ dose | supp, data = ToothGrowth,
#'   type = type_boxplot(boxwex = 0.3, staplewex = 0, outline = FALSE), lty = 1,
#'   legend = list(title = "Supplement"),
#'   main = "Guinea Pigs' Tooth Growth",
#'   xlab = "Vitamin C dose mg", ylab = "tooth length"
#' )
#' @export
type_boxplot = function(
    range = 1.5,
    width = NULL,
    varwidth = FALSE,
    notch = FALSE,
    outline = TRUE,
    boxwex = 0.8,
    staplewex = 0.5,
    outwex = 0.5,
    lighten = TRUE) {
  out = list(
    draw = draw_boxplot(
      range = range,
      width = width,
      varwidth = varwidth,
      notch = notch,
      outline = outline,
      boxwex = boxwex,
      staplewex = staplewex,
      outwex = outwex),
    data = data_boxplot(boxwex = boxwex, lighten = lighten),
    name = "boxplot"
  )
  class(out) = "tinyplot_type"
  return(out)
}



draw_boxplot = function(range, width, varwidth, notch, outline, boxwex, staplewex, outwex) {
    fun = function(iby, ix, iy, ipch, ilty, icol, ibg, x_by = FALSE, facet_by = FALSE, ngrps = 1, flip, ...) {

        at_ix = unique(ix)
        if (isTRUE(x_by)) boxwex = boxwex * 2

        # Handle multiple groups
        if (ngrps > 1 && isFALSE(x_by) && isFALSE(facet_by)) {
            group_offsets = get_environment_variable(".group_offsets")
            boxwex = boxwex / ngrps - 0.01
            at_ix = at_ix + group_offsets[iby]
        }

        boxplot(
            formula = iy ~ ix,
            pch = ipch,
            lty = ilty,
            border = icol,
            col = ibg,
            horizontal = flip,
            add = TRUE, axes = FALSE,
            at = at_ix,
            range = range,
            width = width,
            varwidth = varwidth,
            notch = notch,
            outline = outline,
            boxwex = boxwex,
            staplewex = staplewex,
            outwex = outwex
        )
    }
    return(fun)
}



data_boxplot = function(boxwex = 0.8, lighten = TRUE) {
    fun = function(settings, ...) {
        env2env(settings, environment(), c("datapoints", "by", "facet", "null_facet", "null_palette", "x", "col", "bg", "null_by"))
        settings[["lighten"]] = lighten
        # Convert x to factor if it's not already
        datapoints$x = as.factor(datapoints$x)

        # Handle factor levels and maintain order
        xlvls = levels(datapoints$x)
        xlabs = seq_along(xlvls)
        names(xlabs) = xlvls
        datapoints$x = as.integer(datapoints$x)

        if (null_by && null_facet) {
            xord = order(datapoints$x)
        } else if (null_facet) {
            xord = order(datapoints$by, datapoints$x)
        } else if (null_by) {
            xord = order(datapoints$facet, datapoints$x)
        } else {
            xord = order(datapoints$by, datapoints$facet, datapoints$x)
        }

        # Check if user provided palette before substitute)
        if (length(unique(datapoints[["by"]])) == 1 && null_palette) {
            # With a theme palette active, leave bg = NULL so the fill tracks
            # the resolved border colour (see by_bg). Otherwise use neutral grey.
            if (is.null(bg) && is.null(get_tpar("palette.qualitative", default = NULL))) bg = "lightgray"
        } else {
            if (is.null(bg)) bg = "by"
        }

        # Reorder x, y, ymin, and ymax based on the order determined
        datapoints = datapoints[xord,]

        # Return the result as a list called 'out'
        x = datapoints$x
        y = datapoints$y
        ymin = datapoints$ymin
        ymax = datapoints$ymax
        by = if (length(unique(datapoints$by)) > 1) datapoints$by else by
        facet = if (length(unique(datapoints$facet)) > 1) datapoints$facet else facet
        
        # Compute group offsets for multi-group boxplots
        ngrps = length(unique(datapoints$by))
        if (ngrps > 1 && !settings$x_by) {
            boxwex_grp = boxwex / ngrps - 0.01
            group_offsets = seq(
                -((boxwex - boxwex_grp) / 2),
                ((boxwex - boxwex_grp) / 2),
                length.out = ngrps
            )
        } else {
            group_offsets = rep(0, max(ngrps, 1))
        }
        offsets_axis = "x"

        # legend customizations
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, c(
            "x",
            "y",
            "ymin",
            "ymax",
            "xlabs",
            "datapoints",
            "col",
            "bg",
            "by",
            "facet",
            "group_offsets",
            "offsets_axis"
        ))
    }
    return(fun)
}

bubble = function(settings) {
  # Only process for points and text types
  if (!(settings$type %in% c("p", "text"))) return(invisible())
  
  cex = settings$cex
  
  # Only process if cex is a vector matching data length
  if (is.null(cex) || length(cex) != nrow(settings$datapoints)) return(invisible())
  
  clim = settings$clim %||% c(0.5, 2.5)
  
  bubble = TRUE
  
  ## Identify the pretty break points for our bubble labels
  bubble_labs = pretty(cex, n = 5)
  len_labs = length(bubble_labs)
  cex = rescale_num(sqrt(c(bubble_labs, cex)) / pi, to = clim)
  bubble_cex = cex[1:len_labs]
  cex = cex[(len_labs+1):length(cex)]
  
  # catch for cases where pretty breaks leads to smallest category of 0
  if (bubble_labs[1] == 0) {
    bubble_labs = bubble_labs[-1]
    bubble_cex = bubble_cex[-1]
  }
  names(bubble_cex) = format(bubble_labs)
  
  if (max(clim) > 2.5) {
    settings$legend_args[["x.intersp"]] = max(clim) / 2.5
    settings$legend_args[["y.intersp"]] = sapply(bubble_cex / 2.5, max, 1)
  }

  ## fixme: can't assign pt.cex here b/c of dual legend gotcha (don't want to
  ##   override the "normal" pt.cex too) 
  # legend_args[["pt.cex"]] = legend_args[["pt.cex"]] %||% (settings[["cex"]] %||% par("cex"))
  
  # Must update settings with bubble/bubble_cex/cex before calling sanitize_bubble
  env2env(environment(), settings, c("bubble", "bubble_cex", "cex"))
  
  sanitize_bubble(settings)
}

sanitize_bubble = function(settings) {
  env2env(settings, environment(), c("datapoints", "pch", "alpha", "bg", "cex", "bubble"))
  
  if (!bubble) return(invisible())
  
  datapoints[["cex"]] = cex
  bubble_pch = if (!is.null(pch) && length(pch)==1) pch else par("pch")
  bubble_alpha = if (!is.null(alpha)) alpha else 1
  bubble_bg_alpha = if (!is.null(bg) && length(bg)==1 && is.numeric(bg) && bg > 0 && bg <=1) bg else 1

  env2env(environment(), settings, c("datapoints", "bubble_pch", "bubble_alpha", "bubble_bg_alpha"))
}

#' Format labels
#'
#' @description Function for formatting label appearance, e.g. axis
#' ticks labels. This is what the top-level `xaxl` and `yaxl` arguments
#' from [`tinyplot`] ultimately get passed to.
#' @param x a numeric or character vector
#' @param labeller a formatting function to be applied to `x`, e.g. [`format`],
#'   [`toupper`], [`abs`], or other custom function (including from the popular
#'   **scales** package). Can also be one of the following convenience strings
#'   (symbols), for which common formatting transformations are provided:
#'   `"percent"` (`"%"`), `"comma"` (`","`), `"log"` (`"l"`), `"dollar"`
#'   (`"$"`), `"euro"` (`"€"`), or `"sterling"` (`"£"`).
#' @param na.ignore logical indicating whether the labelling function should
#'   ignore `NA` values in `x`. In other words, should the `NA` values be left
#'   as-is? Default is `TRUE`.
#' @param na.rm logical indicating whether `NA` values should be removed from
#'   `x` and thus the return object too. Default is `TRUE`, but only evaluated
#'   if `na.ignore` is `FALSE`.
#' @return a character vector of the same length as `x` with the transformed
#'   labels.
#' @export
#' @seealso [`tinyplot`]
#' @examples
#' x = 1e4
#' tinylabel(x, "comma")
#' tinylabel(x, ",") # same
#' tinylabel(x, "$") # or "dollar"
#' 
#' # invoke tinylabel from a parent tinyplot call...
#' #   => x/yaxl for adjusting axes tick labels
#' #   => legend = list(labeller = ...) for adjusting the legend labels
#' s77 = transform(data.frame(state.x77), Illiteracy = Illiteracy / 100)
#' tinyplot(Life.Exp ~ Income | Illiteracy, data = s77,
#'          xaxl = '$',
#'          legend = list(labeller = '%'))
#'
#' # log example (combined with axis scaling)
#' tinyplot(x = 10^c(10:0), y = 0:10, type = "b",
#'          log = "x", xaxl = "log")
#'
#' # combine with `x/yaxb` to adjust the actual tick marks ("break points")
#' # at the same time
#' tinyplot(x = 10^c(10:0), y = 0:10, type = "b",
#'          log = "x", xaxl = "log", xaxb = 10^c(1,3,5,7,9))
#'
#' #
#' ## custom function examples
#'
#' ## example I: date formatting
#'
#' dat = data.frame(
#'   date = seq(as.Date("2000/1/1"), by = "month", length.out = 12),
#'   trend = 1:12 + rnorm(12, sd = 1)
#' )
#'
#' tinyplot(trend ~ date, data = dat,
#'          xaxl = function(x) format(x, "%b, %Y"))
#'
#' ## example II: string wrapping
#'
#' # create a "vectorised" version of `base::strwrap` that breaks long
#' # strings into new lines every 18 characters
#' strwrap18 = function(x) sapply(
#'   strwrap(x, width = 18, simplify = FALSE),
#'   paste,
#'   collapse = "\n"
#' )
#'
#' # now demonstrate on a dataset with long y-tick labels
#' dat2 = data.frame(
#'   x = rep(rnorm(100), 3),
#'   y = c(
#'     "tinyplot is a lightweight extension of the base R graphics system.",
#'     "R is a language for statistical computing.",
#'     "Data visualization is an essential skill."
#'   )
#' )
#'
#' tinyplot(y ~ x, data = dat2, type = "j",
#'          yaxl = strwrap18,
#'          theme = "bw") # use theme for horizontal labels + dynamic margin
#' @export
tinylabel = function(x, labeller = NULL, na.ignore = TRUE, na.rm = TRUE) {
  if (is.null(labeller)) {
    return(x)
  }
  assert_logical(na.ignore)
  assert_logical(na.rm)
  xidx = if (na.ignore) {
    which(!is.na(x))
  } else if (na.rm) {
    !is.na(x)
  } else {
    seq_along(x)
  }
  if (is.character(labeller)) {
    labeller = labeller_fun((labeller))
  }
  # don't need to subset if everything is being used. DateTime also require
  # exception logic (e.g., date format needs to be consistent for whole vector)
  if (length(xidx) == length(x) || inherits(x, c("POSIXt", "Date"))) {
    x = labeller(x)
  } else {
    x[xidx] = labeller(x[xidx])
  }
  return(x)
}


labeller_fun = function(label = "percent") {
  labels = c(
    "%"      = "percent",
    ","      = "comma",
    "$"      = "dollar",
    "\u20ac" = "euro",
    "\u00a3" = "sterling",
    "l"      = "log"
  )
  if (label %in% names(labels)) {
    label = labels[label]
  }

  ## all labels plus absolute value version
  # labels = c("percent", "comma", "dollar", "euro", "sterling")
  labels = c(labels, paste0("abs_", labels))

  ## match full label first, then store abs_ info separately
  label = match.arg(label, labels)
  abs_ = substr(label, 1L, 4L) == "abs_"
  if (abs_) {
    label = substr(label, 5L, nchar(label))
  }

  ## actual formatting functions

  # Find the smallest number of decimal places (from 0:max_decimals) that keeps
  # the unique values distinct, so a single consistent format can be applied to
  # the whole vector. Falls back to max_decimals.
  consistent_decimals = function(x, max_decimals = 5L) {
    ux = unique(as.numeric(x))
    Find(
      function(d) {
        length(unique(sprintf(paste0("%.", d, "f"), ux))) == length(ux)
      },
      0:max_decimals
    ) %||%
      max_decimals
  }

  format_percent = function(x) {
    pct = as.numeric(x) * 100
    d = consistent_decimals(pct)
    sprintf(paste0("%.", d, "f%%"), pct)
  }

  # Currency convention: keep clean integers integer-valued, but show at least
  # two decimal places whenever any fractional component is present (e.g.
  # "$0.50" rather than "$0.5"). Negative values place the sign in front of the
  # currency symbol (e.g. "-$1.50" rather than "$-1.50"). The comma formatter is
  # not currency, so it just uses the smallest consistent number of decimals.
  format_currency = function(x, symbol) {
    xn = as.numeric(x)
    d = consistent_decimals(xn)
    has_decimal = any(xn[!is.na(xn)] %% 1 != 0)
    if (has_decimal) d = max(d, 2L)
    fmt = formatC(abs(xn), format = "f", digits = d, big.mark = ",")
    sign = ifelse(!is.na(xn) & xn < 0, "-", "")
    paste0(sign, symbol, fmt)
  }

  format_comma = function(x) {
    d = consistent_decimals(x)
    formatC(as.numeric(x), format = "f", digits = d, big.mark = ",")
  }

  format_dollar = function(x) format_currency(x, "$")

  format_euro = function(x) format_currency(x, "\u20ac")

  format_sterling = function(x) format_currency(x, "\u00a3")

  format_log = function(x) {
    x = as.numeric(x)
    parse(text = paste0(10, "^", format(log10(x), digits = 3)))
  }

  fun = switch(
    label,
    percent  = format_percent,
    comma    = format_comma,
    dollar   = format_dollar,
    euro     = format_euro,
    sterling = format_sterling,
    log      = format_log
  )

  ## combine with absolute value if necessary
  numeric_fun = if (abs_) function(x) fun(abs(x)) else fun

  # The built-in formatters coerce their input via as.numeric(). Guard against
  # non-numeric input (e.g. categorical axis labels reaching a numeric labeller
  # via flip = TRUE) by leaving such values unchanged, which avoids "NAs
  # introduced by coercion" warnings (#622). The is.na(xn) & !is.na(x) test only
  # flags values that became NA through coercion, so genuine NA input still
  # reaches the formatter.
  function(x) {
    xn = suppressWarnings(as.numeric(x))
    if (any(is.na(xn) & !is.na(x))) {
      return(x)
    }
    numeric_fun(x)
  }
}

#' Rectangle plot type
#'
#' @description Type function for plotting rectangles.
#' 
#' @details Contrary to base \code{\link[graphics]{rect}}, rectangles in
#' [tinyplot] must be specified using the `xmin`, `ymin`,`xmax`, and `ymax`
#' arguments. 
#' 
#' @examples
#' i = 4*(0:10)
#' 
#' # "rect" type convenience character string
#' tinyplot(
#'   xmin = 100+i, ymin = 300+i, xmax = 150+i, ymax = 380+i,
#'   by = i, fill = 0.2,
#'   type = "rect"
#' )
#' 
#' # Same result with type_rect()
#' tinyplot(
#'   xmin = 100+i, ymin = 300+i, xmax = 150+i, ymax = 380+i,
#'   by = i, fill = 0.2,
#'   type = type_rect()
#' )
#' 
#' @export
type_rect = function() {
  out = list(
    draw = draw_rect(),
    data = data_rect(),
    name = "rect"
  )
  class(out) = "tinyplot_type"
  return(out)
}


data_rect = function() {
  fun = function(settings, ...) {
    # legend customizations
    settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
    settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% par("lwd")
    settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
    settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
    settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
  }
  return(fun)
}


draw_rect = function() {
    fun = function(ixmin, iymin, ixmax, iymax, ilty, ilwd, icol, ibg, ...) {
        rect(
            xleft = ixmin, ybottom = iymin, xright = ixmax, ytop = iymax,
            lty = ilty,
            lwd = ilwd,
            border = icol,
            col = ibg
        )
    }
    return(fun)
}

#
## Input Sanitization -----
#

#' Sanitize and normalize legend input
#'
#' @description Converts various legend input formats (NULL, character, list,
#'   call) into a standardized legend_args list with an "x" position element.
#'
#' @param legend Legend specification (NULL, character, list, or call)
#' @param legend_args Existing legend_args list to merge with
#'
#' @returns Normalized legend_args list with at least an "x" element
#'
#' @keywords internal
sanitize_legend = function(legend, legend_args) {
  if (is.null(legend_args[["x"]])) {

    # Normalize legend to a list
    largs = if (is.null(legend)) {
      list(x = "right!")
    } else if (is.character(legend)) {
      list(x = legend)
    } else if (is.list(legend)) {
      # Handle unnamed first element as position
      if (length(legend) >= 1 && is.character(legend[[1]]) &&
          (is.null(names(legend)) || names(legend)[1] == "")) {
        names(legend)[1] = "x"
      }
      legend
    } else if (inherits(legend, c("call", "name"))) {
      # Convert call to list and handle unnamed first arg as position
      new_legend = as.list(legend)[-1]  # Remove function name
      if (length(new_legend) >= 1 && (is.null(names(new_legend)) || names(new_legend)[1] == "")) {
        names(new_legend)[1] = "x"
      }
      # Evaluate language elements so e.g. c(0, 1, 2) becomes a real vector
      for (nm in names(new_legend)) {
        if (is.language(new_legend[[nm]])) {
          new_legend[[nm]] = eval(new_legend[[nm]])
        }
      }
      new_legend
    } else {
      list(x = "right!")  # Fallback
    }

    # Ensure position exists
    if (is.null(largs[["x"]])) largs[["x"]] = "right!"

    # Merge
    legend_args = modifyList(legend_args, largs, keep.null = TRUE)
  }

  legend_args
}


#
## Helper Functions -----
#

# Unit conversion helpers (used extensively throughout legend positioning)
lines_to_npc_x = function(val) {
  grconvertX(val, from = "lines", to = "npc") - grconvertX(0, from = "lines", to = "npc")
}

lines_to_user_x = function(val) {
  grconvertX(val, from = "lines", to = "user") - grconvertX(0, from = "lines", to = "user")
}

lines_to_user_y = function(val) {
  grconvertY(val, from = "lines", to = "user") - grconvertY(0, from = "lines", to = "user")
}


# Adjust margins for outer legend placement, measure, and apply soma
legend_outer_margins = function(legend_env, apply = TRUE) {
  omar = legend_env$omar
  ooma = legend_env$ooma
  lmar = legend_env$lmar

  # Step 1: Prepare margins before measuring
  if (legend_env$outer_side) {
    # Extra bump for spineplot if outer_right legend (to accommodate secondary y-axis)
    if (identical(legend_env$type, "spineplot")) {
      lmar[1] = lmar[1] + 1.1
    }

    # Set inner margins before fake legend is drawn
    if (legend_env$outer_right) {
      omar[4] = 0
    } else {
      # For outer left we have to account for the y-axis label too
      omar[2] = par("mgp")[1] + 1 * par("cex.lab")
    }
    par(mar = omar)

    if (legend_env$new_plot) {
      plot.new()
      # For themed + dynamic plots, reinstate adjusted plot margins
      if (legend_env$dynmar) {
        omar = par("mar")
        if (legend_env$outer_right) {
          omar[4] = 0
        } else {
          omar[2] = par("mgp")[1] + 1 * par("cex.lab")
        }
        par(mar = omar)
      }
    }

  } else if (legend_env$outer_end) {
    # Set inner margins before fake legend is drawn
    if (legend_env$outer_bottom) {
      omar[1] = par("mgp")[1] + 1 * par("cex.lab")
      if (legend_env$has_sub && (is.null(.tpar[["side.sub"]]) || .tpar[["side.sub"]] == 1)) {
        omar[1] = omar[1] + 1 * par("cex.sub")
      }
    } else {
      if (is.null(legend_env$dynmar_title_mar)) {
        ooma[3] = ooma[3] + legend_env$topmar_epsilon
        par(oma = ooma)
      }
    }
    par(mar = omar)

    if (legend_env$new_plot) {
      plot.new()
      # For themed + dynamic plots, reinstate adjusted plot margins
      if (legend_env$dynmar) {
        omar = par("mar")
        if (legend_env$outer_bottom) {
          omar[1] = par("mgp")[1] + 1 * par("cex.lab")
          if (legend_env$has_sub && (is.null(.tpar[["side.sub"]]) || .tpar[["side.sub"]] == 1)) {
            omar[1] = omar[1] + 1 * par("cex.sub")
          }
        } else {
          if (is.null(legend_env$dynmar_title_mar)) {
            ooma[3] = ooma[3] + legend_env$topmar_epsilon
            par(oma = ooma)
          }
        }
        par(mar = omar)
      }
    }
  } else {
    if (legend_env$new_plot) plot.new()
  }

  # Update legend environment with prepared margins
  legend_env$omar = omar
  legend_env$ooma = ooma
  legend_env$lmar = lmar

  # Step 2: Measure legend dimensions
  legend_env$dims = measure_fake_legend(legend_env)

  # Step 3: Apply soma if drawing
  if (apply) {
    soma = if (legend_env$outer_side) {
      grconvertX(legend_env$dims$rect$w, to = "lines") - grconvertX(0, to = "lines")
    } else if (legend_env$outer_end) {
      grconvertY(legend_env$dims$rect$h, to = "lines") - grconvertY(0, to = "lines")
    } else {
      0
    }
    soma = soma + sum(legend_env$lmar)

    if (legend_env$outer_side) {
      legend_env$ooma[if (legend_env$outer_right) 4 else 2] = soma
    } else if (legend_env$outer_end) {
      if (legend_env$outer_bottom) {
        legend_env$ooma[1] = soma
        if (legend_env$has_cap) {
          cex_cap = get_tpar("cex.cap", 1)
          legend_env$ooma[1] = legend_env$ooma[1] + cex_cap + 0.2
        }
      } else {
        if (!is.null(legend_env$dynmar_title_mar)) {
          legend_env$omar[3] = legend_env$dynmar_title_mar + soma
        } else {
          legend_env$omar[3] = legend_env$omar[3] + soma - legend_env$topmar_epsilon
        }
        par(mar = legend_env$omar)
        set_environment_variable(.top_legend_soma = soma)
      }
    }
    par(oma = legend_env$ooma)
  }
}


# Calculate legend inset for outer placement
measure_legend_inset = function(legend_env) {
  if (legend_env$outer_side) {
    inset_val = lines_to_npc_x(legend_env$lmar[1])
    # Extra space needed for "left!" because of lhs inner margin
    if (!legend_env$outer_right) {
      inset_val = inset_val + lines_to_npc_x(par("mar")[2])
    }
    c(1 + inset_val, 0)

  } else if (legend_env$outer_end) {
    # Note: Y-direction uses grconvertY (not lines_to_npc_x which is X-only)
    inset_val = grconvertY(legend_env$lmar[1], from = "lines", to = "npc") -
                grconvertY(0, from = "lines", to = "npc")
    if (legend_env$outer_bottom) {
      # Extra space needed for "bottom!" because of lhs inner margin
      inset_bump = grconvertY(par("mar")[1], from = "lines", to = "npc") -
                   grconvertY(0, from = "lines", to = "npc")
      inset_val = inset_val + inset_bump
    } else {
      epsilon_bump = grconvertY(legend_env$topmar_epsilon, from = "lines", to = "npc") -
                     grconvertY(0, from = "lines", to = "npc")
      inset_val = inset_val + epsilon_bump
    }
    c(0, 1 + inset_val)

  } else {
    0
  }
}


compute_ljust_text_width = function(legend_env) {
  if (!isTRUE(legend_env$ljust)) return(invisible(NULL))

  args_oh = modifyList(legend_env$args,
    list(plot = FALSE, title = NULL, text.width = 0), keep.null = TRUE)
  overhead = do.call("legend", args_oh)$rect$w

  args_nat = modifyList(legend_env$args,
    list(plot = FALSE, text.width = NULL), keep.null = TRUE)
  target_w = do.call("legend", args_nat)$rect$w

  tw_needed = target_w - overhead
  if (tw_needed > 0) {
    legend_env$args[["text.width"]] = tw_needed
  }
}


# Internal workhorse function for legend rendering
# This function is called inside recordGraphics() so that all coordinate-dependent
# calculations are re-executed when the plot window is resized
tinylegend = function(legend_env) {
  # Reset to base values (before soma/inset were applied)
  # This is necessary because legend_env is passed by reference and modifications
  # persist across recordGraphics() replays
  legend_env$omar = legend_env$omar_base
  legend_env$ooma = legend_env$ooma_base
  legend_env$args[["inset"]] = legend_env$inset_base

  # Clear stale text.width before measuring (measurement doesn't need it)
  if (isTRUE(legend_env$ljust)) {
    legend_env$args[["text.width"]] = NULL
  }

  # Re-measure legend dimensions (device size may have changed on resize)
  legend_env$dims = measure_fake_legend(legend_env)

  # Calculate and apply soma (outer margin adjustment based on legend size)
  # When soma_target is set (multi-legend), use it directly so all legends
  # share the same outer margin and their left edges align.
  if (!is.null(legend_env$soma_target)) {
    soma = legend_env$soma_target
  } else {
    soma = if (legend_env$outer_side) {
      grconvertX(legend_env$dims$rect$w, to = "lines") - grconvertX(0, to = "lines")
    } else if (legend_env$outer_end) {
      grconvertY(legend_env$dims$rect$h, to = "lines") - grconvertY(0, to = "lines")
    } else {
      0
    }
    soma = soma + sum(legend_env$lmar)
  }

  if (legend_env$outer_side) {
    legend_env$ooma[if (legend_env$outer_right) 4 else 2] = soma
  } else if (legend_env$outer_end) {
    if (legend_env$outer_bottom) {
      legend_env$ooma[1] = soma
      if (legend_env$has_cap) {
        cex_cap = get_tpar("cex.cap", 1)
        legend_env$ooma[1] = legend_env$ooma[1] + cex_cap + 0.2
      }
    } else {
      if (!is.null(legend_env$dynmar_title_mar)) {
        legend_env$omar[3] = legend_env$dynmar_title_mar + soma
      } else {
        legend_env$omar[3] = legend_env$omar[3] + soma - legend_env$topmar_epsilon
      }
      par(mar = legend_env$omar)
      set_environment_variable(.top_legend_soma = soma)
    }
  }
  par(oma = legend_env$ooma)

  # Calculate inset for legend positioning
  legend_env$inset = measure_legend_inset(legend_env)

  # Refresh plot area for exact inset spacing
  # (Uses hook to preserve existing plot with par(new = TRUE))
  oldhook = getHook("before.plot.new")
  setHook("before.plot.new", function() par(new = TRUE), action = "append")
  setHook("before.plot.new", function() par(mar = legend_env$omar), action = "append")
  plot.new()
  setHook("before.plot.new", oldhook, action = "replace")

  # Recompute text.width in the final coordinate context
  if (isTRUE(legend_env$ljust)) {
    compute_ljust_text_width(legend_env)
  }

  # Set the inset in legend args
  legend_env$args[["inset"]] = if (legend_env$user_inset) {
    legend_env$args[["inset"]] + legend_env$inset
  } else {
    legend_env$inset
  }

  # Draw the legend
  if (legend_env$gradient) {
    # Ensure col is set correctly for gradients
    if (!more_than_n_unique(legend_env$args[["col"]], 1)) {
      if (!is.null(legend_env$args[["pt.bg"]]) && length(legend_env$args[["pt.bg"]]) == 100) {
        legend_env$args[["col"]] = legend_env$args[["pt.bg"]]
      }
    }

    draw_gradient_swatch(
      legend_args = legend_env$args,
      fklgnd = legend_env$dims,
      lmar = legend_env$lmar,
      outer_side = legend_env$outer_side,
      outer_end = legend_env$outer_end,
      outer_right = legend_env$outer_right,
      outer_bottom = legend_env$outer_bottom,
      user_inset = legend_env$user_inset
    )
  } else {
    do.call("legend", legend_env$args)
  }

  if (legend_env$outer_bottom && legend_env$has_cap) {
    cex_cap = get_tpar("cex.cap", 1)
    mtext(
      legend_env$cap_text,
      side = 1,
      outer = TRUE,
      line = par("oma")[1] - 1,
      cex = cex_cap,
      col = get_tpar("col.cap", "black"),
      adj = get_tpar(c("adj.cap", "adj")),
      font = get_tpar("font.cap", 1),
      las = 1
    )
  }
}


# Measure legend dimensions using a fake (non-plotted) legend
measure_fake_legend = function(legend_env) {
  fklgnd.args = modifyList(
    legend_env$args,
    list(plot = FALSE),
    keep.null = TRUE
  )

  if (legend_env$gradient) {
    lgnd_labs_tmp = na.omit(fklgnd.args[["legend"]])
    if (length(lgnd_labs_tmp) < 5L) {
      nmore = 5L - length(lgnd_labs_tmp)
      lgnd_labs_tmp = c(lgnd_labs_tmp, rep("", nmore))
    }
    fklgnd.args = modifyList(
      fklgnd.args,
      list(legend = lgnd_labs_tmp),
      keep.null = TRUE
    )
    if (legend_env$outer_end) {
      fklgnd.args = modifyList(
        fklgnd.args,
        list(title = NULL),
        keep.null = TRUE
      )
    }
  }

  do.call("legend", fklgnd.args)
}




#
## Legend Context & Preparation -----
#

#' Prepare legend context from settings
#'
#' @description Main orchestrator that determines:
#'   - Whether to draw legend
#'   - Legend labels and formatting
#'   - Whether multi-legend is needed (for bubble charts)
#'   - Gradient legend setup for continuous grouping
#'
#' @param settings Settings environment from tinyplot
#'
#' @returns NULL (modifies settings environment in-place)
#'
#' @keywords internal
prepare_legend = function(settings) {
  env2env(
    settings,
    environment(),
    c(
      "add",
      "bubble",
      "bubble_cex",
      "by",
      "by_continuous",
      "cap",
      "cex_dep",
      "cex_fct_adj",
      "col",
      "datapoints",
      "legend",
      "legend_args",
      "ngrps",
      "null_by",
      "sub",
      "ylab"
    )
  )

  ncolors = length(col)
  lgnd_labs = rep(NA, times = ncolors)

  # Generate labels for continuous (gradient) legends
  if (isTRUE(by_continuous)) {
    nlabs = 5
    ubyvar = unique(by)
    byvar_range = range(ubyvar)
    pbyvar = pretty(byvar_range, n = nlabs)
    pbyvar = pbyvar[pbyvar >= byvar_range[1] & pbyvar <= byvar_range[2]]
    if (length(ubyvar) == 2 && all(ubyvar %in% pbyvar)) {
      pbyvar = ubyvar
    } else if (length(pbyvar) > nlabs) {
      pbyvar = pbyvar[seq_along(pbyvar) %% 2 == 0]
    }
    pidx = rescale_num(c(byvar_range, pbyvar), to = c(1, ncolors))[-c(1:2)]
    pidx = round(pidx)
    lgnd_labs[pidx] = pbyvar
  }

  has_legend = FALSE
  multi_legend = bubble && !null_by && !isFALSE(legend)
  lgnd_cex = NULL

  # Normalize legend argument
  if (isFALSE(legend)) {
    legend = "none"
  } else if (isTRUE(legend)) {
    legend = NULL
  }

  if (!is.null(legend) && is.character(legend) && legend == "none") {
    legend_args[["x"]] = "none"
    multi_legend = FALSE
  }

  # Handle bubble-only legend (no grouping)
  if (null_by) {
    if (bubble && !multi_legend) {
      legend_args[["title"]] = cex_dep
      lgnd_labs = names(bubble_cex)
      lgnd_cex = bubble_cex * cex_fct_adj
    } else if (is.null(legend)) {
      legend = "none"
      legend_args[["x"]] = "none"
    }
  }

  legend_draw_flag = (is.null(legend) || !is.character(legend) || legend != "none" || bubble) && !isTRUE(add)
  has_sub = text_line_count(sub) > 0L
  has_cap = text_line_count(cap) > 0L

  # Generate labels for discrete legends
  if (legend_draw_flag && isFALSE(by_continuous) && (!bubble || multi_legend)) {
    if (ngrps > 1) {
      lgnd_labs = if (is.factor(datapoints$by)) levels(datapoints$by) else unique(datapoints$by)
    } else {
      lgnd_labs = ylab
    }
  }

  # Grouped (`y_by`) spineplots draw their fills inside draw_spineplot() rather
  # than via the shared `bg` channel, so `settings$bg` is NULL and the legend has
  # nothing to mirror. Resolve the swatch fill here (where `col` is the resolved
  # group palette) into `bg`, so the rest of the legend machinery treats it like
  # any other area type. The fill tracks `lighten`, matching the plotted tiles.
  if (identical(settings$type, "spineplot") && isTRUE(settings$type_info[["y_by"]])) {
    settings$bg = if (isTRUE(settings$lighten)) lighten_fill(col) else col
  }

  env2env(
    environment(),
    settings,
    c(
      "lgnd_labs",
      "has_legend",
      "multi_legend",
      "lgnd_cex",
      "legend",
      "legend_args",
      "legend_draw_flag",
      "has_sub",
      "has_cap"
    )
  )
}


#' Build legend arguments list
#'
#' @description Constructs and configures the legend_args list by:
#'   - Sanitizing legend input
#'   - Setting defaults for all legend parameters
#'   - Computing positioning flags from original position (before transformation)
#'   - Adjusting position anchors for outer legends
#'   - Adjusting for special cases (gradient, horizontal, multi-column)
#'   - Populating legend_env with args and positioning flags
#'
#' @param legend_env Legend environment to populate
#' @param legend Legend placement keyword or list
#' @param legend_args Additional legend arguments
#' @param by_dep The (deparsed) "by" grouping variable name
#' @param lgnd_labs The legend labels
#' @param labeller Character or function for formatting labels
#' @param type Plot type
#' @param pch Plotting character(s)
#' @param lty Line type(s)
#' @param lwd Line width(s)
#' @param col Color(s)
#' @param bg Background fill color(s)
#' @param cex Character expansion(s)
#' @param gradient Logical indicating gradient legend
#'
#' @returns NULL (modifies legend_env in-place)
#'
#' @keywords internal
build_legend_args = function(
  legend_env,

  # Legend specification
  legend,
  legend_args,

  # Labels and grouping
  by_dep,
  lgnd_labs,
  labeller = NULL,

  # Visual aesthetics
  type,
  pch,
  lty,
  lwd,
  col,
  bg,
  cex,

  # Configuration
  gradient
) {
  # Ensure legend_args[["x"]] is populated. When called from the main
  # tinyplot pipeline, this is a no-op (sanitize_legend short-circuits on
  # the is.null guard); when called standalone via the public draw_legend
  # entry point, this normalizes the input.
  legend_args = sanitize_legend(legend, legend_args)

  # Set defaults
  if (!exists("title", where = legend_args)) legend_args[["title"]] = by_dep
  # Treat title = FALSE the same as NULL, i.e. suppress the title (#652).
  if (isFALSE(legend_args[["title"]])) legend_args[["title"]] = NULL
  legend_args[["pch"]] = legend_args[["pch"]] %||% pch
  legend_args[["lty"]] = legend_args[["lty"]] %||% lty
  legend_args[["col"]] = legend_args[["col"]] %||% col
  legend_args[["bty"]] = legend_args[["bty"]] %||% "n"
  legend_args[["horiz"]] = legend_args[["horiz"]] %||% FALSE
  legend_args[["xpd"]] = legend_args[["xpd"]] %||% NA
  legend_args[["lwd"]] = legend_args[["lwd"]] %||% lwd

  # Special handling of pt.cex for bubble plots
  if (is.null(type) || type %in% c("p", "text")) {
    legend_args[["pt.cex"]] = legend_args[["pt.cex"]] %||% (cex %||% par("cex"))
  }

  # Gradient legend adjustments
  if (gradient) {
    legend_args[["pch"]] = 22
    legend_args[["pt.cex"]] = legend_args[["pt.cex"]] %||% 3.5
    legend_args[["y.intersp"]] = legend_args[["y.intersp"]] %||% 1.25
    legend_args[["seg.len"]] = legend_args[["seg.len"]] %||% 1.25
  }

  if (identical(type, "n") && isFALSE(gradient)) {
    legend_args[["pch"]] = legend_args[["pch"]] %||% par("pch")
  }

  # Special pt.bg handling for types that need color-based fills
  if (identical(type, "spineplot")) {
    # The swatch fill comes via `bg` (resolved in prepare_legend for the grouped
    # `y_by` case; NULL otherwise, falling back to the group colour).
    legend_args[["pt.bg"]] = legend_args[["pt.bg"]] %||% bg %||% legend_args[["col"]]
    # Conversely, the border colour is always black
    legend_args[["col"]] = par("fg")
  } else if (identical(type, "ridge") && isFALSE(gradient)) {
    legend_args[["pt.bg"]] = legend_args[["pt.bg"]] %||% sapply(legend_args[["col"]], function(ccol) seq_palette(ccol, n = 2)[2])
  } else {
    legend_args[["pt.bg"]] = legend_args[["pt.bg"]] %||% bg
  }

  # Set legend labels
  legend_args[["legend"]] = legend_args[["legend"]] %||% lgnd_labs
  if (length(lgnd_labs) != length(eval(legend_args[["legend"]]))) {
    warning(
      "\nUser-supplied legend labels do not match the number of groups.\n",
      "Defaulting to automatic labels determined by the group splits in `by`,\n"
    )
    legend_args[["legend"]] = lgnd_labs
  }

  # Apply label formatter if provided
  if (!is.null(legend_args[["labeller"]])) {
    labeller = legend_args[["labeller"]]
    legend_args[["labeller"]] = NULL
    legend_args[["legend"]] = tinylabel(legend_args[["legend"]], labeller = labeller)
  }

  if (isTRUE(gradient)) {
    legend_args[["ncol"]] = NULL
  }

  # Determine positioning flags for anchor adjustment
  outer_side = outer_end = outer_right = outer_bottom = FALSE
  if (grepl("right!$|left!$", legend_args[["x"]])) {
    outer_side = TRUE
    outer_right = grepl("right!$", legend_args[["x"]])
  } else if (grepl("bottom!$|top!$", legend_args[["x"]])) {
    outer_end = TRUE
    outer_bottom = grepl("bottom!$", legend_args[["x"]])
  }

  # Adjust position anchor (we'll position relative to opposite side)
  if (outer_end) {
    if (outer_bottom) {
      legend_args[["x"]] = gsub("bottom!$", "top", legend_args[["x"]])
    } else {
      legend_args[["x"]] = gsub("top!$", "bottom", legend_args[["x"]])
    }
  } else if (outer_side) {
    if (outer_right) {
      legend_args[["x"]] = gsub("right!$", "left", legend_args[["x"]])
    } else {
      legend_args[["x"]] = gsub("left!$", "right", legend_args[["x"]])
    }
  }

  # Additional positioning adjustments
  if (outer_end) {
    # Enforce horizontal legend if user hasn't specified ncol arg
    if (is.null(legend_args[["ncol"]]) || gradient) legend_args[["horiz"]] = TRUE
  } else if (!outer_side) {
    legend_args[["inset"]] = 0
  }

  # Additional tweaks for horizontal and/or multi-column legends
  mcol_flag = !is.null(legend_args[["ncol"]]) && legend_args[["ncol"]] > 1
  user_inset = !is.null(legend_args[["inset"]])

  if (isTRUE(legend_args[["horiz"]]) || mcol_flag) {
    # Tighter horizontal labelling
    if (!gradient) {
      legend_args[["text.width"]] = NA
      # Add a space to all labs except the outermost right ones
      nlabs = length(legend_args[["legend"]])
      nidx = nlabs
      if (mcol_flag) nidx = tail(1:nlabs, (nlabs %/% legend_args[["ncol"]]))
      legend_args[["legend"]][-nidx] = paste(legend_args[["legend"]][-nidx], " ")
    }
    # Catch for horizontal ribbon legend spacing
    if (type == "ribbon") {
      if (legend_args[["pt.lwd"]] == 1) {
        legend_args[["x.intersp"]] = 1
      } else {
        legend_args[["x.intersp"]] = 0.5
      }
    } else if (gradient) {
      legend_args[["x.intersp"]] = 0.5
    }
  }

  # Populate legend environment with args and flags
  legend_env$args = legend_args
  legend_env$mcol = mcol_flag
  legend_env$user_inset = user_inset
  legend_env$outer_side = outer_side
  legend_env$outer_end = outer_end
  legend_env$outer_right = outer_right
  legend_env$outer_bottom = outer_bottom
}


#' Build legend environment
#'
#' @description Creates the legend environment by:
#'   - Initializing environment with metadata
#'   - Calling build_legend_args() to construct legend arguments
#'   - Populating environment with arguments and positioning flags
#'   - Initializing margins and dimensions
#'
#' @param legend Legend placement keyword or list
#' @param legend_args Additional legend arguments
#' @param by_dep The (deparsed) "by" grouping variable name
#' @param lgnd_labs The legend labels
#' @param labeller Character or function for formatting labels
#' @param type Plot type
#' @param pch Plotting character(s)
#' @param lty Line type(s)
#' @param lwd Line width(s)
#' @param col Color(s)
#' @param bg Background fill color(s)
#' @param cex Character expansion(s)
#' @param gradient Logical indicating gradient legend
#' @param lmar Legend margins (inner, outer)
#' @param has_sub Logical indicating presence of sub-caption
#' @param new_plot Logical indicating if plot.new should be called
#'
#' @returns Environment with complete legend specification
#'
#' @keywords internal
build_legend_env = function(
  # Legend specification
  legend,
  legend_args,

  # Labels and grouping
  by_dep,
  lgnd_labs,
  labeller = NULL,

  # Visual aesthetics
  type,
  pch,
  lty,
  lwd,
  col,
  bg,
  cex,

  # Configuration
  gradient,
  lmar,
  has_sub = FALSE,
  has_cap = FALSE,
  cap_text = NULL,
  new_plot = TRUE,
  dynmar_title_mar = NULL
) {
  # Create legend environment
  legend_env = new.env(parent = emptyenv())

  # Initialize metadata
  legend_env$gradient = gradient
  legend_env$type = type
  legend_env$has_sub = has_sub
  legend_env$has_cap = has_cap
  legend_env$cap_text = cap_text
  legend_env$new_plot = new_plot
  legend_env$dynmar = isTRUE(.tpar[["dynmar"]])
  legend_env$dynmar_title_mar = dynmar_title_mar
  legend_env$topmar_epsilon = 0.1

  # Build legend arguments (modifies legend_env in-place)
  build_legend_args(
    legend_env = legend_env,
    legend = legend,
    legend_args = legend_args,
    by_dep = by_dep,
    lgnd_labs = lgnd_labs,
    labeller = labeller,
    type = type,
    pch = pch,
    lty = lty,
    lwd = lwd,
    col = col,
    bg = bg,
    cex = cex,
    gradient = gradient
  )

  # Initialize margins
  legend_env$omar = par("mar")
  legend_env$ooma = par("oma")
  legend_env$lmar = lmar

  # Initialize dimensions and layout
  legend_env$dims = NULL
  legend_env$inset = NULL
  legend_env$rasterbox = NULL

  return(legend_env)
}


#
## Single Legend Rendering -----
#

#' Calculate placement and draw legend
#'
#' @description Main exported function for drawing legends. Supports:
#'   - Inner and outer positioning (with "!" suffix)
#'   - Discrete and continuous (gradient) legends
#'   - Automatic margin adjustment
#'
#' @md
#' @param legend Legend placement keyword or list, passed down from [tinyplot].
#' @param legend_args Additional legend arguments to be passed to
#'   \code{\link[graphics]{legend}}.
#' @param by_dep The (deparsed) "by" grouping variable name.
#' @param lgnd_labs The labels passed to `legend(legend = ...)`.
#' @param labeller Character or function for formatting the labels (`lgnd_labs`).
#'   Passed down to [`tinylabel`].
#' @param type Plotting type(s), passed down from [tinyplot].
#' @param pch Plotting character(s), passed down from [tinyplot].
#' @param lty Plotting linetype(s), passed down from [tinyplot].
#' @param lwd Plotting line width(s), passed down from [tinyplot].
#' @param col Plotting colour(s), passed down from [tinyplot].
#' @param bg Plotting character background fill colour(s), passed down from [tinyplot].
#' @param cex Plotting character expansion(s), passed down from [tinyplot].
#' @param gradient Logical indicating whether a continuous gradient swatch
#'   should be used to represent the colors.
#' @param lmar Legend margins (in lines). Should be a numeric vector of the form
#'   `c(inner, outer)`, where the first number represents the "inner" margin
#'   between the legend and the plot, and the second number represents the
#'   "outer" margin between the legend and edge of the graphics device. If no
#'   explicit value is provided by the user, then reverts back to `tpar("lmar")`
#'   for which the default values are `c(1.0, 0.1)`.
#' @param has_sub Logical. Does the plot have a sub-caption. Only used if
#'   keyword position is "bottom!", in which case we need to bump the legend
#'   margin a bit further.
#' @param has_cap Logical. Does the plot have a caption. Only used if
#'   keyword position is "bottom!", in which case we need to bump the legend
#'   margin a bit further.
#' @param cap_text Character. The caption text to draw below the legend when
#'   position is "bottom!". Ignored otherwise.
#' @param new_plot Logical. Should we be calling plot.new internally?
#' @param draw Logical. If `FALSE`, no legend is drawn but the sizes are
#'   returned. Note that a new (blank) plot frame will still need to be started
#'   in order to perform the calculations.
#' @param soma_target Numeric. Shared outer margin target (in lines) for
#'   multi-legend alignment. If `NULL`, each legend computes its own margin.
#' @param dynmar_title_mar Numeric or `NULL`. The pre-computed `dynmar_computed[3]`
#'   value for "top!" legends under dynmar themes. When set, the legend margin
#'   formula uses this directly to ensure correct title positioning.
#'
#' @returns No return value, called for side effect of producing a(n empty) plot
#'   with a legend in the margin.
#'
#' @importFrom graphics grconvertX grconvertY rasterImage strheight strwidth xinch
#' @importFrom grDevices as.raster recordGraphics
#' @importFrom utils modifyList
#'
#' @examples
#' oldmar = par("mar")
#'
#' draw_legend(
#'   legend = "right!", ## default (other options incl, "left(!)", ""bottom(!)", etc.)
#'   legend_args = list(title = "Key", bty = "o"),
#'   lgnd_labs = c("foo", "bar"),
#'   type = "p",
#'   pch = 21:22,
#'   col = 1:2
#' )
#'
#' # The legend is placed in the outer margin...
#' box("figure", col = "cyan", lty = 4)
#' # ... and the plot is proportionally adjusted against the edge of this
#' # margin.
#' box("plot")
#' # You can add regular plot objects per normal now
#' plot.window(xlim = c(1,10), ylim = c(1,10))
#' points(1:10)
#' points(10:1, pch = 22, col = "red")
#' axis(1); axis(2)
#' # etc.
#'
#' # Important: A side effect of draw_legend is that the inner margins have been
#' # adjusted. (Here: The right margin, since we called "right!" above.)
#' par("mar")
#'
#' # To reset you should call `dev.off()` or just reset manually.
#' par(mar = oldmar)
#'
#' # Note that the inner and outer margin of the legend itself can be set via
#' # the `lmar` argument. (This can also be set globally via
#' # `tpar(lmar = c(inner, outer))`.)
#' draw_legend(
#'   legend_args = list(title = "Key", bty = "o"),
#'   lgnd_labs = c("foo", "bar"),
#'   type = "p",
#'   pch = 21:22,
#'   col = 1:2,
#'   lmar = c(0, 0.1) ## set inner margin to zero
#' )
#' box("figure", col = "cyan", lty = 4)
#'
#' par(mar = oldmar)
#'
#' # Continuous (gradient) legends are also supported
#' draw_legend(
#'   legend = "right!",
#'   legend_args = list(title = "Key"),
#'   lgnd_labs = LETTERS[1:5],
#'   col = hcl.colors(5),
#'   gradient = TRUE ## enable gradient legend
#' )
#'
#' par(mar = oldmar)
#'
#' @export
draw_legend = function(
  legend = NULL,
  legend_args = NULL,
  by_dep = NULL,
  lgnd_labs = NULL,
  labeller = NULL,
  type = NULL,
  pch = NULL,
  lty = NULL,
  lwd = NULL,
  col = NULL,
  bg = NULL,
  cex = NULL,
  gradient = FALSE,
  lmar = NULL,
  has_sub = FALSE,
  has_cap = FALSE,
  cap_text = NULL,
  new_plot = TRUE,
  draw = TRUE,
  soma_target = NULL,
  dynmar_title_mar = NULL
) {
  if (is.null(lmar)) {
    lmar = tpar("lmar")
  } else {
    if (!is.numeric(lmar) || length(lmar) != 2) {
      stop("lmar must be a numeric of length 2.")
    }
  }

  assert_logical(gradient)
  assert_logical(has_sub)
  assert_logical(has_cap)
  assert_logical(new_plot)
  assert_logical(draw)

  # Restore margin defaults
  dynmar = isTRUE(.tpar[["dynmar"]])
  restore_margin_outer()
  if (!dynmar) {
    restore_margin_inner(par("oma"), topmar_epsilon = 0.1)
  }
  set_environment_variable(.top_legend_soma = NULL)

  # Build legend environment
  legend_env = build_legend_env(
    # Legend specification
    legend = legend,
    legend_args = legend_args,

    # Labels and grouping
    by_dep = by_dep,
    lgnd_labs = lgnd_labs,
    labeller = labeller,

    # Visual aesthetics
    type = type,
    pch = pch,
    lty = lty,
    lwd = lwd,
    col = col,
    bg = bg,
    cex = cex,

    # Configuration
    gradient = gradient,
    lmar = lmar,
    has_sub = has_sub,
    has_cap = has_cap,
    cap_text = cap_text,
    new_plot = new_plot,
    dynmar_title_mar = dynmar_title_mar
  )

  # Extract and strip ljust before any legend() calls
  ljust_mode = legend_env$args[["ljust"]] %||% tpar("ljust") %||% "left"
  ljust_mode = match.arg(ljust_mode, c("left", "center", "l", "c"))
  if (ljust_mode == "l") ljust_mode = "left"
  if (ljust_mode == "c") ljust_mode = "center"
  legend_env$args[["ljust"]] = NULL

  # Initial setup: adjust margins, call plot.new, and measure (but don't apply soma yet)
  legend_outer_margins(legend_env, apply = FALSE)

  # Legend justification (vertical, non-gradient, side-positioned only)
  if (!legend_env$gradient && !isTRUE(legend_env$args[["horiz"]]) && !legend_env$outer_end) {

    if (ljust_mode == "left") {
      ttl = legend_env$args[["title"]]
      # Compute title.adj to give 0.5*xch of left padding. Base R places
      # the title at: left + title.adj * (box_w - strwidth(title)), so
      # we solve for title.adj = 0.5*xch / slack.
      if (is.null(legend_env$args[["title.adj"]]) && !is.null(ttl)) {
        xch = par("cex") * xinch(par("cin")[1])
        slack = legend_env$dims$rect$w - strwidth(ttl)
        legend_env$args[["title.adj"]] = if (slack > 0) {
          min(0.5 * xch / slack, 0.5)
        } else {
          0
        }
      } else {
        legend_env$args[["title.adj"]] = legend_env$args[["title.adj"]] %||% 0
      }
      if (is.null(legend_env$args[["text.width"]])) {
        ttl = legend_env$args[["title"]]
        if (!is.null(ttl)) {
          lab_tw = max(strwidth(legend_env$args[["legend"]]))
          ttl_tw = strwidth(ttl)
          if (ttl_tw > lab_tw) {
            legend_env$ljust = TRUE
          }
        }
      }
    } else {
      legend_env$args[["title.adj"]] = legend_env$args[["title.adj"]] %||% 0.5
    }
  }

  if (!draw) {
    return(legend_env$dims)
  }

  legend_env$soma_target = soma_target

  # Store base values AFTER legend_outer_margins setup (before soma/inset are applied)
  # These are needed so tinylegend() can reset to them on each recordGraphics replay
  legend_env$omar_base = legend_env$omar
  legend_env$ooma_base = legend_env$ooma
  legend_env$inset_base = legend_env$args[["inset"]]

  # Wrap margin application, inset calculation, and drawing in recordGraphics()
  # so everything recalculates correctly when the plot window is resized
  recordGraphics(
    tinylegend(legend_env),
    list = list(legend_env = legend_env),
    env = getNamespace("tinyplot")
  )
}

#' Line segments plot type
#'
#' @description Type function for plotting line segments.
#' 
#' @details Contrary to base \code{\link[graphics]{segments}}, line segments in
#' [tinyplot] must be specified using the `xmin`, `ymin`,`xmax`, and `ymax`
#' arguments. 
#' 
#' @examples
#' # "segments" type convenience character string
#' tinyplot(
#'   xmin = c(0,.1), ymin = c(.2,1), xmax = c(1,.9), ymax = c(.75,0),
#'   type = "segments"
#' )
#' 
#' # Same result with type_segments()
#' tinyplot(
#'   xmin = c(0,.1), ymin = c(.2,1), xmax = c(1,.9), ymax = c(.75,0),
#'   type = type_segments()
#' )
#' 
#' @export
type_segments = function() {
  out = list(
    draw = draw_segments(),
    data = NULL,
    name = "segments"
  )
  class(out) = "tinyplot_type"
  return(out)
}


draw_segments = function() {
    fun = function(ixmin, iymin, ixmax, iymax, ilty, ilwd, icol, ...) {
        segments(
            x0 = ixmin, y0 = iymin, x1 = ixmax, y1 = iymax,
            lty = ilty,
            lwd = ilwd,
            col = icol
        )
    }
    return(fun)
}

#' Violin plot type
#' 
#' @md
#' @description Type function for violin plots, which are an alternative to box
#'   plots for visualizing continuous distributions (by group) in the form of
#'   mirrored densities.
#' @inheritParams type_density
#' @param  trim logical indicating whether the violins should be trimmed to the
#'   range of the data. Default is `FALSE`.
#' @param width numeric (ideally in the range `[0, 1]`, although this isn't
#'   enforced) giving the normalized width of the individual violins.
#' @param lighten logical. Should the fills use a lighter, opaque tint of the
#'   series colour(s)? Default is `TRUE`, which keeps single- and multi-group
#'   displays consistent and lets the fill read cleanly over grid lines. Set to
#'   `FALSE` to use the fully-saturated palette colour(s) instead.
#' @inherit stats::density details
#' @details See [`type_density`] for more details and considerations related to
#'   bandwidth selection and kernel types.
#'
#'   
#' @examples
#' # "violin" type convenience string
#' tinyplot(weight ~ feed, data = chickwts, type = "violin")
#' 
#' # to match the defaults of `ggplot2::geom_violin()`, use `trim = TRUE` and
#' # `joint.bw = FALSE`
#' tinyplot(
#'   weight ~ feed, data = chickwts,
#'   # type = type_violin(trim = TRUE, joint.bw = FALSE) # same but see final ex.
#'   type = "violin", trim = TRUE, joint.bw = FALSE
#' )
#' 
#' # For flipped violin plots, it's usually better to use a dynamic theme to
#' # accommodate (horizontal) y-axis labels
#' tinyplot(
#'   weight ~ feed, data = chickwts, type = "violin", flip = TRUE,
#'   theme = "dynamic" # or "clean(2)", "classic", "minimal", etc.
#' )
#' 
#' # you can group by the x var to add colour (here with the original orientation)
#' tinyplot(weight ~ feed | feed, data = chickwts, type = "violin", legend = FALSE)
#' 
#' # dodged grouped violin plot example (different dataset)
#' tinyplot(len ~ dose | supp, data = ToothGrowth, type = "violin")
#' 
#' # note: above we relied on `...` argument passing alongside the "violin"
#' # type convenience string. But this won't work for `width`, since it will
#' # clash with the top-level `tinyplot(..., width = <width>)` arg. To ensure
#' # correct arg passing, it's safer to use the functional `type_violin()` type.
#' tinyplot(
#'   len ~ dose | supp, data = ToothGrowth,
#'   type = type_violin(width = 0.75)
#' )
#' 
#' @importFrom stats density weighted.mean
#' @importFrom stats bw.SJ bw.bcv bw.nrd bw.nrd0 bw.ucv 
#' @export
type_violin = function(
        bw = "nrd0",
        joint.bw =  c("mean", "full", "none"),
        adjust = 1,
        kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"),
        n = 512,
        # more args from density here?
        trim = FALSE,
        width = 0.9,
        lighten = TRUE
    ) {
    kernel = match.arg(kernel, c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight", "cosine", "optcosine"))
    if (is.logical(joint.bw)) {
        joint.bw = ifelse(joint.bw, "mean", "none")
    }
    joint.bw = match.arg(joint.bw, c("mean", "full", "none"))
    out = list(
        data = data_violin(bw = bw, adjust = adjust, kernel = kernel, n = n,
                            joint.bw = joint.bw, trim = trim, width = width,
                            lighten = lighten),
        # draw = NULL,
        # name = "polygon"
        draw = draw_polygon(density = NULL),
        name = "violin"
    )
    class(out) = "tinyplot_type"
    return(out)
}

data_violin = function(bw = "nrd0", adjust = 1, kernel = "gaussian", n = 512,
                        joint.bw = "none", trim = FALSE, width = 0.9,
                        lighten = TRUE) {
    fun = function(settings, ...) {
        env2env(settings, environment(), c("datapoints", "by", "null_palette", "facet", "ylab", "col", "bg", "log", "null_by", "null_facet"))
        settings[["lighten"]] = lighten

        
        # Handle ordering based on by and facet variables
        ngrps = if (null_by) 1 else length(unique(datapoints$by))
        nfacets = if (null_facet) 1 else length(unique(datapoints$facet))
        
        #  catch for special cases
        x_by = y_by = facet_by = FALSE
        if (!null_by) {
            x_by = identical(datapoints$x, datapoints$by)
            y_by = identical(datapoints$y, datapoints$by)
            if (!null_facet) facet_by = identical(datapoints$facet, datapoints$by)
        }
        
        # FIXME (once we add support for gradient fill to draw_polygon)
        if (y_by) {
            warning("\n`y` == `by` is not currently supported for `type_violin`. We hope to support this in a future release, but for now `y` grouping will be turned off automatically.\n")
            by = NULL
            datapoints$by = ""
            ngrps = 1
            null_by = TRUE
        }
        
        # Convert x to factor if it's not already
        datapoints$x = as.factor(datapoints$x)
        if (x_by) datapoints$by = datapoints$x

        # Handle factor levels and maintain order
        xlvls = levels(datapoints$x)
        xlabs = seq_along(xlvls)
        names(xlabs) = xlvls
        # xlabs = levels(datapoints$x)
        datapoints$x = as.integer(datapoints$x)

        if (null_by && null_facet) {
            xord = order(datapoints$x)
        } else if (null_facet) {
            xord = order(datapoints$by, datapoints$x)
        } else if (null_by) {
            xord = order(datapoints$facet, datapoints$x)
        } else {
            xord = order(datapoints$by, datapoints$facet, datapoints$x)
        }

        if (length(unique(datapoints[["by"]])) == 1 && null_palette) {
            # With a theme palette active, leave bg = NULL so the fill tracks
            # the resolved border colour (see by_bg). Otherwise use neutral grey.
            if (is.null(bg) && is.null(get_tpar("palette.qualitative", default = NULL))) bg = "lightgray"
        } else if (is.null(bg)) {
            bg = "by"
        }

        # Reorder x, y, ymin, and ymax based on the order determined
        datapoints = datapoints[xord,]

        
        datapoints = split(datapoints, list(datapoints$x, datapoints$by, datapoints$facet))
        datapoints = Filter(function(k) nrow(k) > 0, datapoints)
        
        if (joint.bw == "none" || is.numeric(bw)) {
            dens_bw = bw
        } else {
            if (joint.bw == "mean") {
                # Use weighted mean of subgroup bandwidths
                bws = sapply(datapoints, function(dat) bw_fun(kernel = bw, dat$y))
                ws = sapply(datapoints, nrow)
                dens_bw = weighted.mean(bws, ws)
            } else if (joint.bw == "full") {
                dens_bw = bw_fun(kernel = bw, unlist(sapply(datapoints, `[[`, "x")))
            }
        }
        
        # Compute group offsets for multi-group violins
        if (ngrps > 1 && isFALSE(x_by) && isFALSE(facet_by)) {
            xwidth_grp = width / ngrps - 0.01
            group_offsets = seq(
                -((width - xwidth_grp) / 2),
                ((width - xwidth_grp) / 2),
                length.out = ngrps
            )
        } else {
            group_offsets = rep(0, max(ngrps, 1))
        }
        offsets_axis = "x"

        datapoints = lapply(seq_along(datapoints), function(d) {
            dat = datapoints[[d]]
            if (trim) {
                yrng = range(dat$y)
                dens = density(dat$y, bw = dens_bw, kernel = kernel, n = n, from = yrng[1], to = yrng[2])
            } else {
                dens = density(dat$y, bw = dens_bw, kernel = kernel, n = n)
            }
            
            x = dens$y
            y = dens$x
            
            
            if (log %in% c("y", "xy")) {
                if (x[1] <= 0) {
                    warning("\nNon-positive density values have been trimmed as part of the logarthmic transformation.\n")
                    xidx = x > 0
                    x = x[xidx]
                    y = y[xidx]
                } 
            }
            
            x = c(x, rev(-x))
            y = c(y, rev(y))
            
            xwidth = xwidth_orig = width
            # dodge groups (if any)
            if ((ngrps > 1)  && isFALSE(x_by) && isFALSE(facet_by)) {
                xwidth = xwidth_orig / ngrps - 0.01
                x = rescale_num(x, to = c(0, xwidth))
                x = x + as.numeric(sub("^([0-9]+)\\..*", "\\1", names(datapoints)[d])) - xwidth/2
                x = x + group_offsets[dat$by[1]]
            } else if (nfacets > 1) {
                x = rescale_num(x, to = c(0, xwidth))
                x = x + as.numeric(sub("^([0-9]+)\\..*", "\\1", names(datapoints)[d])) - xwidth/2
            } else {
                x = rescale_num(x, to = c(0, xwidth))
                x = x + d - xwidth/2
            }
            
            x = c(x, NA)
            y = c(y, NA)
            
            out = data.frame(
                by = dat$by[1], # already split
                facet = dat$facet[1], # already split
                y = y,
                x = x
            )
            return(out)
        })
        datapoints = do.call(rbind, datapoints)
        datapoints = datapoints[1:(nrow(datapoints)-1), ]

        by = if (length(unique(datapoints$by)) == 1) by else datapoints$by
        facet = if (length(unique(datapoints$facet)) == 1) facet else datapoints$facet
        
        # legend customizations
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, c(
            "datapoints",
            "by",
            "facet",
            "ylab",
            "xlabs",
            "col",
            "bg",
            "group_offsets",
            "offsets_axis"
        ))
    }
    return(fun)
}

#' @title Set or query graphical parameters
#'
#' @description Extends \code{\link[graphics]{par}}, serving as a (near) drop-in
#'   replacement for setting or querying graphical parameters. The key
#'   differences is that, beyond supporting the standard group of R graphical
#'   parameters in \code{\link[graphics]{par}}, `tpar` also supports additional
#'   graphical parameters that are provided by `tinyplot`. Similar to
#'   \code{\link[graphics]{par}}, parameters are set by passing appropriate
#'   `key = value` argument pairs, and multiple parameters can be set or queried
#'   at the same time.
#'
#' @param ... arguments of the form `key = value`. This includes all of the
#'   parameters typically supported by \code{\link[graphics]{par}}, as well as
#'   the `tinyplot`-specific ones described in the 'Graphical Parameters'
#'   section below.
#' @param hook Logical. If `TRUE`, base graphical parameters persist across
#'   plots via a hook applied before each new plot (see `?setHook`).
#'
#' @md
#' @details The `tinyplot`-specific parameters are saved in an internal
#'   environment called `.tpar` for performance and safety reasons. However,
#'   they can also be set at package load time via \code{\link[base]{options}},
#'   which may prove convenient for users that want to enable different default
#'   behaviour at startup (e.g., through an `.Rprofile` file). These options all
#'   take a `tinyplot_*` prefix, e.g.
#'   `options(tinyplot_grid = TRUE, tinyplot_facet.bg = "grey90")`.
#'
#' For their part, any "base" graphical parameters are caught dynamically and
#'   passed on to \code{\link[graphics]{par}} as appropriate. Technically, only
#'   parameters that satisfy `par(..., no.readonly = TRUE)` are evaluated.
#'
#' However, note the important distinction: `tpar` only evaluates parameters
#'   from \code{\link[graphics]{par}} if they are passed _explicitly_ by the
#'   user. This means that `tpar` should not be used to capture the (invisible)
#'   state of a user's entire set of graphics parameters, i.e. `tpar()` !=
#'   `par()`. If you want to capture the _all_ existing graphics settings, then
#'   you should rather use `par()` instead.
#'
#' @returns When parameters are set, their previous values are returned in an
#'   invisible named list. Such a list can be passed as an argument to `tpar` to
#'   restore the parameter values.
#'
#'   When just one parameter is queried, the value of that parameter is returned
#'   as (atomic) vector. When two or more parameters are queried, their values
#'   are returned in a list, with the list names giving the parameters.
#'
#'   Note the inconsistency: setting one parameter returns a list, but querying
#'   one parameter returns a vector.
#'
#' @section Additional Graphical Parameters:
#'
#' * `adj.cap`: Numeric value between 0 and 1 controlling the alignment of the plot caption. Defaults to `0.5` (centered) for the default theme, and `1` (right-aligned) for all other themes.
#' * `adj.xlab`: Numeric value between 0 and 1 controlling the alignment of the x-axis label.
#' * `adj.ylab`: Numeric value between 0 and 1 controlling the alignment of the y-axis label.
#' * `cairo`: Logical indicating whether \code{\link[grDevices]{cairo_pdf}} should be used when writing plots to PDF. If `FALSE`, then \code{\link[grDevices]{pdf}} will be used instead, with implications for embedding (non-standard) fonts. Only used if `tinyplot(..., file = "<filename>.pdf")` is called. Defaults to the value of `capabilities("cairo")`.
#' * `cex.cap`: Numeric expansion factor for the plot caption text. Defaults to `1` for the default, basic, and dynamic themes, and `0.8` for clean/classic and their descendants.
#' * `col.cap`: Character specifying the colour of the plot caption. Defaults to `"black"`.
#' * `col.default`: Default colour for single-group displays (i.e. plots without a `by` grouping). Can be `NULL`, a length-1 character colour, or a length-1 numeric index into `palette.qualitative`. Defaults to `NULL`, in which case the first colour of the active qualitative palette is used (or base `palette()[1]`, typically black, if no theme is set). A character value sets the single-group colour independently of the multi-group palette. A numeric value `i` selects the `i`th colour of `palette.qualitative` as the single-group default; a *negative* index additionally drops that colour from the palette used for grouped plots. For example, `col.default = -1` paired with `palette.qualitative = "Okabe-Ito"` uses black (the leading palette colour) for single-group plots and an Okabe-Ito-minus-black palette for grouped plots, avoiding the need to maintain a separate "no-black" palette.
#' * `font.cap`: Integer specifying the font face for the plot caption (`1` = plain, `2` = bold, `3` = italic, `4` = bold italic). Defaults to `1`.
#' * `line.cap`: Numeric specifying the margin line on which to draw the caption. If `NULL` (default), computed automatically based on the available bottom margin.
#' * `dynmar`: Logical indicating whether `tinyplot` should attempt dynamic adjustment of margins to reduce whitespace and/or account for spacing of text elements (e.g., long horizontal y-axis labels). Note that this parameter is tightly coupled to internal `tinythemes()` logic and should _not_ be adjusted manually unless you really know what you are doing or don't mind risking unintended consequences to your plot.
#' * `facet.bg`: Character or integer specifying the facet background colour. If an integer, will correspond to the user's default colour palette (see `palette`). Passed to `rect`. Defaults to `NULL` (none).
#' * `facet.border`: Character or integer specifying the facet border colour. If an integer, will correspond to the user's default colour palette (see `palette`). Passed to `rect`. Defaults to `NA` (none).
#' * `facet.cex`: Expansion factor for facet titles. Defaults to `1`.
#' * `facet.col`: Character or integer specifying the facet text colour. If an integer, will correspond to the user's default global colour palette (see `palette`). Defaults to `NULL`, which is equivalent to "black".
#' * `facet.font`: An integer corresponding to the desired font face for facet titles. For most font families and graphics devices, one of four possible values: `1` (regular), `2` (bold), `3` (italic), or `4` (bold italic). Defaults to `NULL`, which is equivalent to `1` (i.e., regular).
#' * `file.height`: Numeric specifying the height (in inches) of any plot that is written to disk using the `tinyplot(..., file = X)` argument. Defaults to `7`.
#' * `file.res`: Numeric specifying the resolution (in dots per square inch) of any plot that is written to disk in bitmap format (i.e., PNG or JPEG) using the `tinyplot(..., file = X)` argument. Defaults to `300`.
#' * `file.width`: Numeric specifying the width (in inches) of any plot that is written to disk using the `tinyplot(..., file = X)` argument. Defaults to `7`.
#' * `fmar`: A numeric vector of form `c(b,l,t,r)` for controlling the (base) margin padding, in terms of lines, between the individual facets in a faceted plot. Defaults to `c(1,1,1,1)`. If more than three facets are detected, the `fmar` parameter is scaled by 0.75 to reduce excess whitespace. For 2x2 plots, the padding better matches the `cex` expansion logic of base graphics.
#' * `gap.main`: Numeric giving the gap (in margin lines) from the main title baseline to the element below it (plot box when no subtitle; subtitle top when subtitle is present). Only used when `dynmar = TRUE`. Defaults to `0.7`.
#' * `gap.sub`: Numeric giving the gap (in margin lines) from the subtitle baseline to the plot box. Only used when `dynmar = TRUE` and `side.sub = 3`. Defaults to `0.7`.
#' * `grid.col`: Character or (integer) numeric that specifies the color of the panel grid lines. Defaults to `"lightgray"`.
#' * `grid.lty`: Character or (integer) numeric that specifies the line type of the panel grid lines. Defaults to `"dotted"`.
#' * `grid.lwd`: Non-negative numeric giving the line width of the panel grid lines. Defaults to `1`.
#' * `grid`: Logical or character indicating whether a background panel grid should be added to plots automatically. Defaults to `NULL`, which is equivalent to `FALSE`. In addition to logical values, a character string can be used to control axis-specific grids: uppercase letters (`"X"`, `"Y"`, `"XY"`) draw grid lines at the standard axis tick positions (equivalent to `TRUE`), while lowercase letters (`"x"`, `"y"`, `"xy"`) draw a finer grid with additional lines at the midpoints between ticks.
#' * `lmar`: A numeric vector of form `c(inner, outer)` that gives the margin padding, in terms of lines, around the automatic `tinyplot` legend. Defaults to `c(1.0, 0.1)`. The inner margin is the gap between the legend and the plot region, and the outer margin is the gap between the legend and the edge of the graphics device.
#' * `palette.qualitative`: Palette for qualitative colors. See the `palette` argument in `?tinyplot`.
#' * `palette.sequential`: Palette for sequential colors. See the `palette` argument in `?tinyplot`.
#' * `ribbon.alpha`: Numeric factor in the range `[0,1]` for modifying the opacity alpha of "ribbon" and "area" type plots. Default value is `0.2`.
#'
#' @importFrom graphics par
#' @importFrom utils modifyList
#'
#' @seealso [`graphics::par`] which `tpar` builds on top of. [`get_saved_par`]
#' is a convenience function for retrieving graphical parameters at different
#' stages of a `tinyplot` call (and used for internal accounting purposes).
#' [`tinytheme`] allows users to easily set a group of graphics parameters
#' in a single function call, according to a variety of predefined themes.
#'
#' @examples
#' # Return a list of existing base and tinyplot graphic params
#' tpar("las", "pch", "facet.bg", "facet.cex", "grid")
#'
#' # Simple facet plot with these default values
#' tinyplot(mpg ~ wt, data = mtcars, facet = ~am)
#'
#' # Set params to something new. Similar to graphics::par(), note that we save
#' # the existing values at the same time by assigning to an object.
#' op = tpar(
#'   las       = 1,
#'   pch       = 2,
#'   facet.bg  = "grey90",
#'   facet.cex = 2,
#'   grid      = TRUE
#' )
#'
#' # Re-plot with these new params
#' tinyplot(mpg ~ wt, data = mtcars, facet = ~am)
#'
#' # Reset back to original values
#' tpar(op)
#'
#' # Important: tpar() only evaluates parameters that have been passed explicitly
#' #   by the user. So it it should not be used to query and set (restore)
#' #   parameters that weren't explicitly requested, i.e. tpar() != par().
#'
#' # Note: The tinyplot-specific parameters can also be be set via `options`
#' #   with a `tinyplot_*` prefix, which can be convenient for enabling
#' #   different default behaviour at startup time (e.g., via an .Rprofile
#' #   file). Example:
#' # options(tinyplot_grid = TRUE, tinyplot_facet.bg = "grey90")
#'
#' @export
tpar = function(..., hook = FALSE) {

  opts = list(...)
  if (length(opts) == 1 && is.null(names(opts))) {
    if (inherits(opts[[1]], "list") && !is.null(names(opts[[1]]))) {
      opts = opts[[1]]
    }
  }

  ###### Assign parameters

  # assign tinyplot-specific arguments with known names to .tpar
  assign_tpar(opts)

  # return informative error messages if the input is invalid
  assert_tpar(.tpar)

  # if tpar(...) includes arguments that are not known to be tinyplot-specific,
  # we set a hook to set them using par() when the graphic device is started
  nam = names(opts)
  if (!is.null(nam)) {
    base_par = setdiff(nam, known_tpar)
    base_par = opts[base_par]
    if (length(base_par) > 0) {
      if (isTRUE(hook)) {
        # append new hook to existing ones
        new_hooks = list("before.plot.new" = function() par(base_par))
        set_hooks(new_hooks, action = "append")
        # save new hook to tinyplot environment for later removal
        old_hooks = get_environment_variable(".tpar_hooks")
        set_environment_variable(.tpar_hooks = c(old_hooks, new_hooks))
      } else {
        par_names = names(par(no.readonly = TRUE))
        base_par = base_par[names(base_par) %in% par_names]
        base_par_old = par(base_par)
        par(base_par)
      }
    }
  }


  ###### Retrieve parameters

  # User didn't assign any new values, but may have requested explicit (print
  # of) some existing value(s)
  tpar_old = as.list(.tpar)
  if (is.null(nam)) {
    known_par = names(par(no.readonly = TRUE))
    if (!is.null(nam)) {
      used_par = intersect(nam, known_par)
    } else {
      used_par = intersect(opts, known_par)
    }
    if (length(used_par)) {
      if (!is.null(nam)) used_par = opts[used_par]
      used_par_old = par(used_par)
      tpar_old = modifyList(as.list(.tpar), used_par_old, keep.null = TRUE)
    }
    if (!is.null(opts) && length(opts) != 0) {
      # specific values requested
      opts = Filter(is.character, opts)
      ret = (`names<-`(lapply(opts, function(x) .tpar[[x]]), opts))
      if (length(used_par)) {
        ret_par = par(used_par)
        ret = modifyList(ret, ret_par, keep.null = TRUE)
      }
      if (length(ret) == 1) ret = ret[[1]]
      return(ret)
    } else {
      # no specific request; return all existing values invisibly
      return(invisible(tpar_old))
    }
    # assign new values, but still return old values for saving existing settings
    # a la `oldpar = par(param = new_value)`
  } else {
    `names<-`(lapply(nam, function(x) .tpar[[x]]), nam)
    if (length(base_par) > 0 && isFALSE(hook)) tpar_old = modifyList(tpar_old, base_par_old, keep.null = TRUE)
    return(invisible(tpar_old))
  }
}


# Two levels of priority: .tpar[["name"]] -> par("name")
get_tpar = function(opts, default = NULL, tpar_list = NULL) {
  if (is.null(tpar_list)) tpar_list = .tpar
  # parameter priority
  # .tpar[["name"]] -> par("name")
  for (o in opts) {
    tp = tpar_list[[o]]
    if (!is.null(tp)) {
      return(tp)
    } else {
      p = suppressWarnings(par(o))
      if (!is.null(p)) {
        return(p)
      }
    }

  }
  return(default)
}


known_tpar = c(
    "adj.cap",
    "adj.main",
    "adj.sub",
    "adj.xlab",
    "adj.ylab",
    "cex.cap",
    "cex.xlab",
    "cex.ylab",
    "col.cap",
    "col.default",
    "col.xaxs",
    "col.yaxs",
    "cairo",
    "dynmar",
    "font.cap",
    "gap.axis",
    "gap.lab",
    "gap.main",
    "gap.sub",
    "facet.bg",
    "facet.border",
    "facet.cex",
    "facet.col",
    "facet.font",
    "file.height",
    "file.res",
    "file.width",
    "fmar",
    "grid",
    "grid.bg",
    "grid.col",
    "grid.lty",
    "grid.lwd",
    "line.cap",
    "ljust",
    "lmar",
    "lty.xaxs",
    "lty.yaxs",
    "lwd.xaxs",
    "lwd.yaxs",
    "lwd.axis",
    "pch",
    "palette.qualitative",
    "palette.sequential",
    "ribbon.alpha",
    "side.sub",
    "tinytheme",
    "xaxt",
    "yaxt"
)


assign_tpar = function(opts) {
  for (n in intersect(names(opts), known_tpar)) {
    .tpar[[n]] = opts[[n]]
  }
}


assert_tpar = function(.tpar) {
  assert_numeric(.tpar[["adj.cap"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "adj.cap")
  assert_numeric(.tpar[["adj.main"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "adj.main")
  assert_numeric(.tpar[["adj.sub"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "adj.sub")
  assert_numeric(.tpar[["adj.xlab"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "adj.xlab")
  assert_numeric(.tpar[["adj.ylab"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "adj.ylab")
  assert_flag(.tpar[["cairo"]], name = "cairo")
  assert_flag(.tpar[["dynmar"]], null.ok = FALSE, name = "dynmar")
  assert_choice(.tpar[["ljust"]], choice = c("left", "center", "l", "c"), null.ok = TRUE, name = "ljust")
  assert_numeric(.tpar[["lmar"]], len = 2, null.ok = TRUE, name = "lmar")
  assert_numeric(.tpar[["ribbon.alpha"]], len = 1, lower = 0, upper = 1, null.ok = TRUE, name = "ribbon.alpha")
  assert_numeric(.tpar[["grid.lwd"]], len = 1, lower = 0, null.ok = TRUE, name = "grid.lwd")
  assert_grid(.tpar[["grid"]], null.ok = TRUE, name = "grid")
  assert_numeric(.tpar[["file.res"]], len = 1, lower = 0, null.ok = TRUE, name = "file.res")
  assert_numeric(.tpar[["file.height"]], len = 1, lower = 0, null.ok = TRUE, name = "file.height")
  assert_numeric(.tpar[["file.width"]], len = 1, lower = 0, null.ok = TRUE, name = "file.width")
  assert_numeric(.tpar[["facet.font"]], len = 1, null.ok = TRUE, name = "facet.font")
  assert_numeric(.tpar[["facet.cex"]], len = 1, null.ok = TRUE, name = "facet.cex")
  assert_numeric(.tpar[["side.sub"]], len = 1, null.ok = TRUE, name = "side.sub")
  assert_string(.tpar[["grid.bg"]], null.ok = TRUE, name = "grid.bg")
  assert_numeric(.tpar[["fmar"]], len = 4, null.ok = TRUE, name = "fmar")

  col.default = .tpar[["col.default"]]
  if (!is.null(col.default)) {
    if (!is.character(col.default) && !is.numeric(col.default)) {
      stop("col.default needs to be NULL, a (length-1) character colour, or a (length-1) numeric palette index", call. = FALSE)
    }
    assert_true(length(col.default) == 1, name = "length(col.default)==1")
  }

  facet.col = .tpar[["facet.col"]]
  if (!is.null(facet.col)) {
    if (!is.null(facet.col) && !is.numeric(facet.col) && !is.character(facet.col)) {
      stop("facet.col needs to be NULL, or a numeric or character", call. = FALSE)
    }
    assert_true(length(facet.col) == 1, name = "length(facet.col)==1")
  }

  facet.bg = .tpar$facet.bg
  if (!is.null(facet.bg)) {
    if (!is.numeric(facet.bg) && !is.character(facet.bg)) {
      stop("facet.bg needs to be NULL, or a numeric or character", call. = FALSE)
    }
    assert_true(length(facet.bg) == 1, name = "length(facet.bg)==1")
  }

  facet.border = .tpar$facet.border
  if (!is.null(facet.border)) {
    if (!is.numeric(facet.border) && !is.character(facet.border) && !is.na(facet.border)) {
      stop("facet.border needs to be NULL, or a numeric, character, or NA", call. = FALSE)
    }
    assert_true(length(facet.border) == 1, name = "length(facet.border)==1")
  }
}

init_tpar = function(rm_hook = FALSE) {
  rm(list = names(.tpar), envir = .tpar)

  if (isTRUE(rm_hook)) {
    old_hooks = get_environment_variable(".tpar_hooks")
    if (length(old_hooks) > 0) {
      remove_hooks(old_hooks)
      set_environment_variable(.tpar_hooks = NULL)
    }
  }

  .tpar$cairo = if (is.null(getOption("tinyplot_cairo"))) capabilities("cairo") else as.logical(getOption("tinyplot_cairo"))


  .tpar$dynmar = if (is.null(getOption("tinyplot_dynmar"))) FALSE else as.logical(getOption("tinyplot_dynmar"))

  # Figure output options if written to file
  .tpar$file.width = if (is.null(getOption("tinyplot_file.width"))) 7 else as.numeric(getOption("tinyplot_file.width"))
  .tpar$file.height = if (is.null(getOption("tinyplot_file.height"))) 7 else as.numeric(getOption("tinyplot_file.height"))
  .tpar$file.res = if (is.null(getOption("tinyplot_file.res"))) 300 else as.numeric(getOption("tinyplot_file.res"))

  # Facet margin, i.e. gap between the individual facet windows
  .tpar$fmar = if (is.null(getOption("tinyplot_fmar"))) c(1, 1, 1, 1) else as.numeric(getOption("tinyplot_fmar"))

  # Other facet options
  .tpar$facet.cex = if (is.null(getOption("tinyplot_facet.cex"))) 1 else as.numeric(getOption("tinyplot_facet.cex"))
  .tpar$facet.font = if (is.null(getOption("tinyplot_facet.font"))) NULL else as.numeric(getOption("tinyplot_facet.font"))
  .tpar$facet.col = if (is.null(getOption("tinyplot_facet.col"))) NULL else getOption("tinyplot_facet.col")
  .tpar$facet.bg = if (is.null(getOption("tinyplot_facet.bg"))) NULL else getOption("tinyplot_facet.bg")
  .tpar$facet.border = if (is.null(getOption("tinyplot_facet.border"))) NA else getOption("tinyplot_facet.border")

  # Plot grid
  .tpar$grid = if (is.null(getOption("tinyplot_grid"))) FALSE else as.logical(getOption("tinyplot_grid"))
  .tpar$grid.col = if (is.null(getOption("tinyplot_grid.col"))) "lightgray" else getOption("tinyplot_grid.col")
  .tpar$grid.lty = if (is.null(getOption("tinyplot_grid.lty"))) "dotted" else getOption("tinyplot_grid.lty")
  .tpar$grid.lwd = if (is.null(getOption("tinyplot_grid.lwd"))) 1 else as.numeric(getOption("tinyplot_grid.lwd"))

  # Default colour for single-group displays (NULL defers to the first
  # qualitative palette colour, or base palette()[1] if no theme is active)
  .tpar$col.default = if (is.null(getOption("tinyplot_col.default"))) NULL else getOption("tinyplot_col.default")

  # Legend justification
  .tpar$ljust = if (is.null(getOption("tinyplot_ljust"))) "left" else getOption("tinyplot_ljust")

  # Legend margin, i.e. gap between the legend and the plot elements
  .tpar$lmar = if (is.null(getOption("tinyplot_lmar"))) c(1.0, 0.1) else as.numeric(getOption("tinyplot_lmar"))

  # Alpha fill (transparency) default for ribbon and area plots
  .tpar$ribbon.alpha = if (is.null(getOption("tinyplot_ribbon.alpha"))) 0.2 else as.numeric(getOption("tinyplot_ribbon.alpha"))
}

## initialize internal environment for tpar variables
.tpar = new.env()
init_tpar()

# calculate limits of each plot

lim_args = function(settings) {
  env2env(
    settings,
    environment(),
    c(
      "xaxb", "xlabs", "xlim", "null_xlim", "rev_x",
      "yaxb", "ylabs", "ylim", "null_ylim", "rev_y",
      "datapoints", "type"
    )
  )

  # For cases where x/yaxb is provided and corresponding x/ylabs is not null...
  # We can subset these here to provide breaks
  if (!is.null(xaxb) && !is.null(xlabs)) {
    xlabs = xlabs[names(xlabs) %in% xaxb]
    xaxb = NULL # don't need this any more
  }
  if (!is.null(yaxb) && !is.null(ylabs)) {
    ylabs = ylabs[names(ylabs) %in% yaxb]
    yaxb = NULL # don't need this any more
  }

  if (is.null(xlim)) {
    xlim = range(as.numeric(c(
      datapoints[["x"]], datapoints[["xmin"]],
      datapoints[["xmax"]])), finite = TRUE)
  } else if (length(xlim) != 2L || anyNA(xlim)) {
    xdrng = range(as.numeric(c(
      datapoints[["x"]], datapoints[["xmin"]],
      datapoints[["xmax"]])), finite = TRUE)
    xlim = resolve_lim(xlim, xdrng, "xlim")
  }
  if (is.null(ylim)) {
    ylim = range(as.numeric(c(
      datapoints[["y"]], datapoints[["ymin"]],
      datapoints[["ymax"]])), finite = TRUE)
  } else if (length(ylim) != 2L || anyNA(ylim)) {
    ydrng = range(as.numeric(c(
      datapoints[["y"]], datapoints[["ymin"]],
      datapoints[["ymax"]])), finite = TRUE)
    ylim = resolve_lim(ylim, ydrng, "ylim")
  }

  if (identical(type, "boxplot")) {
    xlim = xlim + c(-0.5, 0.5)
  }

  if (null_xlim && !is.null(xaxb) && type != "spineplot") xlim = range(c(xlim, xaxb))
  if (null_ylim && !is.null(yaxb) && type != "spineplot") ylim = range(c(ylim, yaxb))

  # reverse axis direction last, once the range is otherwise finalized
  if (isTRUE(rev_x)) xlim = rev(xlim)
  if (isTRUE(rev_y)) ylim = rev(ylim)

  # update settings
  env2env(
    environment(),
    settings,
    c("xlim", "ylim", "xlabs", "ylabs", "xaxb", "yaxb")
  )
}


#
# x/ylim helpers ----
#

# Resolve a user-supplied x/ylim that may be a scalar or contains a single NA.
# `lim`  : raw user value (already known to be non-NULL)
# `drng` : data range, 2-element numeric, i.e. range(..., finite = TRUE)
# Returns a 2-element numeric vector (raw; window padding applied downstream).
resolve_lim = function(lim, drng, arg = "xlim") {
  if (!is.numeric(lim)) {
    stop(sprintf("`%s` must be numeric (a scalar or length-2 vector).", arg), call. = FALSE)
  }
  n = length(lim)
  if (n == 1L) {
    if (is.na(lim)) stop(sprintf("`%s` cannot be a single `NA`.", arg), call. = FALSE)
    # scalar: ensure the value is covered alongside the data
    return(range(c(drng, lim)))
  }
  if (n == 2L) {
    nas = is.na(lim)
    if (all(nas)) {
      stop(sprintf("`%s` cannot be `c(NA, NA)`; supply at least one finite limit.", arg), call. = FALSE)
    }
    if (!any(nas)) return(lim) # full override: unchanged (reversed axis OK)
    # exactly one NA: fill that side from the data range, pin the other verbatim
    if (nas[1L]) lim[1L] = drng[1L] else lim[2L] = drng[2L]
    return(lim)
  }
  stop(sprintf("`%s` must be length 1 or 2, not length %d.", arg, n), call. = FALSE)
}

# Resolve an axis-reversal keyword passed to x/ylim, e.g. xlim = "reverse" (or
# the "rev" abbreviation). Returns a list with the (possibly NULL-ified) limit
# and a logical flag. When a keyword is detected we strip the limit back to NULL
# so that all the usual auto-range machinery (data range, breaks, free facets,
# type-specific limits) runs untouched; the flag is consumed later to rev() the
# finalized range.
sanitize_lim_rev = function(lim, arg = "xlim") {
  if (is.character(lim)) {
    rev_ok = length(lim) == 1L && !is.na(lim) &&
      nchar(lim) >= 3L && pmatch(tolower(lim), "reverse", nomatch = 0L) == 1L
    if (!rev_ok) {
      stop(sprintf('If `%s` is a character string it must be "reverse" (or "rev").', arg), call. = FALSE)
    }
    return(list(lim = NULL, rev = TRUE))
  }
  list(lim = lim, rev = FALSE)
}

#' Draw gradient (continuous) legend swatch
#'
#' @description For gradient legends, we draw a custom color swatch using
#'   grDevices::as.raster and add labels, tick marks, and title manually.
#'
#' @param legend_args Legend arguments list
#' @param fklgnd Fake legend object (from drawing with plot=FALSE)
#' @param lmar Legend margins
#' @param outer_side Logical flag for outer side placement
#' @param outer_end Logical flag for outer end placement
#' @param outer_right Logical flag for outer right placement
#' @param outer_bottom Logical flag for outer bottom placement
#' @param user_inset Logical flag indicating user-supplied inset
#'
#' @returns NULL (draws gradient legend as side effect)
#'
#' @keywords internal
draw_gradient_swatch = function(
  legend_args,
  fklgnd,
  lmar,
  outer_side,
  outer_end,
  outer_right,
  outer_bottom,
  user_inset = FALSE
) {
  pal = legend_args[["col"]]
  lgnd_labs = legend_args[["legend"]]
  if (!is.null(legend_args[["horiz"]])) {
    horiz = legend_args[["horiz"]]
  } else {
    horiz = FALSE
  }

  # Create raster color swatch
  if (isTRUE(horiz)) {
    rasterlgd = as.raster(matrix(pal, nrow = 1))
  } else {
    rasterlgd = as.raster(matrix(rev(pal), ncol = 1))
  }

  corners = par("usr")
  rasterbox = rep(NA_real_, 4)

  # Determine positioning flags
  inner = !any(c(outer_side, outer_end))
  inner_right = inner_bottom = FALSE
  if (inner) {
    if (!is.null(legend_args[["x"]]) && grepl("left$|right$", legend_args[["x"]])) {
      inner_right = grepl("right$", legend_args[["x"]])
    }
    if (!is.null(legend_args[["x"]]) && grepl("^bottoml|^top", legend_args[["x"]])) {
      inner_bottom = grepl("^bottom", legend_args[["x"]])
    }
  }

  # Calculate raster box coordinates based on position
  if (inner) {
    fklgnd$rect$h = fklgnd$rect$h - lines_to_user_y(1.5 + 0.4)

    rasterbox[1] = fklgnd$rect$left
    if (isFALSE(inner_right)) {
      rasterbox[1] = rasterbox[1] + lines_to_user_x(0.2)
    }
    rasterbox[2] = fklgnd$rect$top - fklgnd$rect$h - lines_to_user_y(1.5 + 0.2)
    rasterbox[3] = rasterbox[1] + lines_to_user_x(1.25)
    rasterbox[4] = rasterbox[2] + fklgnd$rect$h

  } else if (outer_side) {
    rb1_adj = lines_to_user_x(lmar[1] + 0.2)
    rb3_adj = lines_to_user_x(1.25)
    rb2_adj = (corners[4] - corners[3] - lines_to_user_y(5 + 1 + 2.5)) / 2
    # Override if top or bottom
    if (!is.null(legend_args[["x"]])) {
      if (grepl("^bottom", legend_args[["x"]])) {
        rb2_adj = corners[3]
      }
      if (grepl("^top", legend_args[["x"]])) {
        rb2_adj = corners[4] - lines_to_user_y(5 + 1 + 2.5)
      }
    }
    if (user_inset) {
      rb2_adj = rb2_adj + legend_args[["inset"]][2] + 0.05
    }
    rb4_adj = lines_to_user_y(5 + 1)

    if (outer_right) {
      rasterbox[1] = corners[2] + rb1_adj
      if (user_inset) {
        rasterbox[1] = rasterbox[1] - (corners[2] - legend_args[["inset"]][1]) / 2
      }
      rasterbox[2] = rb2_adj
      rasterbox[3] = rasterbox[1] + rb3_adj
      rasterbox[4] = rasterbox[2] + rb4_adj
    } else {
      rb1_adj = rb1_adj + lines_to_user_x(par("mar")[2] + 1)
      rasterbox[1] = corners[1] - rb1_adj
      rasterbox[2] = rb2_adj
      rasterbox[3] = rasterbox[1] - rb3_adj
      rasterbox[4] = rasterbox[2] + rb4_adj
    }

  } else if (outer_end) {
    rb1_adj = (corners[2] - corners[1] - lines_to_user_x(5 + 1)) / 2
    rb3_adj = lines_to_user_x(5 + 1)
    rb2_adj = lines_to_user_y(lmar[1])
    rb4_adj = lines_to_user_y(1.25)

    if (outer_bottom) {
      rb2_adj = rb2_adj + lines_to_user_y(par("mar")[2])
      rasterbox[1] = rb1_adj
      rasterbox[2] = corners[3] - rb2_adj
      rasterbox[3] = rasterbox[1] + rb3_adj
      rasterbox[4] = rasterbox[2] - rb4_adj
    } else {
      rb2_adj = rb2_adj + lines_to_user_y(1.25 + 1)
      rasterbox[1] = rb1_adj
      rasterbox[2] = corners[4] + rb2_adj
      rasterbox[3] = rasterbox[1] + rb3_adj
      rasterbox[4] = rasterbox[2] - rb4_adj
    }
  }

  # Draw the gradient swatch
  rasterImage(
    rasterlgd,
    rasterbox[1], #x1
    rasterbox[2], #y1
    rasterbox[3], #x2
    rasterbox[4], #y2
    xpd = NA
  )

  # Add labels, tick marks, and title
  if (isFALSE(horiz)) {
    draw_gradient_labels_vertical(rasterbox, lgnd_labs, legend_args, inner, outer_right)
  } else {
    draw_gradient_labels_horizontal(rasterbox, lgnd_labs, legend_args)
  }
}


# Draw vertical gradient legend labels, ticks, and title
draw_gradient_labels_vertical = function(rasterbox, lgnd_labs, legend_args, inner, outer_right) {
  labs_idx = !is.na(lgnd_labs)
  lgnd_labs[labs_idx] = paste0(" ", format(lgnd_labs[labs_idx]))

  # Determine anchors based on position
  if (!inner && !outer_right) {
    lbl_x_anchor = rasterbox[1]
    ttl_x_anchor = rasterbox[1] + max(strwidth(lgnd_labs[labs_idx]))
    lbl_adj = c(0, 0.5)
    ttl_adj = c(1, 0)
  } else {
    lbl_x_anchor = rasterbox[3]
    ttl_x_anchor = rasterbox[1]
    lbl_adj = c(0, 0.5)
    ttl_adj = c(0, 0)
  }

  # Draw labels
  text(
    x = lbl_x_anchor,
    y = seq(rasterbox[2], rasterbox[4], length.out = length(lgnd_labs)),
    labels = lgnd_labs,
    xpd = NA,
    adj = lbl_adj
  )

  # Draw tick marks (white dashes)
  lgnd_ticks = lgnd_labs
  lgnd_ticks[labs_idx] = "-   -"
  text(
    x = lbl_x_anchor,
    y = seq(rasterbox[2], rasterbox[4], length.out = length(lgnd_labs)),
    labels = lgnd_ticks,
    cex = 1,
    col = "white",
    family = "sans",
    xpd = NA,
    adj = c(1, 0.5)
  )

  # Draw title
  text(
    x = ttl_x_anchor,
    y = rasterbox[4] + lines_to_user_y(1),
    labels = legend_args[["title"]],
    xpd = NA,
    adj = ttl_adj
  )
}

# Draw horizontal gradient legend labels, ticks, and title
draw_gradient_labels_horizontal = function(rasterbox, lgnd_labs, legend_args) {
  # Legend labels
  text(
    x = seq(rasterbox[1], rasterbox[3], length.out = length(lgnd_labs)),
    y = rasterbox[4],
    labels = lgnd_labs,
    xpd = NA,
    adj = c(0.5, 1.25)
  )

  # Legend tick marks (white dashes)
  lgnd_ticks = lgnd_labs
  lgnd_ticks[!is.na(lgnd_ticks)] = "-   -"
  text(
    x = seq(rasterbox[1], rasterbox[3], length.out = length(lgnd_labs)),
    y = rasterbox[4],
    labels = lgnd_ticks,
    col = "white",
    xpd = NA,
    adj = c(0, 0.5),
    srt = 90
  )

  # Legend title
  text(
    x = rasterbox[1],
    y = rasterbox[4],
    labels = paste0(legend_args[["title"]], " "),
    xpd = NA,
    adj = c(1, -0.5)
  )
}

#' Add new elements to the current `tinyplot`
#'
#' @description
#' This convenience function grabs the preceding `tinyplot` call and updates it
#' with any new arguments that have been explicitly provided by the user. It
#' then injects `add=TRUE` and evaluates the updated call, thereby drawing a new
#' layer on top of the existing plot. `plt_add()` is a shorthand alias for
#' `tinyplot_add()`.
#'
#' @section Limitations:
#' - `tinyplot_add()` works reliably only when adding to a plot originally
#'   created using the [`tinyplot.formula`] method with a valid `data` argument.
#'   We cannot guarantee correct behavior if the original plot was created with
#'   the atomic [`tinyplot.default`] method, due to potential environment
#'   mismatches. (An exception is when the original plot arguments---`x`, `y`,
#'   etc.---are located in the global environment.)
#'
#' - Automatic legends for the added elements will be turned off.
#'
#' @param ... All named arguments override arguments from the previous calls.
#' Arguments not supplied to [tinyplot_add] remain unchanged from the previous
#' call. Arguments are captured unevaluated and spliced into the updated call,
#' so those that rely on non-standard evaluation against `data`---e.g.
#' `subset = cyl == 4` or `weights = wt`---behave the same as in a direct
#' [`tinyplot`] call.
#'
#' @examples
#' tinyplot(Sepal.Width ~ Sepal.Length | Species,
#'   facet = ~Species,
#'   data = iris)
#'
#' tinyplot_add(type = "lm") ## or : plt_add(type = "lm")
#'
#' ## Note: the previous function is equivalent to (but much more convenient
#' ## than) re-writing the full call with the new type and `add=TRUE`:
#'
#' # tinyplot(Sepal.Width ~ Sepal.Length | Species,
#' #          facet = ~Species,
#' #          data = iris,
#' #          type = "lm",
#' #          add = TRUE)
#'
#' ## Arguments relying on non-standard evaluation (e.g. `subset`) work too:
#' tinyplot(mpg ~ wt, data = mtcars)
#' tinyplot_add(subset = cyl == 4, col = "red", pch = 16)
#'
#' @returns No return value, called for side effect of producing a plot.
#'
#' @export
tinyplot_add = function(...) {
  cal = get_environment_variable(".last_call")

  if (is.null(cal)) {
    stop("No previous tinyplot call found.")
  }

  # Capture the dots as *unevaluated* expressions rather than evaluated values.
  # This way, arguments that rely on non-standard evaluation against `data`
  # (e.g. `subset = cyl == 4`, `weights = wt`, or a bare-name formula) are
  # spliced into the rebuilt call and only resolved when it is finally
  # evaluated through the formula method's `model.frame`. Using `list(...)`
  # here would instead force evaluation in this frame, where data-scoped
  # variables are not visible. (#630)
  args = match.call(expand.dots = FALSE)[["..."]]
  nms = names(args)
  for (i in seq_along(args)) {
    n = if (is.null(nms)) "" else nms[[i]]
    if (!is.null(n) && n != "") {
      cal[[n]] = args[[i]]
    }
  }

  # allow first argument in tinyplot_add() to be unnamed
  if (length(args) >= 1L && (is.null(nms) || nms[[1L]] == "")) {
    cal[[2]] = args[[1L]]
  }

  cal[["add"]] = TRUE
  eval(cal, envir = parent.frame())
}



#' @export
#' @name plt_add
#' @rdname tinyplot_add
plt_add = tinyplot_add

# Ensure added layers respect the x-axis order of the original plot layer
# (e.g., when adding lines or ribbons on top of errorbars)
align_layer = function(settings) {
  # Retrieve xlabs and plot/device metadata from original layer
  tinyplot_env = get(".tinyplot_env", envir = parent.env(environment()))
  xlabs_orig = tryCatch(get("xlabs_orig", envir = tinyplot_env), error = function(e) NULL)
  usr_orig = tryCatch(get("usr_orig", envir = tinyplot_env), error = function(e) NULL)
  dev_orig = tryCatch(get("dev_orig", envir = tinyplot_env), error = function(e) NULL)
  
  # Validate that we're adding to the same plot (not a stale xlabs from previous plot)
  if (is.null(usr_orig) || is.null(dev_orig) || dev_orig != dev.cur()) {
    return(invisible())
  }
  # Normalize current usr for comparison (accounting for flipped plots)
  usr_layer = if (isTRUE(settings$flip)) par("usr")[c(3,4,1,2)] else par("usr")
  if (!identical(usr_orig, usr_layer)) {
    return(invisible())
  }
  
  # xlabs of current layer
  xlabs_layer = settings[["xlabs"]]
  
  # Only adjust if original layer has named xlabs
  if (!is.null(names(xlabs_orig))) {
    if (is.factor(settings$datapoints[["x"]])) {
      # Case 1: relevel a factor (e.g., ribbon added to errorbars)
      settings$datapoints[["x"]] = tryCatch(
        factor(settings$datapoints[["x"]], levels = names(xlabs_orig)),
        error = function(e) {
          settings$datapoints[["x"]]
        }
      )
      settings$datapoints = settings$datapoints[order(settings$datapoints[["x"]]), ]
    } else if (!is.null(names(xlabs_layer))) {
      # Case 2: match implicit integer -> label mapping (e.g., lines added to errorbars)
      if (setequal(names(xlabs_layer), names(xlabs_orig))) {
        # If mappings already agree and no dodge, no realignment needed
        if (identical(xlabs_layer, xlabs_orig) && is.null(settings$dodge)) return(invisible())
        orig_order = xlabs_orig[names(xlabs_layer)[settings$datapoints[["x"]]]]
        x_layer = settings$datapoints[["x"]]
        if (is.null(settings$dodge)) {
          x_new = x_layer[orig_order] 
        } else {
          names(x_layer) = names(xlabs_layer)[round(x_layer)]
          x_new = x_layer + (xlabs_orig[names(round(x_layer))] - round(x_layer))
        }
        settings$datapoints[["x"]] = x_new
        # Adjust ancillary variables
        for (v in c("xmin", "xmax")) {
          if (identical(settings$datapoints[[v]], unname(x_layer))) {
            settings$datapoints[[v]] = x_new
          }
        }
        settings$datapoints = settings$datapoints[order(settings$datapoints[["x"]]), ]
        settings$datapoints[["rowid"]] = seq_len(nrow(settings$datapoints))
      }
    }
  }
}

#' @rdname type_errorbar
#' @export
type_pointrange = function(dodge = 0, fixed.dodge = FALSE) {
  out = list(
    draw = draw_pointrange(),
    data = data_pointrange(dodge = dodge, fixed.dodge = fixed.dodge),
    name = "p"
  )
  class(out) = "tinyplot_type"
  return(out)
}


draw_pointrange = function() {
  fun = function(
      ix,
      iy,
      ixmin,
      iymin,
      ixmax,
      iymax,
      icol,
      ibg,
      ipch,
      ilwd,
      icex,
      ...) {
    segments(
      x0 = ixmin,
      y0 = iymin,
      x1 = ixmax,
      y1 = iymax,
      col = icol,
      lwd = ilwd
    )
    draw_points()(
      ix = ix,
      iy = iy,
      icol = icol,
      ibg = ibg,
      ipch = ipch,
      ilwd = ilwd,
      icex = icex
    )
  }
  return(fun)
}


data_pointrange = function(dodge, fixed.dodge) {
  fun = function(settings, ...) {
    env2env(settings, environment(), c("datapoints", "xlabs", "cex", "lty", "lwd"))

    if (is.character(datapoints$x)) {
      datapoints$x = as.factor(datapoints$x)
    }
    if (is.factor(datapoints$x)) {
      ## original data (i.e., no new sorting by factor)
      xlvls = unique(datapoints$x)
      datapoints$x = factor(datapoints$x, levels = xlvls)
      xlabs = seq_along(xlvls)
      names(xlabs) = xlvls
      datapoints$x = as.integer(datapoints$x)
    }
    datapoints$xmin = datapoints$x
    datapoints$xmax = datapoints$x

    # dodge
    if (dodge != 0) {
      datapoints = dodge_positions(datapoints, dodge, fixed.dodge)
    }

    x = datapoints$x
    
    # legend customizations
    settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% (cex %||% par("cex"))
    settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% (lty %||% par("lty"))
    settings$legend_args[["lwd"]] = settings$legend_args[["lwd"]] %||% (lwd %||% par("lwd"))
    
    env2env(environment(), settings, c(
      "x",
      "xlabs",
      "datapoints"
    ))
  }
  return(fun)
}

#' tinyplot Method for Plotting ts Objects (Time Series)
#'
#' @description Convenience interface for visualizing \code{\link[stats]{ts}}
#'   (time series) objects with tinyplot.
#'
#' @details Internally the time series object is converted to a long
#'   data frame with columns `Time` (time index), `Value` (observations),
#'   and `Series` (factor with column labels). Depending on the settings
#'  of `facet` this data frame is visualized either with the formula
#'  `Value ~ Time` or `Value ~ Time | Series`. See the `facet` argument
#'  description for more details and the Examples below for some illustrations.
#'
#'  An exception is made if the user explicitly supplies a distribution `type`
#'  argument (i.e, one of `"histogram"`, `"density"`, `"boxplot"`, `"violin"`,
#'  `"qq"`, `"ridge"`, or `"rug"`). These summarize the series values and
#'  ignore the time index, so they are visualized with the one-sided
#'  formula `~ Value` (or `~ Value | Series` for multivariate series). This
#'  preserves base-R-compatible behaviour such as
#'  `tinyplot(Nile, type = "histogram")`.
#'
#' @param x an object of class `"ts"`.
#' @param facet specification of `facet` for `tinyplot.formula`. The
#'   default in the `tinyplot` method is to use `facet = NULL` for univariate
#'   series and `facet = ~ Series` (equivalent to `facet = "by"`) for
#'   multivariate series.
#' @param type,facet.args,xlab,ylab,... further arguments passed to `tinyplot`.
#'
#' @returns No return value, called for the side effect of producing a plot.
#'
#' @examples
#' ## univariate series
#' tinyplot(Nile)
#' 
#' # exception: expicitly passing a distribution type on a univariate series
#' # still triggers the corresponding transformation
#' tinyplot(Nile, type = "histogram")
#' 
#' ## multivariate series (generally, these also look better with a theme)
#' tinytheme("clean2")
#' tinyplot(EuStockMarkets)                    ## faceted, free scales, same color
#' tinyplot(EuStockMarkets, facet.args = NULL) ## faceted, same scale, same color
#' tinyplot(EuStockMarkets, facet = "by")      ## faceted, free scales, diff colors
#' tinyplot(EuStockMarkets, facet = NULL)      ## single frame, diff colors
#'
#' ## further variations
#' tinyplot(EuStockMarkets, facet = NULL, legend = list("direct", repel = TRUE))
#' tinyplot(EuStockMarkets, facet = "by", facet.args = NULL)
#' tinyplot(EuStockMarkets, facet.args = list(free = TRUE, ncol = 1))
#' 
#' ## pass additional tinyplot args through `...` for further customization
#' tinyplot(EuStockMarkets,
#'          facet.args = NULL,
#'          type = "area", xlab = NA, yaxl = ",",
#'          main = "European stock indices")
#'
#' tinytheme() ## reset
#'
#' @importFrom stats time
#' @export
tinyplot.ts = function(x, facet, type = "l", facet.args = list(free = TRUE), xlab = NULL, ylab = NA, ...) {
  ## basic object properties
  n = NROW(x)
  k = NCOL(x)
  lab = deparse(substitute(x))
  if (k > 1L) lab = paste(lab, 1L:k, sep = ".")
  if (!is.null(colnames(x))) lab = colnames(x)

  single = k == 1L

  ## convert to long data.frame
  df = data.frame(
    Time = rep.int(as.numeric(time(x)), k),
    Value = as.numeric(x),
    Series = factor(rep(1L:k, each = n), labels = lab)
  )

  ## Distribution types summarize the series *values* and ignore the time index,
  ## so they dispatch with a one-sided formula (`~ Value`, or `~ Value | Series`
  ## when multivariate) rather than `Value ~ Time`, and skip the time-series
  ## `ylab` default so tinyplot's own label (e.g. "Frequency") shows. This
  ## preserves base-R-compatible behaviour like `tinyplot(Nile, type = "histogram")`.
  dist_types = c("histogram", "hist", "density", "boxplot", "box",
                 "violin", "qq", "ridge", "rug")
  ## normalize to a type name to catch both the string ("histogram") and the
  ## type object (`type_histogram()`, whose `$name` is the canonical "histogram")
  type_name = if (inherits(type, "tinyplot_type")) {
    type[["name"]]
  } else if (is.character(type) && length(type) == 1L) {
    type
  } else {
    NA_character_
  }
  is_dist = !is.na(type_name) && type_name %in% dist_types

  ## default for facet
  if(missing(facet)) {
    auto = TRUE
    facet = if(single || is_dist) NULL else ~ Series
  } else {
    auto = FALSE
  }
  if (is.null(facet)) facet.args = NULL

  ## dispatch formula and axis labels
  if (is_dist) {
    fml = if (single) ~ Value else ~ Value | Series
    if (single && is.null(xlab)) xlab = lab
    ## use tinyplot's own ylab default (e.g. "Frequency") unless caller set one
    if (missing(ylab)) ylab = NULL
  } else if (single || (!is.null(facet) && auto)) {
    fml = Value ~ Time
  } else {
    fml = Value ~ Time | Series
  }

  ## call tinyplot
  tinyplot(fml, data = df, type = type, facet = facet, facet.args = facet.args,
           xlab = xlab, ylab = ylab, ...)
}

sanitize_datapoints = function(settings) {
  # potentially useful variables
  env2env(
    settings,
    environment(),
    c(
      "x", "xmin", "xmax", "xaxt",
      "y", "ymin", "ymax", "ygroup",
      "facet", "null_by", "by", "type",
      "weights", "labels"
    )
  )

  ## coerce character and logical variables to factors
  ## (aside: we won't risk converting x and y logicals to factors b/c it can
  ##  mess up types that rely on predict underneath the hood, e.g type_lm)
  if (!is.null(x) && is.character(x)) x = factor(x)
  if (!is.null(y) && is.character(y)) y = factor(y)
  if (!null_by && is.logical(by)) by = factor(by)

  if (is.null(x)) {
    ## Special catch for rect and segment plots without a specified y-var
    if (type %in% c("rect", "segments")) {
      x = rep(NA, length(x))
    }
  }

  if (is.null(y)) {
    ## Special catch for area and interval plots without a specified y-var
    if (type %in% c("rect", "segments", "pointrange", "errorbar", "ribbon")) {
      y = rep(NA, length(x))
    } else if (type == "boxplot") {
      y = x
      x = rep.int("", length(y))
      xaxt = "a"
    } else if (!(type %in% c("histogram", "barplot", "density", "function"))) {
      y = x
      x = seq_along(x)
    }
  }

  datapoints = list(
    x = x, xmin = xmin, xmax = xmax,
    y = y, ymin = ymin, ymax = ymax, ygroup = ygroup,
    weights = weights, labels = labels
  )
  datapoints = Filter(function(z) length(z) > 0, datapoints)
  datapoints = data.frame(datapoints)
  if (nrow(datapoints) > 0) {
    datapoints[["rowid"]] = seq_len(nrow(datapoints))
    datapoints[["facet"]] = if (!is.null(facet)) facet else ""
    datapoints[["by"]] = if (!null_by) by else ""
  }

  # potentially modified variables
  env2env(environment(), settings, c("x", "y", "xaxt", "datapoints"))
}

#' @rdname type_ribbon
#' @export
type_area = function(alpha = NULL) {
    out = list(
        draw = NULL,
        data = data_area(alpha = alpha),
        name = "area"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_area = function(alpha = alpha) {
    ribbon.alpha = if (is.null(alpha)) .tpar[["ribbon.alpha"]] else (alpha)
    fun = function(settings, ...) {
        env2env(settings, environment(), "datapoints")
        datapoints$ymax = datapoints$y
        datapoints$ymin = rep.int(0, nrow(datapoints))
        ymax = datapoints$ymax
        ymin = datapoints$ymin
        type = "ribbon"

        # ribbon.alpha comes from parent scope, so assign it locally
        ribbon.alpha = ribbon.alpha

        # legend customizations
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% par("lwd")
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% 0
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25

        env2env(environment(), settings, c(
            "datapoints",
            "ymax",
            "ymin",
            "type",
            "ribbon.alpha"
        ))
    }
    return(fun)
}

# Don't run tests on CRAN (because of test snapshot suite size)
# See https://github.com/grantmcdermott/tinyplot/pull/128
NOT_CRAN = toupper(Sys.getenv("NOT_CRAN")) == "TRUE" || isTRUE(Sys.getenv("NOT_CRAN"))

# Only test on Linux machines (for tinysnapshot fidelity)
ON_LINUX = Sys.info()["sysname"] == "Linux"

if ( NOT_CRAN && ON_LINUX && requireNamespace("tinytest", quietly=TRUE) ){
  tinytest::test_package("tinyplot")
}

# Clean up
rm(ON_LINUX, NOT_CRAN)

#' Jittered points plot type
#'
#' @description Type function for plotting jittered points.
#' Arguments are passed to \code{\link[base]{jitter}}.
#'
#' @inheritParams base::jitter
#' @inherit base::jitter details
#' @examples
#' # "jitter" type convenience string
#' tinyplot(Sepal.Length ~ Species, data = iris, type = "jitter")
#'
#' # Use `type_jitter()` to pass extra arguments for customization
#' tinyplot(Sepal.Length ~ Species, data = iris, type = type_jitter(factor = 0.5))
#' @export
type_jitter = function(factor = 1, amount = NULL) {
    out = list(
        draw = draw_points(),
        data = data_jitter(factor = factor, amount = amount),
        name = "p"
    )
    class(out) = "tinyplot_type"
    return(out)
}


jitter_restore = function(obj, factor, amount) {
    if (inherits(obj, "POSIXlt")) {
        obj = as.POSIXct(obj)
    }
    if (inherits(obj, c("Date", "POSIXt", "yearmon", "yearqtr"))) {
        obj_attrs = attributes(obj)
        out = jitter(unclass(obj), factor = factor, amount = amount)
        attributes(out) = obj_attrs
    } else {
        out = jitter(obj, factor = factor, amount = amount)
    }
    return(out)
}


data_jitter = function(factor, amount) {
    fun = function(settings, ...) {
        env2env(settings, environment(), c("datapoints", "add"))

        x = datapoints$x
        y = datapoints$y
        if (is.factor(x)) {
            xlvls = levels(x)
            xlabs = seq_along(xlvls)
            names(xlabs) = xlvls
            x = as.integer(x)
        } else {
            xlabs = NULL
        }
        if (is.factor(y)) {
            ylvls = levels(y)
            ylabs = seq_along(ylvls)
            names(ylabs) = ylvls
            y = as.integer(y)
        } else {
            ylabs = NULL
        }

        # Apply group offsets from base layer (e.g., boxplot, violin, ridge)
        group_offsets = get_environment_variable(".group_offsets")
        offsets_axis = get_environment_variable(".offsets_axis")
        if (isTRUE(add) && !is.null(group_offsets)) {
            if (identical(offsets_axis, "x") && is.factor(datapoints$by)) {
                # x-axis offsets (boxplot, violin): keyed by group level
                if (is.null(xlabs)) {
                    xf = as.factor(x)
                    xlvls = levels(xf)
                    xlabs = seq_along(xlvls)
                    names(xlabs) = xlvls
                    x = as.integer(xf)
                }
                x = x + group_offsets[as.integer(datapoints$by)]
            } else if (identical(offsets_axis, "y")) {
                # y-axis offsets (ridge): keyed by y-level name
                if (is.null(ylabs)) {
                    yf = as.factor(y)
                    ylvls = levels(yf)
                    ylabs = seq_along(ylvls)
                    names(ylabs) = ylvls
                    y = as.integer(yf)
                }
                y_labels = names(ylabs)[y]
                y = group_offsets[y_labels]
            }
        }

        x = jitter_restore(x, factor = factor, amount = amount)
        y = jitter_restore(y, factor = factor, amount = amount)

        datapoints$x = x
        datapoints$y = y

        env2env(environment(), settings, c(
            "datapoints",
            "x",
            "y",
            "xlabs",
            "ylabs"
        ))
    }
}

#' tinyplot Method for Plotting Data Frames
#'
#' @description Convenience interface for visualizing
#'   \code{\link[base]{data.frame}} objects with tinyplot.
#'
#' @details This is a convenience function for plotting data frames with
#'   or without a formula. The case with the formula mainly facilitates
#'   using `tinyplot()` in combination with pipes. The case without
#'   formula provides a quick way of plotting the variables in a data frame:
#'   a single variable, a pair of variables, or a
#'   \code{\link[graphics]{pairs}}-style grid of all variable combinations for
#'   3 or more variables. The latter can be called directly (also for less than
#'   3 variables) via `tinypairs()`.
#' 
#'   Note that, unlike a base \code{\link[graphics]{pairs}} plot, these
#'   `tinyplot` variants (deliberately) produce a histogram for the univariate
#'   cases along the grid diagonal.
#'
#' @param x an object of class `"data.frame"`.
#' @param formula a \code{\link[stats]{formula}} that is passed on to
#'   \code{\link{tinyplot.formula}}. If `formula` is `NULL` a formula of
#'   type `y ~ 1` or `y ~ x` is set up for 1- and 2-variable data frames,
#'   respectively. For data frames with 3 or more variables, a
#'   \code{\link[graphics]{pairs}}-style grid of all variable combinations is
#'   drawn instead.
#' @param by (pairs-case only) optional grouping variable. Either 1) a character
#'   string giving a column name in `x`, or 2) a vector of equal length to the
#'   main data.frame (_caveat emptor_: this is not checked). The variable is
#'   spliced into each sub-plot's formula as `y ~ x | by` so groups are
#'   distinguished (e.g. by colour). Note that the legend is deliberately
#'   suppressed.
#' @param labs (pairs-case only) logical indicating whether the axes labels
#'   (titles) for each sub-plot in the pairs-style case should be shown. Default
#'   is `FALSE`.
#' @param frames (pairs-case only) logical indicating whether each sub-plot should
#'   be framed by a box. Default is `FALSE`. Note the trailing "s" plural case
#'   to disambiguate from `frame.plot`.
#' @param ... further arguments passed to `tinyplot`.
#'
#' @returns No return value, called for the side effect of producing a plot.
#'
#' @examples
#' ## using tinyplot() with data frames
#' tinyplot(cars)
#' tinyplot(iris, Sepal.Length ~ Petal.Width | Species)
#' tinyplot(
#'   as.data.frame(UCBAdmissions),
#'   Admit ~ Gender, facet = 1 ~ Dept, weights = Freq
#' )
#'
#' ## note that this also enables usage with pipes (in R >= 4.1.0) such as
#' # cars |> tinyplot()
#' # iris |> tinyplot(Sepal.Length ~ Petal.Width | Species)
#' # UCBAdmissions |> 
#' #   as.data.frame() |>
#' #   tinyplot(Admit ~ Gender, facet = 1 ~ Dept, weights = Freq)
#' # etc.
#' 
#' ## pairs-style display for data frames with 3 or more variables
#' ## (works best with a theme)
#' tinytheme("clean2")
#' tinyplot(iris)
#'
#' ## pass `by` arg to group the pairs display (legend is suppressed)
#' ## here, we also add optional frames around the individual sub-plots
#' tinyplot(iris, by = "Species", frames = TRUE)
#' 
#' ## another option (but assumes objects of equal length)
#' tinyplot(iris[, 1:4], by = iris$Species, frames = TRUE)
#'
#' ## tinypairs can also create 2x2 pairs display
#' tinypairs(iris[, c(3, 1)], by = iris$Species)
#'
#' tinytheme() ## reset theme
#'
#' @importFrom stats reformulate
#' @export
tinyplot.data.frame = function (x, formula = NULL, by = NULL, labs = FALSE, frames = FALSE, ...) {
  ## original call
  cl = match.call()

  ## variables
  nm = names(x)
  n = length(nm)

  ## behavior without formula:
  ## 1d: y ~ 1
  ## 2d: y ~ x
  ## 3d: pairs
  if (is.null(formula) && n > 2L) {
    cl[[1L]] = quote(tinyplot::tinypairs)  
  } else {
    ## update call for formula interface
    names(cl)[names(cl) == "x"] = "data"
    cl[[1L]] = quote(tinyplot::tinyplot)
    if (is.null(formula)) cl$formula = . ~ .
    id = which(names(cl) == "formula")
    cl = as.call(as.list(cl)[c(1L, id, setdiff(2L:length(cl), id))])
    ## `by` only applies to the >=3 (pairs-style) case; drop it so it isn't
    ## passed through to tinyplot() here.
    if ("by" %in% names(cl)) cl[["by"]] = NULL
  
    ## backtick-protect names so reformulate() parses non-syntactic names (e.g.
    ## "GDP (2020)") as single symbols rather than as code, matching pairs().
    bt = function(x) paste0("`", x, "`")

    ## default formula
    if (is.null(formula)) {
      cl$formula = if (length(nm) < 2L) reformulate("1", bt(nm)) else reformulate(bt(nm[1L]), bt(nm[2L]))
    }
  }

  ## evaluate updated call
  eval.parent(cl)
}

#' @rdname tinyplot.data.frame
#' @export
tinypairs = function (x, by = NULL, labs = FALSE, frames = FALSE, ...) {
  ## original call
  cl = match.call()

  ## update call for formula interface
  names(cl)[names(cl) == "x"] = "data"
  cl[[1L]] = quote(tinyplot::tinyplot)
  cl$formula = . ~ .
  id = which(names(cl) == "formula")
  cl = as.call(as.list(cl)[c(1L, id, setdiff(2L:length(cl), id))])
  
  ## variables
  nm = names(x)
  n = length(nm)

  ## backtick-protect names so reformulate() parses non-syntactic names (e.g.
  ## "GDP (2020)") as single symbols rather than as code, matching pairs().
  bt = function(x) paste0("`", x, "`")

  op = par(mfrow = c(n, n))
  on.exit(par(op))

  assert_logical(labs)
  assert_logical(frames)
  if (!is.null(by)) {
    if (inherits(by, "character")) {
      ## a column name already in `x`
      assert_choice(by, nm)
    } else {
      ## a standalone vector: splice it into `x` under a unique name so the
      ## formula method can reference it. make.unique() guards against an
      ## existing "__by__" column. The name is added after `n` is computed,
      ## so it is used for grouping only and not drawn as its own row/column.
      by_nm = make.unique(c(nm, "__by__"))[n + 1L]
      x[[by_nm]] = by
      by = by_nm
      cl[["data"]] = x
    }
    ## `by` is spliced directly into each cell formula below, so drop it from
    ## the call. Suppress the legend to avoid repeating it across every cell.
    if ("by" %in% names(cl)) cl[["by"]] = NULL
    cl[["legend"]] = FALSE
  }

  ## To scale cex like faceted plots, we capture the existing theme and apply
  ## temporary cex overrides (restoring on exit). The theme is built as a
  ## language object because `cl` is a matched call, so `cl[["theme"]]` is
  ## unevaluated (e.g. the call `list("dark")`, not a list).
  cex_fct_adj = ifelse(n > 2, 0.66, 0.83) # use same scaling as with faceted plots.
  active_theme = get_tpar("tinytheme", default = "default")
  theme_arg = cl[["theme"]]
  if (is.null(theme_arg)) {
    theme_ij = bquote(list(.(active_theme), cex = .(cex_fct_adj)))
  } else if (is.call(theme_arg) && identical(theme_arg[[1L]], as.name("list"))) {
    theme_arg[["cex"]] = cex_fct_adj
    theme_ij = theme_arg
  } else {
    theme_ij = bquote(list(.(theme_arg), cex = .(cex_fct_adj)))
  }
  ## drop any `theme` from the cell call so it isn't re-applied per cell (see
  ## above). NULL-assignment on a call errors if the element is absent, so
  ## guard with a membership check.
  if ("theme" %in% names(cl)) cl[["theme"]] = NULL

  ## apply temp theme overrides and restore orig theme on exit
  do.call(tinytheme, eval(theme_ij))
  on.exit(if (identical(active_theme, "default")) tinytheme() else tinytheme(active_theme), add = TRUE)

  for (j in 1L:n) {
    for (i in 1L:n) {
      cl_ij = cl
      grp = if (is.null(by)) "" else paste(" |", bt(by))
      if (i == j) {
        cl_ij$formula = reformulate(paste0("1", grp), bt(nm[i]))
        cl_ij$main = nm[i]
      } else {
        cl_ij$formula = reformulate(paste0(bt(nm[i]), grp), bt(nm[j]))
      }
      if (!labs) {
        cl_ij$ylab = NA
        cl_ij$xlab = NA
      }
      eval.parent(cl_ij)
      if (frames) box("figure", lwd = 0.3)
    }
  }
}

## set up empty environment for storing parameters/calls/hooks/...
.tinyplot_env = new.env()

## query environment
get_environment_variable = function(name = NULL) {
  ## either one 'name' or entire environment as list
  if(is.null(name)) return(as.list(.tinyplot_env))
  name = as.character(name)[1L]
  return(.tinyplot_env[[name]])
}

## update environment
set_environment_variable = function(...) {
  ## check for unnamed arguments
  dots = list(...)
  if(is.null(names(dots))) {
    stop("arguments must be named")
  } else if(any(names(dots) == "")) {
    warning("ignoring unnamed arguments")
    dots = dots[names != ""]
  }
  
  ## set environment variables
  if(length(dots) > 0L) {
    for(i in names(dots)) {
      .tinyplot_env[[i]] = dots[[i]]
    }
  }
  invisible(dots)
}

## initialize all environment variables with NULL
set_environment_variable(
  .saved_par_before = NULL,
  .saved_par_after = NULL,
  .saved_par_first = NULL,
  .last_call = NULL,
  .tpar_hooks = NULL,
  .registered_themes = NULL
)

# Copied from https://raw.githubusercontent.com/r-lib/evaluate/refs/heads/main/R/hooks.R
# license: MIT + file LICENSE



#' Set and remove hooks
#'
#' This interface wraps the base [setHook()] function to provide a return
#' value that makes it easy to undo.
#'
#' @param hooks a named list of hooks - each hook can either be a function or
#'   a list of functions.
#' @param action `"replace"`, `"append"` or `"prepend"`
#' @keywords internal
set_hooks = function(hooks, action = "append") {
  old = list()
  for (hook_name in names(hooks)) {
    old[[hook_name]] = getHook(hook_name)
    setHook(hook_name, hooks[[hook_name]], action = action)
  }
  invisible(old)
}

#' @rdname set_hooks
#' @keywords internal
remove_hooks = function(hooks) {
  for (hook_name in names(hooks)) {
    hook = getHook(hook_name)
    if (length(hook) > 0) {
      for (fun in unlist(hooks[hook_name])) {
        hook[sapply(hook, identical, fun)] = NULL
      }
    }
    setHook(hook_name, hook, "replace")
  }
}

setup_device = function(settings) {
  env2env(settings, environment(), c("file", "width", "height"))

  # write to file
  if (!is.null(file)) {
    filepath = file
    filewidth = width
    fileheight = height
    if (is.null(filewidth)) filewidth = .tpar[["file.width"]]
    if (is.null(fileheight)) fileheight = .tpar[["file.height"]]
    fileres = .tpar[["file.res"]]
    # catch to close interactive device if one isn't already open
    fkdev = is.null(dev.list())
    # grab existing device pars to pass on to next one
    dop = par(no.readonly = TRUE)
    # close interactive device if not already open
    if (isTRUE(fkdev)) dev.off()
    exttype = file_ext(filepath)
    if (exttype == "pdf" && .tpar[["cairo"]]) {
      exttype = "cairo"
    } else if (exttype == "jpg") {
      exttype = "jpeg"
    }
    switch(exttype,
      png = png(filepath, width = filewidth, height = fileheight, units = "in", res = fileres),
      jpeg = jpeg(filepath, width = filewidth, height = fileheight, units = "in", res = fileres),
      pdf = pdf(filepath, width = filewidth, height = fileheight),
      cairo = cairo_pdf(filepath, width = filewidth, height = fileheight),
      svg = svg(filepath, width = filewidth, height = fileheight),
      stop("\nUnsupported file extension. Only '.png', '.jpg', '.pdf', or '.svg' are allowed.\n")
    )
    dop$new = FALSE # catch for some interfaces
    par(dop)

    # interactive plot with user-specified width/height
  } else if (!is.null(width) || !is.null(height)) {
    devwidth = width
    devheight = height
    # if one of width or height is missing, set equal to the other
    if (is.null(devwidth)) devwidth = devheight
    if (is.null(devheight)) devheight = devwidth
    # catch to close interactive device if one isn't already open
    fkdev = is.null(dev.list())
    # grab existing device pars to pass on to next one
    dop = par(no.readonly = TRUE)
    # close interactive device if not already open
    if (isTRUE(fkdev)) dev.off()
    dev.new(width = devwidth, height = devheight)
    dop$new = FALSE # catch for some interfaces
    par(dop)
  }
}

#' Spline plot type
#'
#' @description Type function for plotting a cubic (or Hermite) spline interpolation.
#' Arguments are passed to \code{\link[stats]{spline}}; see this latter function
#' for default argument values.
#'
#' @inheritParams stats::spline
#' @details See \code{\link[stats]{spline}} for further details on the available
#'   interpolation methods.
#' @importFrom stats spline
#' @examples
#' # "spline" type convenience string
#' tinyplot(dist ~ speed, data = cars, type = "spline")
#'
#' # Use `type_spline()` to pass extra arguments for customization
#' tinyplot(dist ~ speed,
#'     data = cars, type = type_spline(method = "natural", n = 25),
#'     add = TRUE, lty = 2)
#' @export
type_spline = function(n = NULL,
                       method = "fmm",
                       xmin = NULL,
                       xmax = NULL,
                       xout = NULL,
                       ties = mean) {
    out = list(
        draw = draw_lines(),
        data = data_spline(method = method, ties = ties, n = n, xmin = xmin, xmax = xmax, xout = xout),
        name = "l"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_spline = function(n, method, xmin, xmax, xout, ties, ...) {
    fun = function(settings, ...) {
        env2env(settings, environment(), "datapoints")

        datapoints = split(datapoints, list(datapoints$facet, datapoints$by), drop = TRUE)
        datapoints = lapply(datapoints, function(dat) {
            if (is.null(n)) n = 3 * length(dat$x)
            if (is.null(xmax)) xmax = max(dat$x)
            if (is.null(xmin)) xmin = min(dat$x)
            if (is.null(xout)) {
                fit = spline(x = dat$x, y = dat$y, n = n, method = method, xmin = xmin, xmax = xmax, ties = ties)
            } else {
                fit = spline(x = dat$x, y = dat$y, n = n, method = method, xmin = xmin, xmax = xmax, xout = xout, ties = ties)
            }
            fit = as.data.frame(fit)
            fit$facet = dat$facet[1]
            fit$by = dat$by[1]
            fit
            return(fit)
        })
        datapoints = do.call(rbind, datapoints)
        env2env(environment(), settings, "datapoints")
    }
    return(fun)
}

#' Linear model plot type
#'
#' @description Type function for plotting a linear model fit.
#' Arguments are passed to \code{\link[stats]{lm}}.
#'
#' @inheritParams type_glm
#' @importFrom stats lm predict
#' @examples
#' # "lm" type convenience string
#' tinyplot(Sepal.Width ~ Petal.Width, data = iris, type = "lm")
#'
#' # Grouped model fits (here: illustrating an example of Simpson's paradox)
#' tinyplot(Sepal.Width ~ Petal.Width | Species, data = iris, type = "lm")
#' tinyplot_add(type = "p")
#'
#' # Use `type_lm()` to pass extra arguments for customization
#' tinyplot(Sepal.Width ~ Petal.Width, data = iris, type = type_lm(level = 0.8))
#'
#' # Weighted fit example, either as a top-level argument (which supports
#' # non-standard evaluation in the formula method)...
#' s77 = as.data.frame(state.x77)
#' tinyplot(`Life Exp` ~ Income, data = s77, type = "lm", weights = Population)
#' # ...or directly via the type constructor (requires an evaluated vector)
#' tinyplot(`Life Exp` ~ Income, data = s77, type = type_lm(
#'   weights = s77$Population))
#' @export
type_lm = function(se = TRUE, level = 0.95, weights = NULL) {
    assert_flag(se)
    out = list(
        draw = draw_ribbon(),
        data = data_lm(se = se, level = level, weights = weights),
        name = if (isTRUE(se)) "ribbon" else "l"
    )
    class(out) = "tinyplot_type"
    return(out)
}


data_lm = function(se, level, weights = NULL, ...) {
    fun = function(settings, ...) {
        env2env(settings, environment(), "datapoints")
        # top-level `weights` (carried on datapoints via NSE) take precedence
        # over the constructor-level `weights` argument
        if (is.null(datapoints[["weights"]]) && !is.null(weights)) {
            datapoints[["weights"]] = weights
        }
        has_weights = !is.null(datapoints[["weights"]])
        if (has_weights) settings$weights_used = TRUE
        dat = split(datapoints, list(datapoints$facet, datapoints$by))
        dat = lapply(dat, function(x) {
            if (nrow(x) == 0) {
                return(x)
            }
            if (nrow(x) < 3) {
                x$y = NA
                return(x)
            }
            # .w is NULL when no weights column is present, which lm() treats
            # the same as omitting the argument
            .w = x[["weights"]]
            fit = lm(y ~ x, data = x, weights = .w)
            nd = data.frame(x = seq(min(x$x, na.rm = TRUE), max(x$x, na.rm = TRUE), length.out = 100))
            nd$by = x$by[1]
            nd$facet = x$facet[1]
            if (se == TRUE) {
                p = predict(fit, newdata = nd, se.fit = TRUE)
                p = ci(p$fit, p$se.fit, conf.level = level, fit$df.residual)
                nd$y = p$estimate
                nd$ymax = p$conf.high
                nd$ymin = p$conf.low
            } else {
                nd$y = predict(fit, newdata = nd)
            }
            nd
        })
        datapoints = do.call(rbind, dat)
        datapoints = datapoints[order(datapoints$facet, datapoints$by, datapoints$x), ]
        
        # legend customizations - same as ribbon but add line through square
        settings$legend_args[["pch"]] = settings$legend_args[["pch"]] %||% 22
        settings$legend_args[["pt.cex"]] = settings$legend_args[["pt.cex"]] %||% 3.5
        settings$legend_args[["pt.lwd"]] = settings$legend_args[["pt.lwd"]] %||% 0
        settings$legend_args[["lty"]] = settings$legend_args[["lty"]] %||% par("lty")
        settings$legend_args[["y.intersp"]] = settings$legend_args[["y.intersp"]] %||% 1.25
        settings$legend_args[["seg.len"]] = settings$legend_args[["seg.len"]] %||% 1.25
        
        env2env(environment(), settings, "datapoints")
    }
    return(fun)
}

sanitize_axes = function(settings) {
  env2env(settings, environment(), c("axes", "xaxt", "yaxt", "frame.plot"))
  ## handle defaults of axes, xaxt, yaxt, frame.plot
  ## - convert axes to character if necessary
  ## - set defaults of xaxt/yaxt (if these are NULL) based on axes
  ## - set logical axes based on xaxt/yaxt
  ## - set frame.plot default based on xaxt/yaxt
  if (isFALSE(axes)) {
    axes = xaxt = yaxt = "none"
  } else if (isTRUE(axes)) {
    axes = "standard"
    if (is.null(xaxt)) xaxt = get_tpar("xaxt", default = "standard")
    if (is.null(yaxt)) yaxt = get_tpar("yaxt", default = "standard")
  } else {
    xaxt = yaxt = axes
  }
  axis_types = c("standard", "none", "labels", "ticks", "axis")
  axes = match.arg(axes, axis_types)
  xaxt = match.arg(xaxt, axis_types)
  yaxt = match.arg(yaxt, axis_types)
  xaxt = substr(match.arg(xaxt, axis_types), 1L, 1L)
  yaxt = substr(match.arg(yaxt, axis_types), 1L, 1L)
  axes = any(c(xaxt, yaxt) != "n")
  if (is.null(frame.plot) || !is.logical(frame.plot)) frame.plot = all(c(xaxt, yaxt) %in% c("s", "a"))


  env2env(environment(), settings, c("axes", "xaxt", "yaxt", "frame.plot"))
}

#' @title Generic function for adding an axis to a (tiny)plot
#'   
#' @description Internal function used for adding an axis to a [`tinyplot`]
#'   call.
#' @details `tinyAxis` provides a thin(ish) wrapper around
#'   \code{\link[graphics]{Axis}}, but with enhanced flexibility to (i) match
#'   parameter combinations based on the axis type and plotting theme, (ii)
#'   provide better support for date-time variables, and (iii) enable convenient
#'   formatting of axis tick labels.
#' @inheritParams graphics::Axis
#' @param type the type of axis to be drawn; inherited from the `xaxt` or `yaxt`
#'   arguments of the parent [`tinyplot()`] call. One of either: `"standard"`
#'   (default that draws the axis, ticks, and labels), `"none"` (no axes),
#'   `"ticks"` (only ticks and labels without axis line), `"labels"` (only
#'   labels without ticks and axis line), or `"axis"` (only axis line and labels
#'   but no ticks). Partial matching is allowed, e.g. `type = "s"`.
#' @inheritParams tinylabel
#' @examples
#' \dontrun{
#' 
#' # plot without axes
#' tinyplot(0:10, axes = "n")
#' # add x-axis (labels only)
#' tinyplot:::tinyAxis(x = 0:10, side = 1, type = "l")
#' # add y-axis (with custom label formatting)
#' tinyplot:::tinyAxis(x = 0:10, side = 2, type = "s", labeller = "$")
#' }
#' @keywords internal
tinyAxis = function(x = NULL, ..., type = "standard", labeller = NULL) {
  type = match.arg(type, c("standard", "none", "labels", "ticks", "axis"))
  if (type == "none") {
    invisible(numeric(0L))
  } else {
    args = list(x = x, ...)
    if (type == "labels") {
      args$tick = FALSE
    } else if (type == "ticks") {
      args$lwd = 0
      if (!("lwd.ticks" %in% names(args))) args$lwd.ticks = 1
    } else if (type == "axis") {
      args$lwd.ticks = 0
    } else {
      args$tick = TRUE
    }
    if (!is.null(labeller)) {
      if (!is.null(args$at)) {
        args$labels = if (!is.null(args$labels)) tinylabel(args$labels, labeller) else tinylabel(args$at, labeller)
      } else {
        args$at = if (!inherits(x, c("POSIXt", "Date"))) axTicks(args$side) else axTicksDateTime(args$side, x = x)  
        args$labels = tinylabel(args$at, labeller)
      }
    }
    do.call("Axis", args)
  }
}

# Special case for Date-Time, adapted/simplified from axis.date()
axTicksDateTime = function(side, x, ...) {
  if (inherits(x, "POSIXt")) {
    tz = attr(x, "tz")
    range = extendrange(x)
    rangeDateTime = .POSIXct(range, tz = tz)
  } else {
    range = sort(par("usr")[if (side%%2)  1L:2L else 3:4L])
    range[1L] = ceiling(range[1L])
    range[2L] = floor(range[2L])
    rangeDateTime = range
    class(rangeDateTime) = "Date"
  }
  z = pretty(rangeDateTime, n = par("lab")[2 - side%%2])
  keep = z >= range[1L] & z <= range[2L]
  z = z[keep]
  return(z)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_segments.R
\name{type_segments}
\alias{type_segments}
\title{Line segments plot type}
\usage{
type_segments()
}
\description{
Type function for plotting line segments.
}
\details{
Contrary to base \code{\link[graphics]{segments}}, line segments in
\link{tinyplot} must be specified using the \code{xmin}, \code{ymin},\code{xmax}, and \code{ymax}
arguments.
}
\examples{
# "segments" type convenience character string
tinyplot(
  xmin = c(0,.1), ymin = c(.2,1), xmax = c(1,.9), ymax = c(.75,0),
  type = "segments"
)

# Same result with type_segments()
tinyplot(
  xmin = c(0,.1), ymin = c(.2,1), xmax = c(1,.9), ymax = c(.75,0),
  type = type_segments()
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/repel.R
\name{repel_text}
\alias{repel_text}
\title{Force-directed text repelling}
\usage{
repel_text(
  x,
  y,
  widths,
  heights,
  anchor_x = x,
  anchor_y = y,
  min_gap = 0,
  iterations = 20,
  axis = "both"
)
}
\arguments{
\item{x, y}{Numeric vectors of current label positions.}

\item{widths, heights}{Numeric vectors of label bounding box dimensions (in
the same units as \code{x} and \code{y}).}

\item{anchor_x, anchor_y}{Numeric vectors of original target positions that
labels are pulled back toward during the spring pass. Defaults to \code{x}
and \code{y}.}

\item{min_gap}{Minimum padding between labels. Default \code{0}.}

\item{iterations}{Maximum number of repulsion iterations. Default \code{20}.}

\item{axis}{Movement constraint: \code{"both"} (default), \code{"x"}, or \code{"y"}.}
}
\value{
A list with elements \code{x} and \code{y} giving adjusted positions.
}
\description{
Resolves overlapping text labels using a two-phase algorithm: (1) push
overlapping labels apart symmetrically, (2) spring-pull labels back toward
their anchors without reintroducing overlaps.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_boxplot.R
\name{type_boxplot}
\alias{type_boxplot}
\title{Boxplot type}
\usage{
type_boxplot(
  range = 1.5,
  width = NULL,
  varwidth = FALSE,
  notch = FALSE,
  outline = TRUE,
  boxwex = 0.8,
  staplewex = 0.5,
  outwex = 0.5,
  lighten = TRUE
)
}
\arguments{
\item{range}{this determines how far the plot whiskers extend out
    from the box.  If \code{range} is positive, the whiskers extend
    to the most extreme data point which is no more than
    \code{range} times the interquartile range from the box. A value
    of zero causes the whiskers to extend to the data extremes.}

\item{width}{a vector giving the relative widths of the boxes making
    up the plot.}

\item{varwidth}{if \code{varwidth} is \code{TRUE}, the boxes are
    drawn with widths proportional to the square-roots of the number
    of observations in the groups.}

\item{notch}{if \code{notch} is \code{TRUE}, a notch is drawn in
    each side of the boxes.  If the notches of two plots do not
    overlap this is \sQuote{strong evidence} that the two medians differ
    (Chambers \abbr{et al.}, 1983, p.\ifelse{latex}{\out{~}}{ }{}62).  See \code{\link{boxplot.stats}}
    for the calculations used.}

\item{outline}{if \code{outline} is not true, the outliers are
    not drawn (as points whereas S+ uses lines).}

\item{boxwex}{a scale factor to be applied to all boxes.  When there
    are only a few groups, the appearance of the plot can be improved
    by making the boxes narrower.}

\item{staplewex}{staple line width expansion, proportional to box
    width.}

\item{outwex}{outlier line width expansion, proportional to box
    width.}

\item{lighten}{logical. Should the fills use a lighter, opaque tint of the
series colour(s)? Default is \code{TRUE}, which keeps single- and multi-group
displays consistent and lets the fill read cleanly over grid lines. Set to
\code{FALSE} to use the fully-saturated palette colour(s) instead.}
}
\description{
Type function for producing box-and-whisker plots.
Arguments are passed to \code{\link[graphics]{boxplot}}, although \code{tinyplot}
scaffolding allows added functionality such as grouping and faceting.
Box-and-whisker plots are the default plot type if \code{x} is a factor and \code{y} is
numeric.
}
\examples{
# "boxplot" type convenience string
tinyplot(weight ~ feed, data = chickwts, type = "boxplot")

# Note: Specifying the type here is redundant. Like base plot, tinyplot
# automatically produces a boxplot if x is a factor and y is numeric
tinyplot(weight ~ feed, data = chickwts)

# For flipped boxplots, it's usually better to use a dynamic theme to
# accommodate (horizontal) y-axis labels
tinyplot(weight ~ feed, data = chickwts, flip = TRUE, theme = "dynamic")

# Grouped boxplot example using a different dataset
# (C.f., the final example in `?boxplot`)
tinyplot(
  len ~ dose | supp, data = ToothGrowth, type = "boxplot",
  main = "Guinea Pigs' Tooth Growth",
  legend = list(title = "Supplement"),
  xlab = "Vitamin C dose mg", ylab = "tooth length"
)

# Use `type_boxplot()` to pass extra arguments for customization
tinyplot(
  len ~ dose | supp, data = ToothGrowth,
  type = type_boxplot(boxwex = 0.3, staplewex = 0, outline = FALSE), lty = 1,
  legend = list(title = "Supplement"),
  main = "Guinea Pigs' Tooth Growth",
  xlab = "Vitamin C dose mg", ylab = "tooth length"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_polypath.R
\name{type_polypath}
\alias{type_polypath}
\title{Polypath polygon type}
\usage{
type_polypath(rule = "winding")
}
\arguments{
\item{rule}{character value specifying the path fill mode: either
    \code{"winding"} or \code{"evenodd"}.}
}
\description{
Type function for plotting polygons.
Arguments are passed to \code{\link[graphics]{polypath}}.
}
\examples{
# "polypath" type convenience character string
tinyplot(
    c(.1, .1, .6, .6, NA, .4, .4, .9, .9),
    c(.1, .6, .6, .1, NA, .4, .9, .9, .4),
    type = "polypath", fill = "grey"
)

# Use `type_polypath()` to pass extra arguments for customization
tinyplot(
    c(.1, .1, .6, .6, NA, .4, .4, .9, .9),
    c(.1, .6, .6, .1, NA, .4, .9, .9, .4),
    type = type_polypath(rule = "evenodd"), fill = "grey"
)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{sanitize_legend}
\alias{sanitize_legend}
\title{Sanitize and normalize legend input}
\usage{
sanitize_legend(legend, legend_args)
}
\arguments{
\item{legend}{Legend specification (NULL, character, list, or call)}

\item{legend_args}{Existing legend_args list to merge with}
}
\value{
Normalized legend_args list with at least an "x" element
}
\description{
Converts various legend input formats (NULL, character, list,
call) into a standardized legend_args list with an "x" position element.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tpar.R
\name{tpar}
\alias{tpar}
\title{Set or query graphical parameters}
\usage{
tpar(..., hook = FALSE)
}
\arguments{
\item{...}{arguments of the form \code{key = value}. This includes all of the
parameters typically supported by \code{\link[graphics]{par}}, as well as
the \code{tinyplot}-specific ones described in the 'Graphical Parameters'
section below.}

\item{hook}{Logical. If \code{TRUE}, base graphical parameters persist across
plots via a hook applied before each new plot (see \code{?setHook}).}
}
\value{
When parameters are set, their previous values are returned in an
invisible named list. Such a list can be passed as an argument to \code{tpar} to
restore the parameter values.

When just one parameter is queried, the value of that parameter is returned
as (atomic) vector. When two or more parameters are queried, their values
are returned in a list, with the list names giving the parameters.

Note the inconsistency: setting one parameter returns a list, but querying
one parameter returns a vector.
}
\description{
Extends \code{\link[graphics]{par}}, serving as a (near) drop-in
replacement for setting or querying graphical parameters. The key
differences is that, beyond supporting the standard group of R graphical
parameters in \code{\link[graphics]{par}}, \code{tpar} also supports additional
graphical parameters that are provided by \code{tinyplot}. Similar to
\code{\link[graphics]{par}}, parameters are set by passing appropriate
\code{key = value} argument pairs, and multiple parameters can be set or queried
at the same time.
}
\details{
The \code{tinyplot}-specific parameters are saved in an internal
environment called \code{.tpar} for performance and safety reasons. However,
they can also be set at package load time via \code{\link[base]{options}},
which may prove convenient for users that want to enable different default
behaviour at startup (e.g., through an \code{.Rprofile} file). These options all
take a \verb{tinyplot_*} prefix, e.g.
\code{options(tinyplot_grid = TRUE, tinyplot_facet.bg = "grey90")}.

For their part, any "base" graphical parameters are caught dynamically and
passed on to \code{\link[graphics]{par}} as appropriate. Technically, only
parameters that satisfy \code{par(..., no.readonly = TRUE)} are evaluated.

However, note the important distinction: \code{tpar} only evaluates parameters
from \code{\link[graphics]{par}} if they are passed \emph{explicitly} by the
user. This means that \code{tpar} should not be used to capture the (invisible)
state of a user's entire set of graphics parameters, i.e. \code{tpar()} !=
\code{par()}. If you want to capture the \emph{all} existing graphics settings, then
you should rather use \code{par()} instead.
}
\section{Additional Graphical Parameters}{

\itemize{
\item \code{adj.cap}: Numeric value between 0 and 1 controlling the alignment of the plot caption. Defaults to \code{0.5} (centered) for the default theme, and \code{1} (right-aligned) for all other themes.
\item \code{adj.xlab}: Numeric value between 0 and 1 controlling the alignment of the x-axis label.
\item \code{adj.ylab}: Numeric value between 0 and 1 controlling the alignment of the y-axis label.
\item \code{cairo}: Logical indicating whether \code{\link[grDevices]{cairo_pdf}} should be used when writing plots to PDF. If \code{FALSE}, then \code{\link[grDevices]{pdf}} will be used instead, with implications for embedding (non-standard) fonts. Only used if \code{tinyplot(..., file = "<filename>.pdf")} is called. Defaults to the value of \code{capabilities("cairo")}.
\item \code{cex.cap}: Numeric expansion factor for the plot caption text. Defaults to \code{1} for the default, basic, and dynamic themes, and \code{0.8} for clean/classic and their descendants.
\item \code{col.cap}: Character specifying the colour of the plot caption. Defaults to \code{"black"}.
\item \code{col.default}: Default colour for single-group displays (i.e. plots without a \code{by} grouping). Can be \code{NULL}, a length-1 character colour, or a length-1 numeric index into \code{palette.qualitative}. Defaults to \code{NULL}, in which case the first colour of the active qualitative palette is used (or base \code{palette()[1]}, typically black, if no theme is set). A character value sets the single-group colour independently of the multi-group palette. A numeric value \code{i} selects the \code{i}th colour of \code{palette.qualitative} as the single-group default; a \emph{negative} index additionally drops that colour from the palette used for grouped plots. For example, \code{col.default = -1} paired with \code{palette.qualitative = "Okabe-Ito"} uses black (the leading palette colour) for single-group plots and an Okabe-Ito-minus-black palette for grouped plots, avoiding the need to maintain a separate "no-black" palette.
\item \code{font.cap}: Integer specifying the font face for the plot caption (\code{1} = plain, \code{2} = bold, \code{3} = italic, \code{4} = bold italic). Defaults to \code{1}.
\item \code{line.cap}: Numeric specifying the margin line on which to draw the caption. If \code{NULL} (default), computed automatically based on the available bottom margin.
\item \code{dynmar}: Logical indicating whether \code{tinyplot} should attempt dynamic adjustment of margins to reduce whitespace and/or account for spacing of text elements (e.g., long horizontal y-axis labels). Note that this parameter is tightly coupled to internal \code{tinythemes()} logic and should \emph{not} be adjusted manually unless you really know what you are doing or don't mind risking unintended consequences to your plot.
\item \code{facet.bg}: Character or integer specifying the facet background colour. If an integer, will correspond to the user's default colour palette (see \code{palette}). Passed to \code{rect}. Defaults to \code{NULL} (none).
\item \code{facet.border}: Character or integer specifying the facet border colour. If an integer, will correspond to the user's default colour palette (see \code{palette}). Passed to \code{rect}. Defaults to \code{NA} (none).
\item \code{facet.cex}: Expansion factor for facet titles. Defaults to \code{1}.
\item \code{facet.col}: Character or integer specifying the facet text colour. If an integer, will correspond to the user's default global colour palette (see \code{palette}). Defaults to \code{NULL}, which is equivalent to "black".
\item \code{facet.font}: An integer corresponding to the desired font face for facet titles. For most font families and graphics devices, one of four possible values: \code{1} (regular), \code{2} (bold), \code{3} (italic), or \code{4} (bold italic). Defaults to \code{NULL}, which is equivalent to \code{1} (i.e., regular).
\item \code{file.height}: Numeric specifying the height (in inches) of any plot that is written to disk using the \code{tinyplot(..., file = X)} argument. Defaults to \code{7}.
\item \code{file.res}: Numeric specifying the resolution (in dots per square inch) of any plot that is written to disk in bitmap format (i.e., PNG or JPEG) using the \code{tinyplot(..., file = X)} argument. Defaults to \code{300}.
\item \code{file.width}: Numeric specifying the width (in inches) of any plot that is written to disk using the \code{tinyplot(..., file = X)} argument. Defaults to \code{7}.
\item \code{fmar}: A numeric vector of form \code{c(b,l,t,r)} for controlling the (base) margin padding, in terms of lines, between the individual facets in a faceted plot. Defaults to \code{c(1,1,1,1)}. If more than three facets are detected, the \code{fmar} parameter is scaled by 0.75 to reduce excess whitespace. For 2x2 plots, the padding better matches the \code{cex} expansion logic of base graphics.
\item \code{gap.main}: Numeric giving the gap (in margin lines) from the main title baseline to the element below it (plot box when no subtitle; subtitle top when subtitle is present). Only used when \code{dynmar = TRUE}. Defaults to \code{0.7}.
\item \code{gap.sub}: Numeric giving the gap (in margin lines) from the subtitle baseline to the plot box. Only used when \code{dynmar = TRUE} and \code{side.sub = 3}. Defaults to \code{0.7}.
\item \code{grid.col}: Character or (integer) numeric that specifies the color of the panel grid lines. Defaults to \code{"lightgray"}.
\item \code{grid.lty}: Character or (integer) numeric that specifies the line type of the panel grid lines. Defaults to \code{"dotted"}.
\item \code{grid.lwd}: Non-negative numeric giving the line width of the panel grid lines. Defaults to \code{1}.
\item \code{grid}: Logical or character indicating whether a background panel grid should be added to plots automatically. Defaults to \code{NULL}, which is equivalent to \code{FALSE}. In addition to logical values, a character string can be used to control axis-specific grids: uppercase letters (\code{"X"}, \code{"Y"}, \code{"XY"}) draw grid lines at the standard axis tick positions (equivalent to \code{TRUE}), while lowercase letters (\code{"x"}, \code{"y"}, \code{"xy"}) draw a finer grid with additional lines at the midpoints between ticks.
\item \code{lmar}: A numeric vector of form \code{c(inner, outer)} that gives the margin padding, in terms of lines, around the automatic \code{tinyplot} legend. Defaults to \code{c(1.0, 0.1)}. The inner margin is the gap between the legend and the plot region, and the outer margin is the gap between the legend and the edge of the graphics device.
\item \code{palette.qualitative}: Palette for qualitative colors. See the \code{palette} argument in \code{?tinyplot}.
\item \code{palette.sequential}: Palette for sequential colors. See the \code{palette} argument in \code{?tinyplot}.
\item \code{ribbon.alpha}: Numeric factor in the range \verb{[0,1]} for modifying the opacity alpha of "ribbon" and "area" type plots. Default value is \code{0.2}.
}
}

\examples{
# Return a list of existing base and tinyplot graphic params
tpar("las", "pch", "facet.bg", "facet.cex", "grid")

# Simple facet plot with these default values
tinyplot(mpg ~ wt, data = mtcars, facet = ~am)

# Set params to something new. Similar to graphics::par(), note that we save
# the existing values at the same time by assigning to an object.
op = tpar(
  las       = 1,
  pch       = 2,
  facet.bg  = "grey90",
  facet.cex = 2,
  grid      = TRUE
)

# Re-plot with these new params
tinyplot(mpg ~ wt, data = mtcars, facet = ~am)

# Reset back to original values
tpar(op)

# Important: tpar() only evaluates parameters that have been passed explicitly
#   by the user. So it it should not be used to query and set (restore)
#   parameters that weren't explicitly requested, i.e. tpar() != par().

# Note: The tinyplot-specific parameters can also be be set via `options`
#   with a `tinyplot_*` prefix, which can be convenient for enabling
#   different default behaviour at startup time (e.g., via an .Rprofile
#   file). Example:
# options(tinyplot_grid = TRUE, tinyplot_facet.bg = "grey90")

}
\seealso{
\code{\link[graphics:par]{graphics::par}} which \code{tpar} builds on top of. \code{\link{get_saved_par}}
is a convenience function for retrieving graphical parameters at different
stages of a \code{tinyplot} call (and used for internal accounting purposes).
\code{\link{tinytheme}} allows users to easily set a group of graphics parameters
in a single function call, according to a variety of predefined themes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_rect.R
\name{type_rect}
\alias{type_rect}
\title{Rectangle plot type}
\usage{
type_rect()
}
\description{
Type function for plotting rectangles.
}
\details{
Contrary to base \code{\link[graphics]{rect}}, rectangles in
\link{tinyplot} must be specified using the \code{xmin}, \code{ymin},\code{xmax}, and \code{ymax}
arguments.
}
\examples{
i = 4*(0:10)

# "rect" type convenience character string
tinyplot(
  xmin = 100+i, ymin = 300+i, xmax = 150+i, ymax = 380+i,
  by = i, fill = 0.2,
  type = "rect"
)

# Same result with type_rect()
tinyplot(
  xmin = 100+i, ymin = 300+i, xmax = 150+i, ymax = 380+i,
  by = i, fill = 0.2,
  type = type_rect()
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/get_saved_par.R
\name{get_saved_par}
\alias{get_saved_par}
\title{Retrieve the saved graphical parameters}
\usage{
get_saved_par(when = c("before", "after", "first"))
}
\arguments{
\item{when}{character. From when should the saved parameters be retrieved?
Either "before" (the default) or "after" the preceding \code{tinyplot} call.}
}
\value{
A list of \code{\link[graphics]{par}} settings.
}
\description{
Convenience function for retrieving the graphical parameters
(i.e., the full list of \code{tag = value} pairs held in
\code{\link[graphics]{par}}) from either immediately before or
immediately after the most recent \link{tinyplot} call.
}
\details{
A potential side-effect of \link{tinyplot} is that it can change a user's
\code{\link[graphics]{par}} settings. For example, it may adjust the inner
and outer plot margins to make space for an automatic legend; see
\link{draw_legend}. While it is possible to immediately restore the original
\code{\link[graphics]{par}} settings upon exit via the
\code{tinyplot(..., restore.par = TRUE)} argument, this is not the default
behaviour. The reason being that we need to preserve the adjusted parameter
settings in case users want to add further graphical annotations to their
plot (e.g., \code{\link[graphics]{abline}}, \code{\link[graphics]{text}},
etc.) Nevertheless, it may still prove desirable to recall and reset these
original graphical parameters after the fact (e.g., once all these extra
annotations have been added). That is the purpose of this \link{get_saved_par}
function.

Of course, users may prefer to manually capture and reset graphical
parameters, as per the standard method described in the
\code{\link[graphics]{par}} documentation. For example:

\if{html}{\out{<div class="sourceCode">}}\preformatted{op = par(no.readonly = TRUE)  # save current par settings 
# <do lots of (tiny)plotting>
par(op)                       # reset original pars
}\if{html}{\out{</div>}}

This standard manual approach may be safer than \link{get_saved_par} because it
offers more precise control. Specifically, the value of \link{get_saved_par}
itself will be reset after ever new \link{tinyplot} call; i.e. it may inherit an
already-changed set of parameters. Users should bear these trade-offs in
mind when deciding which approach to use. As a general rule,
\link{get_saved_par} offers the convenience of resetting the original
\code{\link[graphics]{par}} settings even if a user forgot to save them
beforehand. But one should avoid invoking it after a series of consecutive
\link{tinyplot} calls.

Finally, note that users can always call \code{\link[grDevices]{dev.off}}
to reset all \code{\link[graphics]{par}} settings to their defaults.
}
\examples{
#
# Contrived example where we draw a grouped scatterplot with a legend and
# manually add corresponding best fit lines for each group...
#

# First draw the grouped scatterplot
tinyplot(Sepal.Length ~ Petal.Length | Species, iris)

# Preserving adjusted par settings is good for adding elements to our plot
for (s in levels(iris$Species)) {
  abline(
    lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
    col = which(levels(iris$Species)==s)
  )
}

# Get saved par from before the preceding tinyplot call (but don't use yet)
sp = get_saved_par("before")

# Note the changed margins will affect regular plots too, which is probably
# not desirable
plot(1:10)

# Reset the original parameters (could use `par(sp)` here)
tpar(sp)
# Redraw our simple plot with our corrected right margin
plot(1:10)

#
# Quick example going the other way, "correcting" for par.restore = TRUE...
#

tinyplot(Sepal.Length ~ Petal.Length | Species, iris, restore.par = TRUE)
# Our added best lines will be wrong b/c of misaligned par
for (s in levels(iris$Species)) {
  abline(
    lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
    col = which(levels(iris$Species)==s), lty = 2
  )
}
# grab the par settings from the _end_ of the preceding tinyplot call to fix
tpar(get_saved_par("after"))
# now the best lines are correct
for (s in levels(iris$Species)) {
  abline(
    lm(Sepal.Length ~ Petal.Length, iris, subset = Species==s),
    col = which(levels(iris$Species)==s)
  )
}

# reset again to original saved par settings before exit
tpar(sp)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{build_legend_env}
\alias{build_legend_env}
\title{Build legend environment}
\usage{
build_legend_env(
  legend,
  legend_args,
  by_dep,
  lgnd_labs,
  labeller = NULL,
  type,
  pch,
  lty,
  lwd,
  col,
  bg,
  cex,
  gradient,
  lmar,
  has_sub = FALSE,
  has_cap = FALSE,
  cap_text = NULL,
  new_plot = TRUE,
  dynmar_title_mar = NULL
)
}
\arguments{
\item{legend}{Legend placement keyword or list}

\item{legend_args}{Additional legend arguments}

\item{by_dep}{The (deparsed) "by" grouping variable name}

\item{lgnd_labs}{The legend labels}

\item{labeller}{Character or function for formatting labels}

\item{type}{Plot type}

\item{pch}{Plotting character(s)}

\item{lty}{Line type(s)}

\item{lwd}{Line width(s)}

\item{col}{Color(s)}

\item{bg}{Background fill color(s)}

\item{cex}{Character expansion(s)}

\item{gradient}{Logical indicating gradient legend}

\item{lmar}{Legend margins (inner, outer)}

\item{has_sub}{Logical indicating presence of sub-caption}

\item{new_plot}{Logical indicating if plot.new should be called}
}
\value{
Environment with complete legend specification
}
\description{
Creates the legend environment by:
\itemize{
\item Initializing environment with metadata
\item Calling build_legend_args() to construct legend arguments
\item Populating environment with arguments and positioning flags
\item Initializing margins and dimensions
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytheme.R
\name{tinytheme}
\alias{tinytheme}
\title{Set or Reset Plot Themes for \code{tinyplot}}
\usage{
tinytheme(
  theme = c("default", "basic", "dynamic", "clean", "clean2", "bw", "linedraw",
    "classic", "minimal", "ipsum", "ipsum2", "dark", "socviz", "broadsheet", "nber",
    "web", "ridge", "ridge2", "tufte", "float", "void"),
  ...,
  register = NULL
)
}
\arguments{
\item{theme}{A character string specifying the name of the theme to apply.
Themes are arranged in an approximate hierarchy, adding or subtracting
elements in the order presented below. Note that several themes are
\emph{dynamic}, in the sense that they attempt to reduce whitespace in a way
that is responsive to the length of axes labels, tick marks, etc. These
dynamic plots are marked with an asterisk (*) below.
\itemize{
\item \code{"default"}: inherits the user's default base graphics settings.
\item \code{"basic"}: light modification of \code{"default"}, only adding filled points, a panel background grid, and light gray background to facet titles.
\item \code{"dynamic"} (*): builds on \code{"basic"} by enabling dynamic margin adjustment with tighter default margins, horizontal axis labels, and the subtitle moved above the plotting area. Turns off the panel grid. Provides the foundation for all other dynamic themes and is a good starting point for users who wish to build custom dynamic themes.
\itemize{
\item \code{"clean"} (*): builds on \code{"dynamic"} by re-enabling the panel background grid and setting different default palettes ("Tableau 10" for discrete colors and "agSunset" for gradient colors).
\itemize{
\item \code{"clean2"} (*): removes the plot frame (box) from \code{"clean"}.
}
\item \code{"classic"} (*): builds on \code{"dynamic"} with L-shaped axes (removing the top and right-hand edges of the plot frame), smaller axis text, tighter axis spacing, and the "Okabe-Ito" palette as a default for discrete colors. Inspired by the \strong{ggplot2} theme of the same name.
\item \code{"bw"} (*): similar to \code{"clean"}, except uses thinner lines for the plot frame (box), solid fine grid lines, smaller axis text, tighter axis spacing, and sets the "Okabe-Ito" palette as a default for discrete colors. Inspired by the \strong{ggplot2} theme of the same name.
\itemize{
\item \code{"linedraw"} (*): builds on \code{"bw"} with black facet title strips and white strip text. Inspired by the \strong{ggplot2} theme of the same name.
\item \code{"minimal"} (*): removes the plot frame (box) from \code{"bw"}, as well as the background for facet titles. Inspired by the \strong{ggplot2} theme of the same name.
\itemize{
\item \code{"ipsum"} (*): builds on \code{"minimal"} with bold titles, no ticks, fine grid, edge-aligned axis titles, and a custom muted palette. Inspired by the \strong{hrbrthemes} theme of the same name for \strong{ggplot2}.
\item \code{"ipsum2"} (*): a lighter variant of \code{"ipsum"} that retains the original italic subtitle and edge-aligned axis titles, but without the additional spacing and palette changes.
\item \code{"dark"} (*): similar to \code{"minimal"}, but set against a dark background with foreground and a palette colours lightened for appropriate contrast.
\item \code{"socviz"} (*): builds on \code{"minimal"} with L-shaped axes, very light grid lines, and larger axis text. Inspired by Kieran Healy's \strong{socviz} package theme for \strong{ggplot2}.
}
}
\item \code{"broadsheet"} (*): a publication/newspaper style with only horizontal grid lines, no frame, short x-axis ticks, and muted secondary text (subtitle, caption). Compact axis spacing.
\itemize{
\item \code{"nber"} (*): builds on \code{"broadsheet"} for an NBER working paper style with a light blue-grey background, grey text, italic axis titles and captions, and a blue-grey discrete palette.
}
\item \code{"web"} (*): a FiveThirtyEight-inspired style with a light grey device background, no frame or axis lines, and bold grid lines. Suited to web/online publication.
\item \code{"tufte"} (*): floating axes and minimalist plot artifacts in the style of Edward Tufte.
\itemize{
\item \code{"float"} (*): builds on \code{"tufte"} with outward ticks, fewer tick marks, and a "dark" qualitative palette.
}
\item \code{"void"} (*): switches off all axes, titles, legends, etc.
\item \code{"ridge"} (*): a specialized theme for ridge plots (see \code{\link[=type_ridge]{type_ridge()}}). Builds off of \code{"clean"}, but adds ridge-specific tweaks (e.g. default "Zissou 1" palette for discrete colors, solid horizontal grid lines, and minor adjustments to y-axis labels). Not recommended for non-ridge plots.
\itemize{
\item \code{"ridge2"} (*): removes the plot frame (box) from \code{"ridge"}, but retains the x-axis line. Again, not recommended for non-ridge plots.
}
}
}}

\item{...}{Named arguments to override specific theme settings. These
arguments are passed to \code{tpar()} and take precedence over the predefined
settings in the selected theme.}

\item{register}{Optional character string. If provided, the theme (with any
\code{...} overrides) is registered under this name via \code{\link[=tinytheme_register]{tinytheme_register()}}
and simultaneously activated. This is a shortcut for calling
\code{\link[=tinytheme_register]{tinytheme_register()}} and \code{tinytheme()} separately.}
}
\value{
The function returns nothing. It is called for its side effects.
}
\description{
The \code{tinytheme} function sets or resets the theme for plots created with
\code{tinyplot}. Themes control the appearance of plots, such as text alignment,
font styles, axis labels, and even dynamic margin adjustment to reduce
whitespace.
}
\details{
Sets a list of graphical parameters using \code{tpar()}

\strong{Persistent vs. ephemeral themes.} Calling \code{tinytheme("<theme>")} triggers
a persistent theme, which will be applied to all subsequent graphics (incl.
base \code{plot}). To reset the theme to your default settings, call \code{tinytheme()}
without arguments. Alternatively, invoke the \verb{tinyplot(..., theme = <theme>)}
argument for an ephemeral theme that is automatically reset at the end of the
plot call.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Set a persistent theme
tinytheme("clean")
tinyplot(1:10)  # uses "clean"
tinyplot(10:1)  # still uses "clean"
tinytheme()     # reset to default

# Use an ephemeral theme for a single plot (incl. added components)
tinyplot(1:10, theme = "dark")  # uses "dark" ephemerally
tinyplot_add(type = "S")        # same plot, so retains theme
tinyplot(10:1)                  # new plot, so back to default
}\if{html}{\out{</div>}}

\strong{Custom overrides.} To customize a theme's parameters (e.g., \code{mar}, \code{las},
etc.), either pass them directly as additional args to
\verb{tinytheme(<theme>, ...)} or as a list arguments to
\verb{tinyplot(..., theme = list(<theme>, ...))}.
Please note that passing overrides through \code{tpar()} or \code{par()} \emph{won't} work,
because themes work by installing a persistent hook, which means that an
active theme will override any subsequent calls for parameters that the theme
controls.

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Do this
tinytheme("clean", mar = c(5, 5, 2, 2))
<some plot>

# Or this (ephemeral version)
tinyplot(..., theme = list("clean", mar = c(5, 5, 2, 2)))

# Don't do this (the theme hook will overwrite your changes)
tinytheme("clean")
tpar(mar = c(5, 5, 2, 2))
<some plot>
}\if{html}{\out{</div>}}

\strong{Spacing primitives.} Dynamic themes compute \code{mgp} (margin line positions)
automatically from two spacing primitives, rather than requiring users to
reason about how \code{mar}, \code{mgp}, and \code{tcl} combine:
\itemize{
\item \code{gap.axis}: the gap in margin lines between the tick tip and the near edge
of the tick label. Default \code{0.2}.
\item \code{gap.lab}: the gap in margin lines between the far edge of the tick label
and the near edge of the axis title. Default \code{1.0}.
}

These primitives scale automatically with \code{cex.axis} and \code{cex.lab}, so the
visible spacing between elements remains constant regardless of text size.
To adjust spacing, pass them as overrides:

\if{html}{\out{<div class="sourceCode">}}\preformatted{# Tighter spacing between tick labels and axis titles
tinytheme("clean", gap.lab = 0.5)

# More room between ticks and tick labels
tinytheme("clean", gap.axis = 0.5)
}\if{html}{\out{</div>}}

If you supply an explicit \code{mgp} value, it is used as-is and the primitives
are ignored.
}
\examples{
# Reusable plot function
p = function() tinyplot(
  lat ~ long | depth, data = quakes,
  main = "Earthquakes off Fiji",
  sub = "Data courtesy of the Harvard PRIM-H project"
)
p()

# Set a theme
tinytheme("dark")
p()

#  A set theme is persistent and will apply to subsequent plots
tinyplot(0:10)

# Try a different theme
tinytheme("clean")
p()
         
# Customize the theme by overriding default settings
tinytheme("clean",
          adj.xlab = 1, adj.ylab = 1,
          cex.lab = 0.75, cex.axis = 0.9,
          font.sub = 3,
          gap.axis = 0, gap.lab = 0.5,
          tcl = -0.1)
p()

# Another custom theme example, including a different font
tinytheme("bw", font.main = 2, col.axis = "darkcyan", family = "HersheyScript")
p()

# Aside: One or two specialized themes are only meant for certain plot types
tinytheme("ridge2")
tinyplot(I(cut(lat, 10)) ~ depth, data = quakes, type = "ridge")

# Reset the theme
tinytheme()
p()

# For an ephemeral theme, use `tinyplot(..., theme = <theme>)` directly
tinyplot(0:10, theme = "clean", main = "This theme is ephemeral")
tinyplot(10:0, main = "See, no more theme")

# Customize an ephemeral theme by passing arguments as a list
tinyplot(0:10, main = "Custom", theme = list("clean", grid.col = "hotpink"))

# Themes showcase
## We'll use a slightly more intricate plot (long y-axis labs and facets)
## to demonstrate dynamic margin adjustment etc.

thms = eval(formals(tinytheme)$theme)

for (thm in thms) {
  tinytheme(thm)
  tinyplot(
    I(Sepal.Length*1e4) ~ Petal.Length | Species, facet = "by", data = iris,
    yaxl = ",", 
    main = paste0('tinytheme("', thm, '")'),
    sub = "A subtitle",
    cap = "A caption"
  )
  box("outer", lty = 2)
}

# Reset
tinytheme()

}
\seealso{
\code{\link{tpar}} which does the heavy lifting under the hood;
\code{\link[=tinytheme_register]{tinytheme_register()}} for registering custom named themes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot.data.frame.R
\name{tinyplot.data.frame}
\alias{tinyplot.data.frame}
\alias{tinypairs}
\title{tinyplot Method for Plotting Data Frames}
\usage{
\method{tinyplot}{data.frame}(x, formula = NULL, by = NULL, labs = FALSE, frames = FALSE, ...)

tinypairs(x, by = NULL, labs = FALSE, frames = FALSE, ...)
}
\arguments{
\item{x}{an object of class \code{"data.frame"}.}

\item{formula}{a \code{\link[stats]{formula}} that is passed on to
\code{\link{tinyplot.formula}}. If \code{formula} is \code{NULL} a formula of
type \code{y ~ 1} or \code{y ~ x} is set up for 1- and 2-variable data frames,
respectively. For data frames with 3 or more variables, a
\code{\link[graphics]{pairs}}-style grid of all variable combinations is
drawn instead.}

\item{by}{(pairs-case only) optional grouping variable. Either 1) a character
string giving a column name in \code{x}, or 2) a vector of equal length to the
main data.frame (\emph{caveat emptor}: this is not checked). The variable is
spliced into each sub-plot's formula as \code{y ~ x | by} so groups are
distinguished (e.g. by colour). Note that the legend is deliberately
suppressed.}

\item{labs}{(pairs-case only) logical indicating whether the axes labels
(titles) for each sub-plot in the pairs-style case should be shown. Default
is \code{FALSE}.}

\item{frames}{(pairs-case only) logical indicating whether each sub-plot should
be framed by a box. Default is \code{FALSE}. Note the trailing "s" plural case
to disambiguate from \code{frame.plot}.}

\item{...}{further arguments passed to \code{tinyplot}.}
}
\value{
No return value, called for the side effect of producing a plot.
}
\description{
Convenience interface for visualizing
\code{\link[base]{data.frame}} objects with tinyplot.
}
\details{
This is a convenience function for plotting data frames with
or without a formula. The case with the formula mainly facilitates
using \code{tinyplot()} in combination with pipes. The case without
formula provides a quick way of plotting the variables in a data frame:
a single variable, a pair of variables, or a
\code{\link[graphics]{pairs}}-style grid of all variable combinations for
3 or more variables. The latter can be called directly (also for less than
3 variables) via \code{tinypairs()}.

Note that, unlike a base \code{\link[graphics]{pairs}} plot, these
\code{tinyplot} variants (deliberately) produce a histogram for the univariate
cases along the grid diagonal.
}
\examples{
## using tinyplot() with data frames
tinyplot(cars)
tinyplot(iris, Sepal.Length ~ Petal.Width | Species)
tinyplot(
  as.data.frame(UCBAdmissions),
  Admit ~ Gender, facet = 1 ~ Dept, weights = Freq
)

## note that this also enables usage with pipes (in R >= 4.1.0) such as
# cars |> tinyplot()
# iris |> tinyplot(Sepal.Length ~ Petal.Width | Species)
# UCBAdmissions |> 
#   as.data.frame() |>
#   tinyplot(Admit ~ Gender, facet = 1 ~ Dept, weights = Freq)
# etc.

## pairs-style display for data frames with 3 or more variables
## (works best with a theme)
tinytheme("clean2")
tinyplot(iris)

## pass `by` arg to group the pairs display (legend is suppressed)
## here, we also add optional frames around the individual sub-plots
tinyplot(iris, by = "Species", frames = TRUE)

## another option (but assumes objects of equal length)
tinyplot(iris[, 1:4], by = iris$Species, frames = TRUE)

## tinypairs can also create 2x2 pairs display
tinypairs(iris[, c(3, 1)], by = iris$Species)

tinytheme() ## reset theme

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_area.R, R/type_ribbon.R
\name{type_area}
\alias{type_area}
\alias{type_ribbon}
\title{Ribbon and area plot types}
\usage{
type_area(alpha = NULL)

type_ribbon(alpha = NULL, dodge = 0, fixed.dodge = FALSE)
}
\arguments{
\item{alpha}{numeric value between 0 and 1 specifying the opacity of ribbon shading
If no \code{alpha} value is provided, then will default to \code{tpar("ribbon.alpha")}
(i.e., probably \code{0.2} unless this has been overridden by the user in their global
settings.)}

\item{dodge}{Adjustment parameter for dodging overlapping points or ranges in
grouped plots along the x-axis (or y-axis for flipped plots). Either:
\itemize{
\item numeric value in the range \verb{[0,1)}. Note that values are scaled
relative to the spacing of x-axis breaks, e.g. \code{dodge = 0.1} places the
outermost groups one-tenth of the way to adjacent breaks, \code{dodge = 0.5}
places them midway between breaks, etc. Values < 0.5 are recommended.
\item logical. If \code{TRUE}, the dodge width is calculated automatically based on
the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
\code{FALSE} or 0, no dodging is performed.
}

Default value is 0 (no dodging). While we do not check, it is \emph{strongly}
recommended that dodging only be used in cases where the x-axis comprises a
limited number of discrete breaks.}

\item{fixed.dodge}{Logical. If \code{FALSE} (default), dodge positions are
calculated independently for each \code{x} value, based only on the groups
present at that position. If \code{TRUE}, dodge positions are based on all
groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
groups are missing for a particular \code{x} value).}
}
\description{
Type constructor functions for producing polygon ribbons, which
define a \code{y} interval (usually spanning from \code{ymin} to \code{ymax}) for each
\code{x} value. Area plots are a special case of ribbon plot where \code{ymin} is
set to 0 and \code{ymax} is set to \code{y}.
}
\section{Dodging ribbon plots}{


We support dodging for grouped ribbon plots, enabling similar functionality
to dodged errorbar and pointrange plots. However, it is strongly recommended
that dodging is only implemented for cases where the x-axis comprises a
limited number of discrete cases (e.g., coefficient or event-study plots).
See Examples.
}

\examples{
x = 1:100 / 10
y = sin(x)

#
## Ribbon plots

# "ribbon" convenience string
tinyplot(x = x, ymin = y - 1, ymax = y + 1, type = "ribbon")
# Same result with type_ribbon()
tinyplot(x = x, ymin = y-1, ymax = y+1, type = type_ribbon())

# y will be added as a line if it is specified
tinyplot(x = x, y = y, ymin = y-1, ymax = y+1, type = "ribbon")

#
## Area plots

# "area" type convenience string
tinyplot(x, y, type = "area")

# Same result with type_area()
tinyplot(x, y, type = type_area())

# Area plots are often used for time series charts
tinyplot(AirPassengers, type = "area")

#
## Dodged ribbon/area plots

# Dodged ribbon or area plots can be useful in cases where there is strong
# overlap across groups (and a limited number of discrete x-axis values).

dat = data.frame(
  x = rep(c("Before", "After"), each = 2),
  grp = rep(c("A", "B"), 2),
  y = c(10, 10.5, 15, 15.3),
  lwr = c(8, 8.5, 13, 13.3),
  upr = c(12, 12.5, 17, 17.3)
)

tinyplot(
  y ~ x | grp,
  data = dat,
  ymin = lwr, ymax = upr,
  type = type_ribbon(),
  main = "Overlappling ribbons"
)

tinyplot(
  y ~ x | grp,
  data = dat,
  ymin = lwr, ymax = upr,
  type = type_ribbon(dodge = 0.1),
  main = "Dodged ribbons"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend_gradient.R
\name{draw_gradient_swatch}
\alias{draw_gradient_swatch}
\title{Draw gradient (continuous) legend swatch}
\usage{
draw_gradient_swatch(
  legend_args,
  fklgnd,
  lmar,
  outer_side,
  outer_end,
  outer_right,
  outer_bottom,
  user_inset = FALSE
)
}
\arguments{
\item{legend_args}{Legend arguments list}

\item{fklgnd}{Fake legend object (from drawing with plot=FALSE)}

\item{lmar}{Legend margins}

\item{outer_side}{Logical flag for outer side placement}

\item{outer_end}{Logical flag for outer end placement}

\item{outer_right}{Logical flag for outer right placement}

\item{outer_bottom}{Logical flag for outer bottom placement}

\item{user_inset}{Logical flag indicating user-supplied inset}
}
\value{
NULL (draws gradient legend as side effect)
}
\description{
For gradient legends, we draw a custom color swatch using
grDevices::as.raster and add labels, tick marks, and title manually.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_loess.R
\name{type_loess}
\alias{type_loess}
\title{Local polynomial regression plot type}
\usage{
type_loess(
  span = 0.75,
  degree = 2,
  family = "gaussian",
  control = loess.control(),
  se = TRUE,
  level = 0.95,
  weights = NULL
)
}
\arguments{
\item{span}{the parameter \eqn{\alpha} which controls the degree of
    smoothing.}

\item{degree}{the degree of the polynomials to be used, normally 1 or
    2. (Degree 0 is also allowed, but see the \sQuote{Note}.)}

\item{family}{a description of the error distribution and link
    function to be used in the model.  For \code{glm} this can be a
    character string naming a family function, a family function or the
    result of a call to a family function.  For \code{glm.fit} only the
    third option is supported.  (See \code{\link[stats]{family}} for details of
    family functions.)}

\item{control}{control parameters: see \code{\link[stats]{loess.control}}.}

\item{se}{logical. If \code{TRUE} (the default), confidence intervals are drawn.}

\item{level}{the confidence level required if \code{se = TRUE}. Default is 0.95.}

\item{weights}{an optional numeric vector of observation weights for the model
fit, of the same length as the number of data points. Weights can also be
supplied via the top-level \code{weights} argument of \code{\link{tinyplot}} (which is
evaluated with non-standard evaluation in the formula method, and takes
precedence if both are given).}
}
\description{
Type function for plotting a LOESS (LOcal regrESSion) fit.
Arguments are passed to \code{\link[stats]{loess}}.
}
\examples{
# "loess" type convenience string
tinyplot(dist ~ speed, data = cars, type = "loess")

# Use `type_loess()` to pass extra arguments for customization
tinyplot(dist ~ speed, data = cars, type = type_loess(span = 0.5, degree = 1))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_chull.R
\name{type_chull}
\alias{type_chull}
\title{Convex hull plot type}
\usage{
type_chull(density = NULL, angle = 45)
}
\arguments{
\item{density}{the density of shading lines, in lines per inch.  The
    default value of \code{NULL} means that no shading lines are drawn.
    A zero value of \code{density} means no shading nor filling whereas
    negative values and \code{NA} suppress shading (and so allow
    color filling).}

\item{angle}{the slope of shading lines, given as an angle
     in degrees (counter-clockwise).}
}
\description{
Type function for drawing convex hulls around grouped points.
Uses \code{\link[grDevices]{chull}} to compute the convex hull of each
group and draws the result as a polygon.
}
\examples{
# "chull" type convenience character string
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, type = "chull")

# layer filled convex hull(s) on top of points
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic")
tinyplot_add(type = "chull", fill = 0.2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot_add.R
\name{tinyplot_add}
\alias{tinyplot_add}
\alias{plt_add}
\title{Add new elements to the current \code{tinyplot}}
\usage{
tinyplot_add(...)

plt_add(...)
}
\arguments{
\item{...}{All named arguments override arguments from the previous calls.
Arguments not supplied to \link{tinyplot_add} remain unchanged from the previous
call. Arguments are captured unevaluated and spliced into the updated call,
so those that rely on non-standard evaluation against \code{data}---e.g.
\code{subset = cyl == 4} or \code{weights = wt}---behave the same as in a direct
\code{\link{tinyplot}} call.}
}
\value{
No return value, called for side effect of producing a plot.
}
\description{
This convenience function grabs the preceding \code{tinyplot} call and updates it
with any new arguments that have been explicitly provided by the user. It
then injects \code{add=TRUE} and evaluates the updated call, thereby drawing a new
layer on top of the existing plot. \code{plt_add()} is a shorthand alias for
\code{tinyplot_add()}.
}
\section{Limitations}{

\itemize{
\item \code{tinyplot_add()} works reliably only when adding to a plot originally
created using the \code{\link{tinyplot.formula}} method with a valid \code{data} argument.
We cannot guarantee correct behavior if the original plot was created with
the atomic \code{\link{tinyplot.default}} method, due to potential environment
mismatches. (An exception is when the original plot arguments---\code{x}, \code{y},
etc.---are located in the global environment.)
\item Automatic legends for the added elements will be turned off.
}
}

\examples{
tinyplot(Sepal.Width ~ Sepal.Length | Species,
  facet = ~Species,
  data = iris)

tinyplot_add(type = "lm") ## or : plt_add(type = "lm")

## Note: the previous function is equivalent to (but much more convenient
## than) re-writing the full call with the new type and `add=TRUE`:

# tinyplot(Sepal.Width ~ Sepal.Length | Species,
#          facet = ~Species,
#          data = iris,
#          type = "lm",
#          add = TRUE)

## Arguments relying on non-standard evaluation (e.g. `subset`) work too:
tinyplot(mpg ~ wt, data = mtcars)
tinyplot_add(subset = cyl == 4, col = "red", pch = 16)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_text.R
\name{type_text}
\alias{type_text}
\title{Text annotations plot type}
\usage{
type_text(
  labels = NULL,
  labeller = NULL,
  adj = NULL,
  pos = NULL,
  offset = 0.5,
  family = NULL,
  font = NULL,
  vfont = NULL,
  xpd = NULL,
  srt = 0,
  repel = FALSE,
  clim = c(0.5, 2.5)
)
}
\arguments{
\item{labels}{Character vector of length \code{1} or the same length as the
number of \code{x},\code{y} coordinates. If left as \code{NULL}, then the labels will
automatically inherit the corresponding \code{y} values. Can also be supplied as
a top-level \code{\link{tinyplot}} argument, which additionally supports non-standard
evaluation against \code{data} and takes precedence if both are given. See
Examples.}

\item{labeller}{A formatting function (or convenience string) passed to
\code{\link{tinylabel}} for formatting the \code{labels}. Useful for ensuring that the
text labels match the formatting of an axis, e.g. \code{labeller = "\%"} to
display the labels as percentages. Default is \code{NULL}, i.e. no formatting.
See Examples.}

\item{adj}{one or two values in \eqn{[0, 1]} which specify the x
    (and optionally y) adjustment (\sQuote{justification}) of the
    labels, with 0 for left/bottom, 1 for right/top, and 0.5 for
    centered.  On most devices values outside \eqn{[0, 1]} will also
    work.  See below.}

\item{pos}{a position specifier for the text.  If specified this
    overrides any \code{adj} value given.  Values of \code{1},
    \code{2}, \code{3} and \code{4}, respectively indicate
    positions below, to the left of, above and to the right of
    the specified \code{(x,y)} coordinates.}

\item{offset}{when \code{pos} is specified, this value controls the
    distance (\sQuote{offset}) of the text label from the specified
    coordinate in fractions of a character width.}

\item{family}{The name of a font family. Default of \code{NULL} means that the
family will be the same as the main plot text, following
\code{\link[graphics]{par}}. Note that if a \code{family} argument is provided,
then \code{vfont} (below) will automatically be ignored.}

\item{font}{Integer giving the font face to be used,
following \code{\link[graphics]{par}}. On most devices, the mapping is: \code{1} =
regular, \code{2} = bold, \code{3} = italic, \code{4} = bold italic, and \code{5} = symbol.}

\item{vfont}{\code{NULL} for the current font family, or a character
    vector of length 2 for \code{\link[grDevices]{Hershey}} vector
    fonts.  The first element of
    the vector selects a typeface and the second element selects a
    style.  Ignored if \code{labels} is an expression.}

\item{xpd}{Logical value or \code{NA} denoting text clipping behaviour, following
\code{\link[graphics]{par}}.}

\item{srt}{Numeric giving the desired string rotation in degrees.}

\item{repel}{Logical or numeric controlling automatic repulsion of
overlapping text labels. The default \code{FALSE} draws the labels at their
exact (\code{x},\code{y}) coordinates. \code{TRUE} nudges overlapping labels apart using a
force-directed algorithm (labels are pushed off each other and then sprung
back toward their original positions). A numeric value does the same but
additionally enforces that much minimum padding (in user coordinates)
between labels. Works best with the default centered text placement (i.e.
without \code{pos}). \strong{Caveat:} The repulsion logic currently operates
within each group rather than across groups. So the text of different
groups may still overlap. See Examples.}

\item{clim}{Numeric giving the lower and upper limits of the character
expansion (\code{cex}) normalization for bubble charts.}
}
\description{
Type function for adding text annotations to a plot at the
specified (\code{x},\code{y}) coordinates.
}
\examples{
# simplest case (no labels), will auto revert to y labels
tinyplot(1:12, type = "text")

# pass explicit `labels` arg if you want specific text
tinyplot(1:12, type = "text", labels = month.abb)

# you can also use a labeller function (passed to `tinylabel`) to
# customize
tinyplot(1:12, type = "text", labels = month.abb, labeller = toupper)

# tip: use `xpd = NA` to avoid clipping text at the plot region
tinyplot(1:12, type = "text", labels = month.name, xpd = NA)

# for advanced customization, it's safer to pass args through `type_text()`
tinyplot(
  1:12,
  type = type_text(
  labels = month.abb,
  family = "HersheyScript",
  srt = -20)
)

# some formula + data.frame examples...

mt = transform(mtcars[1:20,], cyl = factor(cyl), model = rownames(mtcars[1:20,]))

# as a top-level argument, `labels` supports non-standard evaluation, so a
# bare column name is resolved against `data` (just like the plot variables)

# tinyplot(mpg ~ wt, data = mt, type = type_text(labels = mt$model))
tinyplot(mpg ~ wt, data = mt, type = "text", labels = model) # same

# tip: use `repel = TRUE` to automatically nudge overlapping labels apart
tinyplot(
  mpg ~ wt, data = mt,
  type = "text", labels = model,
  repel = TRUE, xpd = NA
)

# limitation: `repel` logic currently works per group, so grouped text data
# may still overlap
tinyplot(
  mpg ~ wt | cyl, data = mt,
  type = "text", labels = model,
  repel = TRUE, xpd = NA
)

# an advantage of the top-level `labels` approach is that data subsetting and 
# sanitizing ops pass through automatically
tinyplot(hp ~ wt, data = mt)
tinyplot_add(subset = carb==2, col = "red", pch = 16)
tinyplot_add(subset = carb==2, col = "red", type = "text", labels = model, pos = 4)

# fancier plot that combines a theme and some other customizations
tinyplot(
  mpg ~ hp | cyl,
  data = mt,
  legend = list(title = "Cylinders"),
  theme = "web"
)
tinyplot_add(
  labels = model,
  type = type_text(
    family = "HersheySans",
    font = 2,
    pos = 4,
    repel = TRUE,
    xpd = NA
  )
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinylabel.R
\name{tinylabel}
\alias{tinylabel}
\title{Format labels}
\usage{
tinylabel(x, labeller = NULL, na.ignore = TRUE, na.rm = TRUE)
}
\arguments{
\item{x}{a numeric or character vector}

\item{labeller}{a formatting function to be applied to \code{x}, e.g. \code{\link{format}},
\code{\link{toupper}}, \code{\link{abs}}, or other custom function (including from the popular
\strong{scales} package). Can also be one of the following convenience strings
(symbols), for which common formatting transformations are provided:
\code{"percent"} (\code{"\%"}), \code{"comma"} (\code{","}), \code{"log"} (\code{"l"}), \code{"dollar"}
(\code{"$"}), \code{"euro"} (\code{"€"}), or \code{"sterling"} (\code{"£"}).}

\item{na.ignore}{logical indicating whether the labelling function should
ignore \code{NA} values in \code{x}. In other words, should the \code{NA} values be left
as-is? Default is \code{TRUE}.}

\item{na.rm}{logical indicating whether \code{NA} values should be removed from
\code{x} and thus the return object too. Default is \code{TRUE}, but only evaluated
if \code{na.ignore} is \code{FALSE}.}
}
\value{
a character vector of the same length as \code{x} with the transformed
labels.
}
\description{
Function for formatting label appearance, e.g. axis
ticks labels. This is what the top-level \code{xaxl} and \code{yaxl} arguments
from \code{\link{tinyplot}} ultimately get passed to.
}
\examples{
x = 1e4
tinylabel(x, "comma")
tinylabel(x, ",") # same
tinylabel(x, "$") # or "dollar"

# invoke tinylabel from a parent tinyplot call...
#   => x/yaxl for adjusting axes tick labels
#   => legend = list(labeller = ...) for adjusting the legend labels
s77 = transform(data.frame(state.x77), Illiteracy = Illiteracy / 100)
tinyplot(Life.Exp ~ Income | Illiteracy, data = s77,
         xaxl = '$',
         legend = list(labeller = '\%'))

# log example (combined with axis scaling)
tinyplot(x = 10^c(10:0), y = 0:10, type = "b",
         log = "x", xaxl = "log")

# combine with `x/yaxb` to adjust the actual tick marks ("break points")
# at the same time
tinyplot(x = 10^c(10:0), y = 0:10, type = "b",
         log = "x", xaxl = "log", xaxb = 10^c(1,3,5,7,9))

#
## custom function examples

## example I: date formatting

dat = data.frame(
  date = seq(as.Date("2000/1/1"), by = "month", length.out = 12),
  trend = 1:12 + rnorm(12, sd = 1)
)

tinyplot(trend ~ date, data = dat,
         xaxl = function(x) format(x, "\%b, \%Y"))

## example II: string wrapping

# create a "vectorised" version of `base::strwrap` that breaks long
# strings into new lines every 18 characters
strwrap18 = function(x) sapply(
  strwrap(x, width = 18, simplify = FALSE),
  paste,
  collapse = "\n"
)

# now demonstrate on a dataset with long y-tick labels
dat2 = data.frame(
  x = rep(rnorm(100), 3),
  y = c(
    "tinyplot is a lightweight extension of the base R graphics system.",
    "R is a language for statistical computing.",
    "Data visualization is an essential skill."
  )
)

tinyplot(y ~ x, data = dat2, type = "j",
         yaxl = strwrap18,
         theme = "bw") # use theme for horizontal labels + dynamic margin
}
\seealso{
\code{\link{tinyplot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyAxis.R
\name{tinyAxis}
\alias{tinyAxis}
\title{Generic function for adding an axis to a (tiny)plot}
\usage{
tinyAxis(x = NULL, ..., type = "standard", labeller = NULL)
}
\arguments{
\item{x}{an object which indicates the range over which an axis should
    be drawn}

\item{...}{arguments to be passed to methods and perhaps then to
    \code{\link[graphics]{axis}}.}

\item{type}{the type of axis to be drawn; inherited from the \code{xaxt} or \code{yaxt}
arguments of the parent \code{\link[=tinyplot]{tinyplot()}} call. One of either: \code{"standard"}
(default that draws the axis, ticks, and labels), \code{"none"} (no axes),
\code{"ticks"} (only ticks and labels without axis line), \code{"labels"} (only
labels without ticks and axis line), or \code{"axis"} (only axis line and labels
but no ticks). Partial matching is allowed, e.g. \code{type = "s"}.}

\item{labeller}{a formatting function to be applied to \code{x}, e.g. \code{\link{format}},
\code{\link{toupper}}, \code{\link{abs}}, or other custom function (including from the popular
\strong{scales} package). Can also be one of the following convenience strings
(symbols), for which common formatting transformations are provided:
\code{"percent"} (\code{"\%"}), \code{"comma"} (\code{","}), \code{"log"} (\code{"l"}), \code{"dollar"}
(\code{"$"}), \code{"euro"} (\code{"€"}), or \code{"sterling"} (\code{"£"}).}
}
\description{
Internal function used for adding an axis to a \code{\link{tinyplot}}
call.
}
\details{
\code{tinyAxis} provides a thin(ish) wrapper around
\code{\link[graphics]{Axis}}, but with enhanced flexibility to (i) match
parameter combinations based on the axis type and plotting theme, (ii)
provide better support for date-time variables, and (iii) enable convenient
formatting of axis tick labels.
}
\examples{
\dontrun{

# plot without axes
tinyplot(0:10, axes = "n")
# add x-axis (labels only)
tinyplot:::tinyAxis(x = 0:10, side = 1, type = "l")
# add y-axis (with custom label formatting)
tinyplot:::tinyAxis(x = 0:10, side = 2, type = "s", labeller = "$")
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_errorbar.R, R/type_pointrange.R
\name{type_errorbar}
\alias{type_errorbar}
\alias{type_pointrange}
\title{Error bar and pointrange plot types}
\usage{
type_errorbar(length = 0.05, dodge = 0, fixed.dodge = FALSE)

type_pointrange(dodge = 0, fixed.dodge = FALSE)
}
\arguments{
\item{length}{length of the edges of the arrow head (in inches).}

\item{dodge}{Adjustment parameter for dodging overlapping points or ranges in
grouped plots along the x-axis (or y-axis for flipped plots). Either:
\itemize{
\item numeric value in the range \verb{[0,1)}. Note that values are scaled
relative to the spacing of x-axis breaks, e.g. \code{dodge = 0.1} places the
outermost groups one-tenth of the way to adjacent breaks, \code{dodge = 0.5}
places them midway between breaks, etc. Values < 0.5 are recommended.
\item logical. If \code{TRUE}, the dodge width is calculated automatically based on
the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
\code{FALSE} or 0, no dodging is performed.
}

Default value is 0 (no dodging). While we do not check, it is \emph{strongly}
recommended that dodging only be used in cases where the x-axis comprises a
limited number of discrete breaks.}

\item{fixed.dodge}{Logical. If \code{FALSE} (default), dodge positions are
calculated independently for each \code{x} value, based only on the groups
present at that position. If \code{TRUE}, dodge positions are based on all
groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
groups are missing for a particular \code{x} value).}
}
\description{
Type function(s) for producing error bar and pointrange plots.
}
\examples{
tinytheme("basic")

#
## Basic coefficient plot(s)

mod = lm(mpg ~ wt * factor(am), mtcars)
coefs = data.frame(names(coef(mod)), coef(mod), confint(mod))
colnames(coefs) = c("term", "est", "lwr", "upr")

# "errorbar" and "pointrange" type convenience strings
tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "errorbar")
tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "pointrange")

# Use `type_errorbar()` to pass extra arguments for customization
tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs,
         type = type_errorbar(length = 0.2))

#
## Flipped plots

# For flipped errobar / pointrange plots, it is recommended to use a dynamic
# theme that applies horizontal axis tick labels

tinytheme("classic")
tinyplot(est ~ term, ymin = lwr, ymax = upr, data = coefs, type = "errorbar",
         flip = TRUE)
tinyplot_add(type = 'vline', lty = 2)

tinytheme("basic") # back to basic theme for the remaining examples

#
## Dodging groups

models = list(
    "Model A" = lm(mpg ~ wt, data = mtcars),
    "Model B" = lm(mpg ~ wt + cyl, data = mtcars),
    "Model C" = lm(mpg ~ wt + cyl + hp, data = mtcars)
)

models = do.call(
  rbind,
  lapply(names(models), function(m) {
    data.frame(
      model = m,
      term = names(coef(models[[m]])),
      estimate = coef(models[[m]]),
      setNames(data.frame(confint(models[[m]])), c("conf.low", "conf.high"))
    )
  })
)

tinyplot(estimate ~ term | model,
         ymin = conf.low, ymax = conf.high,
         data = models,
         type = type_pointrange(dodge = 0.1))

# Aside 1: relative vs fixed dodge
#  The default dodge position is based on the unique groups (here: models)
#  available to each x value (here: coefficient term). To "fix" the dodge
#  position across all x values, use `fixed.dodge = TRUE`.

tinyplot(estimate ~ term | model,
         ymin = conf.low, ymax = conf.high,
         data = models,
         type = type_pointrange(dodge = 0.1, fixed.dodge = TRUE))

# Aside 2: layering
#  For layering on top of dodged plots, rather pass the dodging arguments
#  through the top-level call if you'd like the dodging behaviour to be
#  inherited automatically by the added layers.

tinyplot(estimate ~ term | model,
         ymin = conf.low, ymax = conf.high,
         data = models,
         type = "pointrange",
         dodge = 0.1, fixed.dodge = TRUE)
tinyplot_add(type = "l", lty = 2)

tinytheme() # reset theme

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_polygon.R
\name{type_polygon}
\alias{type_polygon}
\title{Polygon plot type}
\usage{
type_polygon(density = NULL, angle = 45)
}
\arguments{
\item{density}{the density of shading lines, in lines per inch.  The
    default value of \code{NULL} means that no shading lines are drawn.
    A zero value of \code{density} means no shading nor filling whereas
    negative values and \code{NA} suppress shading (and so allow
    color filling).}

\item{angle}{the slope of shading lines, given as an angle
     in degrees (counter-clockwise).}
}
\description{
Type function for plotting polygons.
Arguments are passed to \code{\link[graphics]{polygon}}.
}
\examples{
# "polygon" type convenience character string
tinyplot(1:9, c(2,1,2,1,NA,2,1,2,1), type = "polygon")

# Use `type_polygon()` to pass extra arguments for customization
tinyplot(1:9, c(2,1,2,1,NA,2,1,2,1), type = type_polygon(density = c(10, 20)))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_abline.R, R/type_hline.R, R/type_vline.R
\name{type_abline}
\alias{type_abline}
\alias{type_hline}
\alias{type_vline}
\title{Add straight lines to a plot}
\usage{
type_abline(a = 0, b = 1)

type_hline(h = 0)

type_vline(v = 0)
}
\arguments{
\item{a, b}{the intercept (default: \code{a} = 0) and slope (default: \code{b} = 1)
terms. Numerics of length 1, or equal to the number of groups or number of
facets (or the product thereof).}

\item{h}{y-value(s) for horizontal line(s). Numeric of length 1, or equal to
the number of groups or number of facets (or the product thereof).}

\item{v}{x-value(s) for vertical line(s). Numeric of length 1, or equal to
the number of groups or number of facets (or the product thereof).}
}
\description{
These functions add straight line(s) through the current plot.
}
\details{
While \code{type_abline}, \code{type_hline}, and \code{type_vline} can be called in a base
plot layer, we expect that they will typically be called as subsequent
layers via \code{\link{tinyplot_add}}.
}
\section{Recycling logic}{

The recycling behaviour of the line parameters (i.e., \code{a}, \code{b}, \code{h}, or \code{v})
is adaptive, depending on whether \code{by} or \code{facet} grouping is detected. While
this leads to different recycling scenarios, the underlying code logic
follows sensible heuristics designed to match user expectations.

Parameter lengths must equal one of four options:
\enumerate{
\item Single value (i.e., length = 1), i.e. simplest case where the same line is
applied uniformly across all groups and facets. Uses the default user colour
(e.g. \code{"black"}, or \code{tpar("palette.qualitative")[1]} if a theme is set).
\item Number of \code{by} groups, i.e. one parameter per group. For example,
\code{tinyplot(mpg ~ wt | factor(cyl), data = mtcars, type = type_hline(h = 21:23))}
will give three horizontal lines, with colours matching the user's qualitative
palette.
\item Number of \code{facet} groups, i.e. one parameter per facet panel. For example:
\code{tinyplot(mpg ~ wt, facet = ~am, data = mtcars, type = type_hline(h = c(20,30)))}
would give separate horizontal lines per facet, but both using the same
default color.
\item Product of \code{by} and \code{facet} groups, i.e. one parameter for each unique
by-facet combination. Orders over facets first and then, within that, by
group. For example:
\code{tinyplot(mpg ~ wt | factor(cyl), facet = ~am, data = mtcars, type = type_hline(h = 21:26))}
will give six separate lines, with the first three (\code{21:23}) coloured by
group in the first facet, and second three (\code{24:26}) coloured by by group
in the second facet.
}

Alongside these general rules, we also try to accommodate special cases when
other aesthetic parameters like \code{lwd} or \code{lty} are invoked by the user. See
Examples.
}

\examples{
#
## abline

tinyplot(x = -10:10, y = rnorm(21) + -10:10, grid = TRUE)
tinyplot_add(type = "abline")
# same as...
# tinyplot_add(type = type_abline(a = 0, b = 1))

# customize by passing bespoke intercept and slope values
tinyplot_add(type = type_abline(a = -1, b = -0.5))

# note that calling as abline & co. as a base plot layer will still lead to
# axes limits that respect the range of the data
tinyplot(x = -10:10, y = -10:10, grid = TRUE, type = "abline")

#
## hline and vline

# Base plot layer
tinyplot(mpg ~ hp | cyl, facet = "by", data = mtcars, ylim = c(0, 40))

# Add horizontal lines at the (default) 0 y-intercept
tinyplot_add(type = "hline", col = "grey")

# Note that group+facet aesthetics will be inherited. We can use this to
# add customized lines (here: the mean `mpg` for each `cyl` group)
tinyplot_add(type = type_hline(with(mtcars, tapply(mpg, cyl, mean))), lty = 2)

# Similar idea for vline
tinyplot_add(type = type_vline(with(mtcars, tapply(hp, cyl, mean))), lty = 2)

#
## Recycling logic

# length(h) == no. of groups
tinyplot(mpg ~ wt | factor(cyl), data = mtcars, type = type_hline(h = 21:23))

# length(h) == no. of facets
tinyplot(mpg ~ wt, facet = ~am, data = mtcars, type = type_hline(h = c(20, 30)))

# length(h) == no. of groups x no. of facets
tinyplot(mpg ~ wt | factor(cyl),
  facet = ~am, data = mtcars,
  type = type_hline(h = 21:26))

# special adjustment case (here: lwd by group)
tinyplot(mpg ~ wt | factor(cyl),
  facet = ~am, data = mtcars,
  type = type_hline(c(20, 30)), lwd = c(21, 14, 7))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend_multi.R
\name{draw_multi_legend}
\alias{draw_multi_legend}
\title{Draw multiple legends with automatic positioning}
\usage{
draw_multi_legend(legend_list, position = "right!")
}
\arguments{
\item{legend_list}{A list of legend arguments, where each element is itself a
list of arguments that can be passed on to \link{draw_legend}. Legends will be
drawn vertically (top to bottom) in the order that they are provided. Note
that we currently only support 2 legends, i.e. the top-level list has a
maximum length of 2.}

\item{position}{String indicating the base keyword position for the
multi-legend. Currently only \code{"right!"} and \code{"left!"} are supported.}
}
\value{
No return value, called for side effect of drawing multiple legends.
}
\description{
Handles multiple legends (e.g., color grouping + bubble size) by:
\enumerate{
\item Extracting dimensions from fake legends
\item Calculating sub-positioning based on dimensions
\item Drawing legends in ascending order of width (widest last)
}
}
\examples{
\dontrun{
oldmar = par("mar")

# Multi-legend example (color + bubble)

l1 = list(
  lgnd_labs = c("Red", "Blue", "Green"),
  legend_args = list(title = "Colors"),
  pch = 16,
  col = c("red", "blue", "green"),
  type = "p"
)

l2 = list(
  lgnd_labs = c("Tiny", "Small", "Medium", "Large", "Huge"),
  legend_args = list(title = "Size"),
  pch = 16,
  col = "black",
  cex = seq(0.5, 2.5, length.out = 5),
  type = "p"
)

# Draw together
draw_multi_legend(list(l1, l2), position = "right!")

par(mar = oldmar)
}

}
\seealso{
\link{draw_legend}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_violin.R
\name{type_violin}
\alias{type_violin}
\title{Violin plot type}
\usage{
type_violin(
  bw = "nrd0",
  joint.bw = c("mean", "full", "none"),
  adjust = 1,
  kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight",
    "cosine", "optcosine"),
  n = 512,
  trim = FALSE,
  width = 0.9,
  lighten = TRUE
)
}
\arguments{
\item{bw}{the smoothing \code{\link[stats:bw.nrd]{bandwidth}} to be used,
see \code{\link[stats]{density}} for details and options.}

\item{joint.bw}{character string indicating whether (and how) the smoothing
bandwidth should be computed from the joint data distribution when there
are multiple subgroups. The options are \code{"mean"} (the default), \code{"full"},
and \code{"none"}. Also accepts a logical argument, where \code{TRUE} maps to
\code{"mean"} and \code{FALSE} maps to \code{"none"}. See the "Bandwidth selection"
section below for a discussion of practical considerations.}

\item{adjust}{the bandwidth used is actually \code{adjust*bw}.
    This makes it easy to specify values like \sQuote{half the default}
    bandwidth.}

\item{kernel}{a character string giving the smoothing kernel to be used. This
must partially match one of \code{"gaussian"}, \code{"rectangular"}, \code{"triangular"},
\code{"epanechnikov"}, \code{"biweight"}, \code{"cosine"} or \code{"optcosine"}, with default
\code{"gaussian"}, and may be abbreviated to a unique prefix (single letter).

\code{"cosine"} is smoother than \code{"optcosine"}, which is the usual 'cosine'
kernel in the literature and almost MSE-efficient. However, \code{"cosine"} is
the version used by S.}

\item{n}{the number of equally spaced points at which the density is
    to be estimated.  When \code{n > 512}, it is rounded up to a power
    of 2 during the calculations (as \code{\link[stats]{fft}} is used) and the
    final result is interpolated by \code{\link[stats]{approx}}.  So it almost
    always makes sense to specify \code{n} as a power of two.
  }

\item{trim}{logical indicating whether the violins should be trimmed to the
range of the data. Default is \code{FALSE}.}

\item{width}{numeric (ideally in the range \verb{[0, 1]}, although this isn't
enforced) giving the normalized width of the individual violins.}

\item{lighten}{logical. Should the fills use a lighter, opaque tint of the
series colour(s)? Default is \code{TRUE}, which keeps single- and multi-group
displays consistent and lets the fill read cleanly over grid lines. Set to
\code{FALSE} to use the fully-saturated palette colour(s) instead.}
}
\description{
Type function for violin plots, which are an alternative to box
plots for visualizing continuous distributions (by group) in the form of
mirrored densities.
}
\details{
See \code{\link{type_density}} for more details and considerations related to
bandwidth selection and kernel types.
}
\examples{
# "violin" type convenience string
tinyplot(weight ~ feed, data = chickwts, type = "violin")

# to match the defaults of `ggplot2::geom_violin()`, use `trim = TRUE` and
# `joint.bw = FALSE`
tinyplot(
  weight ~ feed, data = chickwts,
  # type = type_violin(trim = TRUE, joint.bw = FALSE) # same but see final ex.
  type = "violin", trim = TRUE, joint.bw = FALSE
)

# For flipped violin plots, it's usually better to use a dynamic theme to
# accommodate (horizontal) y-axis labels
tinyplot(
  weight ~ feed, data = chickwts, type = "violin", flip = TRUE,
  theme = "dynamic" # or "clean(2)", "classic", "minimal", etc.
)

# you can group by the x var to add colour (here with the original orientation)
tinyplot(weight ~ feed | feed, data = chickwts, type = "violin", legend = FALSE)

# dodged grouped violin plot example (different dataset)
tinyplot(len ~ dose | supp, data = ToothGrowth, type = "violin")

# note: above we relied on `...` argument passing alongside the "violin"
# type convenience string. But this won't work for `width`, since it will
# clash with the top-level `tinyplot(..., width = <width>)` arg. To ensure
# correct arg passing, it's safer to use the functional `type_violin()` type.
tinyplot(
  len ~ dose | supp, data = ToothGrowth,
  type = type_violin(width = 0.75)
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{draw_legend}
\alias{draw_legend}
\title{Calculate placement and draw legend}
\usage{
draw_legend(
  legend = NULL,
  legend_args = NULL,
  by_dep = NULL,
  lgnd_labs = NULL,
  labeller = NULL,
  type = NULL,
  pch = NULL,
  lty = NULL,
  lwd = NULL,
  col = NULL,
  bg = NULL,
  cex = NULL,
  gradient = FALSE,
  lmar = NULL,
  has_sub = FALSE,
  has_cap = FALSE,
  cap_text = NULL,
  new_plot = TRUE,
  draw = TRUE,
  soma_target = NULL,
  dynmar_title_mar = NULL
)
}
\arguments{
\item{legend}{Legend placement keyword or list, passed down from \link{tinyplot}.}

\item{legend_args}{Additional legend arguments to be passed to
\code{\link[graphics]{legend}}.}

\item{by_dep}{The (deparsed) "by" grouping variable name.}

\item{lgnd_labs}{The labels passed to \code{legend(legend = ...)}.}

\item{labeller}{Character or function for formatting the labels (\code{lgnd_labs}).
Passed down to \code{\link{tinylabel}}.}

\item{type}{Plotting type(s), passed down from \link{tinyplot}.}

\item{pch}{Plotting character(s), passed down from \link{tinyplot}.}

\item{lty}{Plotting linetype(s), passed down from \link{tinyplot}.}

\item{lwd}{Plotting line width(s), passed down from \link{tinyplot}.}

\item{col}{Plotting colour(s), passed down from \link{tinyplot}.}

\item{bg}{Plotting character background fill colour(s), passed down from \link{tinyplot}.}

\item{cex}{Plotting character expansion(s), passed down from \link{tinyplot}.}

\item{gradient}{Logical indicating whether a continuous gradient swatch
should be used to represent the colors.}

\item{lmar}{Legend margins (in lines). Should be a numeric vector of the form
\code{c(inner, outer)}, where the first number represents the "inner" margin
between the legend and the plot, and the second number represents the
"outer" margin between the legend and edge of the graphics device. If no
explicit value is provided by the user, then reverts back to \code{tpar("lmar")}
for which the default values are \code{c(1.0, 0.1)}.}

\item{has_sub}{Logical. Does the plot have a sub-caption. Only used if
keyword position is "bottom!", in which case we need to bump the legend
margin a bit further.}

\item{has_cap}{Logical. Does the plot have a caption. Only used if
keyword position is "bottom!", in which case we need to bump the legend
margin a bit further.}

\item{cap_text}{Character. The caption text to draw below the legend when
position is "bottom!". Ignored otherwise.}

\item{new_plot}{Logical. Should we be calling plot.new internally?}

\item{draw}{Logical. If \code{FALSE}, no legend is drawn but the sizes are
returned. Note that a new (blank) plot frame will still need to be started
in order to perform the calculations.}

\item{soma_target}{Numeric. Shared outer margin target (in lines) for
multi-legend alignment. If \code{NULL}, each legend computes its own margin.}

\item{dynmar_title_mar}{Numeric or \code{NULL}. The pre-computed \code{dynmar_computed[3]}
value for "top!" legends under dynmar themes. When set, the legend margin
formula uses this directly to ensure correct title positioning.}
}
\value{
No return value, called for side effect of producing a(n empty) plot
with a legend in the margin.
}
\description{
Main exported function for drawing legends. Supports:
\itemize{
\item Inner and outer positioning (with "!" suffix)
\item Discrete and continuous (gradient) legends
\item Automatic margin adjustment
}
}
\examples{
oldmar = par("mar")

draw_legend(
  legend = "right!", ## default (other options incl, "left(!)", ""bottom(!)", etc.)
  legend_args = list(title = "Key", bty = "o"),
  lgnd_labs = c("foo", "bar"),
  type = "p",
  pch = 21:22,
  col = 1:2
)

# The legend is placed in the outer margin...
box("figure", col = "cyan", lty = 4)
# ... and the plot is proportionally adjusted against the edge of this
# margin.
box("plot")
# You can add regular plot objects per normal now
plot.window(xlim = c(1,10), ylim = c(1,10))
points(1:10)
points(10:1, pch = 22, col = "red")
axis(1); axis(2)
# etc.

# Important: A side effect of draw_legend is that the inner margins have been
# adjusted. (Here: The right margin, since we called "right!" above.)
par("mar")

# To reset you should call `dev.off()` or just reset manually.
par(mar = oldmar)

# Note that the inner and outer margin of the legend itself can be set via
# the `lmar` argument. (This can also be set globally via
# `tpar(lmar = c(inner, outer))`.)
draw_legend(
  legend_args = list(title = "Key", bty = "o"),
  lgnd_labs = c("foo", "bar"),
  type = "p",
  pch = 21:22,
  col = 1:2,
  lmar = c(0, 0.1) ## set inner margin to zero
)
box("figure", col = "cyan", lty = 4)

par(mar = oldmar)

# Continuous (gradient) legends are also supported
draw_legend(
  legend = "right!",
  legend_args = list(title = "Key"),
  lgnd_labs = LETTERS[1:5],
  col = hcl.colors(5),
  gradient = TRUE ## enable gradient legend
)

par(mar = oldmar)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facet.R
\name{draw_facet_window}
\alias{draw_facet_window}
\alias{facet_layout}
\title{Draw facet windows}
\usage{
draw_facet_window(
  add,
  cex_fct_adj,
  facet.args,
  facet_newlines,
  facet_font,
  facet_rect,
  facet_text,
  facet_col,
  facet_bg,
  facet_border,
  facet,
  facets,
  ifacet,
  nfacets,
  nfacet_cols,
  nfacet_rows,
  axes,
  flip,
  frame.plot,
  oxaxis,
  oyaxis,
  xlabs,
  xlim,
  null_xlim,
  xaxt,
  xaxs,
  xaxb,
  xaxl,
  ylabs,
  ylim,
  null_ylim,
  yaxt,
  yaxs,
  yaxb,
  yaxl,
  rev_x = FALSE,
  rev_y = FALSE,
  asp,
  log,
  dots,
  draw,
  grid,
  has_legend,
  main,
  sub,
  cap,
  type,
  xlab,
  x,
  xmax,
  xmin,
  ylab,
  y,
  ymax,
  ymin,
  tpars = NULL,
  dynmar_computed = NULL,
  dl_overshoot = 0
)

facet_layout(settings)
}
\arguments{
\item{settings}{A list of settings as created by \code{tinyplot()}.}
}
\description{
Internal functions called from \code{tinyplot} in order to draw the
plot window with different facets, grids, axes, etc.

\code{facet_layout} determines the layout of the facets, based on a set of inputs.

\code{draw_facet_window} is the main workhorse function for setting the exterior
plot elements as part of a \code{tinyplot} call, including adjustment of margins
for dynamic themes, etc.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_qq.R
\name{type_qq}
\alias{type_qq}
\title{Quantile-Quantile plot (QQ)}
\usage{
type_qq(distribution = qnorm)
}
\arguments{
\item{distribution}{Distribution function to use.}
}
\description{
Plots the theoretical quantiles of \code{x} on the horizontal axis
against observed values of \code{x} on the vertical axis.
}
\examples{
tinyplot(~mpg, data = mtcars, type = type_qq())

# suppress the line
tinyplot(~mpg, data = mtcars, lty = 0, type = type_qq())
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_histogram.R
\name{type_histogram}
\alias{type_histogram}
\alias{type_hist}
\title{Histogram plot type}
\usage{
type_histogram(
  breaks = "Sturges",
  freq = NULL,
  right = TRUE,
  free.breaks = FALSE,
  drop.zeros = TRUE
)

type_hist(
  breaks = "Sturges",
  freq = NULL,
  right = TRUE,
  free.breaks = FALSE,
  drop.zeros = TRUE
)
}
\arguments{
\item{breaks}{Passed to \code{\link[graphics]{hist}}. One of:
\itemize{
\item a vector giving the breakpoints between histogram cells,
\item a function to compute the vector of breakpoints,
\item a single number giving the number of cells for the histogram,
\item a character string naming an algorithm to compute the number of cells (see ‘Details’ of \code{\link[graphics]{hist}}),
\item a function to compute the number of cells.
In the last three cases the number is a suggestion only; as the breakpoints
will be set to pretty values, the number is limited to 1e6 (with a warning if
it was larger). If breaks is a function, the x vector is supplied to it as
the only argument (and the number of breaks is only limited by the amount of
available memory).
}}

\item{freq}{logical; if \code{TRUE}, the histogram graphic is a
    representation of frequencies, the \code{counts} component of
    the result; if \code{FALSE}, probability densities, component
    \code{density}, are plotted (so that the histogram has a total area
    of one).  Defaults to \code{TRUE} \emph{if and only if} \code{breaks} are
    equidistant (and \code{probability} is not specified).}

\item{right}{logical; if \code{TRUE}, the histogram cells are
    right-closed (left open) intervals.}

\item{free.breaks}{Logical indicating whether the breakpoints should be
computed separately for each group or facet? Default is \code{FALSE}, meaning
that the breakpoints are computed from the full dataset; thus ensuring
common bin widths across each group/facet. Can also use \code{free} as an
acceptable argument alias. Ignored if there are no groups and/or facets.}

\item{drop.zeros}{Logical indicating whether bins with zero counts should be
dropped before plotting. Default is \code{TRUE}. Note that switching to \code{FALSE}
may interfere with faceted plot behaviour if \code{facet.args = list(free)},
since the \code{x} variable is effectively recorded over the full range of the
x-axis (even if it does not extend over this range for every group).}
}
\description{
Type function for histogram plots. \code{type_hist} is an alias for
\code{type_histogram}.
}
\examples{
# "histogram"/"hist" type convenience string(s)
tinyplot(Nile, type = "histogram")

# Use `type_histogram()` to pass extra arguments for customization
tinyplot(Nile, type = type_histogram(breaks = 30))
tinyplot(Nile, type = type_histogram(breaks = 30, freq = FALSE))
# etc.

# Grouped histogram example
tinyplot(
    ~ Petal.Width | Species,
    type = "histogram",
    data = iris
)

# Faceted version
tinyplot(
    ~Petal.Width,
    facet = ~Species,
    type = "histogram",
    data = iris
)

# For visualizing faceted histograms across varying scales, you may also wish
# to impose free histogram breaks too (i.e., calculate breaks separately for
# each group). Compare:

# free facet scales + shared histogram breaks, versus...
tinyplot(
    ~Petal.Width,
    facet = ~Species,
    facet.args = list(free = TRUE),
    type = type_histogram(),
    data = iris
)
# ... free facet scales + free histogram breaks
tinyplot(
    ~Petal.Width,
    facet = ~Species,
    facet.args = list(free = TRUE),
    type = type_histogram(free = TRUE),
    data = iris
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot.ts.R
\name{tinyplot.ts}
\alias{tinyplot.ts}
\title{tinyplot Method for Plotting ts Objects (Time Series)}
\usage{
\method{tinyplot}{ts}(
  x,
  facet,
  type = "l",
  facet.args = list(free = TRUE),
  xlab = NULL,
  ylab = NA,
  ...
)
}
\arguments{
\item{x}{an object of class \code{"ts"}.}

\item{facet}{specification of \code{facet} for \code{tinyplot.formula}. The
default in the \code{tinyplot} method is to use \code{facet = NULL} for univariate
series and \code{facet = ~ Series} (equivalent to \code{facet = "by"}) for
multivariate series.}

\item{type, facet.args, xlab, ylab, ...}{further arguments passed to \code{tinyplot}.}
}
\value{
No return value, called for the side effect of producing a plot.
}
\description{
Convenience interface for visualizing \code{\link[stats]{ts}}
(time series) objects with tinyplot.
}
\details{
Internally the time series object is converted to a long
data frame with columns \code{Time} (time index), \code{Value} (observations),
and \code{Series} (factor with column labels). Depending on the settings
of \code{facet} this data frame is visualized either with the formula
\code{Value ~ Time} or \code{Value ~ Time | Series}. See the \code{facet} argument
description for more details and the Examples below for some illustrations.

An exception is made if the user explicitly supplies a distribution \code{type}
argument (i.e, one of \code{"histogram"}, \code{"density"}, \code{"boxplot"}, \code{"violin"},
\code{"qq"}, \code{"ridge"}, or \code{"rug"}). These summarize the series values and
ignore the time index, so they are visualized with the one-sided
formula \code{~ Value} (or \code{~ Value | Series} for multivariate series). This
preserves base-R-compatible behaviour such as
\code{tinyplot(Nile, type = "histogram")}.
}
\examples{
## univariate series
tinyplot(Nile)

# exception: expicitly passing a distribution type on a univariate series
# still triggers the corresponding transformation
tinyplot(Nile, type = "histogram")

## multivariate series (generally, these also look better with a theme)
tinytheme("clean2")
tinyplot(EuStockMarkets)                    ## faceted, free scales, same color
tinyplot(EuStockMarkets, facet.args = NULL) ## faceted, same scale, same color
tinyplot(EuStockMarkets, facet = "by")      ## faceted, free scales, diff colors
tinyplot(EuStockMarkets, facet = NULL)      ## single frame, diff colors

## further variations
tinyplot(EuStockMarkets, facet = NULL, legend = list("direct", repel = TRUE))
tinyplot(EuStockMarkets, facet = "by", facet.args = NULL)
tinyplot(EuStockMarkets, facet.args = list(free = TRUE, ncol = 1))

## pass additional tinyplot args through `...` for further customization
tinyplot(EuStockMarkets,
         facet.args = NULL,
         type = "area", xlab = NA, yaxl = ",",
         main = "European stock indices")

tinytheme() ## reset

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_function.R
\name{type_function}
\alias{type_function}
\title{Plot a function}
\usage{
type_function(fun = dnorm, args = list(), n = 101, ...)
}
\arguments{
\item{fun}{Function of \code{x} to plot. Defaults to \code{\link[stats]{dnorm}}.}

\item{args}{List of additional arguments to be passed to \code{fun}.}

\item{n}{Number of points to interpolate on the x axis.}

\item{...}{Additional arguments are passed to the \code{lines()} function,
ex: \code{type="p"}, \code{col="pink"}.}
}
\description{
Plot a function
}
\details{
When using \code{type_function()} in a \code{tinyplot()} call, the \code{x} value indicates
the range of values to plot on the x-axis.
}
\examples{
# Plot the normal density (default function)
tinyplot(x = -4:4, type = "function")
# tinyplot(x = -4:4, type = type_function()) # same

# Customize by passing explicit arguments to your function
tinyplot(x = -1:10, type = type_function(
  fun = dnorm, args = list(mean = 3)
))

# Additional arguments are passed to the `lines()` function.
tinyplot(x = -4:4, type = type_function(
  fun = dnorm,
  col = "pink", type = "p", pch = 3
))

# Custom function example
## (Here using `function(x)`, but you could also use the shorter `\(x)`
## anonymous function syntax introduced in R 4.1.0)
tinyplot(x = -4:4, type = type_function(fun = function(x) 0.5 * exp(-abs(x))))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_ridge.R
\name{type_ridge}
\alias{type_ridge}
\title{Ridge plot type}
\usage{
type_ridge(
  scale = 1.5,
  joint.max = c("all", "facet", "by"),
  breaks = NULL,
  probs = NULL,
  ylevels = NULL,
  bw = "nrd0",
  joint.bw = c("mean", "full", "none"),
  adjust = 1,
  kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight",
    "cosine", "optcosine"),
  n = 512,
  gradient = FALSE,
  raster = FALSE,
  col = NULL,
  alpha = NULL
)
}
\arguments{
\item{scale}{Numeric. Controls the scaling factor of each plot.
Values greater than 1 means that plots overlap.}

\item{joint.max}{character indicating how to scale the maximum of the densities:
The default \code{"all"} indicates that all densities are scaled jointly relative to
the same maximum so that the areas of all densities are comparable.
Alternatively, \code{"facet"} indicates that the maximum is computed within
each facet so that the areas of the densities are comparable within each
facet but not necessarily across facets. Finally, \code{"by"} indicates that
each row (in each facet) is scaled separately, so that the areas of the
densities for \code{by} groups in the same row are comparable but not necessarily
across rows.}

\item{breaks}{Numeric. If a color gradient is used for shading, the
breaks between the colors can be modified. The default is to use
equidistant breaks spanning the range of the \code{x} variable.}

\item{probs}{Numeric. Instead of specifying the same \code{breaks} on the
x-axis for all groups, it is possible to specify group-specific quantiles
at the specified \code{probs}. The quantiles are computed based on the density
(rather than the raw original variable). Only one of \code{breaks} or
\code{probs} must be specified.}

\item{ylevels}{a character or numeric vector specifying in which order
the levels of the y-variable should be plotted.}

\item{bw}{the smoothing \code{\link[stats:bw.nrd]{bandwidth}} to be used,
see \code{\link[stats]{density}} for details and options.}

\item{joint.bw}{character string indicating whether (and how) the smoothing
bandwidth should be computed from the joint data distribution. The default
of \code{"mean"} will compute the joint bandwidth as the mean of the individual
subgroup bandwidths (weighted by their number of observations). Choosing
\code{"full"} will result in a joint bandwidth computed from the full
distribution (merging all subgroups). For \code{"none"} the individual bandwidth
will be computed independently for each subgroup. Also accepts a logical
argument, where \code{TRUE} maps to \code{"mean"} and \code{FALSE} maps to \code{"none"}. See
\code{\link{type_density}} for some discussion of practical considerations.}

\item{adjust}{the bandwidth used is actually \code{adjust*bw}.
    This makes it easy to specify values like \sQuote{half the default}
    bandwidth.}

\item{kernel}{a character string giving the smoothing kernel to be used. This
must partially match one of \code{"gaussian"}, \code{"rectangular"}, \code{"triangular"},
\code{"epanechnikov"}, \code{"biweight"}, \code{"cosine"} or \code{"optcosine"}, with default
\code{"gaussian"}, and may be abbreviated to a unique prefix (single letter).

\code{"cosine"} is smoother than \code{"optcosine"}, which is the usual 'cosine'
kernel in the literature and almost MSE-efficient. However, \code{"cosine"} is
the version used by S.}

\item{n}{the number of equally spaced points at which the density is
    to be estimated.  When \code{n > 512}, it is rounded up to a power
    of 2 during the calculations (as \code{\link[stats]{fft}} is used) and the
    final result is interpolated by \code{\link[stats]{approx}}.  So it almost
    always makes sense to specify \code{n} as a power of two.
  }

\item{gradient}{Logical or character. Should a gradient fill be used to
shade the area under the density? If a character specification is used,
then it can either be of length 1 and specify the palette to be used with
\code{gradient = TRUE} corresponding to \code{gradient = "viridis"}. If a character
vector of length greater than 1 is used, then it should specify the
colors in the palette, e.g., \code{gradient = hcl.colors(512)}.}

\item{raster}{Logical. Should the \code{gradient} fill be drawn using
\code{\link[graphics]{rasterImage}}? Defaults to \code{FALSE}, in which case the
\code{gradient} fill will instead be drawn using
\code{\link[graphics]{polygon}}. See the \verb{Technical note on gradient fills}
section below.}

\item{col}{Character string denoting the outline (border) color for all
of the ridge densities. Note that a singular value is expected; if multiple
colors are provided then only the first will be used. This argument is mostly
useful for the aesthetic effect of drawing a common outline color in
combination with gradient fills. See Examples. If left unset here, the
outline color can also be supplied via the top-level \code{tinyplot(..., col =)}
call (the \code{type_ridge(col =)} argument takes precedence if both are given).
When neither is supplied, the outline defaults to the active theme's
\code{col.default} (see \link{tpar}), falling back to the first color of the theme's
qualitative palette, or black if no theme is set.}

\item{alpha}{Numeric in the range \verb{[0,1]} for adjusting the alpha
transparency of the density fills. In most cases, will default to a value of
1, i.e. fully opaque. But for some \code{by} grouped plots (excepting the special
cases where \code{by==y} or \code{by==x}), will default to 0.6.}
}
\description{
Type function for producing ridge plots (also known as joy plots),
which display density distributions for multiple groups with vertical offsets.
This function uses \code{tinyplot} scaffolding, which enables added functionality
such as grouping and faceting.

The line color is controlled by the \code{col} argument in the \code{tinyplot()} call.
The fill color is controlled by the \code{bg} argument in the \code{tinyplot()} call.
}
\section{Technical note on gradient fills}{


\code{tinyplot} uses two basic approaches for drawing gradient fills in ridge line
plots, e.g., if \code{type_ridge(gradient = TRUE)}.

The first (and default) polygon-based approach involves dividing up the main
density region into many smaller polygons along the x-axis. Each of these
smaller polygons inherits a different color "segment" from the underlying
palette swatch, which in turn creates the effect of a continuous gradient
when they are all plotted together. Internally, this polygon-based approach
is vectorized (i.e., all of the sub-polygons are plotted simultaneously). It
is thus efficient from a plotting perspective and generally also performs
well from an aesthetic perspective. However, it can occasionally produce
undesirable plotting artifacts on some graphics devices---e.g., thin but
visible vertical lines---if alpha transparency is being used at the same
time.

For this reason, we also offer an alternative raster-based approach for
gradient fills that users can invoke via
\code{type_ridge(gradient = TRUE, raster = TRUE)}. The essential idea is that we
coerce the density polygon into a raster representation (using
\code{\link[graphics]{rasterImage}}) and achieve the gradient effect via
color interpolation. The trade-off this time is potential smoothness
artifacts around the top of the ridge densities at high resolutions, since we
have converted a vector object into a raster object.

Again, we expect that the choice between these two approaches will only
matter for ridge plots that combine gradient fills with alpha transparency
(and on certain graphics devices). We recommend that users experiment to
determine which approach is optimal for their device.
}

\examples{
aq = transform(
  airquality,
  Month = factor(month.abb[Month], levels = month.abb[5:9]),
  Month2 = factor(month.name[Month], levels = month.name[5:9]),
  Late = ifelse(Day > 15, "Late", "Early")
  )

# default ridge plot (using the "ridge" convenience string)
tinyplot(Month ~ Temp, data = aq, type = "ridge")

# for ridge plots, we recommend pairing with the dedicated theme(s), which
# facilitate nicer y-axis labels, grid lines, etc.

tinytheme("ridge")
tinyplot(Month ~ Temp, data = aq, type = "ridge")

tinytheme("ridge2") # removes the plot frame (but keeps x-axis line)
tinyplot(Month ~ Temp, data = aq, type = "ridge")

# the "ridge(2)" themes are especially helpful for long y labels, due to
# dynamic plot adjustment
tinyplot(Month2 ~ Temp, data = aq, type = "ridge")

# pass customization arguments through type_ridge()... for example, use
# the scale argument to change/avoid overlap of densities (more on scaling
# further below)

tinyplot(Month ~ Temp, data = aq, type = type_ridge(scale = 1))

## by grouping is also supported. two special cases of interest:

# 1) by == y (color by y groups)
tinyplot(Month ~ Temp | Month, data = aq, type = "ridge")

# 2) by == x (gradient coloring along x)
tinyplot(Month ~ Temp | Temp, data = aq, type = "ridge")

# aside: pass explicit `type_ridge(col = <col>)` arg to set a different
# border color
tinyplot(Month ~ Temp | Temp, data = aq, type = type_ridge(col = "white"))

# gradient coloring along the x-axis can also be invoked manually without
# a legend (the next two tinyplot calls are equivalent)

# tinyplot(Month ~ Temp, data = aq, type = type_ridge(gradient = "agsunset"))
tinyplot(Month ~ Temp, data = aq, type = type_ridge(gradient = TRUE))

# aside: when combining gradient fill with alpha transparency, it may be
# better to use the raster-based approach (test on your graphics device)

tinyplot(Month ~ Temp, data = aq,
  type = type_ridge(gradient = TRUE, alpha = 0.5),
  main = "polygon fill (default)")
tinyplot(Month ~ Temp, data = aq,
  type = type_ridge(gradient = TRUE, alpha = 0.5, raster = TRUE),
  main = "raster fill")

# highlighting only the center 50\% of the density (i.e., 25\%-75\% quantiles)
tinyplot(Month ~ Temp, data = aq, type = type_ridge(
  gradient = hcl.colors(3, "Dark Mint")[c(2, 1, 2)],
  probs = c(0.25, 0.75), col = "white"))

# highlighting the probability distribution by color gradient
# (darkest point = median)
tinyplot(Month ~ Temp, data = aq, type = type_ridge(
  gradient = hcl.colors(250, "Dark Mint")[c(250:1, 1:250)],
  probs = 0:500/500))

# faceting also works, although we recommend switching (back) to the "ridge"
# theme for faceted ridge plots

tinytheme("ridge")
tinyplot(Month ~ Ozone, facet = ~ Late, data = aq,
  type = type_ridge(gradient = TRUE))

## use the joint.max argument to vary the maximum density used for
## determining relative scaling...

# jointly across all densities (default) vs. per facet
tinyplot(Month ~ Temp, facet = ~ Late, data = aq,
  type = type_ridge(scale = 1))
tinyplot(Month ~ Temp, facet = ~ Late, data = aq,
  type = type_ridge(scale = 1, joint.max = "facet"))

# jointly across all densities (default) vs. per by row
tinyplot(Month ~ Temp | Late, data = aq,
  type = type_ridge(scale = 1))
tinyplot(Month ~ Temp | Late, data = aq,
  type = type_ridge(scale = 1, joint.max = "by"))

# restore the default theme
tinytheme()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{restore_margin_inner}
\alias{restore_margin_inner}
\title{Restore inner margin defaults}
\usage{
restore_margin_inner(ooma, topmar_epsilon = 0.1)
}
\arguments{
\item{ooma}{Outer margins (from par("oma"))}

\item{topmar_epsilon}{Small epsilon value for top margin adjustment (default 0.1)}
}
\value{
NULL (called for side effect of resetting par("mar"))
}
\description{
Resets inner margins that may have been adjusted for legend
placement. Handles special cases for each margin side and checks for
custom mfrow layouts.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_spline.R
\name{type_spline}
\alias{type_spline}
\title{Spline plot type}
\usage{
type_spline(
  n = NULL,
  method = "fmm",
  xmin = NULL,
  xmax = NULL,
  xout = NULL,
  ties = mean
)
}
\arguments{
\item{n}{if \code{xout} is left unspecified, interpolation takes place
       at \code{n} equally spaced points spanning the interval
       [\code{xmin}, \code{xmax}].}

\item{method}{specifies the type of spline to be used.  Possible
    values are \code{"fmm"}, \code{"natural"}, \code{"periodic"},
    \code{"monoH.FC"} and \code{"hyman"}.  Can be abbreviated.}

\item{xmin, xmax}{left-hand and right-hand endpoint of the
       interpolation interval (when \code{xout} is unspecified).}

\item{xout}{an optional set of values specifying where interpolation
    is to take place.}

\item{ties}{handling of tied \code{x} values.  The string
    \code{"ordered"} or a function (or the name of a function) taking a
    single vector argument and returning a single number or a length-2
    \code{\link{list}} of both, see \code{\link[stats]{approx}} and its
    \sQuote{Details} section, and the example below.}
}
\description{
Type function for plotting a cubic (or Hermite) spline interpolation.
Arguments are passed to \code{\link[stats]{spline}}; see this latter function
for default argument values.
}
\details{
See \code{\link[stats]{spline}} for further details on the available
interpolation methods.
}
\examples{
# "spline" type convenience string
tinyplot(dist ~ speed, data = cars, type = "spline")

# Use `type_spline()` to pass extra arguments for customization
tinyplot(dist ~ speed,
    data = cars, type = type_spline(method = "natural", n = 25),
    add = TRUE, lty = 2)
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/utils.R
\name{restore_margin_outer}
\alias{restore_margin_outer}
\title{Restore outer margin defaults}
\usage{
restore_margin_outer()
}
\value{
NULL (called for side effect of resetting par("omd"))
}
\description{
Resets the outer margin display (omd) to default full device.
Used to clean up after legend drawing that may have adjusted margins.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_glm.R
\name{type_glm}
\alias{type_glm}
\title{Generalized linear model plot type}
\usage{
type_glm(
  family = "gaussian",
  se = TRUE,
  level = 0.95,
  type = "response",
  weights = NULL
)
}
\arguments{
\item{family}{a description of the error distribution and link
    function to be used in the model.  For \code{glm} this can be a
    character string naming a family function, a family function or the
    result of a call to a family function.  For \code{glm.fit} only the
    third option is supported.  (See \code{\link[stats]{family}} for details of
    family functions.)}

\item{se}{logical. If TRUE, confidence intervals are drawn.}

\item{level}{the confidence level required.}

\item{type}{character, partial matching allowed.  Type of weights to
    extract from the fitted model object.  Can be abbreviated.}

\item{weights}{an optional numeric vector of observation weights for the model
fit, of the same length as the number of data points. Weights can also be
supplied via the top-level \code{weights} argument of \code{\link{tinyplot}} (which is
evaluated with non-standard evaluation in the formula method, and takes
precedence if both are given).}
}
\description{
Type function for plotting a generalized model fit.
Arguments are passed to \code{\link[stats]{glm}}.
}
\examples{
# "glm" type convenience string
tinyplot(am ~ mpg, data = mtcars, type = "glm")

# Use `type_glm()` to pass extra arguments for customization
tinyplot(am ~ mpg, data = mtcars, type = type_glm(family = "binomial"))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{build_legend_args}
\alias{build_legend_args}
\title{Build legend arguments list}
\usage{
build_legend_args(
  legend_env,
  legend,
  legend_args,
  by_dep,
  lgnd_labs,
  labeller = NULL,
  type,
  pch,
  lty,
  lwd,
  col,
  bg,
  cex,
  gradient
)
}
\arguments{
\item{legend_env}{Legend environment to populate}

\item{legend}{Legend placement keyword or list}

\item{legend_args}{Additional legend arguments}

\item{by_dep}{The (deparsed) "by" grouping variable name}

\item{lgnd_labs}{The legend labels}

\item{labeller}{Character or function for formatting labels}

\item{type}{Plot type}

\item{pch}{Plotting character(s)}

\item{lty}{Line type(s)}

\item{lwd}{Line width(s)}

\item{col}{Color(s)}

\item{bg}{Background fill color(s)}

\item{cex}{Character expansion(s)}

\item{gradient}{Logical indicating gradient legend}
}
\value{
NULL (modifies legend_env in-place)
}
\description{
Constructs and configures the legend_args list by:
\itemize{
\item Sanitizing legend input
\item Setting defaults for all legend parameters
\item Computing positioning flags from original position (before transformation)
\item Adjusting position anchors for outer legends
\item Adjusting for special cases (gradient, horizontal, multi-column)
\item Populating legend_env with args and positioning flags
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_spineplot.R
\name{type_spineplot}
\alias{type_spineplot}
\title{Spineplot and spinogram types}
\usage{
type_spineplot(
  breaks = NULL,
  tol.ylab = 0.05,
  off = NULL,
  xlevels = NULL,
  ylevels = NULL,
  col = NULL,
  xaxlabels = NULL,
  yaxlabels = NULL,
  weights = NULL,
  lighten = FALSE
)
}
\arguments{
\item{breaks}{if the explanatory variable is numeric, this controls how
    it is discretized. \code{breaks} is passed to \code{\link[graphics]{hist}} and can
    be a list of arguments.}

\item{tol.ylab}{convenience tolerance parameter for y-axis annotation.
    If the distance between two labels drops under this threshold, they are
    plotted equidistantly.}

\item{off}{vertical offset between the bars (in per cent). It is fixed to
    \code{0} for spinograms and defaults to \code{2} for spine plots.}

\item{xlevels, ylevels}{a character or numeric vector specifying the ordering of the
levels of the \code{x} and \code{y} variables (if character) or the corresponding indexes
(if numeric) for the plot.}

\item{col}{a vector of fill colors of the same length as \code{levels(y)}.
    The default is to call \code{\link{gray.colors}}.}

\item{xaxlabels, yaxlabels}{character vectors for annotation of x and y axis.
    Default to \code{levels(y)} and \code{levels(x)}, respectively for the
    spine plot. For \code{xaxlabels} in the spinogram, the breaks are
    used.}

\item{weights}{numeric. A vector of frequency weights for each
    observation in the data. If \code{NULL} all weights are implicitly
    assumed to be 1. If \code{x} is already a 2-way table, the weights
    are ignored.}

\item{lighten}{logical. For grouped spineplots where the \code{y} variable is
itself the grouping variable (i.e. \code{y == by}), should the fills use a
lighter, opaque tint of the series colour(s)? Default is \code{FALSE}, i.e. the
fills use the fully-saturated palette colour(s). (Unlike the other area
types such as \code{\link{type_barplot}}, where lightening is the default, spineplot
tiles abut one another with no gap, so the darker saturated fills read
better against their matching border colours.) Set to \code{TRUE} to opt in to
the lighter tint. Note that \code{lighten} has no effect on other spineplot
displays (single-group or \code{x == by}), which always use a sequential shading
ramp of the base colour.}
}
\description{
Type function(s) for producing spineplots and spinograms, which
are modified versions of histograms or mosaic plots, and particularly
useful for visualizing factor variables. Note that \code{\link{tinyplot}} defaults
to \code{type_spineplot()} if \code{y} is a factor variable.
}
\examples{
# "spineplot" type convenience string
tinyplot(Species ~ Sepal.Width, data = iris, type = "spineplot")

# Aside: specifying the type is redundant for this example, since tinyplot()
# defaults to "spineplot" if y is a factor (just like base plot).
tinyplot(Species ~ Sepal.Width, data = iris)

# Use `type_spineplot()` to pass extra arguments for customization
tinyplot(
  Species ~ Sepal.Width, data = iris,
  type = type_spineplot(breaks = 4)
)

# Passing custom colors to the y-axis categories
tinyplot(
  Species ~ Sepal.Width, data = iris,
  type = type_spineplot(breaks = 4, col = palette.colors(3, "Pastel 1"))
)

# More idiomatic tinyplot way of drawing the previous plot: use y == by
tinyplot(
  Species ~ Sepal.Width | Species, data = iris,
  type = type_spineplot(breaks = 4),
  palette = "Pastel 1", legend = FALSE
)

## Grouped and faceted spineplots

ttnc = as.data.frame(Titanic)

# Note: The Titanic (ttnc) dataset is pre-tabulated, so we pass its frequency
# counts via the top-level `weights` argument (accepted via non-standard
# evaluation in the formula method).
tinyplot(
  Survived ~ Sex, facet = ~ Class, data = ttnc,
  # type_spineplot(weights = ttnc$Freq), ## same thing but not NSE
  type = "spineplot", weights = Freq
)

# Reorder x and y variable categories either by their character levels or
# numeric indexes. (Here we combine a top-level `weights` with constructor-
# level arguments passed through `type_spineplot()`.)
tinyplot(
  Survived ~ Sex, facet = ~ Class, data = ttnc,
  type = type_spineplot(xlevels = c("Female", "Male"), ylevels = 2:1),
  weights = Freq
)

# For (colour) grouped "by" spineplots, it's visually better to facet too
tinyplot(
  Survived ~ Sex | Class, data = ttnc,
  facet = "by",
  type = "spineplot", weights = Freq
)

# Fancier version. Note the smart inheritance of spacing etc.
tinyplot(
  Survived ~ Sex | Class, data = ttnc,
  facet = "by", facet.args = list(nrow = 1),
  type = "spineplot", weights = Freq,
  theme = "void", axes = "t", lty = 0, legend = FALSE,
  main = "Who survived the Titanic disaster?",
  sub = "Frequencies by boarding class and sex"
)

# Aside: It's possible to use "by" on its own (without faceting), but the
# overlaid result isn't great. We will likely overhaul this behaviour in a
# future version of tinyplot...
tinyplot(Survived ~ Sex | Class, data = ttnc,
  type = "spineplot", weights = Freq, alpha = 0.3
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_lines.R
\name{type_lines}
\alias{type_lines}
\title{Lines plot type}
\usage{
type_lines(type = "l", dodge = 0, fixed.dodge = FALSE)
}
\arguments{
\item{type}{1-character string giving the type of plot desired.  The
    following values are possible, for details, see \code{\link[base]{plot}}:
    \code{"p"} for points,
    \code{"l"} for lines,
    \code{"b"} for both points and lines,
    \code{"c"} for empty points joined by lines,
    \code{"o"} for overplotted points and lines,
    \code{"s"} and \code{"S"} for stair steps and
    \code{"h"} for histogram-like vertical lines.  Finally,
    \code{"n"} does not produce any points or lines.}

\item{dodge}{Adjustment parameter for dodging overlapping points or ranges in
grouped plots along the x-axis (or y-axis for flipped plots). Either:
\itemize{
\item numeric value in the range \verb{[0,1)}. Note that values are scaled
relative to the spacing of x-axis breaks, e.g. \code{dodge = 0.1} places the
outermost groups one-tenth of the way to adjacent breaks, \code{dodge = 0.5}
places them midway between breaks, etc. Values < 0.5 are recommended.
\item logical. If \code{TRUE}, the dodge width is calculated automatically based on
the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
\code{FALSE} or 0, no dodging is performed.
}

Default value is 0 (no dodging). While we do not check, it is \emph{strongly}
recommended that dodging only be used in cases where the x-axis comprises a
limited number of discrete breaks.}

\item{fixed.dodge}{Logical. If \code{FALSE} (default), dodge positions are
calculated independently for each \code{x} value, based only on the groups
present at that position. If \code{TRUE}, dodge positions are based on all
groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
groups are missing for a particular \code{x} value).}
}
\description{
Type function for plotting lines.
}
\examples{
# "l" type convenience character string
tinyplot(circumference ~ age | Tree, data = Orange, type = "l")

# Use `type_lines()` to pass extra arguments for customization
tinyplot(circumference ~ age | Tree, data = Orange, type = type_lines(type = "s"))

# Direct legend labels are a good option for grouped lined plots (assuming
# there aren't too many groups and the data are sorted along the x-axis)
tinyplot(
  circumference ~ age | Tree, data = Orange, type = "l",
  legend = "direct"
)

# Fancier version(s) that use a theme and repel overlapping labels
Orange2 = transform(Orange, Tree = paste("Tree", Tree))
tinyplot(
  circumference ~ age | Tree, data = Orange2, type = "l",
  legend = list("direct", repel = TRUE), # auto repel
  theme = "socviz"
)
tinyplot(
  circumference ~ age | Tree, data = Orange2, type = "l",
  legend = list("direct", nudge_y = c("Tree 1" = 3, "Tree 3" = -5)), # manual
  theme = "socviz"
)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_glm.R
\name{ci}
\alias{ci}
\title{Calculate confidence intervals}
\usage{
ci(estimate, std.error, conf.level, df, backtransform = identity)
}
\description{
Calculate confidence intervals
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend.R
\name{prepare_legend}
\alias{prepare_legend}
\title{Prepare legend context from settings}
\usage{
prepare_legend(settings)
}
\arguments{
\item{settings}{Settings environment from tinyplot}
}
\value{
NULL (modifies settings environment in-place)
}
\description{
Main orchestrator that determines:
\itemize{
\item Whether to draw legend
\item Legend labels and formatting
\item Whether multi-legend is needed (for bubble charts)
\item Gradient legend setup for continuous grouping
}
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinytheme.R
\name{tinytheme_register}
\alias{tinytheme_register}
\alias{tinytheme_list}
\alias{tinytheme_unregister}
\title{Register, List, and Unregister Custom Themes}
\usage{
tinytheme_register(name, theme = "default", ...)

tinytheme_list()

tinytheme_unregister(name)
}
\arguments{
\item{name}{Character string. The name for your custom theme. Cannot clash
with or overwrite a built-in theme name (\code{"default"}, \code{"clean"}, etc.)}

\item{theme}{Character string or list. The base theme to inherit from. If a
string, it must reference a built-in or previously-registered theme. If a
list, it is used directly as the base definition. Default is \code{"default"}.}

\item{...}{Named arguments to override specific theme settings. These are
the same parameters accepted by \code{tpar()}.}
}
\value{
\code{tinytheme_register()} returns the theme definition list (invisibly).
\code{tinytheme_list()} returns a named list with character vectors \code{builtin}
and \code{registered}. \code{tinytheme_unregister()} returns \code{NULL} (invisibly).
}
\description{
\code{tinytheme_register()} registers a custom theme so it can be used by name
with \code{tinytheme()} or \code{tinyplot(..., theme = )}. Custom themes inherit from
a base theme and apply user-specified overrides. Registered themes are
session-scoped: they persist across plots but not across R sessions. To make
a custom theme permanently available, register it in your \code{.Rprofile}.

\code{tinytheme_list()} returns the names of all available themes (built-in and
registered).

\code{tinytheme_unregister()} removes a previously registered theme from the
registry. Does not reset an active theme.
}
\examples{
# Register a custom theme based on "float" but with a grid
tinytheme_register("float2", theme = "float", grid = TRUE)

# Use it
tinyplot(1:5, theme = "float2")

# List all themes
tinytheme_list()

# Unregister
tinytheme_unregister("float2")

}
\seealso{
\code{\link[=tinytheme]{tinytheme()}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot-package.R
\docType{package}
\name{tinyplot-package}
\alias{tinyplot-package}
\title{tinyplot: Lightweight Extension of the Base R Graphics System}
\description{
Lightweight extension of the base R graphics system, with support for automatic legends, facets, themes, and various other enhancements.
}
\seealso{
Useful links:
\itemize{
  \item \url{https://grantmcdermott.com/tinyplot/}
  \item Report bugs at \url{https://github.com/grantmcdermott/tinyplot/issues}
}

}
\author{
\strong{Maintainer}: Grant McDermott \email{contact@grantmcdermott.com} (\href{https://orcid.org/0000-0001-7883-8573}{ORCID})

Authors:
\itemize{
  \item Grant McDermott \email{contact@grantmcdermott.com} (\href{https://orcid.org/0000-0001-7883-8573}{ORCID})
  \item Vincent Arel-Bundock \email{vincent.arel-bundock@umontreal.ca} (\href{https://orcid.org/0000-0003-1995-6531}{ORCID})
  \item Achim Zeileis \email{Achim.Zeileis@R-project.org} (\href{https://orcid.org/0000-0003-0918-3766}{ORCID})
}

Other contributors:
\itemize{
  \item Etienne Bacher \email{etienne.bacher@protonmail.com} [contributor]
}

}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_points.R
\name{type_points}
\alias{type_points}
\title{Points plot type}
\usage{
type_points(clim = c(0.5, 2.5), dodge = 0, fixed.dodge = FALSE)
}
\arguments{
\item{clim}{Numeric giving the lower and upper limits of the character
expansion (\code{cex}) normalization for bubble charts.}

\item{dodge}{Adjustment parameter for dodging overlapping points or ranges in
grouped plots along the x-axis (or y-axis for flipped plots). Either:
\itemize{
\item numeric value in the range \verb{[0,1)}. Note that values are scaled
relative to the spacing of x-axis breaks, e.g. \code{dodge = 0.1} places the
outermost groups one-tenth of the way to adjacent breaks, \code{dodge = 0.5}
places them midway between breaks, etc. Values < 0.5 are recommended.
\item logical. If \code{TRUE}, the dodge width is calculated automatically based on
the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
\code{FALSE} or 0, no dodging is performed.
}

Default value is 0 (no dodging). While we do not check, it is \emph{strongly}
recommended that dodging only be used in cases where the x-axis comprises a
limited number of discrete breaks.}

\item{fixed.dodge}{Logical. If \code{FALSE} (default), dodge positions are
calculated independently for each \code{x} value, based only on the groups
present at that position. If \code{TRUE}, dodge positions are based on all
groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
groups are missing for a particular \code{x} value).}
}
\description{
Type function for plotting points, i.e. a scatter plot.
}
\examples{
# "p" type convenience character string
tinyplot(Sepal.Length ~ Petal.Length, data = iris, type = "p")

# Same result with type_points()
tinyplot(Sepal.Length ~ Petal.Length, data = iris, type = type_points())

# Note: Specifying the type here is redundant. Like base plot, tinyplot
# automatically produces a scatter plot if x and y are numeric
tinyplot(Sepal.Length ~ Petal.Length, data = iris)

# Grouped scatter plot example
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris)

# Continuous grouping (with gradient legend)
tinyplot(Sepal.Length ~ Petal.Length | Sepal.Width, data = iris, pch = 19)

# Bubble chart version
tinyplot(Sepal.Length ~ Petal.Length, data = iris, cex = iris$Sepal.Width)

# Fancier version with dual legends and extra customization
tinyplot(Sepal.Length ~ Petal.Length | Species,
  data = iris,
  cex = iris$Sepal.Width, clim = c(1, 5),
  pch = 21, fill = 0.3)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_jitter.R
\name{type_jitter}
\alias{type_jitter}
\title{Jittered points plot type}
\usage{
type_jitter(factor = 1, amount = NULL)
}
\arguments{
\item{factor}{numeric; used in calculation of \code{amount} when
   the provided \code{amount} is 0 or \code{NULL}.}

\item{amount}{number or \code{NULL}; when positive, specifies
    the amount of jitter; see Details, also for other cases.}
}
\description{
Type function for plotting jittered points.
Arguments are passed to \code{\link[base]{jitter}}.
}
\details{
The result of \code{r <-  x + amount * runif(n, -1, 1)}
  where \code{n <- length(x)}.

  When \code{amount} is \code{NULL} (the default)
  \code{amount <- factor * d/5} where \emph{d} is about the smallest
  difference between adjacent unique (up to fuzz) \code{x} values.

  If \code{amount == 0}, \code{amount <- factor * (max(x) - min(x))/50},
  (using only finite \code{x} values), which is the same as in S and the
  references.
  In the above cases, the final \code{amount} is always ensured to be positive.

  The short \code{jitter()} function is itself a succinct definition.
}
\examples{
# "jitter" type convenience string
tinyplot(Sepal.Length ~ Species, data = iris, type = "jitter")

# Use `type_jitter()` to pass extra arguments for customization
tinyplot(Sepal.Length ~ Species, data = iris, type = type_jitter(factor = 0.5))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_lm.R
\name{type_lm}
\alias{type_lm}
\title{Linear model plot type}
\usage{
type_lm(se = TRUE, level = 0.95, weights = NULL)
}
\arguments{
\item{se}{logical. If TRUE, confidence intervals are drawn.}

\item{level}{the confidence level required.}

\item{weights}{an optional numeric vector of observation weights for the model
fit, of the same length as the number of data points. Weights can also be
supplied via the top-level \code{weights} argument of \code{\link{tinyplot}} (which is
evaluated with non-standard evaluation in the formula method, and takes
precedence if both are given).}
}
\description{
Type function for plotting a linear model fit.
Arguments are passed to \code{\link[stats]{lm}}.
}
\examples{
# "lm" type convenience string
tinyplot(Sepal.Width ~ Petal.Width, data = iris, type = "lm")

# Grouped model fits (here: illustrating an example of Simpson's paradox)
tinyplot(Sepal.Width ~ Petal.Width | Species, data = iris, type = "lm")
tinyplot_add(type = "p")

# Use `type_lm()` to pass extra arguments for customization
tinyplot(Sepal.Width ~ Petal.Width, data = iris, type = type_lm(level = 0.8))

# Weighted fit example, either as a top-level argument (which supports
# non-standard evaluation in the formula method)...
s77 = as.data.frame(state.x77)
tinyplot(`Life Exp` ~ Income, data = s77, type = "lm", weights = Population)
# ...or directly via the type constructor (requires an evaluated vector)
tinyplot(`Life Exp` ~ Income, data = s77, type = type_lm(
  weights = s77$Population))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot.matrix.R
\name{tinyplot.matrix}
\alias{tinyplot.matrix}
\title{tinyplot Method for Plotting Matrices}
\usage{
\method{tinyplot}{matrix}(
  x,
  type = NULL,
  legend = NULL,
  facet = NULL,
  xlab = NULL,
  ylab = NULL,
  ...
)
}
\arguments{
\item{x}{an object of class \code{"matrix"}.}

\item{type}{plot type passed on to \code{tinyplot}. Defaults to \code{"p"} (points).}

\item{legend}{specification passed on to \code{tinyplot}. The default is to draw a
legend when the matrix has named columns, and to suppress it otherwise.}

\item{facet}{specification of \code{facet} passed on to \code{tinyplot}. The only
accepted non-\code{NULL} value is the \code{"by"} convenience string, which facets
the plot by matrix column.}

\item{xlab, ylab}{axis labels passed on to \code{tinyplot}. \code{ylab} defaults to the
deparsed matrix name. \code{xlab} defaults to \code{"Index"} when the matrix has no
row names; when it does, the row names already label the ticks so the
x-axis title is suppressed.}

\item{...}{further arguments passed to \code{tinyplot}.}
}
\value{
No return value, called for the side effect of producing a plot.
}
\description{
Convenience interface for visualizing
\code{\link[base]{matrix}} objects with tinyplot.
}
\details{
Internally the matrix is converted to long form and visualized as a
scatter (or other \code{type}) of each column's values against their row index.
Each column is mapped to a separate \code{by} category, so a matrix with
multiple columns produces a grouped plot. Optionally, it can also be
faceted via \code{facet = "by"}. This mirrors the base R
\code{\link[graphics]{matplot}} convention of plotting the columns of a
matrix against the row numbers. If the matrix has column names, these are
used as the group (and legend) labels. Single-column matrices are drawn as
a simple index plot with no grouping or legend.
}
\examples{
# basic use
tinyplot(VADeaths)
tinyplot(VADeaths, type = "b")
tinyplot(VADeaths, type = "b", legend = "direct", theme = "socviz")
tinyplot(VADeaths, type = "b", legend = FALSE, facet = "by", theme = "socviz")

# equivalent plot to an example in `?matplot`
sines = outer(1:20, 1:4, function(x, y) sin(x / 20 * pi * y))
tinyplot(sines, type = "o", pch = "by", lty = "by", col = rainbow(ncol(sines)))

}
\seealso{
\code{\link[graphics]{matplot}}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_density.R
\name{type_density}
\alias{type_density}
\title{Density plot type}
\usage{
type_density(
  bw = "nrd0",
  joint.bw = c("mean", "full", "none"),
  adjust = 1,
  kernel = c("gaussian", "epanechnikov", "rectangular", "triangular", "biweight",
    "cosine", "optcosine"),
  n = 512,
  alpha = NULL
)
}
\arguments{
\item{bw}{the smoothing \code{\link[stats:bw.nrd]{bandwidth}} to be used,
see \code{\link[stats]{density}} for details and options.}

\item{joint.bw}{character string indicating whether (and how) the smoothing
bandwidth should be computed from the joint data distribution when there
are multiple subgroups. The options are \code{"mean"} (the default), \code{"full"},
and \code{"none"}. Also accepts a logical argument, where \code{TRUE} maps to
\code{"mean"} and \code{FALSE} maps to \code{"none"}. See the "Bandwidth selection"
section below for a discussion of practical considerations.}

\item{adjust}{the bandwidth used is actually \code{adjust*bw}.
    This makes it easy to specify values like \sQuote{half the default}
    bandwidth.}

\item{kernel}{a character string giving the smoothing kernel to be used. This
must partially match one of \code{"gaussian"}, \code{"rectangular"}, \code{"triangular"},
\code{"epanechnikov"}, \code{"biweight"}, \code{"cosine"} or \code{"optcosine"}, with default
\code{"gaussian"}, and may be abbreviated to a unique prefix (single letter).

\code{"cosine"} is smoother than \code{"optcosine"}, which is the usual 'cosine'
kernel in the literature and almost MSE-efficient. However, \code{"cosine"} is
the version used by S.}

\item{n}{the number of equally spaced points at which the density is
    to be estimated.  When \code{n > 512}, it is rounded up to a power
    of 2 during the calculations (as \code{\link[stats]{fft}} is used) and the
    final result is interpolated by \code{\link[stats]{approx}}.  So it almost
    always makes sense to specify \code{n} as a power of two.
  }

\item{alpha}{numeric value between 0 and 1 specifying the opacity of ribbon shading
If no \code{alpha} value is provided, then will default to \code{tpar("ribbon.alpha")}
(i.e., probably \code{0.2} unless this has been overridden by the user in their global
settings.)}
}
\description{
Type function for density plots.
}
\details{
The algorithm used in \code{density.default} disperses the mass of the
  empirical distribution function over a regular grid of at least 512
  points and then uses the fast Fourier transform to convolve this
  approximation with a discretized version of the kernel and then uses
  linear approximation to evaluate the density at the specified points.

  The statistical properties of a kernel are determined by
  \eqn{\sigma^2_K = \int t^2 K(t) dt}{sig^2 (K) = int(t^2 K(t) dt)}
  which is always \eqn{= 1} for our kernels (and hence the bandwidth
  \code{bw} is the standard deviation of the kernel) and
  \eqn{R(K) = \int K^2(t) dt}{R(K) = int(K^2(t) dt)}.\cr
  MSE-equivalent bandwidths (for different kernels) are proportional to
  \eqn{\sigma_K R(K)}{sig(K) R(K)} which is scale invariant and for our
  kernels equal to \eqn{R(K)}.  This value is returned when
  \code{give.Rkern = TRUE}.  See the examples for using exact equivalent
  bandwidths.

  Infinite values in \code{x} are assumed to correspond to a point mass at
  \code{+/-Inf} and the density estimate is of the sub-density on
  \code{(-Inf, +Inf)}.
}
\section{Bandwidth selection}{
 While the choice of smoothing bandwidth will
always stand to affect a density visualization, it gains an added
importance when multiple densities are drawn simultaneously (e.g., for
subgroups with respect to \code{by} or \code{facet}). Allowing each subgroup to
compute its own separate bandwidth independently offers greater flexibility
in capturing the unique characteristics of each subgroup, particularly when
distributions differ substantially in location and/or scale. However, this
approach may overemphasize small random variations and make it harder to
visually compare densities across subgroups. Hence, it is often useful to
employ the same ("joint") bandwidth across all subgroups. The following
strategies are available via the \code{joint.bw} argument:
\itemize{
\item The default \code{joint.bw = "mean"} first computes the individual bandwidths
for each group but then computes their mean, weighted by the number of
observations in each group. This will work well when all groups have
similar amounts of scatter (similar variances), even when they have
potentially rather different locations. The weighted averaging stabilizes
potential fluctuations in the individual bandwidths, especially when some
subgroups are rather small.
\item Alternatively, \code{joint.bw = "full"} can be used to compute the joint
bandwidth from the full joint distribution (merging all groups). This will
yield an even more robust bandwidth, especially when the groups overlap
substantially (i.e., have similar locations and scales). However, it may
lead to too large bandwidths and thus too much smoothing, especially when
the locations of the groups differ substantially.
\item Finally, \code{joint.bw = "none"} disables the joint bandwidth so that each
group just employs its individual bandwidth. This is often the best choice
if the amounts of scatter differ substantially between the groups, thus
necessitating different amounts of smoothing.
}
}

\section{Titles}{
 This tinyplot method for density plots differs from the base
\code{\link[stats]{plot.density}} function in its treatment of titles. The
x-axis title displays only the variable name, omitting details about the
number of observations and smoothing bandwidth. Additionally, the main
title is left blank by default for a cleaner appearance.
}

\examples{
# "density" type convenience string
tinyplot(~Sepal.Length, data = iris, type = "density")

# grouped density example
tinyplot(~Sepal.Length | Species, data = iris, type = "density")

# use `bg = "by"` (or, equivalent `fill = "by"`) to get filled densities
tinyplot(~Sepal.Length | Species, data = iris, type = "density", fill = "by")

# use `type_density()` to pass extra arguments for customization
tinyplot(
  ~Sepal.Length | Species, data = iris,
  type = type_density(bw = "SJ"),
  main = "Bandwidth computed using Sheather & Jones (1991)"
)

# The default for grouped density plots is to use the mean of the
# individual subgroup bandwidths (weighted by group size) as the
# joint bandwidth. Alternatively, the bandwidth from the "full"
# data or separate individual bandwidths ("none") can be used.
tinyplot(~Sepal.Length | Species, data = iris,
    ylim = c(0, 1.25), type = "density")        # mean (default)
tinyplot_add(joint.bw = "full", lty = 2)        # full data
tinyplot_add(joint.bw = "none", lty = 3)        # none (individual)
legend("topright", c("Mean", "Full", "None"), lty = 1:3, bty = "n", title = "Joint BW")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_summary.R
\name{type_summary}
\alias{type_summary}
\title{Plot summary values of \code{y} at unique values of \code{x}}
\usage{
type_summary(fun = mean, ...)
}
\arguments{
\item{fun}{summarizing function. Should be compatible with
\code{\link[stats]{ave}}. Defaults to \code{\link[base]{mean}}.}

\item{...}{Additional arguments are passed to the \code{lines()} function,
ex: \code{type="p"}, \code{col="pink"}.}
}
\description{
Applies a summary function to \code{y} along unique values of \code{x}. For example,
plot the mean \code{y} value for each \code{x} value. Internally,
\code{type_summary()} applies a thin wrapper around \code{\link[stats]{ave}} and
then passes the result to \code{\link{type_lines}} for drawing.
}
\examples{
# Plot the mean chick weight over time
tinyplot(weight ~ Time, data = ChickWeight, type = "summary")

# Note: "mean" is the default function, so these are also equivalent:
# tinyplot(weight ~ Time, data = ChickWeight, type = type_summary())
# tinyplot(weight ~ Time, data = ChickWeight, type = type_summary(mean))

# Plot the median instead
tinyplot(weight ~ Time, data = ChickWeight, type = type_summary(median))

# Works with groups and/or facets too
tinyplot(weight ~ Time | Diet, facet = "by", data = ChickWeight, type = "summary")

# Custom/complex function example
tinyplot(
  weight ~ Time | Diet,
  facet = "by", data = ChickWeight,
  type = type_summary(function(y) quantile(y, probs = 0.9) / max(y))
)

}
\seealso{
\code{\link{ave}} which performs the summarizing (averaging) behind the
scenes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/tinyplot.R
\name{tinyplot}
\alias{tinyplot}
\alias{tinyplot.default}
\alias{tinyplot.formula}
\alias{tinyplot.density}
\alias{plt}
\title{Lightweight extension of the base R plotting function}
\usage{
tinyplot(x, ...)

\method{tinyplot}{default}(
  x = NULL,
  y = NULL,
  xmin = NULL,
  xmax = NULL,
  ymin = NULL,
  ymax = NULL,
  weights = NULL,
  labels = NULL,
  by = NULL,
  facet = NULL,
  facet.args = NULL,
  data = NULL,
  type = NULL,
  legend = NULL,
  main = NULL,
  sub = NULL,
  cap = NULL,
  xlab = NULL,
  ylab = NULL,
  ann = par("ann"),
  xlim = NULL,
  ylim = NULL,
  axes = TRUE,
  xaxt = NULL,
  yaxt = NULL,
  xaxs = NULL,
  yaxs = NULL,
  xaxb = NULL,
  yaxb = NULL,
  xaxl = NULL,
  yaxl = NULL,
  log = "",
  flip = FALSE,
  frame.plot = NULL,
  grid = NULL,
  palette = NULL,
  pch = NULL,
  lty = NULL,
  lwd = NULL,
  col = NULL,
  bg = NULL,
  fill = NULL,
  alpha = NULL,
  cex = NULL,
  add = FALSE,
  draw = NULL,
  empty = FALSE,
  restore.par = FALSE,
  file = NULL,
  width = NULL,
  height = NULL,
  asp = NA,
  theme = NULL,
  ...
)

\method{tinyplot}{formula}(
  x = NULL,
  data = parent.frame(),
  facet = NULL,
  facet.args = NULL,
  type = NULL,
  xmin = NULL,
  xmax = NULL,
  ymin = NULL,
  ymax = NULL,
  weights = NULL,
  labels = NULL,
  xlim = NULL,
  ylim = NULL,
  main = NULL,
  sub = NULL,
  cap = NULL,
  xlab = NULL,
  ylab = NULL,
  ann = par("ann"),
  axes = TRUE,
  frame.plot = NULL,
  asp = NA,
  grid = NULL,
  pch = NULL,
  col = NULL,
  lty = NULL,
  lwd = NULL,
  restore.par = FALSE,
  formula = NULL,
  subset = NULL,
  na.action = NULL,
  drop.unused.levels = TRUE,
  ...
)

\method{tinyplot}{density}(x = NULL, type = c("l", "area"), ...)

plt(x, ...)
}
\arguments{
\item{x, y}{the x and y arguments provide the x and y coordinates for the
plot. Any reasonable way of defining the coordinates is acceptable; most
likely the names of existing vectors or columns of data frames. See the
'Examples' section below, or the function
\code{\link[grDevices]{xy.coords}} for details. If supplied separately, \code{x}
and \code{y} must be of the same length.}

\item{...}{other graphical parameters. If \code{type} is a character specification
(such as \code{"hist"}) then any argument names that match those from the corresponding
\verb{type_*()} function (such as \code{\link{type_hist}}) are passed on to that.
All remaining arguments from \code{...} can be further graphical parameters, see
\code{\link[graphics]{par}}).}

\item{xmin, xmax, ymin, ymax}{minimum and maximum coordinates of relevant area
or interval plot types. Only used when the \code{type} argument is one of
\code{"rect"} or \code{"segments"} (where all four min-max coordinates are required),
or \code{"pointrange"}, \code{"errorbar"}, or \code{"ribbon"} (where only \code{ymin} and
\code{ymax} required alongside \code{x}). Can be passed as unquoted (NSE) variable
names when called from the \code{tinyplot.formula} method, e.g. \code{ymin = varname}
where \code{varname} is a variable in \code{data}.}

\item{weights}{an optional numeric vector of observation weights, of the same
length as the number of \code{x},\code{y} coordinates. Can be passed as an unquoted
(NSE) variable name when called from the \code{tinyplot.formula} method, e.g.
\code{weights = varname} where \code{varname} is a variable in \code{data}. Weights are
only consumed by the plot types that support them; specifically, the model
fit and distribution types like \code{\link[=type_lm]{type_lm()}} and \code{\link[=type_spineplot]{type_spineplot()}}.
Because weights trigger a statistical transformation, a warning is emitted
if this argument is supplied to a non-supported type (even though the
actual weights will be ignored, regardless).}

\item{labels}{an optional vector of text labels for use with \code{\link[=type_text]{type_text()}},
of length 1 or the same length as the number of \code{x},\code{y} coordinates. Can be
passed as an unquoted (NSE) variable name when called from the
\code{tinyplot.formula} method, e.g. \code{labels = varname} where \code{varname} is a
variable in \code{data}. Ignored by other plot types. Overrides the \code{labels}
argument of \code{\link[=type_text]{type_text()}} if both are supplied.}

\item{by}{grouping variable(s). The default behaviour is for groups to be
represented in the form of distinct colours, which will also trigger an
automatic legend. (See \code{legend} below for customization options.) However,
groups can also be presented through other plot parameters (e.g., \code{pch},
\code{lty}, or \code{cex}) by passing an appropriate \code{"by"} keyword; see Examples.
Note that continuous (i.e., gradient) colour legends are also supported if
the user passes a numeric or integer to \code{by}. To group by multiple
variables, wrap them with \code{\link[base]{interaction}}.}

\item{facet}{the faceting variable(s) that you want arrange separate plot
windows by. Can be specified in various ways:
\itemize{
\item In "atomic" form, e.g. \code{facet = fvar}. To facet by multiple variables in
atomic form, simply interact them, e.g.
\code{interaction(fvar1, fvar2)} or \code{factor(fvar1):factor(fvar2)}.
\item As a one-sided formula, e.g. \code{facet = ~fvar}. Multiple variables can be
specified in the formula RHS, e.g. \code{~fvar1 + fvar2} or \code{~fvar1:fvar2}. Note
that these multi-variable cases are \emph{all} treated equivalently and
converted to \code{interaction(fvar1, fvar2, ...)} internally. (No distinction
is made between different types of binary operators, for example, and so
\code{f1+f2} is treated the same as \code{f1:f2}, is treated the same as \code{f1*f2},
etc.)
\item As a two-side formula, e.g. \code{facet = fvar1 ~ fvar2}. In this case, the
facet windows are arranged in a fixed grid layout, with the formula LHS
defining the facet rows and the RHS defining the facet columns. At present
only single variables on each side of the formula are well supported. (We
don't recommend trying to use multiple variables on either the LHS or RHS
of the two-sided formula case.)
\item As a special \code{"by"} convenience keyword, in which case facets will match
the grouping variable(s) passed to \code{by} above.
}}

\item{facet.args}{an optional list of arguments for controlling faceting
behaviour. (Ignored if \code{facet} is NULL.) Supported arguments are as
follows:
\itemize{
\item \code{nrow}, \code{ncol} for overriding the default "square" facet window
arrangement. Only one of these should be specified, but \code{nrow} will take
precedence if both are specified together. Ignored if a two-sided formula
is passed to the main \code{facet} argument, since the layout is arranged in a
fixed grid.
\item \code{free} a logical value indicating whether the axis limits (scales) for
each individual facet should adjust independently to match the range of
the data within that facet. Default is \code{FALSE}. Separate free scaling of
the x- or y-axis (i.e., whilst holding the other axis fixed) is not
currently supported.
\item \code{fmar} a vector of form \code{c(b,l,t,r)} for controlling the base margin
between facets in terms of lines. Defaults to the value of \code{tpar("fmar")},
which should be \code{c(1,1,1,1)}, i.e. a single line of padding around each
individual facet, assuming it hasn't been overridden by the user as part
their global \code{\link[tinyplot]{tpar}} settings. Note some automatic
adjustments are made for certain layouts, and depending on whether the plot
is framed or not, to reduce excess whitespace. See
\code{\link[tinyplot]{tpar}} for more details.
\item \code{cex}, \code{font}, \code{col}, \code{bg}, \code{border} for adjusting the facet title text
and background. Default values for these arguments are inherited from
\code{\link[tinyplot]{tpar}} (where they take a "facet." prefix, e.g.
\code{tpar("facet.cex")}). The latter function can also be used to set these
features globally for all \code{tinyplot} plots.
}}

\item{data}{a data.frame (or list) from which the variables in formula
should be taken. A matrix is converted to a data frame.}

\item{type}{character string or call to a \verb{type_*()} function giving the
type of plot desired.
\itemize{
\item NULL (default): Choose a sensible type for the type of \code{x} and \code{y} inputs
(i.e., usually \code{"p"}).
\item 1-character values supported by \code{\link[graphics]{plot}}:
\itemize{
\item \code{"p"} Points
\item \code{"l"} Lines
\item \code{"b"} Both points and lines
\item \code{"c"} Empty points joined by lines
\item \code{"o"} Overplotted points and lines
\item \code{"s"} Stair steps
\item \code{"S"} Stair steps
\item \code{"h"} Histogram-like vertical lines
\item \code{"n"} Empty plot over the extent of the data
}
\item \code{tinyplot}-specific types. These fall into several categories:
\itemize{
\item Shapes:
\itemize{
\item \code{"area"} / \code{\link[=type_area]{type_area()}}: Plots the area under the curve from \code{y} = 0 to \code{y} = f(\code{x}).
\item \code{"errorbar"} / \code{\link[=type_errorbar]{type_errorbar()}}: Adds error bars to points; requires \code{ymin} and \code{ymax}.
\item \code{"pointrange"} / \code{\link[=type_pointrange]{type_pointrange()}}: Combines points with error bars.
\item \code{"polygon"} / \code{\link[=type_polygon]{type_polygon()}}: Draws polygons.
\item \code{"polypath"} / \code{\link[=type_polypath]{type_polypath()}}: Draws a path whose vertices are given in \code{x} and \code{y}.
\item \code{"rect"} / \code{\link[=type_rect]{type_rect()}}: Draws rectangles; requires \code{xmin}, \code{xmax}, \code{ymin}, and \code{ymax}.
\item \code{"ribbon"} / \code{\link[=type_ribbon]{type_ribbon()}}: Creates a filled area between \code{ymin} and \code{ymax}.
\item \code{"segments"} / \code{\link[=type_segments]{type_segments()}}: Draws line segments between pairs of points.
\item \code{"text"} / \code{\link[=type_text]{type_text()}}: Add text annotations.
}
\item Visualizations:
\itemize{
\item \code{"barplot"} / \code{\link[=type_barplot]{type_barplot()}}: Creates a bar plot.
\item \code{"boxplot"} / \code{\link[=type_boxplot]{type_boxplot()}}: Creates a box-and-whisker plot.
\item \code{"chull"} / \code{\link[=type_chull]{type_chull()}}: Draws convex hull(s) around grouped points.
\item \code{"density"} / \code{\link[=type_density]{type_density()}}: Plots the density estimate of a variable.
\item \code{"histogram"} / \code{\link[=type_histogram]{type_histogram()}}: Creates a histogram of a single variable.
\item \code{"jitter"} / \code{\link[=type_jitter]{type_jitter()}}: Jittered points.
\item \code{"qq"} / \code{\link[=type_qq]{type_qq()}}: Creates a quantile-quantile plot.
\item \code{"ridge"} / \code{\link[=type_ridge]{type_ridge()}}: Creates a ridgeline (aka joy) plot.
\item \code{"rug"} / \code{\link[=type_rug]{type_rug()}}: Adds a rug to an existing plot.
\item \code{"spineplot"} / \code{\link[=type_spineplot]{type_spineplot()}}: Creates a spineplot or spinogram.
\item \code{"violin"} / \code{\link[=type_violin]{type_violin()}}: Creates a violin plot.
}
\item Models:
\itemize{
\item \code{"loess"} / \code{\link[=type_loess]{type_loess()}}: Local regression curve.
\item \code{"lm"} / \code{\link[=type_lm]{type_lm()}}: Linear regression line.
\item \code{"glm"} / \code{\link[=type_glm]{type_glm()}}: Generalized linear model fit.
\item \code{"spline"} / \code{\link[=type_spline]{type_spline()}}: Cubic (or Hermite) spline interpolation.
}
\item Functions:
\itemize{
\item \code{\link[=type_abline]{type_abline()}}: line(s) with intercept and slope.
\item \code{\link[=type_hline]{type_hline()}}: horizontal line(s).
\item \code{\link[=type_vline]{type_vline()}}: vertical line(s).
\item \code{\link[=type_function]{type_function()}}: arbitrary function.
\item \code{\link[=type_summary]{type_summary()}}: summarize \code{y} by unique values of \code{x}.
}
}
}}

\item{legend}{one of the following options:
\itemize{
\item NULL (default), in which case the legend will be determined by the
grouping variable. If there is no group variable (i.e., \code{by} is NULL) then
no legend is drawn. If a grouping variable is detected, then an automatic
legend is drawn to the \emph{outer} right of the plotting area. Note that the
legend title and categories will automatically be inferred from the \code{by}
argument and underlying data.
\item A convenience string indicating the legend position. Supported keywords:
\itemize{
\item Standard position keywords from base \code{legend()}, e.g. \code{"right"},
\code{"topleft"}, \code{"bottom"}, etc.
\item Outer positions via a trailing \code{"!"}, e.g. \code{"right!"}, \code{"topleft!"},
or \code{"bottom!"}. This places the legend \emph{outside} the plotting area and
adjusts the margins accordingly.
\item \code{"direct"}: places text labels at the last point of each group's data,
coloured to match. Best suited to line-based plots with x-sorted data,
where "last" corresponds to "rightmost". The right margin is
automatically expanded to prevent clipping. Requires discrete groups
via \code{by}. For faceted plots, labels are drawn in each panel for the
groups present there. Supports additional arguments when passed
via \code{legend(...)}:
\itemize{
\item \code{nudge_x}, \code{nudge_y}: numeric vectors of per-group offsets in data
units. Recycled to the number of groups. Supports named vectors
for targeted adjustment, e.g.
\code{nudge_y = c("Group A" = -10, "Group B" = 20)}.
\item \code{repel}: if \code{TRUE}, automatically separates overlapping labels
vertically. If a positive number, sets the minimum gap between
labels in data units.
}
\item \code{"none"}: turns off legend printing.
}
\item Logical value, where TRUE corresponds to the default case above (same
effect as specifying NULL) and FALSE turns the legend off (same effect as
specifying "none").
\item A list or, equivalently, a dedicated \code{legend()} function with supported
legend arguments, e.g. "bty", "horiz", and so forth. To suppress the
legend title, set \code{title = NULL} or \code{title = FALSE}, e.g.
\code{legend = list(title = FALSE)}.
}}

\item{main}{a main title for the plot, see also \code{title}.}

\item{sub}{a subtitle for the plot.}

\item{cap}{a caption for the plot, drawn at the bottom. Useful for
annotations like data sources. Best paired with a dynamic
\code{\link[tinyplot]{tinytheme}}. For the default theme, should be seen as
a substitute for \code{sub}, since these two will otherwise overlap. Appearance
can be customized via \code{\link[tinyplot]{tpar}} parameters \code{adj.cap},
\code{cex.cap}, \code{col.cap}, \code{font.cap}, and \code{line.cap}.}

\item{xlab}{a label for the x axis, defaults to a description of x.}

\item{ylab}{a label for the y axis, defaults to a description of y.}

\item{ann}{a logical value indicating whether the default annotation (title
and x and y axis labels) should appear on the plot.}

\item{xlim}{the x limits (x1, x2) of the plot. Note that x1 > x2 is allowed
and leads to a reversed axis (although see the \code{"rev(erse)"} keyword option
below). The default value, \code{NULL}, indicates that the range of the \code{finite}
values to be plotted should be used. Alongside the standard length-2 vector
(e.g., \code{xlim = c(0, 1)}), \code{tinyplot} supports three further convenience
forms:
\itemize{
\item a single scalar (e.g. \code{xlim = 0}) ensures that the provided value is
covered by the axis range, irrespective of the data extent.
\item a length-2 vector with one \code{NA} (e.g. \code{xlim = c(0, NA)}) pins the
non-\code{NA} limit and lets the data determine the other limit.
\item the convenience string \code{"rev"} (or the longer \code{"reverse"}) reverses the
automatically-computed axis range. This is equivalent to passing a
descending vector, but without having to know the data extent in advance.
}}

\item{ylim}{the y limits of the plot. Accepts the same input forms as \code{xlim}.}

\item{axes}{logical or character. Should axes be drawn (\code{TRUE} or \code{FALSE})?
Or alternatively what type of axes should be drawn: \code{"standard"} (with
axis, ticks, and labels; equivalent to \code{TRUE}), \code{"none"} (no axes;
equivalent to \code{FALSE}), \code{"ticks"} (only ticks and labels without axis line),
\code{"labels"} (only labels without ticks and axis line), \code{"axis"} (only axis
line and labels but no ticks). To control this separately for the two
axes, use the character specifications for \code{xaxt} and/or \code{yaxt}.}

\item{xaxt, yaxt}{character specifying the type of x-axis and y-axis,
respectively. See \code{axes} for the possible values.}

\item{xaxs, yaxs}{character specifying the style of the interval calculation
used for the x-axis and y-axis, respectively. See
\code{\link[graphics]{par}} for the possible values.}

\item{xaxb, yaxb}{numeric vector (or character vector, if appropriate) giving
the break points at which the axis tick-marks are to be drawn. Break points
outside the range of the data will be ignored if the associated axis
variable is categorical, or an explicit \code{x/ylim} range is given.}

\item{xaxl, yaxl}{a function or a character keyword specifying the format of
the x- or y-axis tick labels. Note that this is a post-processing step that
affects the \emph{appearance} of the tick labels only; use in conjunction with
\code{x/yaxb} if you would like to adjust the position of the tick marks too. In
addition to user-supplied formatting functions (e.g., \code{\link{format}},
\code{\link{toupper}}, \code{\link{abs}}, or other custom function), several convenience
keywords (or their symbol equivalents) are available for common formatting
transformations: \code{"percent"} (\code{"\%"}), \code{"comma"} (\code{","}), \code{"log"} (\code{"l"}),
\code{"dollar"} (\code{"$"}), \code{"euro"} (\code{"€"}), or \code{"sterling"} (\code{"£"}). See the
\code{\link{tinylabel}} documentation for examples.}

\item{log}{a character string which contains \code{"x"} if the x axis is to be
logarithmic, \code{"y"} if the y axis is to be logarithmic and \code{"xy"} or \code{"yx"}
if both axes are to be logarithmic.}

\item{flip}{logical. Should the plot orientation be flipped, so that the
y-axis is on the horizontal plane and the x-axis is on the vertical plane?
Default is FALSE.}

\item{frame.plot}{a logical indicating whether a box should be drawn around
the plot. Can also use \code{frame} as an acceptable argument alias.
The default is to draw a frame if both axis types (set via \code{axes}, \code{xaxt},
or \code{yaxt}) include axis lines.}

\item{grid}{argument for plotting a background panel grid, one of:
\itemize{
\item a logical (i.e., \code{TRUE} to draw the grid on both axes).
\item a character string controlling which axes get grid lines and at what
resolution. Uppercase letters (\code{"X"}, \code{"Y"}, \code{"XY"}) draw grid lines at
the standard axis tick positions (with the latter equivalent to \code{TRUE}),
while lowercase letters (\code{"x"}, \code{"y"}, \code{"xy"}) draw a finer grid with
additional lines at the midpoints between ticks. (Note: the finer grid
has no effect on log-scale axes, since tick positions already include
intermediate values.)
\item a panel grid plotting function like \code{grid()}.
Note that this argument replaces the \code{panel.first} and \code{panel.last}
arguments from base \code{plot()} and tries to make the process more seamless
with better default behaviour. The default behaviour is determined by (and
can be set globally through) the value of \code{tpar("grid")}.
}}

\item{palette}{one of the following options:
\itemize{
\item NULL (default), in which case the palette will be chosen according to
the class and cardinality of the "by" grouping variable. For non-ordered
factors or strings with a reasonable number of groups, this will inherit
directly from the user's default \code{\link[grDevices]{palette}} (e.g.,
"R4"). In other cases, including ordered factors and high cardinality, the
"Viridis" palette will be used instead. Note that a slightly restricted
version of the "Viridis" palette---where extreme color values have been
trimmed to improve visual perception---will be used for ordered factors
and continuous variables. In the latter case of a continuous grouping
variable, we also generate a gradient legend swatch.
\item A convenience string corresponding to one of the many palettes listed by
either \code{palette.pals()} or \code{hcl.pals()}. Note that the string can be
case-insensitive (e.g., "Okabe-Ito" and "okabe-ito" are both valid).
\item A palette-generating function. This can be "bare" (e.g.,
\code{palette.colors}) or "closed" with a set of named arguments (e.g.,
\code{palette.colors(palette = "Okabe-Ito", alpha = 0.5)}). Note that any
unnamed arguments will be ignored and the key \code{n} argument, denoting the
number of colours, will automatically be spliced in as the number of
groups.
\item A vector or list of colours, e.g. \code{c("darkorange", "purple", "cyan4")}.
If too few colours are provided for a discrete (qualitative) set of
groups, then the colours will be recycled with a warning. For continuous
(sequential) groups, a gradient palette will be interpolated.
}}

\item{pch}{plotting "character", i.e., symbol to use. Character, integer, or
vector of length equal to the number of categories in the \code{by} variable.
See \code{pch}. In addition, users can supply a special \code{pch = "by"} convenience
argument, in which case the characters will automatically loop over the
number groups. This automatic looping will begin at the global character
value (i.e., \code{par("pch")}) and recycle as necessary.}

\item{lty}{line type. Character, integer, or vector of length equal to the
number of categories in the \code{by} variable. See \code{lty}. In addition, users
can supply a special \code{lty = "by"} convenience argument, in which case the
line type will automatically loop over the number groups. This automatic
looping will begin at the global line type value (i.e., \code{par("lty")}) and
recycle as necessary.}

\item{lwd}{line width. Numeric scalar or vector of length equal to the
number of categories in the \code{by} variable. See \code{lwd}. In addition, users
can supply a special \code{lwd = "by"} convenience argument, in which case the
line width will automatically loop over the number of groups. This
automatic looping will be centered at the global line width value (i.e.,
\code{par("lwd")}) and pad on either side of that.}

\item{col}{plotting color. Character, integer, or vector of length equal to
the number of categories in the \code{by} variable. See \code{col}. Note that the
default behaviour in \code{tinyplot} is to vary group colors along any variables
declared in the \code{by} argument. Thus, specifying colors manually should not
be necessary unless users wish to override the automatic colors produced by
this grouping process. Typically, this would only be done if grouping
features are deferred to some other graphical parameter (i.e., passing the
"by" keyword to one of \code{pch}, \code{lty}, \code{lwd}, or \code{bg}; see below.)}

\item{bg}{background fill color for the open plot symbols 21:25 (see
\code{points.default}), as well as ribbon and area plot types.
Users can also supply either one of two special convenience arguments that
will cause the background fill to inherit the automatic grouped coloring
behaviour of \code{col}:
\itemize{
\item \code{bg = "by"} will insert a background fill that inherits the main color
mappings from \code{col}.
\item \verb{by = <numeric[0,1]>} (i.e., a numeric in the range \verb{[0,1]}) will insert
a background fill that inherits the main color mapping(s) from \code{col}, but
with added alpha-transparency.
}

For both of these convenience arguments, note that the (grouped) \code{bg}
mappings will persist even if the (grouped) \code{col} defaults are themselves
overridden. This can be useful if you want to preserve the grouped palette
mappings by background fill but not boundary color, e.g. filled points. See
examples.}

\item{fill}{alias for \code{bg}. If non-NULL values for both \code{bg} and \code{fill} are
provided, then the latter will be ignored in favour of the former.}

\item{alpha}{a numeric in the range \verb{[0,1]} for adjusting the alpha channel
of the color palette, where 0 means transparent and 1 means opaque. Use
fractional values, e.g. \code{0.5} for semi-transparency.}

\item{cex}{character expansion. A numerical vector (can be a single value)
giving the amount by which plotting characters and symbols should be scaled
relative to the default. Note that \code{NULL} is equivalent to 1.0, while \code{NA}
renders the characters invisible. There are two additional considerations,
specifically for points-alike plot types (e.g. \code{"p"}):
\itemize{
\item users can also supply a special \code{cex = "by"} convenience argument, in
which case the character expansion will automatically adjust by group
too. The range of this character expansion is controlled by the \code{clim}
argument in the respective types; see \code{\link[=type_points]{type_points()}} for example.
\item passing a \code{cex} vector of equal length to the main \code{x} and \code{y} variables
(e.g., another column in the same dataset) will yield a "bubble"plot with
its own dedicated legend. This can provide a useful way to visualize an
extra dimension of the data; see Examples.
}}

\item{add}{logical. If TRUE, then elements are added to the current plot rather
than drawing a new plot window. Note that the automatic legend for the
added elements will be turned off. See also \link{tinyplot_add}, which provides
a convenient wrapper around this functionality for layering on top of an
existing plot without having to repeat arguments.}

\item{draw}{a function that draws directly on the plot canvas (before \code{x} and
\code{y} are plotted). The \code{draw} argument is primarily useful for adding common
elements to each facet of a faceted plot, e.g.
\code{\link[graphics]{abline}} or \code{\link[graphics]{text}}. Note that
this argument is somewhat experimental and that \emph{no} internal checking is
done for correctness; the provided argument is simply captured and
evaluated as-is within \code{tinyplot()} and thus has access to the local
definition of all variables such as \code{x}, \code{y}, etc. See Examples.}

\item{empty}{logical indicating whether the interior plot region should be
left empty. The default is \code{FALSE}. Setting to \code{TRUE} has a similar effect
to invoking \code{type = "n"} above, except that any legend artifacts owing to a
particular plot type (e.g., lines for \code{type = "l"} or squares for
\code{type = "area"}) will still be drawn correctly alongside the empty plot. In
contrast,\code{type = "n"} implicitly assumes a scatterplot and so any legend
will only depict points.}

\item{restore.par}{a logical value indicating whether the
\code{\link[graphics]{par}} settings prior to calling \code{tinyplot} should be
restored on exit. Defaults to FALSE, which makes it possible to add
elements to the plot after it has been drawn. However, note the the outer
margins of the graphics device may have been altered to make space for the
\code{tinyplot} legend. Users can opt out of this persistent behaviour by
setting to TRUE instead. See also \link{get_saved_par} for another option to
recover the original \code{\link[graphics]{par}} settings, as well as
longer discussion about the trade-offs involved.}

\item{file}{character string giving the file path for writing a plot to disk.
If specified, the plot will not be displayed interactively, but rather sent
to the appropriate external graphics device (i.e.,
\code{\link[grDevices]{png}}, \code{\link[grDevices]{jpeg}},
\code{\link[grDevices]{pdf}}, or \code{\link[grDevices]{svg}}). As a point
of convenience, note that any global parameters held in \verb{(t)par} are
automatically carried over to the external device and don't need to be
reset (in contrast to the conventional base R approach that requires
manually opening and closing the device). The device type is determined by
the file extension at the end of the provided path, and must be one of
".png", ".jpg" (".jpeg"), ".pdf", or ".svg". (Other file types may be
supported in the future.) The file dimensions can be controlled by the
corresponding \code{width} and \code{height} arguments below, otherwise will fall
back to the \code{"file.width"} and \code{"file.height"} values held in
\code{\link[tinyplot]{tpar}} (i.e., both defaulting to 7 inches, and where
the default resolution for bitmap files is also specified as 300
DPI).}

\item{width}{numeric giving the plot width in inches. Together with \code{height},
typically used in conjunction with the \code{file} argument above, overriding the
default values held in \code{tpar("file.width", "file.height")}. If either \code{width}
or \code{height} is specified, but a corresponding \code{file} argument is not
provided as well, then a new interactive graphics device dimensions will be
opened along the given dimensions. Note that this interactive resizing may
not work consistently from within an IDE like RStudio that has an integrated
graphics windows.}

\item{height}{numeric giving the plot height in inches. Same considerations as
\code{width} (above) apply, e.g. will default to \code{tpar("file.height")} if not
specified.}

\item{asp}{the y/xy/x aspect ratio, see \code{plot.window}.}

\item{theme}{keyword string (e.g. \code{"clean"}) or list defining a theme. Passed
on to \code{\link{tinytheme}}, but reset upon exit so that the theme effect is only
temporary. Useful for invoking ephemeral themes.}

\item{formula}{a \code{\link[stats]{formula}} that optionally includes
grouping variable(s) after a vertical bar, e.g. \code{y ~ x | z}. One-sided
formulae are also permitted, e.g. \code{~ y | z}. Only a single \code{y} and \code{x}
variable (if any) must be specified but multiple grouping variables
can be included in different ways, e.g. \code{y ~ x | z1:z2} or
\code{y ~ x | z1 + z2}. (These two representations are treated as equivalent;
both are parsed as \code{interaction(z1, z2)} internally.) If arithmetic
operators are used for transforming variables, they should be wrapped in
\code{I()}, e.g., \code{I(y1/y2) ~ x}. Note that the \code{formula} and \code{x} arguments
should not be specified in the same call.}

\item{subset, na.action, drop.unused.levels}{arguments passed to \code{model.frame}
when extracting the data from \code{formula} and \code{data}.}
}
\value{
No return value, called for side effect of producing a plot.
}
\description{
Enhances the base \code{\link[graphics]{plot}} function. Supported features
include automatic legends and facets for grouped data, additional plot types,
theme customization, and so on. Users can call either \code{tinyplot()}, or its
shorthand alias \code{plt()}.
}
\details{
Disregarding the enhancements that it supports, \code{tinyplot} tries as far as
possible to mimic the behaviour and syntax logic of the original base
\code{\link[graphics]{plot}} function. Users should therefore be able to swap
out existing \code{plot} calls for \code{tinyplot} (or its shorthand alias \code{plt}),
without causing unexpected changes to the output.
}
\examples{
aq = transform(
  airquality,
  Month = factor(Month, labels = month.abb[unique(Month)])
)

# In most cases, `tinyplot` should be a drop-in replacement for regular
# `plot` calls. For example:

op = tpar(mfrow = c(1, 2))
plot(0:10, main = "plot")
tinyplot(0:10, main = "tinyplot")
tpar(op) # restore original layout

# Aside: `tinyplot::tpar()` is a (near) drop-in replacement for `par()`

# Unlike vanilla plot, however, tinyplot allows you to characterize groups
# using either the `by` argument or equivalent `|` formula syntax.

with(aq, tinyplot(Day, Temp, by = Month)) ## atomic method
tinyplot(Temp ~ Day | Month, data = aq) ## formula method

# (Notice that we also get an automatic legend.)

# You can also use the equivalent shorthand `plt()` alias if you'd like to
# save on a few keystrokes

plt(Temp ~ Day | Month, data = aq) ## shorthand alias

# Use standard base plotting arguments to adjust features of your plot.
# For example, change `pch` (plot character) to get filled points and `cex`
# (character expansion) to increase their size.

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  pch = 16,
  cex = 2
)

# Use the special "by" convenience keyword if you would like to map these
# aesthetic features over groups too (i.e., in addition to the default
# colour grouping)

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  pch = "by",
  cex = "by"
)

# We can add alpha transparency for overlapping points

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  pch = 16,
  cex = 2,
  alpha = 0.3
)

# To get filled points with a common solid background color, use an
# appropriate plotting character (21:25) and combine with one of the special
# `bg`/`fill` convenience arguments.
tinyplot(
  Temp ~ Day | Month,
  data = aq,
  pch = 21, # use filled circles
  cex = 2,
  bg = 0.3, # numeric in [0,1] adds a grouped background fill with transparency
  col = "black" # override default color mapping; give all points a black border
)

# Aside: For "bubble" plots, pass an appropriate vector to the `cex` arg.
# This can be useful for depicting an additional dimension of the data (here:
# Wind).
tinyplot(
  Temp ~ Day | Month,
  data = aq,
  pch = 21,
  cex = aq$Wind, # map character size to another feature in the data
  bg = 0.3,
  col = "black"
)

# Converting to a grouped line plot is a simple matter of adjusting the
# `type` argument.

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "l"
)

# Similarly for other plot types, including some additional ones provided
# directly by tinyplot, e.g. density plots or internal plots (ribbons,
# pointranges, etc.)

tinyplot(
  ~ Temp | Month,
  data = aq,
  type = "density",
  fill = "by"
)

# Facet plots are supported too. Facets can be drawn on their own...

tinyplot(
  Temp ~ Day,
  facet = ~Month,
  data = aq,
  type = "area",
  main = "Temperatures by month"
)

# ... or combined/contrasted with the by (colour) grouping.

aq = transform(aq, Summer = Month \%in\% c("Jun", "Jul", "Aug"))
tinyplot(
  Temp ~ Day | Summer,
  facet = ~Month,
  data = aq,
  type = "area",
  palette = "dark2",
  main = "Temperatures by month and season"
)

# Users can override the default square window arrangement by passing `nrow`
# or `ncol` to the helper facet.args argument. Note that we can also reduce
# axis label repetition across facets by turning the plot frame off.

tinyplot(
  Temp ~ Day | Summer,
  facet = ~Month, facet.args = list(nrow = 1),
  data = aq,
  type = "area",
  palette = "dark2",
  frame = FALSE,
  main = "Temperatures by month and season"
)

# Use a two-sided formula to arrange the facet windows in a fixed grid.
# LHS -> facet rows; RHS -> facet columns

aq$hot = ifelse(aq$Temp >= 75, "hot", "cold")
aq$windy = ifelse(aq$Wind >= 15, "windy", "calm")
tinyplot(
  Temp ~ Day,
  facet = windy ~ hot,
  data = aq
)

# To add common elements to each facet, use the `draw` argument

tinyplot(
  Temp ~ Day,
  facet = windy ~ hot,
  data = aq,
  draw = abline(h = 75, lty = 2, col = "hotpink")
)

# The (automatic) legend position and look can be customized using
# appropriate arguments. Note the trailing "!" in the `legend` position
# argument below. This tells `tinyplot` to place the legend _outside_ the plot
# area.

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "l",
  legend = legend("bottom!", title = "Month of the year", bty = "o")
)

# Use legend = "direct" to place text labels at the last point of each
# group's data, coloured to match. Best suited to line-based plots with
# x-sorted data, where "last" corresponds to "rightmost". The right
# margin is automatically expanded to fit the labels. Pairs well with
# dynamic themes for tighter margins overall.

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "l",
  legend = "direct",
  theme = "clean2"
)

# Use `nudge_x/y`` to manually adjust label positions, or `repel` to
# auto-separate overlapping labels:
tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "l",
  legend = legend("direct", nudge_y = c(May = -1, Jun = 2)),
  # legend = legend("direct", repel = TRUE), # another option
  theme = "clean2"
)

# The default group colours are inherited from either the "R4" or "Viridis"
# palettes, depending on the number of groups. However, all palettes listed
# by `palette.pals()` and `hcl.pals()` are supported as convenience strings,
# or users can supply a valid palette-generating function for finer control

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "l",
  palette = "tableau"
)

# It's possible to customize the look of your plots by setting graphical
# parameters (e.g., via `(t)par`)... But a more convenient way is to just use
# built-in themes (see `?tinytheme`).

tinyplot(
  Temp ~ Day | Month,
  data = aq,
  type = "b",
  alpha = 0.5,
  main = "Daily temperatures by month",
  sub = "Brought to you by tinyplot",
  cap = "Source: Base R airquality dataset",
  theme = "clean2"
)

# For more examples and a detailed walkthrough, please see the introductory
# tinyplot tutorial available online:
# https://grantmcdermott.com/tinyplot/vignettes/introduction.html

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_rug.R
\name{type_rug}
\alias{type_rug}
\title{Add a rug to a plot}
\usage{
type_rug(
  ticksize = 0.03,
  side = 1,
  quiet = getOption("warn") < 0,
  jitter = FALSE,
  amount = NULL
)
}
\arguments{
\item{ticksize}{The length of the ticks making up the \sQuote{rug}.
   Positive lengths give inwards ticks.}

\item{side}{On which side of the plot box the rug will be
   plotted. Normally 1 (bottom) or 3 (top).}

\item{quiet}{logical indicating if there should be a warning about
   clipped values.}

\item{jitter}{Logical. Add jittering to separate ties? Default is \code{FALSE}.}

\item{amount}{Numeric. Amount of jittering (see \code{\link[base]{jitter}}).
Only used if \code{jitter} is \code{TRUE}.}
}
\description{
Adds a rug representation (1-d plot) of the data to the plot.
}
\details{
This function should only be used as part of \code{\link[=tinyplot_add]{tinyplot_add()}}, i.e. adding
to an existing plot.

In most cases, determining which variable receives the rug representation
will be based on the \code{side} argument (i.e., x-variable if side is 1 or 3, and
y-variable if side is 2 or 4). An exception is if the preceding plot type was
either \code{"density"} or \code{"histogram"}; for these latter cases, the x-variable
will always be used. See Examples.
}
\examples{
tinyplot(~wt | am, data = mtcars, type = "density", facet = "by", fill = "by")
tinyplot_add(type = "rug")
# use type_rug() to pass extra options
tinyplot_add(type = type_rug(side = 3, ticksize = 0.05))

# For ties, use jittering
tinyplot(eruptions ~ waiting, data = faithful, type = "lm")
tinyplot_add(type = type_rug(jitter = TRUE, amount = 0.3))
tinyplot_add(type = type_rug(jitter = TRUE, amount = 0.1, side = 2))
# Add original points just for reference
tinyplot_add(type = "p")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/dodge.R
\name{dodge_positions}
\alias{dodge_positions}
\title{Dodge positions for grouped data}
\usage{
dodge_positions(
  datapoints,
  dodge,
  fixed.dodge = TRUE,
  cols = NULL,
  settings = NULL
)
}
\arguments{
\item{datapoints}{Data frame containing plot data with at least \code{x} and \code{by}
columns.}

\item{dodge}{Adjustment parameter for dodging overlapping points or ranges in
grouped plots along the x-axis (or y-axis for flipped plots). Either:
\itemize{
\item numeric value in the range \verb{[0,1)}. Note that values are scaled
relative to the spacing of x-axis breaks, e.g. \code{dodge = 0.1} places the
outermost groups one-tenth of the way to adjacent breaks, \code{dodge = 0.5}
places them midway between breaks, etc. Values < 0.5 are recommended.
\item logical. If \code{TRUE}, the dodge width is calculated automatically based on
the number of groups (0.1 per group for 2-4 groups, 0.45 for 5+ groups). If
\code{FALSE} or 0, no dodging is performed.
}

Default value is 0 (no dodging). While we do not check, it is \emph{strongly}
recommended that dodging only be used in cases where the x-axis comprises a
limited number of discrete breaks.}

\item{fixed.dodge}{Logical. If \code{FALSE} (default), dodge positions are
calculated independently for each \code{x} value, based only on the groups
present at that position. If \code{TRUE}, dodge positions are based on all
groups, ensuring "fixed" spacing across x-axis breaks (i.e., even if some
groups are missing for a particular \code{x} value).}

\item{cols}{Character vector of column names to dodge. If \code{NULL} (default),
automatically detects and dodges \code{x}, \code{xmin}, and \code{xmax} if they exist.}

\item{settings}{Environment containing plot settings. If \code{NULL} (default),
retrieved from the calling environment.}
}
\value{
Modified \code{datapoints} data frame with dodged positions.
}
\description{
Adjusts x-coordinates (and optionally xmin/xmax) to dodge overlapping points
or ranges in grouped plots.
}
\details{
When \code{fixed.dodge = TRUE}, all groups are dodged by the same amount across all
x values, which is useful when x is categorical. When \code{fixed.dodge = FALSE},
dodging is calculated independently for each x value, which is useful when
the number of groups varies across x values.

If \code{cols} is not specified, the function automatically dodges \code{x} and any
\code{xmin}/\code{xmax} columns that exist in the data.
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_barplot.R
\name{type_barplot}
\alias{type_barplot}
\title{Barplot type}
\usage{
type_barplot(
  width = 5/6,
  beside = FALSE,
  center = FALSE,
  offset = NULL,
  FUN = NULL,
  xlevels = NULL,
  xaxlabels = NULL,
  drop.zeros = FALSE,
  lighten = TRUE
)
}
\arguments{
\item{width}{numeric, optional vector of bar widths. (The distance between
the midpoints of the bars is always 1.)}

\item{beside}{logical. In case of a \code{by} grouping variable, should bars be
juxtaposed? Default is to use stacked bars instead.}

\item{center}{logical or numeric. In case of stacked barplots (\code{beside = FALSE})
should the bars be centered (or all start at zero, default)? If set to
\code{TRUE} the center is at the mid-point of the middle category (in case of
uneven number of categories) or between the two middle categories (in case
of an even number). Additionally it is possible to set \code{center = 2} or
\code{center = 2.5} to indicate that centering should be after the second category
or the mid-way in the third category, respectively.}

\item{offset}{optional specification for shifting bar baselines, accepting
one of two distinct forms. See the Examples for illustrations of both.
\itemize{
\item \emph{Positions} via an unnamed numeric scalar or vector. Bars start at the
offset value(s) rather than zero, matched per x-level after any \code{xlevels}
reordering (a scalar is applied to all bars). Useful for waterfall charts.
The positional form cannot be combined with \code{center}.
\item \emph{Category} via a character vector such as \code{offset = "Unsure"}, or a
named numeric vector such as \code{offset = c(Unsure = 1.1)}. The named
level(s) of the \code{by} grouping are "set aside", i.e. pulled out of the
(optionally centered) stack and drawn as standalone bars. This is useful
for diverging/Likert plots where a neutral category (e.g. "Unsure") is
shown apart from the diverging stack. The category form requires a \code{by}
grouping and \code{beside = FALSE}, but can be combined with \code{center}.
}}

\item{FUN}{a function to compute the summary statistic for \code{y} within each
group of \code{x} in case of using a two-sided formula \code{y ~ x} (default: mean).}

\item{xlevels}{a character or numeric vector specifying the ordering of the
levels of the \code{x} variable (if character) or the corresponding indexes
(if numeric) for the plot.}

\item{xaxlabels}{a character vector with the axis labels for the \code{x} variable,
defaulting to the levels of \code{x}.}

\item{drop.zeros}{logical. Should bars with zero height be dropped? If set
to \code{FALSE} (default) a zero height bar is still drawn for which the border
lines will still be visible.}

\item{lighten}{logical. Should the fills use a lighter, opaque tint of the
series colour(s)? Default is \code{TRUE}, which keeps single- and multi-group
displays consistent and lets the fill read cleanly over grid lines. Set to
\code{FALSE} to use the fully-saturated palette colour(s) instead.}
}
\description{
Type function for producing barplots. For formulas of type
\code{~ x} (without left-hand side) the barplot visualizes the counts (absolute
frequencies) of the levels of \code{x}. For formulas of type \code{y ~ x} the value
of \code{y} within each level of \code{x} is visualized, if necessary aggregated
using some function (default: mean).
}
\examples{
# Basic examples of frequency tables (without y variable)
tinyplot(~ cyl, data = mtcars, type = "barplot")
tinyplot(~ cyl | vs, data = mtcars, type = "barplot")
tinyplot(~ cyl | vs, data = mtcars, type = "barplot", beside = TRUE)

# Reorder x variable categories either by their character levels or numeric indexes
tinyplot(~ cyl, data = mtcars, type = "barplot", xlevels = c("8", "6", "4"))
tinyplot(~ cyl, data = mtcars, type = "barplot", xlevels = 3:1)

# Note: Above we used automatic argument passing for `beside`. But this
# wouldn't work for `width`, since it would conflict with the top-level
# `tinyplot(..., width = <width>)` argument. It's safer to pass these args
# through the `type_barplot()` functional equivalent.
tinyplot(
  ~ cyl | vs, data = mtcars,
  type = type_barplot(beside = TRUE, drop.zeros = TRUE, width = 0.65)
)

# Example for numeric y aggregated by x (default: FUN = mean) + facets
tinyplot(
  extra ~ ID | group, facet = "by", data = sleep,
  type = "barplot",
  theme = "clean2"
)

# Fancy frequency table:
tinyplot(
  Freq ~ Sex | Survived, data = as.data.frame(Titanic),
  facet = ~ Class, facet.args = list(nrow = 1),
  type = "barplot", flip = TRUE,
  theme = "clean2"
)

# Centered barplot for conditional proportions of dark (black/brown) vs.
# light (red/blond) hair color, conditional on eye color and sex.
# Aside: use `lighten = FALSE` to avoid lightening the bar fill colors.
hec = as.data.frame(proportions(HairEyeColor, 2:3))
hcols = c("black", "sienna", "indianred", "goldenrod")
tinyplot(
  Freq ~ Eye | Hair, data = hec,
  facet = ~ Sex, facet.args = list(ncol = 1),
  type = type_barplot(center = TRUE, lighten = FALSE),
  flip = TRUE, yaxl = "percent",
  theme = list("clean2", palette.qualitative = hcols)
)

# Use cases for the `offset` argument

# 1. Waterfall plot
d = data.frame(item = c("Sales", "Services", "Costs", "Returns", "TOTAL"),
               value = c(100, 40, -80, -10, 50))
d$item = factor(d$item, levels = d$item)
d$offset = c(0, cumsum(d$value[1:3]), 0)
tinyplot(
  value ~ item | I(value < 0), data = d,
  type = type_barplot(offset = d$offset, lighten = FALSE),
  col = NA, # (optional: turn off border)
  legend = FALSE
)
tinyplot_add(type = type_vline(4.5), lty = 2, col = "grey50")

# 2. Diverging/Likert layout: a character (or named numeric) offset "sets
# aside" the named category, pulling it out of the centered stack and drawing
# it as a standalone bar. Here a neutral "Unsure" response is shown apart from
# the diverging agree/disagree scale.
lik = expand.grid(
  question = c("Pay", "Workload", "Manager", "Culture"),
  response = c("Strong disagree", "Disagree", "Agree", "Strong agree", "Unsure")
)
lik$response = factor(lik$response, levels = unique(lik$response))
lik$share = c( # proportions summing to 1 within each question
  .10, .25, .05, .15,
  .20, .30, .15, .20,
  .35, .20, .40, .30,
  .25, .15, .35, .20,
  .10, .10, .05, .15
)
# diverging palette: reds (disagree) -> blues (agree), grey for "Unsure"
pal = c("#b2182b", "#ef8a62", "#67a9cf", "#2166ac", "grey")
tinyplot(
  share ~ question | response, data = lik,
  type = type_barplot(center = TRUE, offset = "Unsure", lighten = FALSE),
  flip = TRUE, xlab = NA, ylab = NA, yaxl = "percent",
  legend = list("top!", title = FALSE),
  theme = list("clean2", palette.qualitative = pal),
  main = "Hypothetical Likert example with category offset"
)
tinyplot_add(type = "vline")

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/type_ellipse.R
\name{type_ellipse}
\alias{type_ellipse}
\title{Confidence ellipse plot type}
\usage{
type_ellipse(level = 0.95, segments = 100, density = NULL, angle = 45)
}
\arguments{
\item{level}{Numeric confidence level for the ellipse. Default is \code{0.95}.}

\item{segments}{Integer number of line segments used to approximate the
ellipse. Default is \code{100}.}

\item{density}{the density of shading lines, in lines per inch.  The
    default value of \code{NULL} means that no shading lines are drawn.
    A zero value of \code{density} means no shading nor filling whereas
    negative values and \code{NA} suppress shading (and so allow
    color filling).}

\item{angle}{the slope of shading lines, given as an angle
     in degrees (counter-clockwise).}
}
\description{
Type function for drawing confidence ellipses around grouped
points.
}
\details{
A confidence ellipse summarizes the spread and correlation of two variables:
its axes align with the principal directions of variation, their lengths
reflect the standard deviations along those directions, and the overall size
is scaled so that (assuming bivariate normality) the specified proportion of
observations falls within the boundary.

Operationally, the ellipse is computed from the covariance matrix
(\code{\link[stats]{cov}}) of each group via eigendecomposition
(\code{\link[base]{eigen}}), scaled by a chi-squared quantile
(\code{\link[stats]{qchisq}}) for the specified confidence level.
Groups with fewer than 3 points are silently dropped.
}
\examples{
# "ellipse" type convenience character string
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, type = "ellipse")

# layer filled ellipse(s) on top of points
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic")
tinyplot_add(type = "ellipse", fill = 0.2)

# to avoid clipping, it's better to reverse the layering (ellipse first)
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris, theme = "basic",
         type = "ellipse", fill = 0.2)
tinyplot_add(type = "p")

# custom confidence level
tinyplot(Sepal.Length ~ Petal.Length | Species, data = iris,
  type = type_ellipse(level = 0.5))

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/legend_multi.R
\name{prepare_legend_multi}
\alias{prepare_legend_multi}
\title{Prepare multi-legend specifications}
\usage{
prepare_legend_multi(settings)
}
\arguments{
\item{settings}{Settings environment from tinyplot}
}
\value{
NULL (modifies settings environment in-place)
}
\description{
Sets up multiple legend specifications for multi-legends
(e.g., color grouping + bubble size). Creates \code{lgby} and \code{lgbub} objects
that will be passed to draw_multi_legend().
}
\keyword{internal}
